/*\

    SEMMS_ARG.C

    Program SEMMS (Spatial EM for Markovian Segmentation) :  arguments treatment

    Van Mo DANG       Janvier 96

    Steps to add an option :
    1 - check unused option characters by running 'semms_exe' without args
    2 - add option fetching in SemmsArgs()
    3 - add field to SemmsParaT in "semms_typ.h"  : OptMeaning
    4 - add eventual enum in "semms_typ.h"      : OptET
    5 - add eventual string array             : OptStrVC
    6 - add option default value              : DEFAULT_OPT
    7 - add comment to this header            : 1.06-k blabla
    8 - add short help in PrintUsage()        : -o [%s ] blabla {%s|%s}
    9 - add long help/vers in README

Vers-mod  Date         Who Description

1.03-a    22-AUG-1997  MD  '-l' to output logfile (default now is NO log)
1.04-a    04-OCT-1997  MD  '-s f myfile.zz' instead of just '-s f'
1.04-b    05-OCT-1997  MD  '-B heu <bstep>' implemented
1.04-c    09-OCT-1997  MD  if "-o -" log name is input base name
1.04-d    10-OCT-1997  MD  '-c cvthres' implemented
1.04-e    10-OCT-1997  MD  '-S seed' implemented
1.04-f    10-OCT-1997  MD  '-O order' implemented
1.04-g    13-OCT-1997  MD  '-D' implemented
1.04-h    13-OCT-1997  MD  '-R' implemented (reference class)
1.04-i    13-OCT-1997  MD  '-C crit'
1.04-j    04-NOV-1997  MD  '-I eiter'
1.04-k    11-JAN-1997  MD  now '-h general/filein/fileout/examples/vers'
1.05-a    12-JAN-1998  MD  '-M miss' implemented
1.05-b    05-FEB-1998  MD  add 'M' in '-C crit'
1.06-a    17-JUN-1998  MD  ReadStrFile and alloc parameters <- from nem_exe.c
1.06-b    23-JUN-1998  MD  use new StatModel structure (old NoiseModel)
1.06-b    01-AUG-1998  MD  model switch now '-m norm/lapl p* s**'
1.06-c    03-AUG-1998  MD  add '-s mi' and '-s mf'
1.06-d    10-SEP-1998  MD  add '-U para|seq'
1.06-e    10-SEP-1998  MD  add '-t random|first'
1.06-f    15-SEP-1998  MD  add '-G nbit conv step'
1.06-g    21-SEP-1998  MD  change '-c 0.01' to '-c {none|[clas]|crit} 0.01'
1.06-h    05-OCT-1998  MD  default nb of random inits = D * 5
1.06-i    05-OCT-1998  MD  default crit = CRIT_M
          22-JUN-2001  FC  add choice em for -V option
	  02-JUL-2001  FC  add choice beta for -V option
	  12-JUL-2001  FC  suppressed the message "zeroin for potentials
	                   estimation not yet implemented"
	  20-JUL-2001  FC  trick for not loading .str file -> nk = 0
	                   in AlgoArgs
	  20-JUL-2001  FC  improving -g options -> arguments for generation
	                   of a Potts model representation
	  06-AUG-2001  FC  added -d option so as to make possible the
	                   loop simulation-noise-estimation/restauration
	  07-MAR-2002  NP  add the option pi of argument -s (init with given params)
          27-JAN-2005  NP remove PrintUsage, PrintHelp (now use README)
\*/

#include "semms_typ.h"    /* DataT, ... */
#include "lib_io.h"     /* ReadOpeningComments */ /*V1.06-a*/
#include "genmemo.h"    /* GenAlloc, ... */ /*V1.06-a*/
#include <stdio.h>      /* printf, ... */
#include <stdlib.h>     /* atof, ... */
#include <string.h>     /* strncpy, ... */
#include <time.h>       /* time() */          /*V1.04-e*/

#include "semms_arg.h"    /* prototypes */

#define DEFAULT_ALGO         ALGO_NEM
#define DEFAULT_RESTO        RESTO_MPM
#define DEFAULT_MODE         NORMAL             /* FC */
#define DEFAULT_Opt          ModelSelec
#define DEFAULT_BETA         1.0
#define DEFAULT_BTAMODE      BETA_FIX           /*V1.04-b*/
#define DEFAULT_PARAMODE     PARAM_ESTIM
#define DEFAULT_SINGLMODE    SINGL_NONE
#define DEFAULT_BTAHEUSTEP   0.1                /*V1.04-b*/
#define DEFAULT_BTAHEUMAX    2.0                /*V1.04-b*/
#define DEFAULT_BTAHEUDDROP  0.8             	/*V1.04-b*/
#define DEFAULT_BTAHEUDLOSS  0.5             	/*V1.04-b*/
#define DEFAULT_BTAHEULLOSS  0.02            	/*V1.04-b*/
#define DEFAULT_BTAGRADNIT   10               	/*V1.06-f*/
#define DEFAULT_BTAGRADCVTH  0.001           	/*V1.06-f*/
#define DEFAULT_BTAGRADSTEP  0.0            	/*V1.06-f*/
#define DEFAULT_BTAGRADRAND  0            	/*V1.06-f*/
#define DEFAULT_CRIT         CRIT_M             /*V1.04-h*/
#define DEFAULT_CVTHRES      0.01               /*V1.04-d*/
#define DEFAULT_CVTEST       CVTEST_HARDCLASS   /*V1.06-g*/
#define DEFAULT_FAMILY       FAMILY_NORMALDIAG  /*V1.06-b*/
#define DEFAULT_DISPER       DISPER__D          /*V1.06-b*/
#define DEFAULT_PROPOR       PROPOR_K           /*V1.06-b*/
#define DEFAULT_FORMAT       FORMAT_HARD
#define DEFAULT_INIT         INIT_SORT
#define DEFAULT_INITfpe      INITfpe_GIVEN
#define DEFAULT_MISSING      MISSING_IGNORE    /*V1.05-a*/
#define DEFAULT_SORTEDVAR    0
#define DEFAULT_NEIGHSPEC    NEIGH_FOUR
#define DEFAULT_NBITERS      1000
#define DEFAULT_NBEITERS     1
#define DEFAULT_NBEMITERS    1
#define DEFAULT_NBFPEITERS   50
#define DEFAULT_NBGIBBSITERS 100
#define DEFAULT_NBSIMULS     1
#define DEFAULT_NBGIBBS      1
#define DEFAULT_NBRANDINITS  5                  /*V1.06-h*/
#define DEFAULT_ORDER        ORDER_DIRECT       /*V1.04-f*/
#define DEFAULT_UPDATE       UPDATE_SEQ         /*V1.06-d*/
#define DEFAULT_TIE          TIE_RANDOM         /*V1.06-e*/
#define DEFAULT_CompF        FALSE
#define DEFAULT_True         0
#define DEFAULT_Label        0
#define DEFAULT_CPUmax       10000
#define DEFAULT_bmin         0
#define DEFAULT_bmax         0
#define DEFAULT_bstep        1
#define DEFAULT_Kmin         0
#define DEFAULT_Kmax         0

/* typedef enum */
/* { */

/*   HELP_GENERAL, */
/*   HELP_OPTIONS, */
/*   HELP_EXAMPLES, */
/*   HELP_FILEIN, */
/*   HELP_FILEOUT, */
/*   HELP_VERSIONS, */
/*   HELP_NB */

/* } HelpET ; */


static const char *TypeStrC[ TYPE_NB ] = { "S", "I", "N" } ; /*V1.06-a*/

static const char *AlgoStrVC[ ALGO_NB ] = {"meanf", "gsf","icm",
                                             "nem" ,"ncem" } ;
static const char *RestoStrVC[ RESTO_NB ] = { "icm", "mpm" } ;
static const char *OptStrVC[ Opt_NB ] = {"ModelSelection", "GibbsSampler","AddGaussianNoise" } ;
static const char *BtaStrVC[ BETA_NB ] = { "fix" , "psgrad", "zeroin",
					   "heu_d" , "heu_l", "sa" } ;
static const char *ParamStrVC[ PARAM_NB ] = {"fix" , "estim"} ;
static const char *SinglStrVC[ SINGL_NB ] = {"fix" , "zeroin", "none", "em",
                                             "beta"} ;
       const char *CritStrVC[ CRIT_NB ] = { "U" , "M", "D" , "L" } ;
static const char *CvTestStrVC[ CVTEST_NB ] = { "none", "class" , "hardclass", "crit", "beta", "params", "func" } ;
static const char *FormatStrVC[ FORMAT_NB ] = { "hard", "fuzzy","uf" } ;
/* static const char *HelpStrVC[ HELP_NB ] = { */
/*   "general", */
/*   "options", */
/*   "examples", */
/*   "filein", */
/*   "fileout", */
/*   "versions" */
/* } ; /\* V1.04-k*\/ */

static const char *InitStrVC[ INIT_NB ] = { "s", "r", "pi", "mi", "mf", "f", "l","ind" } ;
/*V1.05-a*/
static const char *MissStrVC[ MISSING_NB ] = { "replace" , "ignore" } ;
/*V1.06-b*/
static const char *FamilyStrVC[ FAMILY_NB ] = { "normgen", "normdiag","lapl" } ;
static const char *DisperStrVC[ DISPER_NB ] = { "s__", "sk_", "s_d", "skd" } ;
static const char *ProporStrVC[ PROPOR_NB ] = { "p_", "pk" } ;
static const char *NeighStrVC[ NEIGH_NB ] = { "4", "f" } ;
static const char *OrderStrVC[ ORDER_NB ] = { "direct", "random" } ;/*V1.04-f*/
static const char *UpdateStrVC[ UPDATE_NB ] = { "seq", "para" } ;/*V1.06-d*/
static const char *TieStrVC[ TIE_NB ] = { "random", "first" } ;/*V1.06-e*/
static const char *InitFPEStrVC[ INITfpe_NB ] = { "hard", "given" } ;
static const char *SaveTypeStrVC[SAVETYPE_NB] = { "all", "last" } ;
/*static const char *TrueDataStatusStrVC[ Status_NB ] = { "known",  "unknown" } ; */


/* ==================== LOCAL FUNCTION PROTOTYPING =================== */

/* Called by AlgoArgs */

StatusET GetMixPara        /* STS_OK or STS_E_ARG */
 (
  const FamilyET  Family,     /* I : distribution family */
  const int       K,          /* I : number of classes */
  const int       D,          /* I : number of variables */
  const char*     Opts_Q[],   /* I : array of options */
  const int       Q,          /* I : number of options */
  int*            IoptP,      /* I/O : index of current option */
  ModelParaT*     ParaP       /* O : read parameters */
 ) ;

/* Idem */
StatusET GetSaveModalities (
			     const char*      Option,
			    AlgoParaT* AlgoParaP
);

/* ==================== GLOBAL FUNCTION DEFINITION =================== */


/* ------------------------------------------------------------------- */
int AlgoArgs
        (
          int           Argc,
          const char    *Argv[],
          char          *Fname,         /* O */
          StatModelT    *StatModelP,    /* O */
          AlgoParaT      *AlgoParaP,      /* O */
	  char*         DatadescS,      /* O */
	  DataT*        DataP,          /* O */
	  ImageNeighT*  ImageP,         /* O */
	  TypeET*       SpatialTypeP    /* O */
        )
/* ------------------------------------------------------------------- */
{
  /* Local declarations */
  StatusET     err = STS_OK ;
  const char*  func = "AlgoArgs" ;
  int          nk = 0 ;
  int          ik ;
  int          nbopt ;
  int          iopt ;
  const char** opts;

  static int  ReadStrFile (  /*V1.06-a*/
			   const char*  BaseName,     /* I */
			   char*        Comment,      /* O */
			   DataT*       DataP,        /* O */
			   ImageNeighT* ImageP,       /* O */
			   TypeET*      TypeP         /* O */
			   );

  static char GetOptionSwitch (const char *Arg);
  static int  GetEnum (const char* S, const char* SV[], int SizeV); /*V1.04-b*/
  /* static StatusET PrintHelp (const char* HelpType); */

  switch ( Argc ) /*V1.04-k*/
    {
    case 1:  /* No args : print error message */
    case 2:
    case 3:
      fprintf(stderr,"use Readme_use file to see how to build a command line\n");
      return STS_I_DONE ;

    default:  /* 3 or more args : continue processing args */
      break ;
    }

  strncpy( Fname, Argv[1], LEN_FILENAME ) ;
  //sscanf( Argv[2], "%d", &nk ) ;
  //StatModelP->Spec.K = nk ;
  //if ( nk <= 0 )
  //  {
  //    fprintf( stderr, "Nb of classes must be > 0 (here %s)\n",
  //             Argv[2] ) ;
  //   return STS_E_ARG ;
  // }

  /* Set default value of optional parameters */
  StatModelP->Spec.ClassFamily = DEFAULT_FAMILY ;
  StatModelP->Spec.ClassDisper = DEFAULT_DISPER ;
  StatModelP->Spec.ClassPropor = DEFAULT_PROPOR ;

  AlgoParaP->MSparam.bmin  = DEFAULT_bmin;
  AlgoParaP->MSparam.bmax  = DEFAULT_bmax;
  AlgoParaP->MSparam.bstep = DEFAULT_bstep ;
  AlgoParaP->MSparam.Kmin  = DEFAULT_Kmin;
  AlgoParaP->MSparam.Kmax  = DEFAULT_Kmax;
  AlgoParaP->Algo          = DEFAULT_ALGO ;
  AlgoParaP->Resto         = DEFAULT_RESTO;
  AlgoParaP->Mode          = DEFAULT_MODE;
  AlgoParaP->Opt           = DEFAULT_Opt ;
  StatModelP->Para.Beta    = DEFAULT_BETA ;          /*V1.06-b*/
  StatModelP->Spec.BetaModel = DEFAULT_BTAMODE ;       /*V1.04-b*/
  AlgoParaP->BtaHeuStep    = DEFAULT_BTAHEUSTEP ;    /*V1.04-b*/
  AlgoParaP->BtaHeuMax     = DEFAULT_BTAHEUMAX ;
  AlgoParaP->BtaHeuDDrop   = DEFAULT_BTAHEUDDROP ;
  AlgoParaP->BtaHeuDLoss   = DEFAULT_BTAHEUDLOSS ;
  AlgoParaP->BtaHeuLLoss   = DEFAULT_BTAHEULLOSS ;
  AlgoParaP->BtaPsGrad.NbIter    = DEFAULT_BTAGRADNIT  ;/*V1.06-g*/
  AlgoParaP->BtaPsGrad.ConvThres = DEFAULT_BTAGRADCVTH ;
  AlgoParaP->BtaPsGrad.Step      = DEFAULT_BTAGRADSTEP ;
  AlgoParaP->BtaPsGrad.RandInit  = DEFAULT_BTAGRADRAND ;
  AlgoParaP->ParamMode     = DEFAULT_PARAMODE ;
  AlgoParaP->SinglMode     = DEFAULT_SINGLMODE ;
  AlgoParaP->MissMode      = DEFAULT_MISSING ;
  AlgoParaP->Crit          = DEFAULT_CRIT ;          /*V1.04-h*/
  AlgoParaP->CvThres       = DEFAULT_CVTHRES ;       /*V1.04-d*/
  AlgoParaP->CvTest        = DEFAULT_CVTEST ;           /*V1.06-g*/
  AlgoParaP->NbIters       = DEFAULT_NBITERS ;
  AlgoParaP->NbEIters      = DEFAULT_NBEITERS ;
  AlgoParaP->NbEMIters     = DEFAULT_NBEMITERS ;
  AlgoParaP->NbfpeIters    = DEFAULT_NBFPEITERS ;
  AlgoParaP->NbGibbsIters  = DEFAULT_NBGIBBSITERS ;
  AlgoParaP->NbSimuls      = DEFAULT_NBSIMULS ;
  AlgoParaP->NbGibbs       = DEFAULT_NBGIBBS ;
  AlgoParaP->NbRandomInits = DEFAULT_NBRANDINITS*DataP->NbVars ;  /*V1.06-h*/
  AlgoParaP->Seed          = time( NULL ) ;          /*V1.04-e*/
  AlgoParaP->Format        = DEFAULT_FORMAT ;
  AlgoParaP->InitMode      = DEFAULT_INIT ;
  AlgoParaP->InitFPE       = DEFAULT_INITfpe ;
  AlgoParaP->SortedVar     = DEFAULT_SORTEDVAR ;
  AlgoParaP->NeighSpec     = DEFAULT_NEIGHSPEC ;
  AlgoParaP->VisitOrder    = DEFAULT_ORDER ;         /*V1.04-f*/
  AlgoParaP->SiteUpdate    = DEFAULT_UPDATE ;        /*V1.06-d*/
  AlgoParaP->TieRule       = DEFAULT_TIE ;           /*V1.06-e*/
  AlgoParaP->CompF         = DEFAULT_CompF ;
  AlgoParaP->Debug         = FALSE ;         /*V1.04-g*/
  AlgoParaP->True.K  = DEFAULT_True ;
  AlgoParaP->Label        = DEFAULT_Label ;
  AlgoParaP->CPUmax       = DEFAULT_CPUmax;
  strncpy( AlgoParaP->OutBaseName, Fname, LEN_FILENAME ) ;
  strncpy( AlgoParaP->NeighName, Fname, LEN_FILENAME ) ;
  strncat( AlgoParaP->NeighName, ".nei", LEN_FILENAME ) ;
  strncpy( AlgoParaP->RefName, "", LEN_FILENAME ) ;
  strncpy( AlgoParaP->ParamName, Fname, LEN_FILENAME ) ;
  strncat( AlgoParaP->ParamName, ".param", LEN_FILENAME ) ;
  strncpy( AlgoParaP->ListName, Fname, LEN_FILENAME ) ;
  strncpy( AlgoParaP->SinglName, Fname, LEN_FILENAME ) ;
  strncat( AlgoParaP->SinglName, ".singl", LEN_FILENAME ) ;
  strncpy( AlgoParaP->ListName, Fname, LEN_FILENAME ) ;
  strncat( AlgoParaP->ListName, ".list", LEN_FILENAME ) ;
  strncpy( AlgoParaP->TrueName, Fname, LEN_FILENAME ) ;
  strncat( AlgoParaP->TrueName, ".true", LEN_FILENAME ) ;

  /* Treat options */
  nbopt = Argc - 2 ;
  opts  = & ( Argv[ 2 ] ) ;

  for ( iopt = 0, err = STS_OK ;
        ( iopt < nbopt ) && ( err == STS_OK ) ;
        iopt ++ )
    {
      char copt = GetOptionSwitch( opts[ iopt ] ) ;

      if ( copt != '\0' )
        switch( copt )
        {
        case 'a' :    /* next arg is type of algorithm */
          iopt ++ ;
          if ( iopt < nbopt )
            {
	      AlgoParaP->Algo = GetEnum( opts[ iopt ], AlgoStrVC, ALGO_NB ) ;
              if ( AlgoParaP->Algo == -1 )
                {
                  fprintf( stderr, " Unknown type of algorithm %s\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG) ;
                }

            }
          else
            {
              fprintf( stderr, " Expected value after switch %s\n",
                       opts[ iopt - 1 ] ) ;
              return( STS_E_ARG) ;
            }


          break ;



        case 'c' :    /* next arg : conv test */  /*V1.04-d*/ /*V1.06-g*/
	  iopt ++ ;
          if ( iopt < nbopt ) /* conv test given */ {
	    AlgoParaP->CvTest=GetEnum( opts[ iopt ], CvTestStrVC, CVTEST_NB );
	    if ( AlgoParaP->CvTest == -1 ) {
	      fprintf( stderr, " Unknown convergence test %s\n", opts[iopt] ) ;
	      return (STS_E_ARG );
	    }
	    else if ( AlgoParaP->CvTest != CVTEST_NONE ) /* get threshold */ {
	      iopt ++ ;
	      if ( iopt < nbopt ) /* threshold given */ {
		AlgoParaP->CvThres = atof( opts[ iopt ] ) ;
		if ( AlgoParaP->CvThres <= 0 ) {
		  fprintf( stderr, " Conv threshold must be > 0 (here %s)\n",
			   opts[ iopt ] ) ;
		  return( STS_E_ARG );
		} /* else threshold > 0 : OK */
	      }
	      else /* threshold not given */ {
		fprintf( stderr, " Expected threshold for conv test %s\n",
			 opts[ iopt - 1 ] ) ;
		return( STS_E_ARG );
	      }
            } /* threshold checked */
	  } /* conv test given */
          else {
	    fprintf( stderr, " Expected value after switch %s\n",
		     opts[ iopt - 1 ] ) ;
	    return( STS_E_ARG) ;
	  }
          break ;

	case 'd' :  /* If present : loop simulation-noise-restauration
		       with parameters given
		    */
	  if (iopt + 3 < nbopt)
	    {
	      iopt++;
	      AlgoParaP->LoopParam.NbIters = atof(opts[iopt]);
	      if (AlgoParaP->LoopParam.NbIters <= 0)
		{
		  fprintf(stderr,"Nb of loop iters must be strictly positive.\n");
		  return(STS_E_ARG);
		}
	      iopt++;
	      if (GetSaveModalities(opts[iopt],AlgoParaP) != STS_OK)
		return( STS_E_ARG);
	      iopt++;
	      AlgoParaP->LoopParam.SaveType = GetEnum(opts[iopt],SaveTypeStrVC,
						      SAVETYPE_NB);
	      /* Important : we want a loop behaviour */
	      AlgoParaP->Mode = LOOP;
	      AlgoParaP->Opt  = GibbsSampler;
	    }
	  else
	    {
	      fprintf(stderr,"Expected 3 values after switch %s\n",
		      opts[iopt-1]);
	      return( STS_E_ARG);
	    }
	  break;

	case 'e' :    /* next arg is status of true data */
        iopt ++ ;
          if ( iopt < nbopt )
            {
	      AlgoParaP->True.K = atof( opts[ iopt ] );
	        if ( AlgoParaP->True.K <= 0 )
                {
                  fprintf( stderr, " true K must be >0, here  %s\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG );
		  }
            }
          else
            {
              fprintf( stderr, " Expected  true data status after switch %s\n",
                       opts[ iopt - 1 ] ) ;
              return( STS_E_ARG) ;
            }
          break ;

        case 'f' :    /* next arg is partition input/output format */
          iopt ++ ;
          if ( iopt < nbopt )
            {
	      AlgoParaP->Format=GetEnum( opts[ iopt ], FormatStrVC, FORMAT_NB );
              if ( AlgoParaP->Format == -1 )
                {
                  fprintf( stderr, " Unknown format %s\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG) ;
                }
            }
          else
            {
              fprintf( stderr, " Expected value after switch %s\n",
                       opts[ iopt - 1 ] ) ;
              return( STS_E_ARG) ;
            }
          break ;

      case 'g':
         iopt++;
         StatModelP->Spec.K = atoi( opts[ iopt ] ) ;
              if ( StatModelP->Spec.K  <= 0 )
                {
                  fprintf( stderr, "Nb of classes  be >0, here  %s\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG );
		}

         if ( iopt + StatModelP->Spec.K + 4 < nbopt )
           {
            int boucle;
            /* the Potts singleton potentials */
            if ( iopt + StatModelP->Spec.K < nbopt ) {
		StatModelP->Para.Singl  = (double*) malloc(sizeof(double) *
							  StatModelP->Spec.K);
		for (boucle = 0;  boucle < StatModelP->Spec.K; boucle++)
                    {
		       iopt++;
		       StatModelP->Para.Singl [boucle] = atof(opts[iopt]);
		    }
	      }
             else {
		fprintf(stderr, " Expected 5 + K values after switch %s\n",
			opts[ iopt - 1] ) ;
		return( STS_E_ARG);
	          }
	    /* beta */
           iopt++;
           StatModelP->Para.Beta = atof( opts[ iopt ] ) ;
           if ( StatModelP->Para.Beta < 0 )
                {
                  fprintf( stderr, " beta (Potts param) be >=0, here  %s\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG );
		  }

           /* Nl Nc */
	    iopt++;
            ImageP->Nl= atoi( opts[ iopt ] ) ;
            if ( ImageP->Nl < 0 )
                {
                  fprintf( stderr, "Nb of lignes must   be >=0, here  %s\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG );
		  }

              iopt++;
              ImageP->Nc = atoi( opts[ iopt ] ) ;
              if ( ImageP->Nc < 0 )
                {
                  fprintf( stderr, " Nb of columns be >=0, here  %s\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG );
		  }

	      /* nb of GS iters */
              iopt++;
	      AlgoParaP->NbGibbsIters = atoi(opts[iopt]);
              if ( AlgoParaP->NbGibbsIters <= 0 )
                {
                  fprintf( stderr, " Nb of Gibbs sampling iters be >0, here  %s\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG );
		  }

              AlgoParaP->Opt = GibbsSampler;
	      nk             = StatModelP->Spec.K;
	      DataP->NbVars  = 1;
	      DataP->NbMiss  = 0;
	      DataP->NbPts   = ImageP->Nc * ImageP->Nl;

           } /* end of if  iopt + StatModelP->Spec.K + 4 < nbopt  */

        else
           {
              fprintf( stderr, " Expected 5  + K values after switch %s\n",
                       opts[ iopt - 1 ] ) ;
              return( STS_E_ARG) ;
	   }

          break ;


        case 'i' :    /* next arg is number of iterations */
          iopt ++ ;
          if ( iopt < nbopt )
            {
              AlgoParaP->NbIters = atoi( opts[ iopt ] ) ;
              if ( AlgoParaP->NbIters < 0 )
                {
                  fprintf( stderr, "Nb iterations must be >= 0 (here %s)\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG );
                }
            }
          else
            {
              fprintf( stderr, " Expected value after switch %s\n",
                       opts[ iopt - 1 ] ) ;
              return( STS_E_ARG) ;
            }
          break ;


        case 'm' :    /* next 3 args is stat model */ /*V1.06-b*/
          if ( iopt + 3 < nbopt )
            {
	      iopt ++ ;
	      StatModelP->Spec.ClassFamily =
		GetEnum( opts[ iopt ], FamilyStrVC, FAMILY_NB );
              if ( StatModelP->Spec.ClassFamily == -1 )
                {
                  fprintf( stderr, " Unknown family %s\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG) ;
                }
              if ( StatModelP->Spec.ClassFamily == FAMILY_LAPLACE)
		{
                  fprintf( stderr, "****Be carefull**** the program  have bugs for Laplace density \n");
                  return( STS_E_ARG);
		}

              if ( (DataP->NbVars == 1) &&
		   (StatModelP->Spec.ClassFamily == FAMILY_NORMALGEN) )
		  {
                   fprintf( stderr, "if D = 1, choose normdiag and not normgen\n");
                    return( STS_E_ARG);
                  }

	      iopt ++ ;
	      StatModelP->Spec.ClassPropor =
		GetEnum( opts[ iopt ], ProporStrVC, PROPOR_NB );
              if ( StatModelP->Spec.ClassPropor == -1 )
                {
                  fprintf( stderr, " Unknown proportion %s\n",
                           opts[ iopt ] ) ;
                 return( STS_E_ARG) ;
                }
	      iopt ++ ;
	      StatModelP->Spec.ClassDisper =
		GetEnum( opts[ iopt ], DisperStrVC, DISPER_NB );
              if ( StatModelP->Spec.ClassDisper == -1 )
                {
                  fprintf( stderr, " Unknown dispersion %s\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG) ;
                }

              if ( StatModelP->Spec.ClassFamily == FAMILY_NORMALGEN )
		{
                  if ( (StatModelP->Spec.ClassDisper == DISPER___ )||
                   (StatModelP->Spec.ClassDisper == DISPER_K_ ))
		    {
                      fprintf( stderr, " invalid dispersion for model normalgen\n");
                      return( STS_E_ARG) ;
                     }
		}
            }
          else
            {
              fprintf( stderr, " Expected 3 values after switch %s\n",
                       opts[ iopt  ] ) ;
              return( STS_E_ARG) ;
            }
          break ;


        case 'n' :    /* next arg is neighborhood specification */
          iopt ++ ;
          if ( iopt < nbopt )
            {
	      AlgoParaP->NeighSpec=GetEnum( opts[ iopt ], NeighStrVC, NEIGH_NB);
              if ( AlgoParaP->NeighSpec == -1 )
                {
                  fprintf( stderr, " Unknown neighborhood specification %s\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG) ;
                }
            }
          else
            {
              fprintf( stderr, " Expected value after switch %s\n",
                       opts[ iopt - 1 ] ) ;
             return( STS_E_ARG) ;
            }
          break ;


        case 'o' :    /* next arg is output file name */
          iopt ++ ;
          if ( iopt < nbopt )
            {
              strncpy( AlgoParaP->OutBaseName, opts[ iopt ], LEN_FILENAME ) ;
            }
          else
            {
              fprintf( stderr, " Expected file name after switch %s\n",
                       opts[ iopt - 1 ] ) ;
              return( STS_E_ARG) ;
            }
          break ;


        case 'r' :  /* next argument is the nb of Gibbs Samplers at each
                  iteration of the GSF */
          if ( iopt + 1 < nbopt )
	  {

              iopt ++ ;
                 AlgoParaP->NbGibbs = atoi( opts[ iopt ] ) ;
                 if ( AlgoParaP->NbGibbs < 1 )
                  {
                  fprintf( stderr, "Nb Gibbs samplers must be >= 1 (here %s)\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG) ;
                  }

	  }
          else
          {
              fprintf( stderr, " Expected 2 values after switch %s\n",
                       opts[ iopt  ]) ;
              err = STS_E_ARG ;
          }
          break ;


        case 's' :    /* next arg is initialization mode */
          iopt ++ ;
          if ( iopt < nbopt )
            {
	      AlgoParaP->InitMode = GetEnum( opts[ iopt ], InitStrVC, INIT_NB );

              switch( AlgoParaP->InitMode )  {
	        case INIT_EMIND:
		  break;

                case INIT_SORT:  /* next arg is sorted var */
                  iopt ++ ;
                  if ( iopt < nbopt )
                    {
                      AlgoParaP->SortedVar = atoi( opts[ iopt ] ) - 1 ;
                      if ( AlgoParaP->SortedVar < 0 )
                            {
                             fprintf( stderr, "variable dimension to sort must be >= 1(here  %s)\n",
                             opts[ iopt ] ) ;
                             return( STS_E_ARG) ;
                            }

                    }
                  else
                    {
                      fprintf( stderr,
                               " Expected variable index after switch %s\n",
                               opts[ iopt - 1 ] ) ;
                      return (STS_E_ARG) ;
                    }
                  break ;

                case INIT_RANDOM:  /* next arg is number of initializations */
                  iopt ++ ;
                  if ( iopt < nbopt )
                    {
                      AlgoParaP->NbRandomInits = atoi( opts[ iopt ] ) ;
                      if (AlgoParaP->NbRandomInits  <= 0 )
                            {
                             fprintf( stderr, " Nb Random Inits must be >=1,  (here   %s)\n",
                             opts[ iopt ] ) ;
                             return( STS_E_ARG) ;
                            }
                    }
                  else
                    {
                      fprintf( stderr,
                               " Expected nb of trials after switch %s\n",
                               opts[ iopt - 1 ] ) ;
                      return(STS_E_ARG) ;
                    }
                  switch (AlgoParaP->Algo)
		    {
		    case ALGO_MEANF:
		    case ALGO_ICM:
		    case ALGO_GSF:
                          fprintf( stderr,
                               " this init mode is not available for this program\n") ;
                         return(STS_E_ARG) ;
                         break ;
                    default: break;
		    }
                  break ;

                case INIT_FILE:  /* next arg is name of initial partition */
		  iopt ++ ;   /*V1.04-a*/
		  if ( iopt < nbopt )
		    {
		      strncpy( AlgoParaP->StartName, opts[ iopt ],
			       LEN_FILENAME ) ;
		    }
		  else
		    {
		      fprintf( stderr,
			       " Expected file name after switch %s\n",
			       opts[ iopt - 1 ] ) ;
		      return(STS_E_ARG) ;
		    }
                  break ;



	        case INIT_PARAM: /* next args are parameter values */ /* Nath's version */
                  iopt ++ ;
		  if ( iopt < nbopt )
		    {
		      strncpy( AlgoParaP->StartParamName, opts[ iopt ],
			       LEN_FILENAME ) ;
		    }
		  else
		    {
		      fprintf( stderr,
			       " Expected file name after switch %s\n",
			       opts[ iopt - 1 ] ) ;
		      return(STS_E_ARG) ;
		    }
                  break ;

		  break;

	      case INIT_MIXINI: /* next args are parameter values */ /* Mo's version */
	      case INIT_MIXFIX: /* next args are parameter values *//* Mo's version */
		  /*V1.06-c*/
                     fprintf( stderr,
                               " this init mode has to be adapted to the use of -K bmin bmax bstep kmin kmax \n") ;
                         return( STS_E_ARG) ;

                  switch (AlgoParaP->Algo)
		    {
		    case ALGO_MEANF:
		    case ALGO_ICM:
		    case ALGO_GSF:
                          fprintf( stderr,
                               " this init mode is not available for this program\n") ;
                         return( STS_E_ARG) ;
                         break ;
                    default:
                          err = GetMixPara( StatModelP->Spec.ClassFamily, nk,
				    DataP->NbVars, opts, nbopt,
				    & iopt, & StatModelP->Para ) ;
                          if (err != STS_OK) return ( STS_E_ARG) ;
		    }
		  break ;

                default:
                  fprintf( stderr, " Unknown init mode %s\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG) ;
	      }

            }
          else
            {
              fprintf( stderr, " Expected init mode after switch %s\n",
                       opts[ iopt - 1 ] ) ;
              return( STS_E_ARG) ;
            }
          break ;

        case 'u' : /* when using the option AddGaussNoise */
	   iopt++ ;
	   StatModelP->Spec.K= atof( opts[ iopt ] );
           nk = StatModelP->Spec.K;
	   if ( StatModelP->Spec.K < 1 )
             {
                fprintf( stderr, "Nb of classes must be >=1  (here  %s)\n",
                             opts[ iopt ] ) ;
                             return( STS_E_ARG) ;
                            }


	  AlgoParaP->Opt = AddGaussianNoise ;

          if ( (2*StatModelP->Spec.K +iopt) < nbopt )
            {

             if ((AlgoParaP->GaussNoise.unique
              = GenAlloc(StatModelP->Spec.K, sizeof(double), 0, "AlgoArgs","unique"))  == NULL )
            return STS_E_MEMORY ;
            if ((AlgoParaP->GaussNoise.stdev
             = GenAlloc(StatModelP->Spec.K, sizeof(double), 0, "AlgoArgs","unique"))  == NULL )
            return STS_E_MEMORY ;

	    for (ik = 0; ik < StatModelP->Spec.K; ik ++)
              {
                iopt++ ;
                AlgoParaP->GaussNoise.unique[ik] = atof( opts[ iopt ] );
              }

	    for (ik = 0; ik < StatModelP->Spec.K; ik ++)
              {
                iopt++ ;
                AlgoParaP->GaussNoise.stdev[ik] = atof( opts[ iopt ] );
              }
	    }


         else
            {
              fprintf( stderr, " Expected value after switch %s\n",
                       opts[ iopt] ) ;
              return( STS_E_ARG) ;
            }

        break ;


        case 'w' : /* next arg is the max of CPU to be used by the algo */
        iopt ++ ;
        if ( iopt < nbopt )
            {
             AlgoParaP->CPUmax=atoi( opts[ iopt ] ) ;
              if ( AlgoParaP->CPUmax < 0 )
                {
                  fprintf( stderr, "CPU max must be >= 0 (here %s)\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG) ;
                }
            }
          else
            {
              fprintf( stderr, " Expected value after switch %s\n",
                       opts[ iopt - 1 ] ) ;
              return( STS_E_ARG) ;
            }
          break ;


        case 'B' :    /* next arg is beta estimation method */ /*V1.04-b*/
          iopt ++ ;
          if ( iopt < nbopt )
            {
              StatModelP->Spec.BetaModel =
		GetEnum( opts[ iopt ], BtaStrVC, BETA_NB );
	      if ( StatModelP->Spec.BetaModel == -1 )
		{
                  fprintf( stderr, " Unknown type of beta estimation %s\n",
                           opts[ iopt ] ) ;
                  return(STS_E_ARG );
		}
	      else {
		if ( ( StatModelP->Spec.BetaModel == BETA_HEUL ) |
		     ( StatModelP->Spec.BetaModel == BETA_HEUD ) )
		  AlgoParaP->Crit = CRIT_U ;  /* force to choose U criterion */
	      }
            }
          else
            {
              fprintf( stderr, " Expected value after switch %s\n",
                       opts[ iopt - 1 ] ) ;
              return( STS_E_ARG) ;
            }

          if ( (AlgoParaP->Algo == ALGO_NEM) ||(AlgoParaP->Algo == ALGO_NCEM))
            {
             /* make sure the proportions are supposed to be equal if BETA_PSGRAD*/
             if ( (StatModelP->Spec.BetaModel == BETA_ZEROIN) &&
                (StatModelP->Spec.ClassPropor != PROPOR__ ))
	      {
                fprintf( stderr, "proportions must be equal if beta is estimated in NEM, NCEM or GEM \n");
                return( STS_E_ARG );
	       }
	    }
          if (StatModelP->Spec.BetaModel != BETA_FIX)
	     {
               if ( (AlgoParaP->Algo == ALGO_GSF) )
		 {
                   if (AlgoParaP->NbGibbs != 1)
                    {
		      /* prior approx not well defined in that case */
                     fprintf( stderr, "beta must be fixed if GSF  is used with nbGibbs >1 \n");
                     return( STS_E_ARG );
	            }
		 }
	     }

          break ;



        case 'C' :    /* next arg is crit to choose local max */ /*V1.04-i*/
          iopt ++ ;
          if ( iopt < nbopt )
            {
              AlgoParaP->Crit = GetEnum( opts[ iopt ], CritStrVC, CRIT_NB );
	      if ( AlgoParaP->Crit == -1 )
		{
                  fprintf( stderr, " Unknown criterion name %s\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG) ;
		}
            }
          else
            {
              fprintf( stderr, " Expected value after switch %s\n",
                       opts[ iopt - 1 ] ) ;
              return( STS_E_ARG );
            }
          break ;

        case 'E' :    /* next arg is the initialization mode for ComputePriorMean */
         iopt ++ ;
          if ( iopt < nbopt )  {
	    AlgoParaP->InitFPE = GetEnum( opts[ iopt ], InitFPEStrVC, INITfpe_NB );
            if ( AlgoParaP->InitFPE  == -1 )
                {
                  fprintf( stderr, " Unknown init mode for theFPE %s\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG) ;
                }
            }
          else
            {
              fprintf( stderr, " Expected value after switch %s\n",
                       opts[ iopt - 1 ] ) ;
              return( STS_E_ARG );
            }
          break ;



	case 'G' :   /* next 4 args are parameters of beta gradient */
	  if ( iopt + 4 < nbopt )
	    {
	      AlgoParaP->BtaPsGrad.NbIter    = atof( opts[ iopt + 1 ] ) ;
	      AlgoParaP->BtaPsGrad.ConvThres = atof( opts[ iopt + 2 ] ) ;
	      AlgoParaP->BtaPsGrad.Step      = atof( opts[ iopt + 3 ] ) ;
	      AlgoParaP->BtaPsGrad.RandInit  = atoi( opts[ iopt + 4 ] ) ;
	      iopt += 4 ;
	    }
	  else
	    {
	      fprintf( stderr, " Expected 4 values after switch %s\n",
		       opts[ iopt  ] ) ;
	      return( STS_E_ARG) ;
	    }
	  break ;




	case 'H' :   /* next 5 args are parameters of beta heuristic */
	  if ( iopt + 5 < nbopt )
	    {
	      AlgoParaP->BtaHeuStep = atof( opts[ iopt + 1 ] ) ;
	      AlgoParaP->BtaHeuMax  = atof( opts[ iopt + 2 ] ) ;
	      AlgoParaP->BtaHeuDDrop = atof( opts[ iopt + 3 ] ) ;
	      AlgoParaP->BtaHeuDLoss = atof( opts[ iopt + 4 ] ) ;
	      AlgoParaP->BtaHeuLLoss = atof( opts[ iopt + 5 ] ) ;
	      iopt += 5 ;
	    }
	  else
	    {
	      fprintf( stderr, " Expected 5 values after switch %s\n",
		       opts[ iopt ] ) ;
	      err = STS_E_ARG ;
	    }
	  break ;


        case 'I' :    /* next arg is number of iterations for ICm or MeanField eq.*/ /*V1.04-j*/
          iopt ++ ;
          if ( iopt < nbopt )
            {
              AlgoParaP->NbEIters = atoi( opts[ iopt ] ) ;
              if ( AlgoParaP->NbEIters < 0 )
                {
                  fprintf( stderr, "Nb iterations must be >= 0 (here %s)\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG) ;
                }
            }
          else
            {
              fprintf( stderr, " Expected value after switch %s\n",
                       opts[ iopt - 1 ] ) ;
              return( STS_E_ARG) ;
            }
          break ;

        case 'K' : /* parameters for simultaneous selection of beta and K */
          if ( iopt + 5 < nbopt )
            {
              iopt++;
              AlgoParaP->MSparam.bmin = atof( opts[ iopt ] ) ;
              if ( AlgoParaP->MSparam.bmin < 0 )
                {
                  fprintf( stderr, "bmin must be >=0  (here %s)\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG) ;
                }

              iopt++;
              AlgoParaP->MSparam.bmax = atof( opts[ iopt ] ) ;
              if (AlgoParaP->MSparam.bmax < 0 )
                {
                  fprintf( stderr, " bmax must be >= 0 (here %s)\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG) ;
                }


              if (AlgoParaP->MSparam.bmax < AlgoParaP->MSparam.bmin )
		{
		fprintf( stderr, " beta max must be >= beta min\n");
                err = STS_E_ARG ;
		}
              else
		{
                 iopt++;
                 AlgoParaP->MSparam.bstep = atof( opts[ iopt ] ) ;
                 if (AlgoParaP->MSparam.bstep < 0 )
                    {
                       fprintf( stderr, " bstep must be >=0  (here %s)\n",
                           opts[ iopt ] ) ;
                         return( STS_E_ARG) ;
                     }


                 iopt++;
                 AlgoParaP->MSparam.Kmin = atof( opts[ iopt ] ) ;
                 nk = AlgoParaP->MSparam.Kmin ;
                 if (AlgoParaP->MSparam.Kmin <1 )
                     {
                         fprintf( stderr, " Kmin must be >=1(here %s)\n",
                           opts[ iopt ] ) ;
                         return( STS_E_ARG) ;
                      }


                 iopt++;
                 AlgoParaP->MSparam.Kmax = atof( opts[ iopt ] ) ;
                 if (AlgoParaP->MSparam.Kmax<1 )
                  {
                       fprintf( stderr, "Kmax must be >=1(here %s)\n",
                           opts[ iopt ] ) ;
                       return( STS_E_ARG) ;
                   }

                 if (AlgoParaP->MSparam.Kmax < AlgoParaP->MSparam.Kmin )
		  {
		  fprintf( stderr, " K max must be >= K min\n");
                  return( STS_E_ARG) ;
		  }
                }

              AlgoParaP->Opt = ModelSelec ;
	    }

          else
            {
              fprintf( stderr, " Expected 5 values after switch %s\n",
                       opts[ iopt] ) ;
              return( STS_E_ARG );
            }
          break ;

        case 'L' :    /* next arg is about available labels */     /*V1.05-a*/
          iopt ++ ;
          if ( iopt < nbopt )
          {
             AlgoParaP->Label = 1  ;
	     strncpy( AlgoParaP->LabelName, opts[ iopt ],
			       LEN_FILENAME ) ;

	  }
          else
            {
              fprintf( stderr, " Expected value after switch %s\n",
                       opts[ iopt - 1 ] ) ;
               return(STS_E_ARG );
            }
          break;

        case 'M' :    /* next arg is missing data mode */     /*V1.05-a*/
          iopt ++ ;
          if ( iopt < nbopt )
            {
	      AlgoParaP->MissMode=GetEnum(opts[ iopt ], MissStrVC, MISSING_NB);
              if ( AlgoParaP->MissMode == -1 )
                {
                  fprintf( stderr, " Unknown missing mode %s\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG) ;
                }
            }
          else
            {
              fprintf( stderr, " Expected value after switch %s\n",
                       opts[ iopt - 1 ] ) ;
              return( STS_E_ARG) ;
            }
          break ;



        case 'N' :    /* next arg is number of iterations for EM within each iteration
                        of the mean field-like algorithms */
          iopt ++ ;
          if ( iopt < nbopt )
            {
              AlgoParaP->NbEMIters = atoi( opts[ iopt ] ) ;
              if ( AlgoParaP->NbEMIters < 0 )
                {
                  fprintf( stderr, "Nb iterations must be >= 0 (here %s)\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG) ;
                }
            }
          else
            {
              fprintf( stderr, " Expected value after switch %s\n",
                       opts[ iopt - 1 ] ) ;
              return( STS_E_ARG) ;
            }
          break ;

        case 'O' :    /* next arg is visit order type */     /*V1.04-f*/
          iopt ++ ;
          if ( iopt < nbopt )
            {
	      AlgoParaP->VisitOrder=GetEnum(opts[ iopt ], OrderStrVC, ORDER_NB);
              if ( AlgoParaP->VisitOrder == -1 )
                {
                  fprintf( stderr, " Unknown visit order type %s\n",
                           opts[ iopt ] ) ;
                  return(STS_E_ARG) ;
                }
            }
          else
            {
              fprintf( stderr, " Expected value after switch %s\n",
                       opts[ iopt - 1 ] ) ;
              err = STS_E_ARG ;
            }
          break ;

	case 'P' :  /* next arg is non-Beta parameters estimation mode */
          iopt ++ ;
          if ( iopt < nbopt )
            {
	      AlgoParaP->ParamMode = GetEnum( opts[ iopt ], ParamStrVC, PARAM_NB );
              if ( AlgoParaP->ParamMode == -1 )
                {
                  fprintf( stderr, " Unknown type of non-beta params estimation %s\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG );
                }
            }
          else
            {
              fprintf( stderr, " Expected init mode after switch %s\n",
                       opts[ iopt - 1 ] ) ;
              return( STS_E_ARG );
            }
          break ;



        case 'R' :    /* next arg is reference class file */  /*V1.04-h*/
          iopt ++ ;
          if ( iopt < nbopt )
            {
              strncpy( AlgoParaP->RefName, opts[ iopt ], LEN_FILENAME ) ;
            }
          else
            {
              fprintf( stderr, " Expected value after switch %s\n",
                       opts[ iopt - 1 ] ) ;
              return( STS_E_ARG) ;
            }
          break ;



        case 'S' :    /* next arg is random generator seed */  /*V1.04-e*/
          iopt ++ ;
          if ( iopt < nbopt )
            {
              AlgoParaP->Seed = atol( opts[ iopt ] ) ;
            }
          else
            {
              fprintf( stderr, " Expected value after switch %s\n",
                       opts[ iopt - 1 ] ) ;
              return( STS_E_ARG) ;
            }
          break ;


        case 'U' :    /* next arg is site update mode */  /*V1.06-d*/
          iopt ++ ;
          if ( iopt < nbopt )
            {
	      AlgoParaP->SiteUpdate=GetEnum(opts[ iopt ],UpdateStrVC,UPDATE_NB);
              if ( AlgoParaP->SiteUpdate == -1 )
                {
                  fprintf( stderr, " Unknown site update mode %s\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG) ;
                }
            }
          else
            {
              fprintf( stderr, " Expected value after switch %s\n",
                       opts[ iopt - 1 ] ) ;
              return(STS_E_ARG) ;
            }
          break ;


	case 'V':  /* next arg singleton potentiel estimation mode */
          iopt ++ ;
          if ( iopt < nbopt )
            {
	      AlgoParaP->SinglMode = GetEnum( opts[ iopt ], SinglStrVC, SINGL_NB );
	      if ( AlgoParaP->SinglMode == -1 )
                {
                  fprintf( stderr, " Unknown type of singleton potentiel estimation %s\n",
                           opts[ iopt ] ) ;
                  return( STS_E_ARG) ;
                }
            }
          else
            {
              fprintf( stderr, " Expected init mode after switch %s\n",
                       opts[ iopt - 1 ] ) ;
              return( STS_E_ARG) ;
            }



            if (AlgoParaP->SinglMode != SINGL_NONE)
                   if( (AlgoParaP->Algo == ALGO_NEM) ||
                       (AlgoParaP->Algo == ALGO_NCEM) )
		     {
                        fprintf( stderr, "Must use -V none with NEM, NCEM or GEM\n");
                        return( STS_E_ARG) ;

                      }


          break ;


          break;

         case 'Z' :    /* next arg is restoration method */ /*V1.04-b*/
          iopt ++ ;
          if ( iopt < nbopt )
            {
              AlgoParaP->Resto =
		GetEnum( opts[ iopt ], RestoStrVC, RESTO_NB );
	      if ( AlgoParaP->Resto == -1 )
		{
                  fprintf( stderr, " Unknown type of restoration method %s\n",
                           opts[ iopt ] ) ;
                  return(  STS_E_ARG) ;
		}

             if ( (AlgoParaP->Resto == RESTO_ICM) && (AlgoParaP->Algo == ALGO_NEM)
                  && (StatModelP->Spec.BetaModel == BETA_FIX)  && (StatModelP->Para.Beta == 0 ) )
		{
                  fprintf( stderr, "EM for mixture is not valid to estimate the parameter when  the segmentation is recovered using ICM  \n");

                 return( STS_E_ARG) ;
		}


	    }
          else
            {
              fprintf( stderr, " Expected value after switch %s\n",
                       opts[ iopt - 1 ] ) ;
              return( STS_E_ARG) ;
            }


          break ;

        case 'v' :    /* print out version information */
          return( STS_I_DONE) ;
          break ;


        default :
          fprintf( stderr, "Unknown option switch '%s'\n", opts[ iopt ] ) ;
           return(STS_E_ARG) ;
        } /* end if ( copt != '\0' ) switch ( copt ) */
      else
        {
          fprintf( stderr, " Expected an option switch at arg %d (%s)\n",
                   iopt + 3 , opts[ iopt ] ) ;
          return( STS_E_ARG) ;
        }
    } /* end for ( iopt ... ) */

  strncpy( AlgoParaP->OutName, AlgoParaP->OutBaseName, LEN_FILENAME ) ;
  strncat( AlgoParaP->OutName, EXT_OUTNAMEHARD,
	   /*  AlgoParaP->Format == FORMAT_HARD ? EXT_OUTNAMEHARD : EXT_OUTNAMEFUZZY,*/
           LEN_FILENAME ) ;

      strcpy( AlgoParaP->LogName, "" ) ;
 

  /* !!! Read structure file */ /*V1.06-a*/
  if (AlgoParaP->Opt != GibbsSampler)
    if ( ( err = ReadStrFile( Fname,
			      DatadescS,
			      DataP,
			      ImageP,
			      SpatialTypeP ) ) != STS_OK )
      return err ;



  /* !!! Allocate model parameters */ /*V1.06-a*/

  if ( (StatModelP->Para.Ztilde    = GenAlloc( DataP->NbPts * nk, sizeof(double),
                                         1, func, "Ztilde")) == NULL )
               return STS_E_MEMORY;
  if ( (StatModelP->Para.Prop_K    = GenAlloc( nk, sizeof(double),
					 1, func, "Prop_K" ))== NULL )
               return STS_E_MEMORY ;
  if ( (StatModelP->Para.Center_KD = GenAlloc( nk * DataP->NbVars, sizeof(double),
					 1, func, "Center_KD" ))== NULL )
               return STS_E_MEMORY ;

  if ( (StatModelP->Para.NbObs_K   = GenAlloc( nk, sizeof(double),
					 1, func, "NbObs_K" ))== NULL )
               return STS_E_MEMORY ;
  if ( (StatModelP->Para.NbObs_KD  = GenAlloc( nk * DataP->NbVars, sizeof(double),
					 1, func, "NbObs_KD" )) == NULL )
               return STS_E_MEMORY;

  if ( (StatModelP->Desc.DispSam_D = GenAlloc( DataP->NbVars, sizeof(double),
					  1, func, "DispSam_D" ))== NULL )
               return STS_E_MEMORY;
  if ( (StatModelP->Desc.MiniSam_D = GenAlloc( DataP->NbVars, sizeof(double),
					  1, func, "MiniSam_D" ))== NULL )
               return STS_E_MEMORY;
  if ( (StatModelP->Desc.MaxiSam_D = GenAlloc( DataP->NbVars, sizeof(double),
					  1, func, "MaxiSam_D" ))== NULL )
               return STS_E_MEMORY;

  /* !!! Allocate the last  model parameters */
  if ( (StatModelP->Spec.ClassFamily == FAMILY_NORMALGEN )
                          && (StatModelP->Spec.ClassDisper == DISPER__D) )
  {
    if ((StatModelP->Para.Disp_KD   = GenAlloc( DataP->NbVars* (DataP->NbVars+1)/2,
                                    sizeof(double), 1, func, "Disp_KD" )) == NULL )
               return STS_E_MEMORY;
    if ((StatModelP->Para.NbObs_KDD = GenAlloc( DataP->NbVars * (DataP->NbVars+1)/2,
                                    sizeof(double),  1, func, "NbObs_KDD" )) == NULL )
               return STS_E_MEMORY;
    if ((StatModelP->Para.Iner_KD   = GenAlloc( DataP->NbVars* (DataP->NbVars+1)/2,
                                    sizeof(double), 1, func, "Iner_KD" ))== NULL )
               return STS_E_MEMORY ;
  }

  else
  {
     if (StatModelP->Spec.ClassFamily == FAMILY_NORMALDIAG )
     {
      if ((StatModelP->Para.Disp_KD   = GenAlloc( nk * DataP->NbVars,
                                    sizeof(double), 1, func, "Disp_KD" ))== NULL )
               return STS_E_MEMORY ;
      if ((StatModelP->Para.NbObs_KDD = GenAlloc( 1,
                                    sizeof(double),  1, func, "NbObs_KDD" ))== NULL )
               return STS_E_MEMORY ;
      if ((StatModelP->Para.Iner_KD   = GenAlloc( nk * DataP->NbVars,
                                    sizeof(double), 1, func, "Iner_KD" ))== NULL )
               return STS_E_MEMORY ;

     }
     else
     {
      if ((StatModelP->Para.Disp_KD   = GenAlloc( nk * DataP->NbVars* (DataP->NbVars+1)/2,
                                    sizeof(double), 1, func, "Disp_KD" ))== NULL )
               return STS_E_MEMORY ;
      if ((StatModelP->Para.NbObs_KDD = GenAlloc( nk * DataP->NbVars * (DataP->NbVars+1)/2,
                                    sizeof(double),  1, func, "NbObs_KDD") )== NULL )
               return STS_E_MEMORY ;
      if ((StatModelP->Para.Iner_KD   = GenAlloc( nk * DataP->NbVars* (DataP->NbVars+1)/2,
                                    sizeof(double), 1, func, "Iner_KD" ))== NULL )
               return STS_E_MEMORY ;
     }
  }


  return err ;

} /* end of AlgoArgs() */



/* ==================== LOCAL FUNCTION DEFINITION =================== */



/* ------------------------------------------------------------------- */
static int  ReadStrFile
            (
                const char*  BaseName,          /* I */
                char*        CommentS,          /* O */
                DataT*       DataP,             /* O */
                ImageNeighT* ImageP,            /* O */
                TypeET*      TypeP              /* O */
            )
/*\
  Read structure file
       expected format :
         type size nbvars
           type : N | I | S
           size : (if type == I) lx ly (if type == N or S) nbpts
           nbvars : integer
\*/
/* ------------------------------------------------------------------- */
{
    char    infname[ LEN_FILENAME + 1 ] ;
    FILE    *fstr ;
    int     nbelts ;
    char    type[100] ;  /* N | I | S */
    int     n1, n2, n3 ;

    static void my_strupr( char *s ) ;

    strncpy( infname, BaseName, LEN_FILENAME ) ;
    strncat( infname, ".str" , LEN_FILENAME ) ;

    if ( ReadOpeningComments( infname , "#" , LEN_LINE ,
                              & fstr , CommentS ) == -1 )
    {
        fprintf( stderr, "File %s does not exist\n", infname ) ;
        return STS_E_FILEIN ;
    }


    /* Now we expect a line having structure : "I nl nc d" or "S n d" or
       "N n d"
    */
    nbelts = fscanf( fstr, "%s %d %d %d", type, &n1, &n2, &n3 ) ;
    fclose( fstr ) ;  /*V1.05-h*/

    if ( nbelts < 3 )
    {
        fprintf( stderr, "Structure file (%s) not enough fields\n",
                 infname ) ;
        return STS_E_FILE ;
    }

    my_strupr( type ) ;

    if ( ! strcmp( type, TypeStrC[ TYPE_NONSPATIAL ] ) )
    {
        /* No spatial information */
        DataP->NbPts  = n1 ;
        DataP->NbVars = n2 ;
        *TypeP        = TYPE_NONSPATIAL ;
    }
    else if ( ! strcmp( type, TypeStrC[ TYPE_SPATIAL ] ) )
    {
        /* Spatial irregular */
        DataP->NbPts  = n1 ;
        DataP->NbVars = n2 ;
        *TypeP        = TYPE_SPATIAL ;
    }
    else if ( ! strcmp( type, TypeStrC[ TYPE_IMAGE ] ) )
    {
        /* Image */
        if ( nbelts < 4 )
        {
            fprintf( stderr, "Structure file (%s) not enough fields\n",
                     infname ) ;
            return STS_E_FILE ;
        }
        ImageP->Nl    = n1 ;
        ImageP->Nc    = n2 ;
        DataP->NbPts  = ImageP->Nl * ImageP->Nc ;
        DataP->NbVars = n3 ;
        *TypeP = TYPE_IMAGE ;
    }
    else
    {
        /* default : Error in structure file */
        fprintf( stderr, "Data type %s unknown in file %s\n",
                 type, infname ) ;
        return STS_E_FILE ;
    }

    return STS_OK ;

}  /* end of ReadStrFile() */




/* ------------------------------------------------------------------- */
static char GetOptionSwitch( const char *Arg )
/* ------------------------------------------------------------------- */
{
    if ( ( Arg[ 0 ] == '-' ) || ( Arg[ 0 ] == '/' ) )
      {
        return Arg[ 1 ] ;
      }
    else
      {
        return '\0' ;
      }
} /* end of GetOptionSwitch() */


/* ------------------------------------------------------------------- */
static int GetEnum( const char* S, const char* SV[], int SizeV ) /*V1.04-b*/
/*\

    Returns the position of string S in the array SV consisting of
    SizeV strings (indiced from 0 to SizeV - 1).  Returns -1 if S is
    not found in SV.

\*/
/* ------------------------------------------------------------------- */
{
  int pos ;
  int found ;

  for ( pos = 0 , found = FALSE ;
	( pos < SizeV ) && ( ! found ) ;
	pos ++ )
    {
      if ( ! strcmp( S , SV[ pos ] ) )
	found = TRUE ;
    }

  if ( found )
    return ( pos - 1 ) ;  /* '- 1' because for() adds 1 after finding */
  else
    return -1 ;
}





/* ------------------------------------------------------------------- */
StatusET GetMixPara        /* STS_OK or STS_E_ARG */
 (
  const FamilyET  Family,     /* I : distribution family */
  const int       K,          /* I : number of classes */
  const int       D,          /* I : number of variables */
  const char*     Opts_Q[],   /* I : array of options */
  const int       Q,          /* I : number of options */
  int*            IoptP,      /* I/O : index of current option */
  ModelParaT*     ParaP       /* O : read parameters */
 )
/* ------------------------------------------------------------------- */
{
  StatusET  sts = STS_OK ;
  int       k ;   /* class counter : 0..K-1 */
  int       d ;   /* variable counter : 0..D-1 */
  double     pK ;  /* remaining proportion for class K */

  int       nbpara = K - 1 + K * D + K * D ;

  if ( (*IoptP) + nbpara >= Q ) {
    fprintf( stderr, "Need at least %d parameters (%d provided)\n",
	     nbpara, Q - (*IoptP) ) ;

    return( STS_E_ARG) ;
  }

  /* Read proportions */
  for ( k = 0, pK = 1 ;
	k < K - 1 ;
	k ++ ) {
    (*IoptP) ++ ;
    ParaP->Prop_K[ k ] = atof( Opts_Q[ (*IoptP) ] ) ;
    if (ParaP->Prop_K[ k ] <= 0 ){
           fprintf( stderr, "init prop must be > 0, (here %s ) \n",
	     Opts_Q[ (*IoptP)  ] ) ;

    return( STS_E_ARG) ;
  }

    pK = pK - ParaP->Prop_K[ k ] ;
  }
  ParaP->Prop_K[ K - 1 ] = pK ;
  if ( pK <= 0.0 ) {
    fprintf( stderr, "Last class has pK = %5.2f <= 0\n", pK ) ;
    return STS_E_ARG ;
  }


  /* Read centers */
  for ( k = 0 ; k < K ; k ++ ) {

    for ( d = 0 ; d < D ; d ++ ) {

      (*IoptP) ++ ;
      ParaP->Center_KD[ k * D + d ] = atof( Opts_Q[ (*IoptP) ] ) ;
    }
  }

  /* Read dispersions */
  for ( k = 0 ; k < K ; k ++ ) {

    for ( d = 0 ; d < D ; d ++ ) {

      (*IoptP) ++ ;
      if ( (Family == FAMILY_NORMALGEN) || (Family == FAMILY_NORMALDIAG)) {
	double x = atof( Opts_Q[ (*IoptP) ] ) ;
	ParaP->Disp_KD[k*D*(D+1)/2+d*(2*D+1-d)/2 ] = x * x ;
      }
      else
	ParaP->Disp_KD[k*D*(D+1)/2+d*(2*D+1-d)/2 ] = atof( Opts_Q[ (*IoptP) ] ) ;

      if ( ParaP->Disp_KD[k*D*(D+1)/2+d*(2*D+1-d)/2] <= 0 ) {

	fprintf( stderr, "Dispersion(k=%d, d=%d) = %5.3f <= 0\n",
		 k+1, d+1, ParaP->Disp_KD[k*D*(D+1)/2+d*(2*D+1-d)/2 ] ) ;
	sts = STS_E_ARG ;
      }
    }
  }

  return sts ;

}   /* end of GetMixPara() */

/* ------------------------------------------------------------------- */
StatusET GetSaveModalities (
			    const char*      Option,
			    AlgoParaT* AlgoParaP
)
/* ------------------------------------------------------------------- */
{
  if (Option[0] == 's')
    {
      AlgoParaP->LoopParam.SavePotts = (Option[1] == 'm');
      AlgoParaP->LoopParam.SaveNoise = (Option[2] == 'b');
      AlgoParaP->LoopParam.SaveRest  = (Option[3] == 'r');
      return STS_OK;
    }
  else
    {
      fprintf(stderr,"Bad syntax in setting the saving options.\n");
      return STS_E_ARG;
    }
}






/*V1.06-a*/
/* ------------------------------------------------------------------- */
static void my_strupr( char *s )
/* ------------------------------------------------------------------- */
{
    if ( s == NULL ) return ;
    for ( ; (*s) != '\0' ; s ++ )
    {
        if ( ( (*s) > 'a' ) && ( (*s) < 'z' ) )
        {
            (*s) = (*s) + ( 'A' - 'a' ) ;
        }
    }
}


/* ======================================================================= */
