/*\

    SEMMS_EXE.C

    Spatial EM for Markovian Segmentation project :
    main program and input/output functions.

    From June 96 to October 2001

    Van Mo DANG
    Universite de Technologie de Compiegne
    URA CNRS 817

    Peyrard Nathalie
    Unit de Biomtrie - INRA Avignon

    Forbes Florence
    INRIA Rhne-Alpes

    
\*/

/*\
    SEMMS_EXE.C

Vers-mod  Date         Who  Description

1.03-a    03-NOV-1996  MD   Add prototype of ReadLabelFile() in main()
1.03-b    03-NOV-1996  MD   Save Pk, Vk, and Ck in don.mf
1.03-c    22-AUG-1997  MD   Unknown labels : memberships set to 0
1.03-d    22-AUG-1997  MD   Save criteria to don.mf file
1.03-e    30-SEP-1997  MD   Save criterion L to don.mf file
1.03-f    30-SEP-1997  MD   Save algorithm parameters into don.mf file
1.04-a    04-OCT-1997  MD   Any name for initial partition file
1.04-b    09-OCT-1997  MD   If "-o -" classification saved to standard output
1.04-c    09-OCT-1997  MD   If "-s f -" initial classification from stdin
1.04-d    10-OCT-1997  MD   Random seed set by option
1.04-e    10-OCT-1997  MD   Visit order ("-O random")
1.04-f    13-OCT-1997  MD   Read reference classification
1.04-g    05-NOV-1997  MD   Call to srand48() replaced by srandom()
1.04-h    02-DEC-1997  MD   Add "GEM" to AlgoDesC
1.05-a    12-JAN-1998  MD   Count missing data
1.05-b    17-JAN-1998  MD   Program exit values changed
1.05-c    25-JAN-1998  MD   main() now mainfunc()
1.05-d    26-JAN-1998  MD   DataP->LabelV init to NULL by default
1.05-e    26-JAN-1998  MD   Free allocated data after processing
1.05-f    26-JAN-1998  MD   GenAlloc/GenFree instead of malloc/calloc/free
1.05-g    30-JAN-1998  MD   mainfunc() prototyped in mainfunc.h
1.05-h    30-JAN-1998  MD   Fixed bug : str file was opened and not closed
1.05-i    05-FEB-1998  MD   my_strupr returns void
1.05-j    05-FEB-1998  MD   save M (Markov pseudo-likelihood) into don.mf
1.06-a    17-JUN-1998  MD   use new StatModel structure (old NoisModel)
1.06-b    23-JUN-1998  MD   LEN_LINE -> nem_typ.h
1.06-c    20-SEP-1998  MD   change format of ref / label file (!!!)
1.06-d    20-SEP-1998  MD   struct ErrInfo added to CriterT
1.06-e    21-SEP-1998  MD   TieRule copied in Errinfo
          15-JUIL-2000 NP   Modification of ReadLabelFile:
	                    the non-labeled pixels are set
                            according to an initial segmentation
          01-AUG-2000  NP   Add the argument AlgoParaP.Opt. 2 values :
	                    EstimParam, Model Selection
	  06-APR-2001  NP   Add the type FORMAT_UF for the output segmentation
	                    Available in SaveMSResults but not SaveResults.
          12-JUL-2001  FC   SaveMSResults enhanced with the values of the
	                    potentials on singletons written in the .mf file
          17-JUL-2001  NP   add the functions GaussNoise and AddGaussNoise
	  20-JUL-2001  FC   modify GetInputPara so as not to load .dat and
	                    .nei file and to initialize correctly DataP
	  20-JUL-2001  FC   add the GibbsSampling function to create
	                    representations of Potts models (with of course,
			    potentials on singletons)
          01-AUG-2001  NP   add the selection of K in the mode LOOP
	  06-AUG-2001  FC   Loop mode is to be finished. Bugs fixed by Nathalie
	                    and some improvements needed
	  16-AUG-2001  FC   Loop cleared with some functions
	                    to modify StatParam
	  16-AUG-2001  FC   ErrorRate for each class took into account
	  16-AUG-2001  FC   Prop for each class when Loop or true picture
	                    took into account
	  18-MAY-2004  BA   in GaussNoise() possibility to have
                            1.different mean than 0,1,2, now
                            1.mean(K)!=K. 
          28-JAN-2005  NP   replace 'nem' by 'semms'
 
\*/

#include "mainfunc.h"   /* Prototype of exported mainfunc() */


#include "semms_typ.h"    /* DataT, ... */
#include "semms_arg.h"    /* SemmsArg, ... */
#include "semms_alg.h"    /* ClassifyByAlgo, ... */
#include "semms_rnd.h"    /* RandomInteger, ... */
#include "lib_io.h"     /* ReadOpeningComments, ... */
#include "genmemo.h"    /* GenAlloc, ... */
#include <stdlib.h>

#include <stdio.h>      /* printf, ... */
#ifdef __TURBOC__
#include <alloc.h>      /* coreleft, ... */
#endif
#include <string.h>     /* strncpy, ... */
#include <math.h>       /* sqrt, ... */
#include <time.h>



static const char *TypeDesC[ TYPE_NB ] = { "Spatial", "Image", "NoSpatial" } ;
static const char *AlgoDesC[ ALGO_NB ] = { "MEANF",  "GSF", "ICM",
                                           "NEM" , "NCEM (C-step)"  } ;
static const char *FamilyDesVC[ FAMILY_NB ] = { "NormalGen", "NormalDiag", "Laplace" } ;
static const char *DisperDesVC[ DISPER_NB ] = { "S__", "SK_", "S_D", "S_KD" } ;
static const char *ProporDesVC[ PROPOR_NB ] = { "P_", "Pk" } ;
/* static const char *InitDesC[ INIT_NB ] = { "Sort_variables" , */
/* 					   "Random_centers" , */
/* 					   "Mixture_initial" , */
/* 					   "Mixture_fixed" , */
/* 					   "Full_labels" , */
/* 					   "Partial_labels" } ;    /\*V1.03-f*\/ */

/* static const char *OrderDesC[ ORDER_NB ] = { "Direct_order" , "Random_order" }; */




/* ==================== LOCAL FUNCTION PROTOTYPING =================== */
static int GibbsSampling (
 AlgoParaT*   AlgoParaP,
 StatModelT*  StatModelP,
 SpatialT*    SpatialP,
 DataT*       DataP
);

static int AddGaussNoise (
			  int            Nk,
			  AlgoParaT*     AlgoParaP,
			  DataT*         DataP
			  );


static int GaussNoise (
		       double*              NoisedData,
		       const int            Npt,
		       const int            Nk,
		       const double*           Unique,
		       const double*        Stdev
		       );


  static  int ModelSelection
        (
	  const int           Argc,
	  const char*         Argv[],
          AlgoParaT           *AlgoParaP,        /* I */
          const SpatialT      *SpatialP,        /* I */
          DataT               *DataP,           /* I/O */
          StatModelT          *StatModelP,      /* I/O */
          double               *ClassifM,
          WorkingT            *WorkingP,
          MSCritT             *MSCritP,
          VarParamT           *VarParamP,
          double               *Increase,
	  CriterT             *Criteria,          /* O */  /*V1.03-f*/
          double               *EBlist
	 );


/* Called by ClassifyByAlgo */

    static int GetInputPara
        (
          int           Argc,
          const char    *Argv[],
          DataT         *DataP,         /* O and allocated */
          AlgoParaT     *AlgoParaP,      /* O */
          SpatialT      *SpatialP,      /* O and allocated */
          StatModelT    *StatModelP,    /* O and allocated */
	  ErrinfoT      *ErrinfoP,      /* O and allocated */
	  ErrcurT       *ErrcurP,       /* O and allocated */
          double         **ClassifMP,     /* O and allocated */
          MSCritT       *MSCritP,         /* O and allocated */
          double        **VarCenter_KD,    /* O and allocated */
          double        **VarDisp_KD,       /* O and allocated */
          double         **Increase,       /* O and allocated */
          double         **EBlist
        ) ;



   static int SaveMSResults
    (
         const AlgoParaT*     AlgoParaP,
         const ModelParaT*    ModelParaP,
         const DataT*         DataP,
         const SpatialT*      SpatialP,
         const StatModelT*    ModelP,
         double*              ClassifM,
         double*              LogFikM,
         const int            Npt,
         const int            K,
         const MSCritT*       MSCritP,
         const CriterT*       CriterP,
         FILE*                fmf,
         FILE*                fout
    ) ;


   static int SaveClassif
    (
        const AlgoParaT*     AlgoParaP,
        const DataT*         DataP,
        const SpatialT*      SpatialP,
        const StatModelT*    ModelP,
        double*              ClassifM,
        const int            Npt,
        const int            K,
        FILE*                fout
	) ;

static int      AllocKvar
(
 AlgoParaT*   AlgoParaP,
 StatModelT*  StatModelP,
 VarParamT*   VarParamP,
 const int    K,
 const int    NbVars,
 const int    NbPts
);


static void FreeAllocatedData
(
 DataT*       DataP,       /* O and deallocated */
 SpatialT*    SpatialP,    /* O and deallocated */
 StatModelT*  StatModelP,  /* O and deallocated */  /*V1.06-a*/
 AlgoParaT*   AlgoParaP,   /* DataTrue: O and deallocated */
 VarParamT*   VarParamP,   /* O and deallocated */
 CriterT*     CriterP,     /* O and deallocated */  /*V1.06-d*/
 double*       ClassifM,     /* O and deallocated */
 double*       Increase,     /* O and deallocated */
 double*       EBlist     /* O and deallocated */
) ;


int FreeStatParam
(
   StatParamT* StatParam
 );

/* Called by GetInputPara */


    static int SetVisitOrder  /*V1.04-e*/
        (
	     int         Npt,          /* I */
	     OrderET     VisitOrder,   /* I */
	     int**       SiteVisitVP   /* O and allocated (Npt) */
	) ;

    static int  ReadMatrixFile
         (
             const char  *FileName,     /* I */      /*V1.04-a*/
             int         Nl,            /* I */
             int         Nc,            /* I */
             double       **MatP         /* O and allocated (size : Nl * Nc) */
         ) ;



    static int  ReadLabelFile  /*V1.03-a*/
         (
             const char  *LabelName,    /* I */
             int         Npt,           /* I */
	     int         *KfileP,       /* O : file # classes */ /*V1.06-c*/
             int         **LabelVP,     /* O and allocated (Npt) */
             double       **ClassifMP    /* O and allocated (Npt*Nk) */
         );

    static int  MakeErrinfo
         (
             const char* RefName,       /* I : filename of reference class */
             int         N,             /* I : number of objects */
             int         Kc,            /* I : user number of classes */
	     TieET       TieRule,       /* I : specified MAP tie rule */
	     ErrinfoT*   ErrinfoP,      /* O and allocated */
	     ErrcurT*    ErrcurP        /* O and allocated */
         ) ;

    static int  ReadNeiFile
         (
             const char  *BaseName,     /* I */
             int         NbPts,         /* I */
             NeighET     NeighSpec,     /* I */
             char*       NeiDescS,      /* O [LEN_LINE+1] */
             SpatialT    *SpatialP      /* O and allocated */
         ) ;



/* Called by ReadNeiFile */

    static int  ReadPtsNeighs
                (
                    FILE        *Fnei,          /* I/O */
                    int         NbPts,          /* I */
                    int         *MaxNeiP,       /* O */
                    NeighDataT  *NeighDataP     /* O and allocated */
                ) ;

    static int  ReadImageNeigh
                (
                    FILE        *Fnei,          /* I/O */
                    NeighDataT  *NeighDataP     /* O and allocated */
                ) ;

    static int  SetImageNeigh
                (
                    NeighET     NeighSpec,          /* I */
                    char*       NeiDescS,           /* O [LEN_LINE+1] */
                    NeighDataT* NeighDataP          /* O and allocated */
                ) ;


/* Called by MakeErrinfo */

static int factorial(int n);

static int compute_permutations    /* ret 0 if OK, -1 if memory error */
(
 const int Startval,         /* I : start value of integer suite */
 const int K,                /* I : size of integer suite to permute > 0 */
 int*      perms_Kfact_K_p[] /* O : matrix to store permuted values */
) ;


/* Called by compute_permutations */

static int rec_permutations        /* ret 0 if OK, -1 if memory error */
(
 const int array_A[],       /* I : remaining array to permute */
 const int A,               /* I : length of the remaining array : 0..K */
 const int K,               /* I : length of original array */
 int       offset,          /* I : first line of storage to use */
 int       perms_Kfact_K[]  /* O : matrix to store permuted values, use
			       lines :   offset -> offset + A! - 1
			       columns : K - A  -> K - 1 */
) ;


/* Called by */

int CreateName (
		char* FinalName,
		char* BaseName,
		int   Number,
		char* Ext
		);

int ClearParams (
		 double* Singl_SaveP,
		 double* Beta_SaveP,
		 StatModelT* StatModelP
		 );

int AllocStatParam(
		   int i,
		   StatParamT* StatParam
		   );

int InitializeStatParam(
			int i,
			StatParamT* StatParam
			);

int FinalStatParams(
		    int i,
		    int nbIters,
		    StatParamT* StatParam
		    );

int SaveStatParams(
		   int NbClasses,
                   int Kmin,
                   int Kmax,
		   FILE* fsf,
		   MSparamT*  MSparamP,
                   AlgoET     Algo,
		   SinglET    SinglMode
		   );

/*int SaveStatParams(
		   int NbClasses,
		   char* BaseName,
		   StatParamT* StatParamP
		   );
*/

/* ==================== GLOBAL FUNCTION DEFINITION =================== */


/* ------------------------------------------------------------------- */
int mainfunc( int argc, const char * * argv )
/*\
  Main program for
    - estimating parameters and get segmented image
    - or adding a gaussien noise to an image
    - or simulating a potts model with potentials on singletons
    - or model selection
\*/
/* ------------------------------------------------------------------- */
{
  StatusET                err ;
  static  DataT           Data = {0} ;
  static  AlgoParaT       AlgoPara = {0} ;
  static  SpatialT        Spatial = {{{0}}} ;
  static  StatModelT      StatModel = {{0}} ;
  static  double          *ClassifM = 0 ;
  static  double          *EBlist = 0 ;
  static  MSCritT         MSCritP = {0} ;
  static  VarParamT       VarParamP = {0} ;
  static  WorkingT        WorkingP  = {NULL};
  //static  StatParamT      StatParam = {0};
  static  StatParamT*     StatParam;
  static  double          *Increase = 0;
  CriterT                 Criteria = {0} ; /*V1.03-d*/
  int i,j;                /* loops, loops, loops... */

  /* main program algorithm :
     - read all necessary data into memory
     - compute  param estimation and partition by Algo method
       or do what must be done
     - save results to file
  */

  //InitListMem();

  fprintf( stderr , " * * * SEMMS (Spatial EM for Markov Segmentation)  * * *\n");

#ifdef __TURBOC__
  fprintf( stderr, "\n Initial free memory : %lu bytes\n\n",
             (unsigned long) coreleft());
#endif

  /* RandomSeedByTime( ) ;  V1.04-d*/

  if ((err = GetInputPara (argc, argv, &Data, &AlgoPara, &Spatial,
			   &StatModel, &Criteria.Errinfo, &Criteria.
			   Errcur, &ClassifM,
			   &MSCritP, &(VarParamP.VarCenter_KD),
			   &(VarParamP.VarStd_KD),
			   &Increase, &EBlist))
      == STS_OK) {
#ifdef __TURBOC__
    srand((unsigned) AlgoPara.Seed);
#else
    srandom(AlgoPara.Seed);  /*V1.04-g*/
#endif

    if (AlgoPara.Mode == NORMAL)
      switch(AlgoPara.Opt)
	{
	case ModelSelec : /* run the algo for differents values
			     of K and/or beta */
	  ModelSelection(argc, argv,&AlgoPara, &Spatial, &Data,
			 &StatModel, ClassifM, &WorkingP,
			 &MSCritP,&VarParamP,Increase,
			 &Criteria, EBlist );
	  break;

	case GibbsSampler :
	  /* Create a representation of a Potts model with values transmitted
	     through parameters */
	  GibbsSampling(&AlgoPara, &StatModel, &Spatial, &Data);
	  break;

        case AddGaussianNoise : /* Add gaussian noise N(0,sigma^2) to original values */
	  AddGaussNoise(StatModel.Spec.K,&AlgoPara,&Data);
	  break;

        default :
	  return STS_E_FUNCARG;
	}
    else
      {
	int nk = StatModel.Spec.K;
	/* For saving simulation parameters */
	double  oldBeta;
	double* oldSingl = GenAlloc(nk,sizeof(double),
				    0,"mainfunc","oldSingl");
	int nbIters = AlgoPara.LoopParam.NbIters;

	AlgoPara.MSparam.BICw = GenAlloc(
	        (AlgoPara.MSparam.Kmax - AlgoPara.MSparam.Kmin + 1),
		sizeof(double), 0,"mainfunc","BICw");
	AlgoPara.MSparam.BICp = GenAlloc(
                (AlgoPara.MSparam.Kmax - AlgoPara.MSparam.Kmin + 1),
		sizeof(double), 0,"mainfunc","BICp");
	AlgoPara.MSparam.ICLw = GenAlloc(
                (AlgoPara.MSparam.Kmax - AlgoPara.MSparam.Kmin + 1),
		sizeof(double), 0,"mainfunc","ICLw");
	AlgoPara.MSparam.ICLp = GenAlloc(
		(AlgoPara.MSparam.Kmax - AlgoPara.MSparam.Kmin + 1),
		sizeof(double), 0,"mainfunc","ICLp");

	/* Allocate the StatParam structure */
	AlgoPara.MSparam.StatParam = (StatParamT*)
	  GenAlloc(AlgoPara.MSparam.Kmax - AlgoPara.MSparam.Kmin + 1,
		   sizeof(StatParamT),0,"mainfunc","StatParam");
	for (i = AlgoPara.MSparam.Kmin ; i <= AlgoPara.MSparam.Kmax ; i++)
	  {
	    StatParam = &(AlgoPara.MSparam.StatParam[i -
						    AlgoPara.MSparam.Kmin]);
	    AllocStatParam(i,StatParam);

	  }

	/* Initialise the StatParam structure */
	for (i = 0;  i <= AlgoPara.MSparam.Kmax - AlgoPara.MSparam.Kmin; i++)
	  {
	    StatParam = &(AlgoPara.MSparam.StatParam[i]);
	    InitializeStatParam(i + AlgoPara.MSparam.Kmin, StatParam);
	  }
	/* End of initialisation */

	//int Errors[nbIters];

	/* Important : for saving pictures or not */
	if (AlgoPara.LoopParam.SaveType == LAST)
	  {AlgoPara.LoopParam.SaveNow = 0;}

	for (i = 0; i < nbIters; i++)
	  {
	    if ((AlgoPara.LoopParam.SaveType == LAST)
		&& (i == nbIters - 1))
	      AlgoPara.LoopParam.SaveNow = 1;
            fprintf(stderr, "\n\n ITER = %d\n", i);
	    CreateName(AlgoPara.OutName,AlgoPara.OutBaseName,
		       i,EXT_OUTNAMEPOTTS);
	    GibbsSampling(&AlgoPara, &StatModel, &Spatial, &Data);
	    /* Copy the datas in DataTrue for comparison */
	    memcpy(AlgoPara.True.DataTrue,Data.PointsM,
		   Data.NbPts * Data.NbVars* sizeof( double ));
	    /* End of copy */
	    CreateName(AlgoPara.OutName,AlgoPara.OutBaseName,
		       i,EXT_OUTNAMENOISE);
	    AddGaussNoise(nk,&AlgoPara, &Data);
	    CreateName(AlgoPara.OutName,AlgoPara.OutBaseName,
		       i,EXT_OUTNAMEHARD);
	    ClearParams(oldSingl,&oldBeta,&StatModel);
	    ModelSelection(argc, argv,&AlgoPara, &Spatial, &Data,
			   &StatModel, ClassifM, &WorkingP,
			   &MSCritP,&VarParamP,Increase,
			   &Criteria, EBlist ); 

           
           
	    /* The StatParam structure is updated in Model Selection */
	    StatModel.Para.Beta = oldBeta;
	    for (j = 0; j < nk; j++)
	      StatModel.Para.Singl[j] = oldSingl[j];
	  }

	/* Giving good final values to StatParam */
	for (i = AlgoPara.MSparam.Kmin ; i <= AlgoPara.MSparam.Kmax; i++)
	  {
	    StatParam = &(AlgoPara.MSparam.StatParam[i -
						    AlgoPara.MSparam.Kmin]);
	    FinalStatParams(i, nbIters, StatParam);
	  }

        for ( i = 0 ;
              i <( AlgoPara.MSparam.Kmax - AlgoPara.MSparam.Kmin +1);
              i ++ )
          {
	    switch (AlgoPara.Algo)
	      {
	      case ALGO_MEANF:
	      case ALGO_GSF:
	      case ALGO_ICM:
		AlgoPara.MSparam.BICw[i] = AlgoPara.MSparam.BICw[i]*
		  100 / nbIters;
		AlgoPara.MSparam.BICp[i] = AlgoPara.MSparam.BICp[i]*
		  100 / nbIters;
		AlgoPara.MSparam.ICLw[i] = AlgoPara.MSparam.ICLw[i]*
		  100 / nbIters;
		AlgoPara.MSparam.ICLp[i] = AlgoPara.MSparam.ICLp[i]*
		  100 / nbIters;
		break ;

	      case ALGO_NEM:
	      case ALGO_NCEM:
		AlgoPara.MSparam.BICp[i] = AlgoPara.MSparam.BICp[i]*
		  100 / nbIters;
		AlgoPara.MSparam.ICLp[i] = AlgoPara.MSparam.ICLp[i]*
		  100 / nbIters;
		break ;

	      default:
		return -1;
	      }
	  }
 

	/* Saving the StatParams in a .sf file */
	{
	  FILE* SFFile;
	  char  RealName[LEN_FILENAME];
	  /* Opens the file */
	  strcpy(RealName,AlgoPara.OutBaseName);
	  strcat(RealName,EXT_OUTNAMESTATS);
	  if ((SFFile = fopen(RealName, "w")) == NULL)
	    {
	      fprintf( stderr, "Could not open file '%s' in write mode\n",
		       RealName);
	      return STS_E_FILEOUT;
	    }
	  SaveStatParams(nk,AlgoPara.MSparam.Kmin ,
			 AlgoPara.MSparam.Kmax,SFFile,&(AlgoPara.MSparam),
			 AlgoPara.Algo, AlgoPara.SinglMode); 
	  /* Close the file */
	  fclose(SFFile);
	}
        /* free param of case LOOP */
        /*GenFree(&(AlgoPara.MSparam));*/ /* should be here, but leads to a bug */
	GenFree(oldSingl) ;


      } 
  }


  FreeAllocatedData( &Data, &Spatial, &StatModel, &AlgoPara,
		     &VarParamP, &Criteria, ClassifM,
		     Increase,EBlist ) ;


    //  DisplayListMem();
  switch( err )
    {
    case STS_OK :
      if (AlgoPara.Mode == NORMAL)
	switch(AlgoPara.Opt)
	  {
	  case ModelSelec :
	    if ( strcmp( AlgoPara.OutBaseName , "-" ) != 0 )  /*V1.04-b*/
	      {
		fprintf( stderr, "Algo completed, classification(s) in %s\n",
			 AlgoPara.OutName ) ;
		fprintf( stderr, " criteria and parameters in %s%s\n",
			 AlgoPara.OutBaseName, EXT_MFNAME ) ;

		
	      }
	    else
	      {
		fprintf( stderr, "Algo completed, classification to screen\n");
	      }
	    return EXIT_OK ;
	    break ;

	  case GibbsSampler :
	    if (strcmp(AlgoPara.OutBaseName,"-") != 0) {
	      fprintf(stderr,"Algo completed, potts image in %s\n",
		      AlgoPara.OutName);
	    }
	    return EXIT_OK;
	    break;

	  case AddGaussianNoise :
	    if ( strcmp( AlgoPara.OutBaseName , "-" ) != 0 )  /*V1.04-b*/
	      {
		fprintf( stderr, "Algo completed, noised data in %s\n",
			 AlgoPara.OutName ) ;
	      }
	    return EXIT_OK ;
	    break ;

	  default :
	    return EXIT_OK ;
	  }
      else
	{
	  /* fprintf(stderr, "End of loop successfull.\n"); */
	  return EXIT_OK ;
	}

    case STS_W_EMPTYCLASS :
      fprintf( stderr, "*** SEMMS warning status : empty class\n" );
      return EXIT_W_RESULT ;

    case STS_I_DONE :
      return EXIT_OK ;

    case STS_E_ARG :
      return EXIT_E_ARGS ;

    case STS_E_MEMORY :
      fprintf( stderr, "*** SEMMS error status : not enough memory\n" );
#ifdef __TURBOC__
      fprintf( stderr, "\n Memory left : %lu bytes\n",
	       (unsigned long) coreleft() );
#endif
      return EXIT_E_MEMORY ;

    case STS_E_FILEIN :
      fprintf( stderr, "*** SEMMS error status : can't open a file\n" );
      return EXIT_E_FILE ;

    case STS_E_FILE :
      fprintf( stderr, "*** SEMMS error status : wrong file format\n" );
      return EXIT_E_FILE ;

    case STS_E_FUNCARG :
      fprintf( stderr, "*** SEMMS internal error : bad arguments\n" );
      return EXIT_E_BUG ;

    default :
      fprintf( stderr, "*** SEMMS unknown error status (%d)\n", err );
      return EXIT_E_BUG ;
    }
} /* end of mainfunc() */



/* ==================== LOCAL FUNCTION DEFINITION =================== */

/* ------------------------------------------------------------------- */
static int GibbsSampling (
			  AlgoParaT*   AlgoParaP,
			  StatModelT*  StatModelP,
			  SpatialT*    SpatialP,
			  DataT*       DataP
)
/* ------------------------------------------------------------------- */
{
  /* A few declarations */
  int         i,j;
  int         err = STS_OK;
  int         Npt = DataP->NbPts ;
  FILE*       fout;
  int         ipt;



  /* Creating a representation of a Potts model with parameters transmitted */
  if ((DataP->PointsM  = GenAlloc(Npt, sizeof(double), 0, "GibbsSampling",
				  "PointsM"))
      == NULL )
    return STS_E_MEMORY ;

  /* fprintf(stderr,"Calling CreateGibbs.\n"); */
/*   fprintf(stderr,"Value of Beta : %f\n", StatModelP->Para.Beta); */
  CreateGibbs(DataP,AlgoParaP,StatModelP,SpatialP);

  /* Compute the proportions of the different classes */
  for (i = 0; i < AlgoParaP->True.K; i++)
    AlgoParaP->True.Prop[i] = 0;
  for (ipt = 0; ipt < Npt; ipt++)
    AlgoParaP->True.Prop[(int) (DataP->PointsM[ipt])]++;
  for (i = 0; i < AlgoParaP->True.K; i++)
    AlgoParaP->True.Prop[i] =  AlgoParaP->True.Prop[i] / Npt;
  /* Update the statistics about it
     if, of course, we are in loop mode ...
     Otherwise, write in stdout the proportions between the different
     classes ...
  */
  if (AlgoParaP->Mode == LOOP)
    for (i = AlgoParaP->MSparam.Kmin ; i <= AlgoParaP->MSparam.Kmax ; i++)
      {
	StatParamT* StatParam =
	  &(AlgoParaP->MSparam.StatParam[i - AlgoParaP->MSparam.Kmin]);
	for (j = 0; j < AlgoParaP->True.K; j++)
	  {
	    StatParam->MeanProp [j] += AlgoParaP->True.Prop[j];
	    StatParam->VarProp  [j] += pow(AlgoParaP->True.Prop[j],2);
	  }
	 // exit(1);
      }
  else
    for(j = 0; j < AlgoParaP->True.K; j++)
      printf("Proportion on class %d : %f\n",j,AlgoParaP->True.Prop[j]);
  /* Saving the representation */
  if ((AlgoParaP->Mode == NORMAL) ||
      ((AlgoParaP->Mode == LOOP) && (AlgoParaP->LoopParam.SavePotts)))
    if ((AlgoParaP->LoopParam.SaveType == ALL) ||
	((AlgoParaP->LoopParam.SaveType == LAST) &&
	 (AlgoParaP->LoopParam.SaveNow)))
      {
	if (strcmp(AlgoParaP->OutBaseName,"-") != 0)  /*V1.04-b*/
	  {
	    if ((fout = fopen(AlgoParaP->OutName,"w")) == NULL)
	      {
		fprintf(stderr, "Could not open file '%s' in write mode\n",
			AlgoParaP->OutName);
		return STS_E_FILEOUT ;
	      }
	  }
	else
	  fout = stdout;

	for (ipt = 0; (ipt < Npt) && (err == STS_OK) ; ipt ++)
	  {
	    if (ipt % SpatialP->NeighData.Image.Nc == 0)
	      fprintf(fout, "\n");
	    if (fprintf(fout, " %2.0f ",DataP->PointsM[ipt]) == EOF)
	      {
		fprintf(stderr, "Cannot write yi, i = %d, \n",
			ipt + 1);
		err = STS_E_FILEOUT;
	      }
	  } /* end for ( ipt ... ) */

	if (fout != stdout)
	  fclose(fout);
      }

  return err;
} /* End of GibbsSampling */

/* ------------------------------------------------------------------- */
static int AddGaussNoise
(
 int            Nk,
 AlgoParaT*     AlgoParaP ,
 DataT*         DataP

)
/* ------------------------------------------------------------------- */
{
  int         err = STS_OK ;
  FILE*       fout ;
  int         ipt  ;
  int         Npt = DataP->NbPts ;




    /*if ( ( err = ReadMatrixFile(AlgoParaP->TrueName ,
                                 DataP->NbPts,
                                 1,
                                 & AlgoParaP->True.DataTrue ) ) != STS_OK )
				 return err ; */


 /* adding gaussian noise to the original values */


 err = GaussNoise(DataP->PointsM, Npt, Nk,
		  AlgoParaP->GaussNoise.unique, AlgoParaP->GaussNoise.stdev);

 /*saving the noised values */
 /*fprintf(stderr, " saving noised values in  %s\n",
   AlgoParaP.OutName );*/
 if ((AlgoParaP->Mode == NORMAL) ||
     ((AlgoParaP->Mode == LOOP) && (AlgoParaP->LoopParam.SaveNoise)))
   if ((AlgoParaP->LoopParam.SaveType == ALL)
       || ((AlgoParaP->LoopParam.SaveType == LAST) &&
	   (AlgoParaP->LoopParam.SaveNow)))
     {
       if ( strcmp( AlgoParaP->OutBaseName, "-" ) != 0 )  /*V1.04-b*/
	 {
	   if ( ( fout = fopen( AlgoParaP->OutName, "w" ) ) == NULL )
	     {
	       fprintf( stderr, "Could not open file '%s' in write mode\n",
			AlgoParaP->OutName ) ;
	       return STS_E_FILEOUT ;
	     }
	 }
       else
	 fout = stdout ;

       for ( ipt = 0 ; ( ipt < Npt ) && ( err == STS_OK ) ; ipt ++ )
	 {
	   if ( fprintf( fout, " %5.3f ",
			 DataP->PointsM[ ipt ] ) == EOF )
	     {
	       fprintf( stderr, "Cannot write yi i = %d, \n",
			ipt + 1 ) ;

	       err = STS_E_FILEOUT ;
	     }

	 } /* end for ( ipt ... ) */

       if ( fout != stdout )
	 fclose( fout ) ;

     }
 return err ;

} /* end of AddGaussNoise */


/* ------------------------------------------------------------------- */
static int GaussNoise
(
   double*              NoisedData, /* original values then noised values */
   const int            Npt,
   const int            Nk,
   const double*        Unique, /*  vector of the labels in Simul */
   const double*        Stdev  /*  vector of the stdevs of the N(0,sigma^2) noise added to
				    the original values according to their label */

)
/* ------------------------------------------------------------------- */
{
  int    ipt ;
  double g = 0 ;
  int    err = STS_OK ;
  int    ik ;

  for (ipt = 0 ; ipt < Npt ; ipt ++)
    {
      g = normalpol();
      for (ik = 0 ; ik < Nk ; ik ++ )
		if ( NoisedData[ipt] == ik)	
		 {
		 	NoisedData[ipt] = Unique[ik];
		 	g = g*Stdev[ik];
			ik = Nk;
		 }
      NoisedData[ipt] = NoisedData[ipt] + g ; /* realisation of N(NoisedData[ipt], Stdev[k])
                                            where k is the label of ipt */
    }

       return err ;

} /* end of AddGaussNoise */


/* ------------------------------------------------------------------- */
static int ModelSelection
(
	  const int           Argc,
	  const char*         Argv[],
          AlgoParaT           *AlgoParaP,        /* I */
          const SpatialT      *SpatialP,        /* I */
          DataT               *DataP,           /* I/O */
          StatModelT          *StatModelP,      /* I/O */
          double              *ClassifM,
          WorkingT            *WorkingP,
          MSCritT             *MSCritP,
          VarParamT           *VarParamP,
          double              *Increase,
	  CriterT             *Criteria,          /* O */  /*V1.03-f*/
          double              *EBlist
)
/* ------------------------------------------------------------------- */
{
 int   sts = STS_OK ;
 int        Nl =  DataP->NbPts;
 int        Kmin = AlgoParaP->MSparam.Kmin ;
 int        Kmax = AlgoParaP->MSparam.Kmax ;
 int        k,l,j, ik;
 double     bmin = AlgoParaP->MSparam.bmin ;
 double     bmax = AlgoParaP->MSparam.bmax ;
 double     bstep = AlgoParaP->MSparam.bstep ;
 ParamET    ParamMode = AlgoParaP->ParamMode;
 AlgoET     Algo = AlgoParaP->Algo;
 CvemET     CvTest =AlgoParaP->CvTest;
 double     CvThres =AlgoParaP->CvThres;
 BetaET     BetaModel = StatModelP->Spec.BetaModel;
 double     oldBICp = -EPSILON_INV  ;
 double     oldBICw = -EPSILON_INV;
 double     oldICLp = -EPSILON_INV;
 double     oldICLw = -EPSILON_INV;
 int        selecKBICw = 0 ;
 int        selecKBICp = 0 ;
 int        selecKICLw = 0 ;
 int        selecKICLp = 0 ;
 double     b;
 double     btest = -1;
 double*    ClassifInit = 0 ;
 double*    Hclassif = 0;
 int*       LastCPUtime = 0;
 WorkingT   ICMwork ;
 FILE*      fp = NULL; /* initial segmentation or initial params */
 FILE*      fs = NULL; /* file toto.singl */
 FILE*      fmf; /* file toto.mf */
 FILE*      fout; /* file toto.cf */
 char       mfname[ LEN_FILENAME ] ;
 int        il, i, ic ;
 float      m = 0.0;
 float      v = 0.0 ;
 float      s = 0.0 ;
 int         nbscan = 0;
 float       x, beta ;
 int         iarg ;          /* arg counter   : 0..Argc-1 */
 StatParamT* StatParam = {0};
 int         oldspeck;



   /* open the file with initial segmentations */
   if (AlgoParaP->InitMode == INIT_FILE)
	{
     if ( strcmp( AlgoParaP->StartName , "-" ) != 0 )   /*V1.04-c*/
      {
	if ( ( fp = fopen(AlgoParaP->StartName , "r" ) ) == NULL )
	  {
	    fprintf( stderr, "File %s does not exist\n", AlgoParaP->StartName ) ;
	    GenFree( ClassifM ) ;
	    return STS_E_FILEIN ;
	  }
     }
    else
      fp = stdin ;
	}


    /* open the file with initial segmentations */
   if (AlgoParaP->InitMode == INIT_PARAM)
     {
     if ( strcmp( AlgoParaP->StartParamName , "-" ) != 0 )   /*V1.04-c*/
      {
	if ( ( fp = fopen(AlgoParaP->StartParamName , "r" ) ) == NULL )
	  {
	    fprintf( stderr, "File %s does not exist\n", AlgoParaP->StartParamName ) ;
	    GenFree( ClassifM ) ;
	    return STS_E_FILEIN ;
	  }
     }
    else
      fp = stdin ;
	}






   /* open the file with singleton potentiels */
   if (AlgoParaP->SinglMode == SINGL_FIX)
	{
     if ( strcmp( AlgoParaP->SinglName , "-" ) != 0 )   /*V1.04-c*/
      {
	if ( ( fs = fopen(AlgoParaP->SinglName , "r" ) ) == NULL )
	  {
	    fprintf( stderr, "File %s does not exist\n", AlgoParaP->SinglName ) ;
	    return STS_E_FILEIN ;
	  }
     }
    else
      fs = stdin ;
	}



    /* open the output files */
   if ((AlgoParaP->Mode == NORMAL) ||
       ((AlgoParaP->Mode == LOOP) && (AlgoParaP->LoopParam.SaveRest)))
     if ((AlgoParaP->LoopParam.SaveType == ALL)
	 || ((AlgoParaP->LoopParam.SaveType == LAST) &&
	     (AlgoParaP->LoopParam.SaveNow)))
       {
	 if ( strcmp( AlgoParaP->OutBaseName , "-" ) != 0 )  /*V1.04-b*/
	   {
	     strncpy( mfname, AlgoParaP->OutBaseName, LEN_FILENAME ) ;
	     strncat( mfname, EXT_MFNAME, LEN_FILENAME ) ;
	     if ( ( fmf = fopen( mfname, "w" ) ) == NULL )
	       {
		 fprintf( stderr, "Could not open file '%s' in write mode\n",
			  mfname ) ;
		 return STS_E_FILEOUT ;
	       }
	   }
	 else
	   fmf = stderr ;
       }
     else
       fmf = fopen("/dev/null","w");
   else
     fmf = fopen("/dev/null","w");

   /* command line */
   fprintf( fmf, "Command line : \n\n  " ) ;
   for ( iarg = 0 ; iarg < Argc ; iarg ++ )
     fprintf( fmf, "%s " , Argv[ iarg ] ) ;

   if ((AlgoParaP->Mode == NORMAL) ||
       ((AlgoParaP->Mode == LOOP) && (AlgoParaP->LoopParam.SaveRest)))
     if ((AlgoParaP->LoopParam.SaveType == ALL)
	 || ((AlgoParaP->LoopParam.SaveType == LAST) &&
	     (AlgoParaP->LoopParam.SaveNow)))
       {
	 if ( strcmp( AlgoParaP->OutBaseName, "-" ) != 0 )  /*V1.04-b*/
	   {
	     if ( ( fout = fopen( AlgoParaP->OutName, "w" ) ) == NULL )
	       {
		 fprintf( stderr, "Could not open file '%s' in write mode\n",
			  AlgoParaP->OutName ) ;
		 return STS_E_FILEOUT ;
	       }
	   }
	 else
	   fout = stdout ;
       }
     else
       fout = fopen("/dev/null", "w");
   else
     fout = fopen("/dev/null", "w");

   /* Allocate all k-depending variables for k=Kmax */
   if ( (sts = AllocKvar(AlgoParaP, StatModelP, VarParamP, Kmax, DataP->NbVars, DataP->NbPts)) != STS_OK)
     { return sts ;}

   /* Allocate initial classification matrixs for k=Kmax */
       if ( ( ClassifM =
	       GenAlloc( DataP->NbPts * Kmax , sizeof( double ),
			 0, "ModelSelection" , "ClassifM" ) ) == NULL )
	   return STS_E_MEMORY ;
       if ( ( ClassifInit =
	       GenAlloc( DataP->NbPts * Kmax, sizeof( double ),
			 0, "ModelSelection" , "ClassifInit" ) ) == NULL )
	   return STS_E_MEMORY ;
      if ((Hclassif =
              GenAlloc( DataP->NbPts * Kmax, sizeof(double),
                             0, "ModelSelection", "Hclassif"))  == NULL )
            return STS_E_MEMORY ;


 /* Allocate all working variables for k=Kmax   */

          WorkingP->ColdM           = GenAlloc(Nl * Kmax, sizeof( double ), 0, "ModelSelection", "ColdM" );
          WorkingP->FikM            = GenAlloc(Nl * Kmax, sizeof( double ), 0, "ModelSelection", "FikM" ) ;
          WorkingP->PkFkiM          = GenAlloc(Nl * Kmax, sizeof( double ), 0, "ModelSelection", "PkFikM" ) ;
          WorkingP->LogFikM         = GenAlloc(Nl *  Kmax, sizeof( double ), 0, "ModelSelection", "LogFikM" ) ;
          WorkingP->LogPkFkiM       = GenAlloc( Nl * Kmax, sizeof( double ), 0, "ModelSelection", "LogPkFkiM" ) ;
          WorkingP->Neighs.NbNeigh  = SpatialP->MaxNeighs ;
          WorkingP->Neighs.NeighsV  = GenAlloc( SpatialP->MaxNeighs, sizeof( NeighT ),
  					    0, "ModelSelection", "NeighsV" ) ;
          WorkingP->CtmpM           = GenAlloc(Nl *  Kmax, sizeof( double ), 0, "ModelSelection", "CtmpM" ) ;
          WorkingP->Ztilde          = GenAlloc(Nl *  Kmax, sizeof( double ), 0, "ModelSelection", "Ztilde" ) ;



          #ifdef __TURBOC__
             fprintf( stderr, "\n Remaining memory after allocating : %lu bytes\n",
               (unsigned long) coreleft() );
          #endif

          if ( ( WorkingP->ColdM == NULL ) ||
           ( WorkingP->CtmpM == NULL ) ||
           ( WorkingP->FikM  == NULL ) ||
  	   ( WorkingP->LogFikM  == NULL ) ||
           ( ( WorkingP->Neighs.NeighsV == NULL ) &&
             ( SpatialP->MaxNeighs > 0 ) ) )
         {
          fprintf( stderr, "Could not allocate ModelSelection working variables\n" );
          return STS_E_MEMORY ;
         }




 /* run the algo for K = Kmin to K = Kmax */
    oldspeck = StatModelP->Spec.K;

    for (k=Kmin ; k<(Kmax+1); k++)
   {

     fprintf( stderr, "\n Run algo for %d classes\n",k ) ;
     StatModelP->Spec.K = k ;


     /* now just cleaning, in case...*/
        for ( i = 0; i <DataP->NbPts *k; i++)
              {
                ClassifM[i] = 0 ;
                ClassifInit[i] = 0 ;
                Hclassif[i] = 0 ;
	      }

   /* Read initial classification matrix in k clusters, if available */
      if (AlgoParaP->InitMode == INIT_FILE)
	{
         fprintf( stderr, "Reading initial partition for %d classes\n",k ) ;
         for ( il = 0, ic = 0 ; ( il < Nl ) && (! feof( fp )) ; il ++ )
         {
            /* printf( "\n pt %d : ", il ) ; */
            for ( ic = 0 ; ( ic < k ) && (! feof( fp )) ; ic ++ )
            {

            if ( fscanf( fp, "%g", &x) == 1 )
	      {
                ClassifInit[ ( il * k ) + ic ] = x ;
                /* printf( " %g", x ) ; */
            }
            else    ic -- ;
	    }
	 }
	}


 /* Read initial parameters for k clusters, if available */
      if (AlgoParaP->InitMode == INIT_PARAM)
	{
         fprintf( stderr, "Reading initial parameters for %d classes\n",k ) ;
          for (ik = 0; ik<k; ik++)
           {

              nbscan= fscanf( fp, "%f", &m) ;
              if (nbscan == 1)
                {
                  StatModelP->Para.Center_KD[ik] = m ;
                }


               nbscan= fscanf( fp, "%f", &v) ;
               if (nbscan == 1)
                {
                   StatModelP->Para.Disp_KD[ik] = v ;
                }

               nbscan= fscanf( fp, "%f", &s) ;
               if (nbscan == 1)
                {
                   StatModelP->Para.Singl[ik] =  s ;
                }


	   }

           nbscan= fscanf( fp, "%f", &beta) ;


      /* for security  */
      if (AlgoParaP->SinglMode == SINGL_NONE)
      {
	  for (ik = 0; ik<k; ik++)
           {
	     StatModelP->Para.Singl[ik] =  0 ;
	   }
      }
    }


    /* Read singleton potentiels if they are available */
     if ( AlgoParaP->SinglMode == SINGL_FIX)
       {
        fprintf( stderr, "Reading singleton potentiels for %d classes\n",k ) ;
        for ( l = 0;(l<k) && (! feof( fs )); l++)
	  {
            if ( fscanf( fs, "%f", &x) == 1 )
	      StatModelP->Para.Singl[l] = x;

          }

       }



      /* now just cleaning, in case...*/

        for ( i = 0; i <DataP->NbPts *k; i++)
              {
                WorkingP->ColdM[i] = 0 ;
                WorkingP->FikM[i] = 0 ;
                WorkingP->LogFikM[i] = 0 ;
                WorkingP->LogPkFkiM[i] = 0 ;
                WorkingP->PkFkiM [i] = 0 ;
                WorkingP->CtmpM[i] = 0 ;
                WorkingP->Ztilde[i] = 0 ;
	      }


    /* run the algo for selected values of beta */
     for (b = bmin; b<(bmax+bstep); b= b+bstep)
       {
         StatModelP->Para.Beta = b ;

         /* if  beta is estimated and not fixed, and we have an initialization
	    for this param (arg -s pi)*/
         if ((StatModelP->Spec.BetaModel == BETA_ZEROIN) &&(AlgoParaP->InitMode == INIT_PARAM))
            {
	           StatModelP->Para.Beta = beta ;
	           fprintf(stderr,"beta init = %f",StatModelP->Para.Beta);
	    }


      memcpy( ClassifM, ClassifInit, Nl * k * sizeof(double) ) ;

         if ((StatModelP->Spec.BetaModel == BETA_FIX) &&(b == btest))
           {
            fprintf(stderr,
			"algo not run for K=%d, and beta=%f, since already empty class for K=%d\n",k,b,k-1);
           }
         else
	   {

         if ((StatModelP->Spec.BetaModel == BETA_FIX) &&(b!=btest))
	  { fprintf( stderr, "Run algo for beta = %f\n",b ) ; }

	 sts = ClassifyByAlgo(AlgoParaP, SpatialP, DataP,
                                    StatModelP, ClassifM, WorkingP,
                                    MSCritP,VarParamP,Increase,
				    Criteria, EBlist);

         /* if the restoration is obtained with ICM supervised, for the param etimated previously */
         if (AlgoParaP->Resto == RESTO_ICM)
	   {

	    ICMwork.ColdM    = GenAlloc(Nl * k, sizeof( double ), 0, "ModelSelection", "ICMColdM" );
            ICMwork.FikM     = GenAlloc(Nl * k, sizeof( double ), 0, "ModelSelection", "ICMFikM" ) ;
            ICMwork.PkFkiM     = GenAlloc(Nl * k, sizeof( double ), 0, "ModelSelection", "ICMPkFikM" ) ;
            ICMwork.LogFikM  = GenAlloc(Nl * k , sizeof( double ), 0, "ModelSelection", "ICMLogFikM" ) ;
            ICMwork.LogPkFkiM = GenAlloc( Nl * k, sizeof( double ), 0, "ModelSelection", "ICMLogPkFkiM" ) ;
            ICMwork.Neighs.NbNeigh  = SpatialP->MaxNeighs ;
            ICMwork.Neighs.NeighsV  = GenAlloc( SpatialP->MaxNeighs, sizeof( NeighT ),
  					    0, "ModelSelection", "ICMNeighsV" ) ;
            ICMwork.CtmpM    = GenAlloc(Nl * k , sizeof( double ), 0, "ModelSelection", "ICMCtmpM" ) ;
            ICMwork.Ztilde   = GenAlloc(Nl * k , sizeof( double ), 0, "ModelSelection", "ICMZtilde" ) ;



           #ifdef __TURBOC__
             fprintf( stderr, "\n Remaining memory after allocating : %lu bytes\n",
               (unsigned long) coreleft() );
           #endif

           if ( ( ICMwork.ColdM == NULL ) ||
             ( ICMwork.CtmpM == NULL ) ||
             ( ICMwork.FikM  == NULL ) ||
  	     ( ICMwork.LogFikM  == NULL ) ||
             ( ( ICMwork.Neighs.NeighsV == NULL ) &&
             ( SpatialP->MaxNeighs > 0 ) ) )
           {
              fprintf( stderr, "Could not allocate ModelSelection working variables\n" );
              return STS_E_MEMORY ;
           }


            memcpy(ICMwork.CtmpM, WorkingP->CtmpM, Nl * k * sizeof( double ) ) ;
            memcpy(ICMwork.FikM , WorkingP->FikM, Nl * k * sizeof( double ) );
	    memcpy(ICMwork.PkFkiM, WorkingP->PkFkiM, Nl * k * sizeof( double ) ) ;
            memcpy(ICMwork.LogFikM, WorkingP->LogFikM, Nl * k * sizeof( double ) ) ;
            memcpy(ICMwork.LogPkFkiM, WorkingP->LogPkFkiM, Nl * k * sizeof( double ) ) ;
            memcpy(ICMwork.ColdM, WorkingP->ColdM, Nl * k * sizeof( double ) ) ;
            memcpy(ICMwork.Ztilde, WorkingP->Ztilde, Nl * k * sizeof( double )) ;

	    /* initialise ICM with param and resto estimated previously */
            sts = HardClassif(AlgoParaP,DataP,Nl,k,ClassifM,Hclassif);
            memcpy(Hclassif , ClassifM, Nl * k * sizeof( double ) ) ;
            /* modify the options for the restoration algo */
            AlgoParaP->ParamMode = PARAM_FIX;
            StatModelP->Spec.BetaModel = BETA_FIX;
            AlgoParaP->Algo = ALGO_ICM;
            AlgoParaP->CvTest = CVTEST_HARDCLASS ;
            AlgoParaP->CvThres = 0.01;


            fprintf( stderr, "Running ICM to find the restoration \n");
            sts = FieldAlgo(DataP, AlgoParaP, SpatialP, &StatModelP->Spec,
                          &StatModelP->Desc,
                          &StatModelP->Para, ClassifM, MSCritP, Increase,  &ICMwork, LastCPUtime);

           /* Free ICMwork  */
            GenFree( ICMwork.CtmpM ) ;
	    /* GenFree( ICMwork.CiNumV ) ;*/
            GenFree( ICMwork.PkFkiM ) ;
            GenFree( ICMwork.FikM ) ;
            GenFree( ICMwork.LogFikM ) ;
            GenFree( ICMwork.LogPkFkiM ) ;
            GenFree( ICMwork.ColdM ) ;
            GenFree( ICMwork.Ztilde ) ;
            GenFree( ICMwork.Neighs.NeighsV ) ;

	   } /* end if( AlgoParaP->Resto==RESTO_ICM) */

         /* else restoration obtained by MPM on the fuzzy resto ClassifM  */

       /* TO BE CHECKED : */
       /* Set StatParam to the Good value */
       if (AlgoParaP->Mode == LOOP)
	 {
	   for (j = 0; j < k; j++)
	     {
	       StatParam = &(AlgoParaP->MSparam.StatParam[k - Kmin]);
	       StatParam->MeanUk    [j] += StatModelP->Para.Center_KD[j];
	       StatParam->VarUk     [j] += pow(StatModelP->Para.Center_KD[j],
					       2);
	       StatParam->MeanSigmaK[j] += sqrt(StatModelP->Para.Disp_KD[j]);
	       StatParam->VarSigmaK [j] += StatModelP->Para.Disp_KD[j];
	       StatParam->MeanPotK  [j] += StatModelP->Para.Singl[j];
	       StatParam->VarPotK   [j] += pow(StatModelP->Para.Singl[j],2);
	     }
	   for (j = 0; j < k+1; j++)
	     {
	       StatParam->MeanErr[j] += abs(AlgoParaP->True.ErrorRate[j]);
	       StatParam->VarErr[j]  += pow(AlgoParaP->True.ErrorRate[j],2);
	     }
	   StatParam->MeanBeta  += StatModelP->Para.Beta;
	   StatParam->VarBeta   += pow(StatModelP->Para.Beta,2);
	 }
       /* Restauring good parameters for simulations */



         if (sts ==  STS_OK){
	   sts = SaveMSResults(AlgoParaP,&(StatModelP->Para), DataP, SpatialP, StatModelP,ClassifM,
                          WorkingP->LogFikM,Nl,k,MSCritP, Criteria,fmf,fout);}
         if( sts == STS_W_EMPTYCLASS)
	   {
            if (StatModelP->Spec.BetaModel == BETA_FIX)
	      {
              fprintf(stderr,"for K=%d, and beta=%f, there is an empty class in the segmentation\n",k,b);
	      btest=b;
              }
           else
              fprintf(stderr,"for K=%d,  there is an empty class in the segmentation\n",k);
	     }
            }
       /* come back to the options for the param estimation algo */
       AlgoParaP->ParamMode = ParamMode;
       StatModelP->Spec.BetaModel = BetaModel;
       AlgoParaP->Algo = Algo;
       AlgoParaP->CvTest = CvTest ;
       AlgoParaP->CvThres = CvThres;

       } /* end of for (b = bmin; ...*/

 

     if (bmin == bmax) /* we don't want beta to vary when comparing crit to select K */
       {
	 switch (AlgoParaP->Algo)
	   {
             case ALGO_MEANF:
             case ALGO_GSF:
             case ALGO_ICM:

                if (MSCritP->BICWinf > oldBICw ) { selecKBICw = k ;
	                    oldBICw   = MSCritP->BICWinf ;}

                if (MSCritP->BICPmf > oldBICp ) { selecKBICp = k ;
	                    oldBICp   = MSCritP->BICPmf  ;}

                if (MSCritP->ICLMarg.ICLHWinf > oldICLw ) { selecKICLw = k ;
	                    oldICLw   = MSCritP->ICLMarg.ICLHWinf ;}

                 if (MSCritP->ICLZtilde.ICLHmfWmf > oldICLp ) { selecKICLp = k ;
	                    oldICLp   = MSCritP->ICLZtilde.ICLHmfWmf ;}
		 break;

            case ALGO_NEM:
            case ALGO_NCEM:
               if (MSCritP->BICPmf > oldBICp ) { selecKBICp = k ;
	                    oldBICp   = MSCritP->BICPmf  ;}

               if (MSCritP->ICLZtilde.ICLHmfWmf > oldICLp ) { selecKICLp = k ;
	                    oldICLp   = MSCritP->ICLZtilde.ICLHmfWmf ;}
               break;

	   default :
		return -1;
	   }

       }

     } /* end of for (k=Kmin ; ...*/

    /* Free Working variables */
    GenFree( WorkingP->KmaxesV )  ;
    GenFree( WorkingP->CtmpM ) ;
    GenFree( WorkingP->CiNumV ) ;
    GenFree( WorkingP->PkFkiM ) ;
    GenFree( WorkingP->FikM ) ;
    GenFree( WorkingP->LogFikM ) ;
    GenFree( WorkingP->LogPkFkiM ) ;
    GenFree( WorkingP->Neighs.NeighsV ) ;
    GenFree( WorkingP->ColdM ) ;
    GenFree( WorkingP->Ztilde ) ;


    GenFree(Hclassif);

   /* save the class selected by each criterium */
  if (AlgoParaP->Mode == LOOP)
	 {
          switch (AlgoParaP->Algo)
	   {
             case ALGO_MEANF:
             case ALGO_GSF:
             case ALGO_ICM:
                AlgoParaP->MSparam.BICw[selecKBICw - Kmin] += 1 ;
                AlgoParaP->MSparam.BICp[selecKBICp - Kmin] += 1 ;
                AlgoParaP->MSparam.ICLw[selecKICLw - Kmin] += 1 ;
                AlgoParaP->MSparam.ICLp[selecKICLp - Kmin] += 1 ;
		 break;

            case ALGO_NEM:
            case ALGO_NCEM:
                AlgoParaP->MSparam.BICp[selecKBICp - Kmin] += 1 ;
                AlgoParaP->MSparam.ICLp[selecKICLp - Kmin] += 1 ;
               break;

	   default :
		return -1;
	   }
	 }

   /*fprintf(stderr,"KBICw = %6d, KBICp = %6d\n", selecKBICw,selecKBICp);*/

 /* To check : restauring the good k */
 StatModelP->Spec.K = oldspeck;

 if (fmf != stderr) fclose(fmf);
 if ( fout != stdout ) fclose( fout ) ;

 if (AlgoParaP->InitMode == INIT_FILE)
 {
   if ( strcmp( AlgoParaP->StartName , "-" ) != 0 )
      fclose( fp ) ;
 }

 if (AlgoParaP->SinglMode == SINGL_FIX)
 {
   if ( strcmp( AlgoParaP->SinglName , "-" ) != 0 )
      fclose( fs ) ;
 }

 GenFree(ClassifInit);

 

 return sts ;


} /* end of ModelSelection */

/* ------------------------------------------------------------------- */
static int GetInputPara
    (
          int           Argc,
          const char    *Argv[],
          DataT         *DataP,         /* O and allocated */
          AlgoParaT     *AlgoParaP,      /* O */
          SpatialT      *SpatialP,      /* O and allocated */
          StatModelT    *StatModelP,    /* O and allocated */
	  ErrinfoT      *ErrinfoP,      /* O and allocated */
	  ErrcurT       *ErrcurP,       /* O and allocated */
          double        **ClassifMP,     /* O and allocated */
          MSCritT       *MSCritP ,       /* O and allocated */
          double        **VarCenter_KD ,     /* O and allocated */
          double        **VarDisp_KD ,     /* O and allocated */
          double        **Increase,      /* O and allocated */
          double        **EBlist
    )
/* ------------------------------------------------------------------- */
{
    StatusET    err;
    const char* func = "GetInputPara";
    char        fname     [LEN_FILENAME + 1];
    char        namedat   [LEN_FILENAME + 1];   /*V1.04-a*/
    char        datadescS [LEN_LINE + 1];
    char        neidescS  [LEN_LINE + 1];

    /* Start of GetInputPara() */
    if ( ( err = AlgoArgs( Argc, Argv, fname, StatModelP, AlgoParaP,
			  datadescS, DataP, &SpatialP->NeighData.Image,
			  &SpatialP->Type ) )
        != STS_OK )
      {
        return err ;
      }

    /* Allocate and set sites visit order */
    if ( ( err = SetVisitOrder( DataP->NbPts,   /*V1.04-e*/
				AlgoParaP->VisitOrder,
				& DataP->SiteVisitV ) ) != STS_OK )
      return err ;


    /* Read known labels file, if available */
    if ( AlgoParaP->Label == 1)
    {
      if ((err = ReadLabelFile( AlgoParaP->LabelName, DataP->NbPts,
				& StatModelP->Spec.K,
				& DataP->LabelV,
				ClassifMP ) ) != STS_OK )
        return err;
    }


    /* If not GibbsSampling */
    if (AlgoParaP->Opt != GibbsSampler) {
      /* Read points */
      fprintf( stderr, "Reading points \n" ) ;
      strncpy( namedat , fname , LEN_FILENAME ) ;  /*V1.04-a*/
      strncat( namedat , ".dat", LEN_FILENAME ) ;

      if ( ( err = ReadMatrixFile( namedat,        /*V1.04-a*/
				   DataP->NbPts,
				   DataP->NbVars,
				   &DataP->PointsM ) ) != STS_OK )
	return err ;

      /* Count missing data */ /*V1.05-a*/
      {
	int i, j;

	DataP->NbMiss = 0 ;
	 for ( i = 0 ; i < DataP->NbPts ; i ++ )
	   {
	     for ( j = 0 ; j < DataP->NbVars ; j ++ )
	       {
		 if ( isnan( DataP->PointsM[ i * DataP->NbVars + j ] ) )
		   DataP->NbMiss ++ ;
	       }
	   }
      }

      /* read true data if known */
      if (AlgoParaP->True.K != 0)
	{
	  /* Read true data */
	  if ((err = ReadMatrixFile(AlgoParaP->TrueName ,        /*V1.04-a*/
				    DataP->NbPts,
				    1,
				    & AlgoParaP->True.DataTrue)) != STS_OK)
	    return err ;
	  /* Allocate memory for the error rates */
	  AlgoParaP->True.ErrorRate = GenAlloc(AlgoParaP->True.K+1,
					       sizeof(double),0,"GetInputPara",
					       "ErrorRate");
	  AlgoParaP->True.Prop      = GenAlloc(AlgoParaP->True.K,
					       sizeof(double),0,"GetInputPara",
					       "Prop");
	  /* Compute the proportions */
	  {
	    int j;
	    for (j = 0; j < AlgoParaP->True.K; j++)
	      AlgoParaP->True.Prop[j] = 0;
	    for (j = 0; j < DataP->NbPts; j++)
	      AlgoParaP->True.Prop[((int) AlgoParaP->True.DataTrue[j])]++;
	    for (j = 0; j < AlgoParaP->True.K; j++)
	      AlgoParaP->True.Prop[j] = AlgoParaP->True.Prop[j] / DataP->NbPts;
	  }
	}



      /* Eventually read reference class file */  /*V1.04-f*/
      if ((err = MakeErrinfo( AlgoParaP->RefName, DataP->NbPts,
			      StatModelP->Spec.K, AlgoParaP->TieRule,
			      ErrinfoP, ErrcurP)) != STS_OK)
	return err ;

      /* Read neighborhood file */
      if (SpatialP->Type != TYPE_NONSPATIAL)
	{
	  fprintf( stderr, "Reading neighborhood information \n" ) ;
	  if ( ( err = ReadNeiFile( fname,
				    DataP->NbPts,
				    AlgoParaP->NeighSpec ,
				    neidescS,
				    SpatialP)) != STS_OK)
	    return err ;
	}
      else
	{
	  StatModelP->Para.Beta = 0.0 ; /*V1.06-a*/
	  SpatialP->MaxNeighs   = 0 ;
	}
    } else
      {
	/* Allocate classification matrix */
	if (((*ClassifMP) =
	     GenAlloc(DataP->NbPts * StatModelP->Spec.K,
		      sizeof( double ),
		      0, func, "(*ClassifMP)")) == NULL)
	  return STS_E_MEMORY ;
	SpatialP->Type = TYPE_IMAGE; // Very important
	/* Set Correct Neighborhood system */
	if ((err = ReadNeiFile( fname,
				DataP->NbPts,
				AlgoParaP->NeighSpec ,
				neidescS,
				SpatialP)) != STS_OK)
	   return err ;
	 /* Define true data */
	 AlgoParaP->True.K         = StatModelP->Spec.K;
	 AlgoParaP->True.DataTrue  = GenAlloc(DataP->NbPts * DataP->NbVars,
					      sizeof(double),0,"DataTrue",
					      "GetInputPara");
	 AlgoParaP->True.ErrorRate = GenAlloc(StatModelP->Spec.K+1,
					      sizeof(double),0,"ErrorRate",
					      "GetInputPara");
	 AlgoParaP->True.Prop      = GenAlloc(StatModelP->Spec.K,
					      sizeof(double),0,"True",
					      "GetInputPara");
       }


    switch(AlgoParaP->Opt)
      {
      case ModelSelec :
	fprintf( stderr, "\nData : " ) ;
	if ( strcmp( datadescS, "" ) )
	  fprintf( stderr, "%s\n", datadescS ) ;
	else
	  fprintf( stderr, "\n" ) ;

	fprintf( stderr, "  file names =  %10s   |   nb points   = %10d\n",
		 fname, DataP->NbPts ) ;
	fprintf( stderr, "  type       =  %10s   |   dim         = %10d\n",
		 TypeDesC[ SpatialP->Type ], DataP->NbVars ) ;
	if ( SpatialP->Type == TYPE_IMAGE )
	  {
	    fprintf( stderr, "  image size =  (%4d,%4d)\n",
		     SpatialP->NeighData.Image.Nl,
		     SpatialP->NeighData.Image.Nc ) ;
	  }
	if ( DataP->NbMiss > 0 ) /*V1.05-a*/
	  {
	    fprintf( stderr, "  %d missing values / %d\n",
		     DataP->NbMiss, DataP->NbPts * DataP->NbVars );
	  }
	if ( SpatialP->Type != TYPE_NONSPATIAL )
	  {
	    fprintf( stderr, "Neighborhood system :\n  max neighb =  %10d\n",
		     SpatialP->MaxNeighs ) ;
	    fprintf( stderr, "%s\n", neidescS ) ;
	  }

	fprintf( stderr, "\n" ) ;
	fprintf( stderr, "Algo parameters :\n" ) ;
	fprintf( stderr, "Type of algorithm : '%s'\n" ,
		 AlgoDesC[ AlgoParaP->Algo ] ) ;
	//fprintf( stderr,
	//	 "   nk                    = %3d\n",
	//	 StatModelP->Spec.K) ;
	fprintf( stderr,
	      "  model                 = %s %s %s \n",
		 FamilyDesVC[ StatModelP->Spec.ClassFamily ],
		 ProporDesVC[ StatModelP->Spec.ClassPropor ],
		 DisperDesVC[ StatModelP->Spec.ClassDisper ]);
	fprintf( stderr, "\n" ) ;
	break ;

      case AddGaussianNoise :
	fprintf( stderr, "\nData : " ) ;
	if ( strcmp( datadescS, "" ) ) fprintf( stderr, "%s\n", datadescS ) ;
	else fprintf( stderr, "\n" ) ;
	break ;

      case GibbsSampler : break;

      default :
		return -1;

      } /*end switch (AlgoParaP->Opt) */


     /* If loop mode : display parameters */
    if (AlgoParaP->Mode == LOOP)
      {
	/* fprintf(stderr, "Loop mode activated :\n"); */
	fprintf(stderr, " Number of iterations : %d\n",
		AlgoParaP->LoopParam.NbIters);
      }
    else
      {
	/* fprintf(stderr, "Normal mode activated.\n"); */
      }

    /* initialization of Increase*/
    if (((*Increase) = GenAlloc( 1000 , sizeof( double ),0, func,
				 "(*Increase)"))
	== NULL )
      return STS_E_MEMORY;

    /* initialization of VarParam */
    if ((AlgoParaP->Algo == ALGO_GSF))
      {
	if (((*VarCenter_KD)=GenAlloc((StatModelP->Spec.K)* (DataP->NbVars),
				      sizeof( double ),1, func,
				      "(*VarCenter_KD)")) == NULL )
	  return STS_E_MEMORY;
	if ( (StatModelP->Spec.ClassFamily == FAMILY_NORMALGEN )
	     && (StatModelP->Spec.ClassDisper == DISPER__D) )
	  {
	    if (((*VarDisp_KD)=GenAlloc(DataP->NbVars ,
					sizeof( double ),1, func,
					"(*VarDisp_KD)")) == NULL )
	      return STS_E_MEMORY;
	  }
	else
	  {
	    if (((*VarDisp_KD)=GenAlloc((StatModelP->Spec.K)* (DataP->NbVars),
					sizeof( double ),1, func,
					"(*VarDisp_KD)")) == NULL )
	      return STS_E_MEMORY;
	  }
      }


    return STS_OK ;
} /* end of GetInputPara() */


/* ------------------------------------------------------------------- */
static int SetVisitOrder   /*V1.04-e*/
        (
	     int         Npt,          /* I */
	     OrderET     VisitOrder,   /* I */
	     int**       SiteVisitVP   /* O and allocated (Npt) */
	)
/* ------------------------------------------------------------------- */
{
  int ivis ;


  if (( *SiteVisitVP = GenAlloc( Npt, sizeof( int ), 0,
				 "SetVisitOrder", "SiteVisitVP" ) ) == NULL )
      return STS_E_MEMORY ;

  for ( ivis = 0 ; ivis < Npt ; ivis ++ )
    (*SiteVisitVP)[ ivis ] = ivis ;

  if ( VisitOrder == ORDER_RANDOM )
    {
      RandomPermutationAlgo( (*SiteVisitVP) , Npt ) ;
    }

  return STS_OK ;
}


/* ------------------------------------------------------------------- */
static int ReadMatrixFile
         (
             const char  *FileName,    /* I */      /*V1.04-a*/
             int         Nl,           /* I */
             int         Nc,           /* I */
             double      **MatP        /* O and allocated (size : Nl * Nc) */
         )
/* ------------------------------------------------------------------- */
{
    double       *Mat ;
    FILE        *fp ;
    int         il ;
    int         ic ;
    float       x ;


    if ( ( *MatP = GenAlloc( Nl *  Nc, sizeof( double ),
			     0, "ReadMatrixFile", FileName ) ) == NULL )
      return STS_E_MEMORY ;
    Mat = *MatP ;

    if ( strcmp( FileName , "-" ) != 0 )   /*V1.04-c*/
      {
	if ( ( fp = fopen( FileName, "r" ) ) == NULL )
	  {
	    fprintf( stderr, "File %s does not exist\n", FileName ) ;
	    GenFree( Mat ) ;
	    return STS_E_FILEIN ;
	  }
     }
    else
      fp = stdin ;

    for ( il = 0, ic = 0 ; ( il < Nl ) && (! feof( fp )) ; il ++ )
    {
      /* printf( "\n pt %d : ", il ) ; */
        for ( ic = 0 ; ( ic < Nc ) && (! feof( fp )) ; ic ++ )
        {

            if ( fscanf( fp, "%g", &x) == 1 )
	      {
                Mat[ ( il * Nc ) + ic ] = x ;
                /* printf( " %g", x ) ; */
            }
            else    ic -- ;
        }
    }

    if ( strcmp( FileName , "-" ) != 0 )   /*V1.04-c*/
      fclose( fp ) ;

    if ( ( il < Nl ) || ( ic < Nc ) )
    {
        if ( ic == 0 )
        {
            ic = Nc ;
            il -- ;
        }
        fprintf( stderr, "%s : short file (%d/%d lines and %d/%d columns)\n",
                 FileName, il, Nl, ic, Nc ) ;
        GenFree( Mat ) ;
        return STS_E_FILE ;
    }
    else
        return STS_OK ;

}  /* end of ReadMatrixFile() */



/* ------------------------------------------------------------------- */
static int  ReadLabelFile
         (
             const char  *LabelName,    /* I */
             int         Npt,           /* I */
	     int         *KfileP,       /* O : file # classes */ /*V1.06-c*/
             int         **LabelVP,     /* O and allocated (Npt) */
             double       **ClassifMP    /* O and allocated (Npt*Nk) */
         )
/* ------------------------------------------------------------------- */
{
    FILE  *fp;
    int   pt ;

    if ( ( fp = fopen( LabelName, "r" ) ) == NULL )
    {
        fprintf( stderr, "File %s does not exist\n", LabelName ) ;
        return STS_E_FILEIN ;
    }


    if ( ( *LabelVP =
	   GenAlloc( Npt, sizeof( int ),
		     0, "ReadLabelFile", LabelName ) ) == NULL )
        return STS_E_MEMORY ;

    /* Read labels and store in matrix and vector */
    for ( pt = 0 ; ( pt < Npt ) && (! feof( fp )) ; pt ++ ) {
        /* Read label indication */
        fscanf( fp , "%d" , & (*LabelVP)[ pt ] ) ;

        /* If label is known and valid */
        if ( ( 0 < (*LabelVP)[ pt ] ) && ( (*LabelVP)[ pt ] <= (*KfileP) ) )
	  LabelToClassVector( (*KfileP), (*LabelVP)[ pt ] - 1,
			      & (*ClassifMP)[ pt * (*KfileP) ] ) ;
	/* Else (label unknown or invalid) */

    }

    fclose( fp ) ;


    if ( pt < Npt )
    {
        fprintf( stderr, "%s : short file (%d/%d labels)\n",
                 LabelName, pt , Npt ) ;
        GenFree( (*ClassifMP) ) ; (*ClassifMP) = NULL ;
        GenFree( (*LabelVP) ) ;   (*LabelVP)   = NULL ;

        return STS_E_FILE ;
    }
    else
        return STS_OK ;

}  /* end of ReadLabelFile() */



/* ------------------------------------------------------------------- */
static int  MakeErrinfo
         (
             const char* RefName,       /* I : filename of reference class */
             int         N,             /* I : number of objects */
             int         Kc,            /* I : user number of classes */
	     TieET       TieRule,       /* I : specified MAP tie rule */
	     ErrinfoT*   ErrinfoP,      /* O and allocated */
	     ErrcurT*    ErrcurP        /* O and allocated */
         )
/* ------------------------------------------------------------------- */
{
  StatusET err ;
  int      *tmpV ;
  int      ipt ;

  if ( strcmp( RefName, "" ) != 0 ) {
    ErrinfoP->Kc = Kc ;
    if ( ( err = ReadLabelFile( RefName, N,
				& ErrinfoP->Kr,
				& tmpV,
				& ErrinfoP->Refclas_N_Kr ) ) != STS_OK )
      return err ;

    /* Check all reference labels ok */
    for ( ipt = 0, err = STS_OK ;
	  ( ipt < N ) && ( err == STS_OK ) ; ipt ++ ) {
      if ( ( tmpV[ ipt ] <= 0 ) || ( tmpV[ ipt ] > ErrinfoP->Kr ) ) {
	fprintf( stderr,
		 "Reference class for point %d not in 1..%d \n",
		 ipt + 1, ErrinfoP->Kr ) ;
	err = STS_E_FILE ;
      }
    }
    GenFree( tmpV ) ; tmpV = NULL ;
    if ( err != STS_OK )
      return err ;

    /* Compute greatest number of classes and permutations of classes */
    ErrinfoP->Km = ( ErrinfoP->Kc > ErrinfoP->Kr ) ?
      ErrinfoP->Kc : ErrinfoP->Kr ;

    ErrinfoP->Kmfac = factorial( ErrinfoP->Km ) ;

    ErrinfoP->TieRule = TieRule ;

    compute_permutations( 0, ErrinfoP->Km, & ErrinfoP->Perm_Kmfac_Km ) ;

    /* Allocate and initialize later computed stuff */
    if ( ( ErrcurP->Agree_Km_Km =
	   GenAlloc( ErrinfoP->Km * ErrinfoP->Km, sizeof( double ),
		     0, "MakeErrinfo", "Agree_Km_Km" ) ) == NULL )
        return STS_E_MEMORY ;

    if ( ( ErrcurP->Loclas_N_Kc =
	   GenAlloc( N * ErrinfoP->Kc, sizeof( double ),
		     0, "MakeErrinfo", "Loclas_N_Kc" ) ) == NULL )
        return STS_E_MEMORY ;

    ErrcurP->Ibestpermut = -1 ;
    ErrcurP->Errorrate   = -2.0 ;

    return STS_OK ;
  }
  else {
    ErrinfoP->Kr = 0; /*NULL ;*/
    ErrinfoP->Refclas_N_Kr = NULL ;
    ErrcurP->Errorrate     = -1.0 ;
    return STS_OK ;
  }

}  /* end of MakeErrinfo() */





/* ------------------------------------------------------------------- */
static int factorial(int n)
/* ------------------------------------------------------------------- */
{
  int result = 1;

  for (; n>0; n--)
    result *= n;

  return result;
}


/* ------------------------------------------------------------------- */
static int compute_permutations    /* ret 0 if OK, -1 if memory error */
(
 const int Startval,         /* I : start value of integer suite */
 const int K,                /* I : size of integer suite to permute > 0 */
 int*      perms_Kfact_K_p[] /* O : matrix to store permuted values */
)
/* ------------------------------------------------------------------- */
{
  int   ik ;      /* current index and value of integer suite : 0..K-1 */
  int*  array_K ; /* integer suite to permute : [ 0 ... (K-1) ] */
  int   Kfact ;   /* K! */
  int   err ;     /* 0 if OK, -1 if memory error */

  if ( K <= 0 )
    return 1 ;

  Kfact = factorial( K ) ;

  if( ( (*perms_Kfact_K_p) = malloc( Kfact * K * sizeof( int ) ) ) == NULL )
    return -1 ;

  if( ( array_K = malloc( K * sizeof( int ) ) ) == NULL ) {
    free( (*perms_Kfact_K_p) ) ;
    (*perms_Kfact_K_p) = NULL ;
    return -1 ;
  }

  for ( ik = 0 ; ik < K ; ik ++ )
    array_K[ ik ] = Startval + ik ;

  err = rec_permutations( array_K, K, K, 0, (*perms_Kfact_K_p) ) ;

  free( array_K ) ;

  return err ;
}


/* ------------------------------------------------------------------- */
static int rec_permutations        /* ret 0 if OK, -1 if memory error */
(
 const int array_A[],       /* I : remaining array to permute */
 const int A,               /* I : length of the remaining array : 0..K */
 const int K,               /* I : length of original array */
 int       offset,          /* I : first line of storage to use */
 int       perms_Kfact_K[]  /* O : matrix to store permuted values, use
			       lines :   offset -> offset + A! - 1
			       columns : K - A  -> K - 1 */
)
/* ------------------------------------------------------------------- */
{
  int    err ;         /* 0 if currently OK */
  int    ia ;          /* array element currently removed : 0..A-1 */
  int    ja ;          /* array element currently copied : 0..A-1 (neq ia) */
  int    am1fact ;     /* (A-1)! */
  int    iam1fact ;    /* current perms line (without offset) : 0..am1fact-1 */
  int*   redarr_am1 ;  /* array with one element removed (A-1) */

  am1fact = factorial( A - 1 ) ;

  /* Check against out of bounds offset */
  if ( ( offset < 0 ) || ( factorial( K ) < ( offset + A * am1fact ) ) )
    return 1 ;

  if ( ( redarr_am1 = malloc( ( A - 1 ) * sizeof( int ) ) ) == NULL )
    return -1;

  /* For each element of given array */
  for ( ia = 0, err = 0 ; ( ia < A ) && ( err == 0 ) ; ia ++ ) {

    /* Copy (A-1)! times this element into the column ( K - A ) of perms,
       starting from line ( offset + ia * (A-1)! ) (skip previous ia's) */
    for ( iam1fact = 0 ; iam1fact < am1fact ; iam1fact ++ )
      perms_Kfact_K[ ( offset + ia * am1fact + iam1fact ) * K + ( K - A ) ] =
	array_A[ ia ] ;

    /* Copy array into reduced array without this element */
    for ( ja = 0 ; ja < ia ; ja ++ )
      redarr_am1[ ja ] = array_A[ ja ] ;
    for ( ja = ia + 1 ; ja < A ; ja ++ )
      redarr_am1[ ja - 1 ] = array_A[ ja ] ;

    /* Recursive call with reduced array */
    err = rec_permutations( redarr_am1, A-1, K, offset + ia * am1fact,
			    perms_Kfact_K ) ;
  }

  free( redarr_am1 ) ;
  return err ;
}




/* ------------------------------------------------------------------- */
static int  ReadNeiFile
         (
             const char  *BaseName,          /* I */
             int         NbPts,              /* I */
             NeighET     NeighSpec,          /* I */
	     char*       NeiDescS,           /* O [LEN_LINE+1] */
             SpatialT    *SpatialP           /* I/O and allocated */
         )
/* ------------------------------------------------------------------- */
{
    char        infname[ LEN_FILENAME + 1 ] ;
    FILE        *fnei ;
    StatusET    err = STS_OK ;

    /* Read neighbourhood data depending on spatial type */
    if ( SpatialP->Type == TYPE_SPATIAL )
    {
        strncpy( infname, BaseName, LEN_FILENAME ) ;
        strncat( infname, ".nei" , LEN_FILENAME ) ;

        if ( ReadOpeningComments( infname , "#" , LEN_LINE ,
                                  & fnei , NeiDescS ) == -1 )
        {
            fprintf( stderr, "File %s does not exist\n", infname ) ;
            return STS_E_FILEIN ;
        }

        err = ReadPtsNeighs( fnei, NbPts,
                             &SpatialP->MaxNeighs,
                             &SpatialP->NeighData ) ;
        fclose( fnei ) ;
    }
    else    /* Type is image */
    {
      if ( NeighSpec == NEIGH_FILE )
        {
            strncpy( infname, BaseName, LEN_FILENAME ) ;
            strncat( infname, ".nei" , LEN_FILENAME ) ;
            if ( ReadOpeningComments( infname , "#" , LEN_LINE ,
		                      & fnei , NeiDescS ) == -1 )
            {
                fprintf( stderr, "File %s does not exist\n", infname ) ;
                return STS_E_FILEIN ;
            }

            err = ReadImageNeigh( fnei, &SpatialP->NeighData ) ;
            fclose( fnei ) ;
        }
        else
        {
            err = SetImageNeigh( NeighSpec, NeiDescS, &SpatialP->NeighData ) ;
        }
      // Added by FC the 25/07/01
      SpatialP->MaxNeighs = SpatialP->NeighData.Image.NbNeigh;
      /* Very very important too */
      SpatialP->NeighData.PtsNeighsV =
	GenAlloc(1,sizeof(PtNeighsT),
		 0, "ReadNeiFile", "PtsNeighsV");
      SpatialP->NeighData.PtsNeighsV->NbNeigh = SpatialP->MaxNeighs;
      SpatialP->NeighData.PtsNeighsV->NeighsV =
	GenAlloc(SpatialP->MaxNeighs, sizeof(NeighT),
		 0, "ReadNeiFile", "NeighsV");
    }


    return err ;

}  /* end of ReadNeiFile() */


/* ------------------------------------------------------------------- */
static int      ReadPtsNeighs
                (
                    FILE        *Fnei,          /* I/O */
                    int         NbPts,          /* I */
                    int         *MaxNeiP,       /* O */
                    NeighDataT  *NeighDataP     /* O and allocated */
                )
/* ------------------------------------------------------------------- */
{
    const char* func = "ReadPtsNeighs" ;
    int       weighted ;
    int       l ;
    int       ipt ;
    StatusET  err = STS_OK ;
    PtNeighsT *ptsneighsV ;
    int       nmax ;


    /* Indicator of weights presence */
    fscanf( Fnei, "%d", &weighted ) ;

    /* Allocate structure of all points' neighbours */
    if ( ( ptsneighsV = GenAlloc( NbPts, sizeof( PtNeighsT ),
				  0, func, "ptsneighsV" ) )
           == NULL )
    {
        fprintf( stderr, "Cannot allocate list of neighbours of %d points\n" ,
                 NbPts ) ;
        err = STS_E_MEMORY ;
        return err ;
    }
    NeighDataP->PtsNeighsV = ptsneighsV ;

    /* For each point, default nb of neighbours is zero */
    for ( ipt = 0 ; ipt < NbPts ; ipt ++ )
    {
        ptsneighsV[ ipt ].NbNeigh = 0 ;
    }

    /* Read neighbours of each point */
    nmax = 0 ;
    for( l = 0 ; ( err == STS_OK ) && ( ! feof( Fnei ) ) ; l ++ )
    {
        int     ipt ;

        if ( fscanf( Fnei, "%d", &ipt ) == 1 )
        {
            int nbv ;   /* Declared nb of neighbours */
            int nv ;    /* Effective nb of neighbours */

            if ( fscanf( Fnei, "%d", &nbv ) == 1 )
            {
                NeighT   *neighsV ;
                int      iv ;

                if ( ( neighsV = GenAlloc( nbv, sizeof( NeighT ),
					   0, func, "neighsV" ) ) == NULL )
                {
                    fprintf( stderr, "Can't allocate %d neighb. for pt %d\n" ,
                             nbv, ipt ) ;
                    err = STS_E_MEMORY ;
                    return  err ;
                }

                ptsneighsV[ ipt - 1 ].NeighsV = neighsV ;

				fprintf(stderr, "nb voisins = %d\n", nbv);
                for ( iv = 0, nv = 0 ;
                      ( iv < nbv ) && ( ! feof( Fnei ) ) ;
                      iv ++ )
                {
                    int     iptv ;
                    if ( fscanf( Fnei, "%d", &iptv ) == 1 )
                    {
                        if ( ( 1 <= iptv ) && ( iptv <= NbPts ) )
                        {
                            neighsV[ nv ].Index = iptv - 1 ;
                            nv ++ ;
                        }
                    }
                    else
                    {
                        fprintf( stderr,
                                 "Error in neighb. file l.%d : neighbor %d\n",
                                 l, iv ) ;
                        err = STS_E_FILE ;
                    }
                } /* end  for ( iv ... ) */

                if ( weighted )
                {
                    for ( iv = 0, nv = 0 ;
                         ( iv < nbv ) && ( ! feof( Fnei ) ) ;
                         iv ++ )
                    {
                        float   weight ;

                        if ( fscanf( Fnei, "%g", &weight) == 1 )
                        {
                            if ( weight != 0.0 )
                            {
                                neighsV[ nv ].Weight = weight ;
                                nv ++ ;
                            }
                        }
                        else
                        {
                          fprintf( stderr,
                                  "Error in neighb. file l.%d : weight %d\n",
                                  l, iv ) ;
                            err = STS_E_FILE ;
                        }
                    } /* end  for ( iv ... ) */
                } /* end if weighted */
                else
                {
                    for ( iv = 0 ; iv < nv ; iv ++ )
                        neighsV[ iv ].Weight = 1.0 ;
                }

                ptsneighsV[ ipt - 1 ].NbNeigh = nv ;

                if ( nv > nmax )     nmax = nv ;

            } /* end if ( fscanf( ... , &nbv ) == 1 ) */
            else
                 {}
        } /* end if ( fscanf( ..., &ipt ) == 1 ) */
        else
            {}
    } /* end for( l = 0 ; ( ! feof( Fnei ) ) ; l ++ ) */

    *MaxNeiP = nmax ;

    return err ;

}   /* end of ReadPtsNeighs() */


/* ------------------------------------------------------------------- */
static int      ReadImageNeigh
                (
                    FILE        *Fnei,          /* I/O */
                    NeighDataT  *NeighDataP     /* O and allocated */
                )
/* ------------------------------------------------------------------- */
{
    StatusET  err = STS_OK ;
    int       dlmin, dlmax, dcmin, dcmax ;
    int       dl, dc ;
    int       tnb ;     /* window size */
    int       nv ;      /* nb of non-zero weights */
    INeighT   *tneighV;  /* to be allocated */

    fscanf( Fnei, "%d %d %d %d", &dlmin, &dlmax, &dcmin, &dcmax ) ;
    tnb = ( dlmax - dlmin + 1 ) * ( dcmax - dcmin + 1 ) ;

    if ( ( tneighV = GenAlloc( tnb, sizeof( INeighT ),
			       0, "ReadImageNeigh", "tneighV" ) ) == NULL )
        return STS_E_MEMORY ;

    NeighDataP->Image.NeighsV = tneighV ;

    nv = 0 ;
    for ( dl = dlmin ; dl <= dlmax ; dl ++ )
    {
        for ( dc = dcmin ; dc <= dcmax ; dc ++ )
        {
            float   weight ;

            if ( fscanf( Fnei, "%g", &weight ) == 1 )
            {
                if ( weight != 0.0 )
                {
                    tneighV[ nv ].Dl = dl ;
                    tneighV[ nv ].Dc = dc ;
                    tneighV[ nv ].Weight = weight ;
                    nv ++ ;
                }
                else /* else this neighbour has no weight : skip it */
                {
                }
            }
            else
            {
                fprintf( stderr, "Neighbors file error (dl = %d, dc = %d)\n",
                         dl, dc ) ;
                err = STS_E_FILE ;
            }
        } /* for dc */
    } /* for dl */

    NeighDataP->Image.NbNeigh = nv ;
    return err ;

}   /* end of ReadImageNeighs() */


/* ------------------------------------------------------------------- */
static int  SetImageNeigh
            (
                NeighET     NeighSpec,          /* I */
                char*       NeiDescS,           /* O [LEN_LINE+1] */
                NeighDataT* NeighDataP          /* O and allocated */
            )
/* ------------------------------------------------------------------- */
{
    INeighT   *neighV;  /* to be allocated */

    switch( NeighSpec )
    {
    case NEIGH_FOUR :
        if ( ( neighV = GenAlloc( 4, sizeof( INeighT ),
				  0, "SetImageNeigh", "neighV" ) ) == NULL )
        {
            fprintf( stderr, "Could not allocate %d image neighbours\n",
                     4 ) ;
            return STS_E_MEMORY ;
        }
        NeighDataP->Image.NeighsV = neighV ;
        NeighDataP->Image.NbNeigh = 4 ;

        neighV[ 0 ].Dl = -1 ;
        neighV[ 0 ].Dc = 0 ;
        neighV[ 0 ].Weight = 1.0 ;

        neighV[ 1 ].Dl = 0 ;
        neighV[ 1 ].Dc = -1 ;
        neighV[ 1 ].Weight = 1.0 ;

        neighV[ 2 ].Dl = 0 ;
        neighV[ 2 ].Dc = +1 ;
        neighV[ 2 ].Weight = 1.0 ;

        neighV[ 3 ].Dl = +1 ;
        neighV[ 3 ].Dc = 0 ;
        neighV[ 3 ].Weight = 1.0 ;

        strncpy( NeiDescS ,
                 "  Default 1st-order neighbors (horizontal and vertical)\n",
                 LEN_LINE ) ;
        break ;

    default :
        fprintf( stderr, "Unknown neighborhood type (%d)\n", NeighSpec ) ;
        return STS_E_FUNCARG ;
    }

    return STS_OK ;
}   /* end of SetNeigh() */



/* ------------------------------------------------------------------- */
static int SaveClassif
(
 const AlgoParaT*     AlgoParaP,
 const DataT*         DataP,
 const SpatialT*      SpatialP,
 const StatModelT*    ModelP,
 double*              ClassifM,
 const int            Npt,
 const int            K,
 FILE*                fout
 )
/* ------------------------------------------------------------------- */
{

 StatusET   sts =  STS_OK ;
 StatusET   err =  STS_OK ;
 int        il, ic, ipt, i; /* point counters */
 int        k;           /* class and dim counters */
 int        Nl = SpatialP->NeighData.Image.Nl ;
 int        Nc = SpatialP->NeighData.Image.Nc ;
 /*int        Nd = DataP->NbVars;*/
 int        nk = ModelP->Spec.K ;
 int        test = 0; /* to check if empty class in the hard classification */


 switch (SpatialP->Type)
  	  {
            case TYPE_SPATIAL:
             if (fout == NULL)
               {
               sts = STS_E_FILEOUT;
               return sts;
                }
             if ( K != 1)
             {
               fprintf(fout,"K=%d, beta = %f\n",K,ModelP->Para.Beta);
               if ( AlgoParaP->Format == FORMAT_HARD )
                {

                 int*   kmaxesV;  /* classes having same maximum probabilites */

                 if ( ( kmaxesV = GenAlloc( nk, sizeof( int ),
				   0, "SaveMSResults", "kmaxesV" ) ) == NULL )
                 return STS_E_MEMORY ;

                /* For each record */
                for ( i = 0 ; ( i < Npt ) && ( err == STS_OK ) ; i ++ ){
                int   kmax ;
                /* If label is not known */
                if ( ( AlgoParaP->Label   != 1 ) ||
		 ( DataP->LabelV[ i ] == 0 ) )
                {

                /* Compute its MAP class */
                kmax = ComputeMAP( ClassifM, i, nk, AlgoParaP->TieRule,
			       kmaxesV ) ;
	        }
                else kmax = DataP->LabelV[ i ] -1 ;

                if ((kmax+1) == K) test = 1;


                 /* Save the record's class (add 1 to have k in 1..Nk) */
                 if ( fprintf( fout, "%d ", kmax+1 ) == EOF )
                 {
                  fprintf( stderr, "Cannot write Ci i = %d\n",
                            i + 1 ) ;
                  err = STS_E_FILEOUT ;
                 }
                } /* end for ( i ... ) */

                GenFree( kmaxesV ) ;

                if (test == 0) sts = STS_W_EMPTYCLASS ;

               } /* end if Format == HARD */



             if ( AlgoParaP->Format == FORMAT_UF) /* format required to use ClassifM as initial segmentation */
              /* matrix of the z_ik */
              {

               int*   kmaxesV;  /* classes having same maximum probabilites */

               if ( ( kmaxesV = GenAlloc( nk, sizeof( int ),
				   0, "SaveMSResults", "kmaxesV" ) ) == NULL )
               return STS_E_MEMORY ;

               /* For each record */
               for ( i = 0 ; ( i < Npt ) && ( err == STS_OK ) ; i ++ ){
               int   kmax ;
               /* If label is not known */
               if ( ( AlgoParaP->Label   != 1 ) ||
		 ( DataP->LabelV[ i ] == 0 ) )
                {

                /* Compute its MAP class */
                kmax = ComputeMAP( ClassifM, i, nk, AlgoParaP->TieRule,
			       kmaxesV ) ;
	         }
               else kmax = DataP->LabelV[ i ] -1 ;

               if ((kmax+1) == K) test = 1;

               for (k = 0; k<nk; k++){
                         if (k == kmax) ClassifM[i*nk+k] =1;
                         else ClassifM[i*nk+k] =0;
		    }

               for (k = 0; k<nk; k++)
                {
                 if ( fprintf( fout, " %5.3f ",
                              ClassifM[ (i* nk) + k ] ) == EOF )
                  {
                    fprintf( stderr, "Cannot write Zk i = %d, k =%d\n",
                            i + 1, k + 1 ) ;

                    err = STS_E_FILEOUT ;
                  }
	         }

                fprintf( fout, "\n" ) ;

               } /* end for ( il ... ) */

               GenFree( kmaxesV ) ;

               if (test == 0) sts = STS_W_EMPTYCLASS ;

 
               } /* end  Format == FORMAT_UF */

               if ( AlgoParaP->Format == FORMAT_FUZZY )
               {

                for ( ipt = 0 ; ( ipt < Npt ) && ( err == STS_OK ) ; ipt ++ )
                {

                 for ( k = 0 ; ( k < nk ) && ( err == STS_OK ) ; k ++ )
                 {
                   if ( fprintf( fout, " %5.3f ",
                              ClassifM[ (ipt * nk) + k ] ) == EOF )
                    {
                    fprintf( stderr, "Cannot write Uik i = %d, k =%d\n",
                            ipt + 1, k + 1 ) ;

                    err = STS_E_FILEOUT ;
                    }
                   } /* end for ( k = 0 ... ) */

                  } /* end for ( ipt ... ) */
                } /* end else Format == FORMAT_FUZZY */

                } /* end (if  K != 1) */

	      break; /* break for case spatial */




	    case TYPE_IMAGE:
             if (fout == NULL)
               {
               sts = STS_E_FILEOUT;
               return sts;
                }
             if ( K != 1)
             {
               fprintf(fout,"K=%d, beta = %f\n",K,ModelP->Para.Beta);
               if ( AlgoParaP->Format == FORMAT_HARD )
                {

                 int*   kmaxesV;  /* classes having same maximum probabilites */

                 if ( ( kmaxesV = GenAlloc( nk, sizeof( int ),
				   0, "SaveMSResults", "kmaxesV" ) ) == NULL )
                 return STS_E_MEMORY ;

                /* For each record */
                for ( il = 0 ; ( il < Nl ) && ( err == STS_OK ) ; il ++ ){
	        for ( ic = 0 ; ( ic < Nc ) && ( err == STS_OK ) ; ic ++ ){
                int i = il*Nc +ic;
                int   kmax ;
                /* If label is not known */
                if ( ( AlgoParaP->Label   != 1 ) ||
		 ( DataP->LabelV[ i ] == 0 ) )
                {

                /* Compute its MAP class */
                kmax = ComputeMAP( ClassifM, i, nk, AlgoParaP->TieRule,
			       kmaxesV ) ;
	        }
                else kmax = DataP->LabelV[ i ] -1 ;

                if ((kmax+1) == K) test = 1;


                /* Save the record's class (add 1 to have k in 1..Nk) */
                if ( fprintf( fout, "%d ", kmax+1 ) == EOF )
                {
                  fprintf( stderr, "Cannot write Ci i = %d\n",
                            i + 1 ) ;
                  err = STS_E_FILEOUT ;
                }

                 if ( ( SpatialP->Type == TYPE_IMAGE ) &&
                 ( ((i + 1) % SpatialP->NeighData.Image.Nc) == 0 ) )
                 {
                 fprintf( fout, "\n" ) ;
                 }
                } /* end for ( ic ... ) */
                } /* end for ( il ... ) */

                GenFree( kmaxesV ) ;

                if (test == 0) sts = STS_W_EMPTYCLASS ;

               } /* end if Format == HARD */



             if ( AlgoParaP->Format == FORMAT_UF) /* format required to use ClassifM as initial segmentation */
              /* matrix of the z_ik */
              {

               int*   kmaxesV;  /* classes having same maximum probabilites */

               if ( ( kmaxesV = GenAlloc( nk, sizeof( int ),
				   0, "SaveMSResults", "kmaxesV" ) ) == NULL )
               return STS_E_MEMORY ;

               /* For each record */
               for ( il = 0 ; ( il < Nl ) && ( err == STS_OK ) ; il ++ ){
	       for ( ic = 0 ; ( ic < Nc ) && ( err == STS_OK ) ; ic ++ ){
               int i = il*Nc +ic;
               int   kmax ;
               /* If label is not known */
               if ( ( AlgoParaP->Label   != 1 ) ||
		 ( DataP->LabelV[ i ] == 0 ) )
                {

                /* Compute its MAP class */
                kmax = ComputeMAP( ClassifM, i, nk, AlgoParaP->TieRule,
			       kmaxesV ) ;
	         }
               else kmax = DataP->LabelV[ i ] -1 ;

               if ((kmax+1) == K) test = 1;

               for (k = 0; k<nk; k++){
                         if (k == kmax) ClassifM[i*nk+k] =1;
                         else ClassifM[i*nk+k] =0;
		    }

               for (k = 0; k<nk; k++)
                {
                 if ( fprintf( fout, " %5.3f ",
                              ClassifM[ (i* nk) + k ] ) == EOF )
                  {
                    fprintf( stderr, "Cannot write Zk i = %d, k =%d\n",
                            i + 1, k + 1 ) ;

                    err = STS_E_FILEOUT ;
                  }
	         }

                fprintf( fout, "\n" ) ;

               } /* end for ( ic ... ) */
               } /* end for ( il ... ) */

               GenFree( kmaxesV ) ;

               if (test == 0) sts = STS_W_EMPTYCLASS ;

 
               } /* end  Format == FORMAT_UF */

               if ( AlgoParaP->Format == FORMAT_FUZZY )
               {

                for ( ipt = 0 ; ( ipt < Npt ) && ( err == STS_OK ) ; ipt ++ )
                {

                 for ( k = 0 ; ( k < nk ) && ( err == STS_OK ) ; k ++ )
                 {
                   if ( fprintf( fout, " %5.3f ",
                              ClassifM[ (ipt * nk) + k ] ) == EOF )
                    {
                    fprintf( stderr, "Cannot write Uik i = %d, k =%d\n",
                            ipt + 1, k + 1 ) ;

                    err = STS_E_FILEOUT ;
                    }
                   } /* end for ( k = 0 ... ) */

                  } /* end for ( ipt ... ) */
                } /* end else Format == FORMAT_FUZZY */

                } /* end (if  K != 1) */

	     break; /* break for case IMAGE */

            default:
	      fprintf( stderr, "Error,the saving of result is only available for spatial and image types\n");
	  }

 return(err);

}






/* ------------------------------------------------------------------- */
static int SaveMSResults
(
 const AlgoParaT*     AlgoParaP,
 const ModelParaT*    ModelParaP,
 const DataT*         DataP,
 const SpatialT*      SpatialP,
 const StatModelT*    ModelP,
 double*              ClassifM,
 double*              LogFikM,
 const int            Npt,
 const int            K,
 const MSCritT*       MSCritP,
 const CriterT*       CriterP,
 FILE*                fmf,
 FILE*                fout
 )
/* ------------------------------------------------------------------- */
{
 StatusET   sts =  STS_OK ;
 StatusET   err =  STS_OK ;
 int        k, d ;           /* class and dim counters */
 int        Nd = DataP->NbVars;
 int        nk = ModelP->Spec.K ;


 /* save the hard or fuzzy classification */
 err = SaveClassif(AlgoParaP,DataP,SpatialP,ModelP,ClassifM, Npt, K,fout);

 /*  save criteria and parameter estimation for model with K classes*/

if (fmf == NULL)
   {
     sts = STS_E_FILEOUT;
     return sts;
   }

  switch( AlgoParaP->Algo )
      {
      case ALGO_GSF:
      case ALGO_MEANF:
      case ALGO_ICM:
	  fprintf( fmf, "\n\n K = %d:  \n",K );
          fprintf( fmf, " LogVcond %4s BICp %3s BICw  \n"," ", " ");
          fprintf( fmf,"%6.0f %4s  %6.0f  %1s %6.0f  \n\n",MSCritP->LogVCond," ",MSCritP->BICPmf, " ", MSCritP->BICWinf);



	   /* Write the proportions of the different classes */
	   if (AlgoParaP->Mode == LOOP)
          {
	     int k;
	     for (k = 0; k < K; k++)
	       fprintf( fmf,"Proportion of sites in class %d : %6.4f\n",
			k, AlgoParaP->True.Prop[k]);
	   }
      	   /* Write the error criterion */
  if (AlgoParaP->True.K == K )
    {
      int k;
      fprintf( fmf, "Global Error = %6.4f \n", AlgoParaP->True.ErrorRate[0]);
      for (k = 1; k <= K; k++)
	fprintf( fmf, "Error on class %d = %6.4f \n",k,
		 AlgoParaP->True.ErrorRate[k]);
    }

  fprintf( fmf, "beta  = %6.4f \n", ModelP->Para.Beta ) ;



	 /*  write different heading according to the model */
         switch( ModelP->Spec.ClassFamily )
           {
           case FAMILY_NORMALGEN:
           case FAMILY_NORMALDIAG:
	     fprintf( fmf, " k  %5s mean  %5s std \n", " ", " ") ;

	       break;

           case FAMILY_LAPLACE:
	       fprintf( fmf, " k   %6s Mu %6s lambda \n", " ", " ") ;
	       break;


           default:
		return -1;
	   }

          /*  write different parameter according to the model */
	 /* for each class */
	  for ( k = 0 ; k < nk ; k ++ )
              {

		 fprintf( fmf," %d",k);

              /* Save mean, and  variance matrix */
              /* V1.03-b*/    /*V1.06-a*/
	  for ( d = 0 ; d < Nd ; d ++ )
               {
                /* print the mean */
		  fprintf( fmf, " %10.3g ", ModelP->Para.Center_KD[ (k*Nd) + d ] ) ;


	      /* print the sqrt of the diag of the covariance matrix */

		  switch( ModelP->Spec.ClassFamily )
	         {
                 case FAMILY_NORMALDIAG:
		     fprintf( fmf, " %10.3g\n",
		       sqrt(ModelP->Para.Disp_KD[k*Nd + d]) ) ;
			break;

	         case FAMILY_NORMALGEN:
		   if ( ModelP->Spec.ClassDisper == DISPER__D)
		     {
	               fprintf( fmf, " %10.3g\n",
		        sqrt(ModelP->Para.Disp_KD[d*(2*Nd +1 -d)/2]) ) ;
		     }
                   if ( ModelP->Spec.ClassDisper == DISPER_KD)
		     {
	               fprintf( fmf, " %10.3g\n",
		        sqrt(ModelP->Para.Disp_KD[k*Nd*(Nd+1)/2 + d*(2*Nd +1 -d)/2  ]) ) ;
		     }
		   break;

	         case FAMILY_LAPLACE:
	               fprintf( fmf, " %10g  ",
                       ModelP->Para.Disp_KD[k*Nd*(Nd+1)/2+ d*(2*Nd +1 -d)/2  ] ) ;
	               break;

	         default:
	               fprintf( fmf, "ClassFamily unknown" ) ;
	         }
	       }

	      }

	  fprintf(fmf, "Potentials for the singletons :\n");
	  for(k = 0; k < nk; k++) {
	    fprintf(fmf,"%d : %f \n", k, ModelParaP->Singl[k]);
	  }
	      break;


	      default:
	/* EM, NEM, NCEM, GEM */
           fprintf( fmf, "\n\n K = %d:  \n",K );
           fprintf( fmf, "LogVCond  %3s BIC  \n", " ");
           fprintf( fmf, "%6.0f  %3s %6.0f \n",
                           MSCritP->LogVCond, " ", MSCritP->BICPmf);

      	   /* Write the error criterion */
            if (AlgoParaP->True.K == K )
            {
             int k;
             fprintf( fmf, "Global Error = %6.4f \n", AlgoParaP->True.ErrorRate[0]);
             for (k = 1; k <= K; k++)
	     fprintf( fmf, "Error on class %d = %6.4f \n",k,
		 AlgoParaP->True.ErrorRate[k]);
	    }



           fprintf( fmf, "beta  = %6.4f \n", ModelP->Para.Beta ) ;

	 /*  write different heading according to the model */
         switch( ModelP->Spec.ClassFamily )
           {
           case FAMILY_NORMALGEN:
           case FAMILY_NORMALDIAG:
	       fprintf( fmf, " k %6s prop %6s mean  %5s std \n", " ", " ", " ") ;

	       break;

           case FAMILY_LAPLACE:
	       fprintf( fmf, " k  %3s prop %6s Mu %6s lambda \n", " ", " ", " ") ;
	       break;

           default:
		return -1;
	   }

         /*  write different parameter according to the model */
          /* for each class */
           for ( k = 0 ; k < nk ; k ++ )
              {

	     fprintf( fmf," %d",k);
             fprintf( fmf, "  %10.3g ", ModelP->Para.Prop_K[ k ] ) ;

            /* Save mean, and  variance matrix */
            for ( d = 0 ; d < Nd ; d ++ )
            {
                fprintf( fmf, " %10.3g ", ModelP->Para.Center_KD[ (k*Nd) + d ] ) ;

	        switch( ModelP->Spec.ClassFamily )
	         {
	         case FAMILY_NORMALGEN:
		   if ( ModelP->Spec.ClassDisper == DISPER__D)
		     {
	               fprintf( fmf, " %10.3g\n",
		        sqrt(ModelP->Para.Disp_KD[d*(2*Nd +1 -d)/2])) ;
		     }
                   if ( ModelP->Spec.ClassDisper == DISPER_KD)
		     {
	               fprintf( fmf, " %10.3g\n",
		        sqrt(ModelP->Para.Disp_KD[k*Nd*(Nd+1)/2 + d*(2*Nd +1 -d)/2  ]) ) ;
		     }
		      break;

                 case FAMILY_NORMALDIAG:
	               fprintf( fmf, " %10.3g\n",
		       sqrt( ModelP->Para.Disp_KD[k*Nd + d] ) ) ;
	               break;

	         case FAMILY_LAPLACE:
	               fprintf( fmf, " %10g\n", ModelP->Para.Disp_KD[k*Nd*(Nd+1)/2 + d*(2*Nd +1 -d)/2] ) ;
	               break;

	         default:
	               fprintf( fmf, " %10g\n", ModelP->Para.Disp_KD[k*Nd*(Nd+1)/2 + d*(2*Nd +1 -d)/2] ) ;
	         }
                }

	      }
      }





    return sts ;



}  /* end of SaveMSResults() */

/* ------------------------------------------------------------------- */
static int      AllocKvar
(
 AlgoParaT*   AlgoParaP,
 StatModelT*  StatModelP,
 VarParamT*   VarParamP,
 const int    K,
 const int    NbVars,
 const int    NbPts
)
/* ------------------------------------------------------------------- */
{
  StatusET  sts = STS_OK;

  const char* func = "AllocKvar";
  /*double*      VarCenter_KD = VarParamP->VarCenter_KD;
    double*      VarDisp_KD = VarParamP->VarDisp_KD;*/

  /* !!! Allocate model parameters */ /*V1.06-a*/

  if ( (StatModelP->Para.Ztilde    = GenAlloc( NbPts * K, sizeof(double),
                                         1, func, "Ztilde")) == NULL )
               return STS_E_MEMORY;
  if ( (StatModelP->Para.Singl    = GenAlloc( K, sizeof(double),
                                         0, func, "Singl")) == NULL )
               return STS_E_MEMORY;

  if ( (StatModelP->Para.Prop_K    = GenAlloc( K, sizeof(double),
					 1, func, "Prop_K" ))== NULL )
               return STS_E_MEMORY ;
  if ( (StatModelP->Para.Center_KD = GenAlloc( K * NbVars, sizeof(double),
					 1, func, "Center_KD" ))== NULL )
               return STS_E_MEMORY ;

  if ( (StatModelP->Para.NbObs_K   = GenAlloc( K, sizeof(double),
					 1, func, "NbObs_K" ))== NULL )
               return STS_E_MEMORY ;
  if ( (StatModelP->Para.NbObs_KD  = GenAlloc( K * NbVars, sizeof(double),
					 1, func, "NbObs_KD" )) == NULL )
               return STS_E_MEMORY;

  if ( (StatModelP->Desc.DispSam_D = GenAlloc( NbVars, sizeof(double),
					  1, func, "DispSam_D" ))== NULL )
               return STS_E_MEMORY;
  if ( (StatModelP->Desc.MiniSam_D = GenAlloc( NbVars, sizeof(double),
					  1, func, "MiniSam_D" ))== NULL )
               return STS_E_MEMORY;
  if ( (StatModelP->Desc.MaxiSam_D = GenAlloc( NbVars, sizeof(double),
					  1, func, "MaxiSam_D" ))== NULL )
               return STS_E_MEMORY;

  /* !!! Allocate the last  model parameters */
  if ( (StatModelP->Spec.ClassFamily == FAMILY_NORMALGEN )
                          && (StatModelP->Spec.ClassDisper == DISPER__D) )
  {
    if ((StatModelP->Para.Disp_KD   = GenAlloc( NbVars* (NbVars+1)/2,
                                    sizeof(double), 1, func, "Disp_KD" )) == NULL )
               return STS_E_MEMORY;
    if ((StatModelP->Para.NbObs_KDD = GenAlloc( NbVars * (NbVars+1)/2,
                                    sizeof(double),  1, func, "NbObs_KDD" )) == NULL )
               return STS_E_MEMORY;
    if ((StatModelP->Para.Iner_KD   = GenAlloc( NbVars* (NbVars+1)/2,
                                    sizeof(double), 1, func, "Iner_KD" ))== NULL )
               return STS_E_MEMORY ;
  }

  else
  {
     if (StatModelP->Spec.ClassFamily == FAMILY_NORMALDIAG )
     {
      if ((StatModelP->Para.Disp_KD   = GenAlloc( K * NbVars,
                                    sizeof(double), 1, func, "Disp_KD" ))== NULL )
               return STS_E_MEMORY ;
      if ((StatModelP->Para.NbObs_KDD = GenAlloc( 1,
                                    sizeof(double),  1, func, "NbObs_KDD" ))== NULL )
               return STS_E_MEMORY ;
      if ((StatModelP->Para.Iner_KD   = GenAlloc( K * NbVars,
                                    sizeof(double), 1, func, "Iner_KD" ))== NULL )
               return STS_E_MEMORY ;

     }
     else
     {
      if ((StatModelP->Para.Disp_KD   = GenAlloc( K * NbVars* (NbVars+1)/2,
                                    sizeof(double), 1, func, "Disp_KD" ))== NULL )
               return STS_E_MEMORY ;
      if ((StatModelP->Para.NbObs_KDD = GenAlloc( K * NbVars * (NbVars+1)/2,
                                    sizeof(double),  1, func, "NbObs_KDD") )== NULL )
               return STS_E_MEMORY ;
      if ((StatModelP->Para.Iner_KD   = GenAlloc( K * NbVars* (NbVars+1)/2,
                                    sizeof(double), 1, func, "Iner_KD" ))== NULL )
               return STS_E_MEMORY ;
     }
  }


  /* initialization of VarParam */
    if ( (AlgoParaP->Algo == ALGO_GSF) )
    {

      if ( (VarParamP->VarCenter_KD=GenAlloc(K* NbVars,
                                   sizeof( double ),1, func, "VarCenter_KD")) == NULL )
        return STS_E_MEMORY;

      if ( (StatModelP->Spec.ClassFamily == FAMILY_NORMALGEN )
                          && (StatModelP->Spec.ClassDisper == DISPER__D) )
      {
                  if ( (VarParamP->VarStd_KD=GenAlloc(NbVars ,
                                  sizeof( double ),1, func, "VarDisp_KD")) == NULL )
                   return STS_E_MEMORY;
      }
      else
      {
                  if (  (VarParamP->VarStd_KD = GenAlloc(K* NbVars ,
					  sizeof( double ),1, func, "VarDisp_KD")) == NULL )
                   return STS_E_MEMORY;
      }
      }

 return sts;

} /* end of AllocKvar() */

/* ------------------------------------------------------------------- */
static void FreeAllocatedData
(
 DataT*       DataP,       /* O and deallocated */
 SpatialT*    SpatialP,    /* O and deallocated */
 StatModelT*  StatModelP,  /* O and deallocated */  /*V1.06-a*/
 AlgoParaT*   AlgoParaP,   /* DataTrue: O and deallocated */
 VarParamT*   VarParamP,   /* O and deallocated */
 CriterT*     CriterP,     /* O and deallocated */  /*V1.06-d*/
 double*       ClassifM,     /* O and deallocated */
 double*       Increase,     /* O and deallocated */
 double*       EBlist     /* O and deallocated */

)
/* ------------------------------------------------------------------- */
{
  int ipt ;  /* counter 0..Npt-1 to free each point neighbors */


  /* Free components of DataP */
  GenFree( DataP->PointsM ) ;     DataP->PointsM    = NULL ;
  GenFree( DataP->LabelV ) ;      DataP->LabelV     = NULL ;
  GenFree( DataP->SiteVisitV ) ;  DataP->SiteVisitV = NULL ;
  GenFree( DataP->SortPos_ND ) ;  DataP->SortPos_ND = NULL ;

  /* Free components of SpatialP */
  switch( SpatialP->Type )
    {
    case TYPE_SPATIAL:
      /* deallocate each point's neighbors */
      for ( ipt = 0; ipt < DataP->NbPts ; ipt ++ )
	GenFree( SpatialP->NeighData.PtsNeighsV[ ipt ].NeighsV ) ;

      /* deallocate array of array of neighbors */
      GenFree( SpatialP->NeighData.PtsNeighsV ) ;

      break ;

    case TYPE_IMAGE:
      /* deallocate neighborhood window */
      GenFree( SpatialP->NeighData.Image.NeighsV ) ;
      break ;

    default: /* non spatial : no allocated spatial info */
             fprintf(stderr, " wrong data format \n");
    }


  /* Free True Data matrix */
 if (AlgoParaP->True.K != 0)
   {
     GenFree(AlgoParaP->True.DataTrue );
     GenFree(AlgoParaP->True.ErrorRate);
     GenFree(AlgoParaP->True.Prop);
   }


  /* Free components of Desc*/
  GenFree( StatModelP->Desc.DispSam_D ) ;
  GenFree( StatModelP->Desc.MiniSam_D ) ;
  GenFree( StatModelP->Desc.MaxiSam_D ) ;

  /* Free components of Para */
  GenFree( StatModelP->Para.Ztilde ) ;
  GenFree( StatModelP->Para.Singl ) ;
  GenFree( StatModelP->Para.Center_KD ) ;
  GenFree( StatModelP->Para.Disp_KD ) ;
  GenFree( StatModelP->Para.Prop_K ) ;

  GenFree( StatModelP->Para.NbObs_K ) ;
  GenFree( StatModelP->Para.NbObs_KD ) ;
  GenFree( StatModelP->Para.NbObs_KDD ) ;
  GenFree( StatModelP->Para.Iner_KD ) ;

  /* Free components of VarParamP */
  GenFree( VarParamP->VarCenter_KD ) ;
  GenFree( VarParamP->VarStd_KD ) ;
  //GenFree( StatModelP->Para.Prop_K );

  /* Free components of CriterP */
  GenFree( CriterP->Errinfo.Refclas_N_Kr  ) ;
  GenFree( CriterP->Errinfo.Perm_Kmfac_Km ) ;
  GenFree( CriterP->Errcur.Agree_Km_Km    ) ;
  GenFree( CriterP->Errcur.Loclas_N_Kc    ) ;

  CriterP->Errinfo.Refclas_N_Kr  = NULL ;
  CriterP->Errinfo.Perm_Kmfac_Km = NULL ;
  CriterP->Errcur.Agree_Km_Km    = NULL ;
  CriterP->Errcur.Loclas_N_Kc    = NULL ;

  /* Free classification matrix */
  GenFree( ClassifM ) ;

  /* Free Increase */
  GenFree( Increase ) ;

  /* Free list of (NRJ,Beta) */
  GenFree( EBlist ) ;

  /* Free the GaussNoise vectors if necessary */
  if ((AlgoParaP->Mode == LOOP) || (AlgoParaP->Opt == AddGaussianNoise))
    {
      GenFree(AlgoParaP->GaussNoise.unique);
      GenFree(AlgoParaP->GaussNoise.stdev);
    }
}

/* ------------------------------------------------------------------- */
int CreateName (
		char* FinalName,
		char* BaseName,
		int   Number,
		char* Ext
		)
/* ------------------------------------------------------------------- */
{
  char NumFile[3];
  sprintf(NumFile,"%03d",Number);
  strcpy(FinalName,BaseName);
  strcat(FinalName,NumFile);
  strcat(FinalName,Ext);

  return(0);
}

/* ------------------------------------------------------------------- */
int ClearParams (
		 double* Singl_SaveP,
		 double* Beta_SaveP,
		 StatModelT* StatModelP
		 )
/* ------------------------------------------------------------------- */
{
  int i;
  *Beta_SaveP            = StatModelP->Para.Beta;
  StatModelP->Para.Beta  = 0;
  for (i = 0; i < StatModelP->Spec.K; i++)
    {
      Singl_SaveP[i]            = StatModelP->Para.Singl[i];
      StatModelP->Para.Singl[i] = 0;
      StatModelP->Para.Center_KD[i] = 0;
      StatModelP->Para.Disp_KD[i] = 0;
    }
  return(0);

}

/* ------------------------------------------------------------------- */
int AllocStatParam(
		   int i,
		   StatParamT* StatParam
		   )
/* ------------------------------------------------------------------- */
{
  StatParam->MeanUk     = GenAlloc(i,sizeof(double),
				   0,"mainfunc","MeanUk");
  StatParam->VarUk      = GenAlloc(i,sizeof(double),
				   0,"mainfunc","VarUk");
  StatParam->MeanSigmaK = GenAlloc(i,sizeof(double),
				   0,"mainfunc","MeanSigmaK");
  StatParam->VarSigmaK  = GenAlloc(i,sizeof(double),
				   0,"mainfunc","VarSigmaK");
  StatParam->MeanPotK   = GenAlloc(i,sizeof(double),
				   0,"mainfunc","MeanPotK");
  StatParam->VarPotK    = GenAlloc(i,sizeof(double),
				   0,"mainfunc","VarPotK");
  StatParam->MeanErr    = GenAlloc(i+1,sizeof(double),
				   0,"mainfunc","MeanErr");
  StatParam->VarErr     = GenAlloc(i+1,sizeof(double),
				   0,"mainfunc","VarErr");
  StatParam->MeanProp   = GenAlloc(i,sizeof(double),
				   0,"mainfunc","MeanProp");
  StatParam->VarProp    = GenAlloc(i,sizeof(double),
				   0,"mainfunc","VarProp");
  return 0;
}

/* ------------------------------------------------------------------- */
int FreeStatParam(
		   StatParamT* StatParam
		   )
/* ------------------------------------------------------------------- */
{
 /* Free stat params */
 GenFree(StatParam->MeanUk ) ;
 GenFree(StatParam->VarUk ) ;
 GenFree(StatParam->MeanPotK ) ;
 GenFree(StatParam->VarPotK ) ;
 GenFree(StatParam->MeanErr ) ;
 GenFree(StatParam->VarErr ) ;
 GenFree(StatParam->MeanProp ) ;
 GenFree(StatParam->VarProp ) ;
 GenFree(StatParam->MeanSigmaK ) ;
 GenFree(StatParam->VarSigmaK ) ;

 return 0;
}


/* ------------------------------------------------------------------- */
int InitializeStatParam(
			int i,
			StatParamT* StatParam
			)
/* ------------------------------------------------------------------- */
{
  int j;
  for (j = 0 ; j < i ; j++)
    {
      StatParam->MeanUk    [j] = 0;
      StatParam->VarUk     [j] = 0;
      StatParam->MeanSigmaK[j] = 0;
      StatParam->VarSigmaK [j] = 0;
      StatParam->MeanPotK  [j] = 0;
      StatParam->VarPotK   [j] = 0;
      StatParam->MeanProp  [j] = 0;
      StatParam->VarProp   [j] = 0;
    }
  for(j = 0; j < i + 1; j++)
    {
      StatParam->MeanErr [j] =  0;
      StatParam->VarErr  [j] =  0;
    }
  StatParam->MeanBeta  = 0;
  StatParam->VarBeta   = 0;
  return 0;
}

/* -------------------------------------------------------------------- */
int FinalStatParams(
		    int i,
		    int nbIters,
		    StatParamT* StatParam
		    )
/* -------------------------------------------------------------------- */
{
  int j;
  for (j = 0; j < i ; j++)
    {
      StatParam->VarUk     [j] =
	(StatParam->VarUk[j] - pow(StatParam->MeanUk[j],2)
	 / nbIters) / nbIters;
      StatParam->MeanUk    [j] = StatParam->MeanUk[j] / nbIters;
      StatParam->VarSigmaK [j] =
	(StatParam->VarSigmaK[j] - pow(StatParam->MeanSigmaK[j],2)
	 /nbIters) / nbIters;
      StatParam->MeanSigmaK[j] = StatParam->MeanSigmaK[j] / nbIters;
      StatParam->VarPotK   [j] =
	(StatParam->VarPotK[j] - pow(StatParam->MeanPotK[j],2)
	 / nbIters) / nbIters;
      StatParam->MeanPotK  [j] = StatParam->MeanPotK[j] / nbIters;
      StatParam->VarProp   [j] =
	(StatParam->VarProp[j] - pow(StatParam->MeanProp[j],2)
	 / nbIters) / nbIters;
      StatParam->MeanProp  [j] = StatParam->MeanProp[j] / nbIters;
    }
  for (j = 0; j < i + 1; j++)
    {
      StatParam->VarErr[j]    = (StatParam->VarErr[j] - pow(StatParam->MeanErr[j],2)/nbIters) / nbIters;
      StatParam->MeanErr[j]   = StatParam->MeanErr[j] / nbIters;
    }
  StatParam->VarBeta   =
    (StatParam->VarBeta - pow(StatParam->MeanBeta,2)
     / nbIters) / nbIters;
  StatParam->MeanBeta  = StatParam->MeanBeta / nbIters;
  return 0;
}

/* ------------------------------------------------------------------- */
int SaveStatParams(
		   int        nbClasses,
                   int        Kmin,
                   int        Kmax,
		   FILE*      fsf,
		   MSparamT*  MSparam,
                   AlgoET     Algo,
		   SinglET    SinglMode
		   )
/* ------------------------------------------------------------------- */
{
  int j;
  char *AlgoStr[ ALGO_NB ]   = {"meanf",  "gsf",
				"icm","nem","ncem"};
  char *SinglStr[ SINGL_NB ] = {"fix" , "zeroin", "none","em","beta"} ;

  StatParamT*   StatParamP = &(MSparam->StatParam[nbClasses -
						      MSparam->Kmin]);
  fprintf(fsf, "Type of algorithm : %s\n", AlgoStr[Algo]);
  fprintf(fsf, "Type of potential estimation : %s\n\n", SinglStr[SinglMode]);
  /* Write all the statistics if Kmin <true K < Kmax (otherwise, not estimated) */
  if ((nbClasses >= Kmin) &&(nbClasses<=Kmax))
  {
    for (j = 0; j < nbClasses ; j++)
    {
      fprintf(fsf,"Class %d :\n",j);
      fprintf(fsf," Mean Uk : %15.10f\n",StatParamP->MeanUk[j]);
      fprintf(fsf," Sig  Uk : %15.10f\n",sqrt(StatParamP->VarUk[j]));
      fprintf(fsf," Mean SigmaK : %15.10f\n",StatParamP->MeanSigmaK[j]);
      fprintf(fsf," Sig  SigmaK : %15.10f\n",
	      sqrt(StatParamP->VarSigmaK[j]));
      fprintf(fsf," Mean PotK   : %15.10f\n",StatParamP->MeanPotK[j]);
      fprintf(fsf," Sig  PotK   : %15.10f\n",
	      sqrt(StatParamP->VarPotK[j]));
      fprintf(fsf," Mean Err    : %15.10f\n",StatParamP->MeanErr[j+1]);
      fprintf(fsf," Sig  Err    : %15.10f\n",sqrt(StatParamP->VarErr[j+1]));
      fprintf(fsf," Mean Prop   : %15.10f\n",StatParamP->MeanProp[j]);
      fprintf(fsf," Sig  Prop   : %15.10f\n\n",sqrt(StatParamP->VarProp[j]));
    }
  fprintf(fsf,"Mean Beta : %15.10f\n",StatParamP->MeanBeta);
  fprintf(fsf,"Sig  Beta : %15.10f\n",sqrt(StatParamP->VarBeta));
  fprintf(fsf,"MeanErr   : %15.10f\n",StatParamP->MeanErr[0]);
  fprintf(fsf,"SigErr    : %15.10f\n",sqrt(StatParamP->VarErr[0]));
  }
  else fprintf(fsf," True K not in [Kmin,Kmax] so true parameters were not estimated\n\n");

     /*write all the criteria selection */
     switch(Algo)
	{
        case ALGO_NEM:
        case ALGO_NCEM:

        fprintf(fsf,"\n\n\n");
        fprintf(fsf,"%9s"," ");
	for (j = Kmin; j <=Kmax ; j++)
            fprintf(fsf,"K = %d %3s ",j," ");

        fprintf(fsf,"\n");
        fprintf(fsf,"BIC %2s"," ");

        for (j = Kmin; j <=Kmax ; j++)
            fprintf(fsf,"%3.5f %2s",MSparam->BICp[j - Kmin]," ");

        fprintf(fsf,"\n");
        

	  break ;

        case ALGO_MEANF:
        case ALGO_GSF:
        case ALGO_ICM:

        fprintf(fsf,"\n\n\n");
        fprintf(fsf,"%9s"," ");
	for (j = Kmin; j <=Kmax ; j++)
            fprintf(fsf,"K = %d %3s ",j," ");

        fprintf(fsf,"\n");
        fprintf(fsf,"BICw %2s"," ");

        for (j = Kmin; j <=Kmax ; j++)
            fprintf(fsf,"%3.5f %2s",MSparam->BICw[j - Kmin]," ");

        fprintf(fsf,"\n");
        fprintf(fsf,"BICp %2s"," ");

        for (j = Kmin; j <=Kmax ; j++)
            fprintf(fsf,"%3.5f %2s",MSparam->BICp[j - Kmin]," ");

        fprintf(fsf,"\n");

	break ;

        default :
		return -1;

	}

     return(0);
}



/* ~~~~~~~~~~~~~~~~~~~~~~~~ END OF FILE ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */



