/*\

    NEM_NOI.C

    Programme SEMMS (Spatial EM for Markovian Segmentation: routines calcul parametres classes

    Van Mo DANG       Janvier 96


    Vers-mod  Date         Description

    1.03-a    01-NOV-1996  Add function SetIdMatrix + calls by ParaPk.. 
                           to output variance matrix in file don.mf
    1.04-a    08-JAN-1998  Fixed error in DensPkVkI log(vk) needs * Nd
    1.05-a    12-JAN-1998  MissMode in ParaP*V*I
    1.05-b    12-JAN-1998  Process missing data in DensPkVkI
    1.05-c    12-JAN-1998  Add ParaPkV*I_Missing() (empty)
    1.05-d    15-JAN-1998  Add CommonGaussDiagMissing() <= ParaPkV*I_Missing()
    1.05-e    26-JAN-1998  Call to GenAlloc
\*/

#include "genmemo.h"    /* GenAlloc */
#include "semms_typ.h"    /* DataT, ... */
#include "semms_noi.h"    /* ParaP_V_I, ... */
#include <stdio.h>      /* printf, ... */
#include <stdlib.h>     /* malloc, ... */
#include <string.h>     /* memcpy, ... */
#include <math.h>       /* exp, ... */
#include <values.h>     /* MAXDOUBLE */


#define TWO_PI    2 * 3.14159 

/*V1.05-d*/
#define _IJ       ( ( i * D ) + j )    /* access a (.,D) matrix by (i,j) */
#define _HJ       ( ( h * D ) + j )    /* access a (.,D) matrix by (h,j) */
#define _IH       ( ( i * K ) + h )    /* access a (.,K) matrix by (i,h) */
#define sqr(x)    ((x)*(x))            /* macro for x^2 */


/* static int      ParaPkV_I_Complete          /* ret : nb of empty classes */
/* 		   (
/* 		     const double       *UM,        /* I : classification matrix */
/* 		     const DataT       *DataP,     /* I : observed points */
/* 		     int               Nk,         /* I */
/* 		     MissET            MissMode,   /* I */
/* 		     NoiseParaT        *NoiseParaP /* O : estimated parameters */
/* 		   ) ;
/* 
/* static int      ParaPkV_I_Missing           /* ret : nb of empty classes */
/* 		   (
/* 		     const double       *UM,        /* I : classification matrix */
/* 		     const DataT       *DataP,     /* I : observed points */
/* 		     int               Nk,         /* I */
/* 		     MissET            MissMode,   /* I */
/* 		     NoiseParaT        *NoiseParaP /* O : estimated parameters */
/* 		   ) ;
/* 
/* 
/* static int      ParaPkVkI_Complete          /* ret : nb of empty classes */
/* 		   (
/* 		     const double       *UM,        /* I : classification matrix */
/* 		     const DataT       *DataP,     /* I : observed points */
/* 		     int               Nk,         /* I */
/* 		     MissET            MissMode,   /* I */
/* 		     NoiseParaT        *NoiseParaP /* O : estimated parameters */
/* 		   ) ;
/* 
/* static int      ParaPkVkI_Missing           /* ret : nb of empty classes */
/* 		   (
/* 		     const double       *UM,        /* I : classification matrix */
/* 		     const DataT       *DataP,     /* I : observed points */
/* 		     int               Nk,         /* I */
/* 		     MissET            MissMode,   /* I */
/* 		     NoiseParaT        *NoiseParaP /* O : estimated parameters */
/* 		   ) ;
 */

static EstimNoiseFT ParaPkV_I_Complete ;
static EstimNoiseFT ParaPkV_I_Missing ;
static EstimNoiseFT ParaPkVkI_Complete ;
static EstimNoiseFT ParaPkVkI_Missing ;


static void SetIdMatrix ( int Nd , double* M ) ;  /*V1.03-a*/


/*V1.05-d*/
static StatusET          /* Return status : OK, EMPTY or MEMORY */
CommonGaussDiagMissing
 (
  const double*  Xij_nd,        /* I : data matrix (N,D) */
  int           N,             /* I : nb of observation vectors */
  int           D,             /* I : nb of variables */
  const double*  Cih_nk,        /* I : classification matrix (N,K) */
  int           K,             /* I : number of clusters */
  MissET        Miss,          /* I : how to treat missing data */
  const double*  OldVh_k,       /* I : old volumes (K) */
  const double*  OldChjj_kdd,   /* I : old normalized covariances (K,D,D) */

  double*        OldNewMhj_kd,  /* I/O : old then updated means (K,D) */
  double*        NewP_k,        /* O : updated proportions (K) */
  int*          EmptyK_P,      /* O : index of empty class (0 or 1 ..K) */
  double*        Nh_k,          /* O : size of a class (K) */
  double*        Nhj_kd,        /* O : size of a class and variable (K,D) */
  double*        Disphj_kd      /* O : dispersion of a class/variable (K,D) */
 ) ; 





/* ------------------------------------------------------------------- */
int DensPkVkI          /* ret : 0 if OK, -1 if zero density */
        (
            int                Nd,         /* I : point dimension */
            int                Ik,         /* I : class number : 0..Nk-1 */
            const NoiseParaT*  NoiseP,     /* I : noise parameters */
            const double*       XV,         /* I : point (dim d) */
            double*             FkP,        /* O : density for class Ik */
            double*             LogFkP      /* O : log of density */
        )
/* ------------------------------------------------------------------- */
{
    int     d ;     /* 0..Nd-1 : current variable */
    double   dk ;    /* >= 0.0  : squared distance of XV to mean Ik */
    int     nbobs ; /* 0..Nd-1 : nb of observed variables */ /*V1.05-b*/
    double   vk = NoiseP->Vk[ Ik ] ;

    /* Gaussian mixture density, model [ Pk , Vk . I ] :

       Complete data :
       fk(x) = [(2*pi * vk )^(-Nd * 0.5) * exp( -0.5 * dist_y_mk )
       where  dist_y_mk = sum[d=1,Nd]( (y(d) - mk(d))^2 ) / vk

       Incomplete data :
       fk(x) = [(2*pi * vk )^(-Card(oi) * 0.5) * exp( -0.5 * dist_y_mk )
       where   dist_y_mk = sum[d in oi]  (y(d) - mk(d))^2 ) / vk

    */

    for ( d = 0, dk = 0.0, nbobs = 0 ; d < Nd ; d ++ )
      {
	if ( ! isnan( XV[ d ] ) )   /*V1.05-b*/
	  {
	    double dif = XV[ d ] - NoiseP->Mk[ (Ik * Nd) + d ] ;

	    dk = dk + (dif * dif) ;
	    nbobs ++ ;
	  }
    }

    if ( vk > 0 )
      {
	/*V1.04-a*//*V1.05-b*/
        *LogFkP = -0.5 * ( nbobs * log( TWO_PI * vk ) +  dk / vk ) ;
        *FkP = exp( *LogFkP ) ;
        return 0 ;
      }
    else
      {
        *LogFkP = - MAXDOUBLE ;
        *FkP = 0.0 ;
        return -1 ;
      }


}   /* end of DensPkVkI() */



/* ------------------------------------------------------------------- */
StatusET                    /* ret : OK, W_EMPTYCLASS or E_MEMORY *//*V1.05-c*/
ParaP_V_I
 (
  const double   *Cih_nk,    /* I : classification matrix (N,K) */
  const DataT   *DataP,     /* I : observed points */
  int           K,          /* I : number of classes */
  MissET        MissMode,   /* I : how to treat missing data */ /*V1.05-b*/

  int           *EmptyK_P,  /* O : which empty class (1..K) or 0 *//*V1.05-c*/
  NoiseParaT    *NoiseParaP /* O : estimated parameters */
 ) 
/* ------------------------------------------------------------------- */
{
  int      k ;
  StatusET sts ;


  sts = ParaPkV_I( Cih_nk, DataP, K, MissMode, EmptyK_P, NoiseParaP ) ;

  for ( k = 0 ; k < K ; k ++ )
    {
        NoiseParaP->Pk[ k ] = 1.0 / K ;
    }
  return sts ;
}

/* ------------------------------------------------------------------- */
StatusET ParaPkV_I          /* ret : OK, W_EMPTYCLASS or E_MEMORY *//*V1.05-c*/
 (
  const double   *Cih_nk,    /* I : classification matrix (N,K) */
  const DataT   *DataP,     /* I : observed points */
  int           K,          /* I : number of classes */
  MissET        MissMode,   /* I : how to treat missing data */ /*V1.05-b*/

  int           *EmptyK_P,  /* O : which empty class (1..K) or 0 *//*V1.05-c*/
  NoiseParaT    *NoiseParaP /* O : estimated parameters */
 ) 
/* ------------------------------------------------------------------- */
{

    if ( DataP->NbMiss == 0 )
      return ParaPkV_I_Complete( Cih_nk, DataP, K, MissMode, 
				 EmptyK_P, NoiseParaP ) ;
    else
      return ParaPkV_I_Missing( Cih_nk, DataP, K, MissMode, 
				EmptyK_P, NoiseParaP ) ;

}   /* end of ParaPkV_I() */


/* ------------------------------------------------------------------- */
static StatusET             /* ret : OK, W_EMPTYCLASS or E_MEMORY *//*V1.05-c*/
ParaPkV_I_Complete
 (
  const double   *Cih_nk,    /* I : classification matrix (N,K) */
  const DataT   *DataP,     /* I : observed points */
  int           K,          /* I : number of classes */
  MissET        MissMode,   /* I : how to treat missing data */ /*V1.05-b*/

  int           *EmptyK_P,  /* O : which empty class (1..K) or 0 *//*V1.05-c*/
  NoiseParaT    *NoiseParaP /* O : estimated parameters */
 ) 
/* ------------------------------------------------------------------- */
{
    int         h ;         /* class counter : 0 .. K - 1 */
    double       sumh_trTh ; /* sum over h of class h's dispersion matrix's trace */
    double       V ;         /* common volume = sum[h] trTh / (N * Nd) */
    int         nd  = DataP->NbVars ;   /* Nd : nb of variables */
    int         npt = DataP->NbPts ;    /* N : nb of points */

    /* For each class */
    for ( h = 0 , (*EmptyK_P) = 0, sumh_trTh = 0.0 ; 
          h < K ; 
          h ++ )
    {
        int     ipt ;
        int     d ;
        double   sumi_uih ;
        double   sumi_uih_yi2 ;

        /* 'fuzzy' cardinal : sizeh = sum[i]( uih ) */
        /* proportion :       ph    = 1 / K */
        /* mean :             mh(d) = sum[i]( uih * yi(d) ) / sizeh */
        /* trace of its dispersion matrix : 
           trTh = sum[i]( uih * sum[d] yi(d)? ) - sizeh * sum[d] mh(d)? */
        /* volume :           Vh = V = sum[h] trTh / (N * Nd) */

        /* Initialize means to zero */
        for ( d = 0 ; d < nd ; d ++ )
        {
            NoiseParaP->Mk[ ( h * nd ) + d ] = 0 ;
        }

        /* Do cumulated sums over all points */
        for ( ipt = 0, sumi_uih = 0.0, sumi_uih_yi2 = 0.0 ; 
              ipt < npt ; 
              ipt ++ )
        {
            double   sumd_uih_yid2 ;
            double   uih                             = Cih_nk[ ( ipt * K ) + h ] ;

            for ( d = 0, sumd_uih_yid2 = 0.0 ; 
                  d < nd ; 
                  d ++ )
            {
                double   yid   = DataP->PointsM[ (ipt * nd) + d ] ;

                NoiseParaP->Mk[ ( h * nd ) + d ]    += uih * yid ;
                sumd_uih_yid2                       += uih * yid * yid ;
            }

            sumi_uih        += uih ;
            sumi_uih_yi2    += sumd_uih_yid2 ;
        }

        /* Normalize the sums by the cardinal of the class */
        if ( sumi_uih != 0.0 )
        {
            double   sumd_mhd2 ;
            double   trTh ;
            double   invsizeh = 1 / sumi_uih ;

            for ( d = 0, sumd_mhd2 = 0.0 ; 
                  d < nd ; 
                  d ++ )
            {
                double   mhd         = NoiseParaP->Mk[ (h * nd) + d ] ;

                mhd                 *= invsizeh ;
                sumd_mhd2           += mhd * mhd ;

                NoiseParaP->Mk[ (h * nd) + d ]  = mhd ;
            }

            trTh                    = sumi_uih_yi2 - sumi_uih * sumd_mhd2 ;
            sumh_trTh               += trTh ;
        }
        else
        {
            (*EmptyK_P) = h + 1 ;
        }

        NoiseParaP->Pk[ h ] = sumi_uih / npt ;

        /* Set normalized variance matrix to identity */
        SetIdMatrix( nd , & NoiseParaP->Ck[ h*nd*nd ] ) ;  /*V1.03-a*/

    } /* end for ( h = 0 , (*EmptyK_P) = 0 ; h < K ; h ++ ) */

    V = sumh_trTh / ( npt * nd ) ;

    for ( h = 0 ; h < K ; h ++ )
    {
        NoiseParaP->Vk[ h ] = V ;
    }

    if ( (*EmptyK_P) == 0 )
      return STS_OK ;
    else
      return STS_W_EMPTYCLASS ;

}   /* end of ParaPkV_I_Complete() */



/* ------------------------------------------------------------------- */
static StatusET             /* ret : OK, W_EMPTYCLASS or E_MEMORY *//*V1.05-c*/
ParaPkV_I_Missing
 (
  const double   *Cih_nk,    /* I : classification matrix (N,K) */
  const DataT   *DataP,     /* I : observed points */
  int           K,          /* I : number of classes */
  MissET        MissMode,   /* I : how to treat missing data */ /*V1.05-b*/

  int           *EmptyK_P,  /* O : which empty class (1..K) or 0 *//*V1.05-c*/
  NoiseParaT    *NoiseParaP /* O : estimated parameters */
 ) 
/* ------------------------------------------------------------------- */
{
  const char* func = "ParaPkV_I_Missing" ;

  int      N = DataP->NbPts ;
  int      D = DataP->NbVars ;

  int      h ;         /* current class    0..K-1 */
  int      j ;         /* current variable 0..D-1 */

  double*   nh_k ;      /* size / class (K) local alloc (unused) */
  double*   nhj_kd ;    /* size / obs class & var (K,D) local alloc (unused)*/
  double*   disphj_kd ; /* dispersion / class & var (K,D) local alloc */
  double    vol ;       /* common volume */

  StatusET sts ;


  /* Allocate local data */
  nh_k      = GenAlloc( K,     sizeof( double ), 1, func, "nh_k" ) ;
  nhj_kd    = GenAlloc( K * D, sizeof( double ), 1, func, "nhj_kd" ) ;
  disphj_kd = GenAlloc( K * D, sizeof( double ), 1, func, "disphj_kd" ) ;

  /* Common computation for any diagonal gaussian model with missing data */
  sts = CommonGaussDiagMissing
    ( DataP->PointsM, DataP->NbPts, DataP->NbVars, Cih_nk, K, MissMode,
      NoiseParaP->Vk, NoiseParaP->Ck, 
      NoiseParaP->Mk, NoiseParaP->Pk, EmptyK_P, nh_k, nhj_kd, disphj_kd ) ;

  /* Compute common volume : 
     sum dispersions over the classes and variables 
     then divide by all/observed data size */
  
  vol = 0.0 ;

  for ( h = 0; h < K ; h ++ )
    {
      /* If not empty class */
      if ( nh_k[ h ] > 0 )
	{
	  for ( j = 0 ; j < D ; j ++ )
	    {
	      vol += disphj_kd[ _HJ ] ;
	    }
	}
    }

  if ( MissMode == MISSING_REPLACE )
    vol /= ( N * D ) ;
  else
    vol /= ( N * D - DataP->NbMiss ) ;

  /* For each class h */
  for ( h = 0; h < K ; h ++ )
    {
      /* Set its volume to the common volume */
      NoiseParaP->Vk[ h ] = vol ;

      /* Set its normalized covariance to identity */
      SetIdMatrix( D, & NoiseParaP->Ck[ h * D * D ] ) ;
    }
      
  /* Free local data */
  GenFree( disphj_kd ) ;
  GenFree( nhj_kd ) ;
  GenFree( nh_k ) ;

  return sts ;
}   /* end of ParaPkV_I_Missing() */



/* ------------------------------------------------------------------- */
StatusET             /* ret : OK, W_EMPTYCLASS or E_MEMORY *//*V1.05-c*/
ParaP_VkI
 (
  const double   *Cih_nk,    /* I : classification matrix (N,K) */
  const DataT   *DataP,     /* I : observed points */
  int           K,          /* I : number of classes */
  MissET        MissMode,   /* I : how to treat missing data */ /*V1.05-b*/

  int           *EmptyK_P,  /* O : which empty class (1..K) or 0 *//*V1.05-c*/
  NoiseParaT    *NoiseParaP /* O : estimated parameters */
 ) 
/* ------------------------------------------------------------------- */
{
  StatusET sts ;
  int k ;

  sts = ParaPkVkI( Cih_nk, DataP, K, MissMode, EmptyK_P, NoiseParaP ) ;

  for ( k = 0 ; k < K ; k ++ )
    {
        NoiseParaP->Pk[ k ] = 1.0 / K ;
    }
  return sts ;
}


/* ------------------------------------------------------------------- */
StatusET             /* ret : OK, W_EMPTYCLASS or E_MEMORY *//*V1.05-c*/
ParaPkVkI
 (
  const double   *Cih_nk,    /* I : classification matrix (N,K) */
  const DataT   *DataP,     /* I : observed points */
  int           K,          /* I : number of classes */
  MissET        MissMode,   /* I : how to treat missing data */ /*V1.05-b*/

  int           *EmptyK_P,  /* O : which empty class (1..K) or 0 *//*V1.05-c*/
  NoiseParaT    *NoiseParaP /* O : estimated parameters */
 ) 
/* ------------------------------------------------------------------- */
{

    if ( DataP->NbMiss == 0 )
      return ParaPkVkI_Complete( Cih_nk, DataP, K, MissMode, 
				 EmptyK_P, NoiseParaP ) ;
    else
      return ParaPkVkI_Missing( Cih_nk, DataP, K, MissMode, 
				EmptyK_P, NoiseParaP ) ;

}   /* end of ParaPkV_I() */



/* ------------------------------------------------------------------- */
static StatusET             /* ret : OK, W_EMPTYCLASS or E_MEMORY *//*V1.05-c*/
ParaPkVkI_Complete
 (
  const double   *Cih_nk,    /* I : classification matrix (N,K) */
  const DataT   *DataP,     /* I : observed points */
  int           K,          /* I : number of classes */
  MissET        MissMode,   /* I : how to treat missing data */ /*V1.05-b*/

  int           *EmptyK_P,  /* O : which empty class (1..K) or 0 *//*V1.05-c*/
  NoiseParaT    *NoiseParaP /* O : estimated parameters */
 ) 
/* ------------------------------------------------------------------- */
{
    int         k ;

    /* For each class */
    for ( k = 0 , (*EmptyK_P) = 0 ; k < K ; k ++ )
    {
        int     ipt ;                  
        int     id ;
        int     nd  = DataP->NbVars ;
        int     npt = DataP->NbPts ;    /* = n nb points */
        double   sizek = 0.0 ;            /* = nk fuzzy cardinal [0,n] */
        double   trWk = 0.0 ;

        /* Compute its 'fuzzy' cardinal : nk = sum[i=1,n]( uik ) */
        /* Compute its proportion : pk = nk / n */
        /* Compute its mean (mk) : mk = sum[i=1,n]( uik * Y(i,:) ) / nk */
        /* Compute its intra-class inertia matrix : 
           Wk = sum[i=1,n]( uik * ( Y(i,:) - mk ) * ( Y(i,:) - mk )' ) / nk */
        /* Compute its volume : Vk = tr(Wk) / d */

        /* Initialize means to zero */
        for ( id = 0 ; id < nd ; id ++ )
        {
            NoiseParaP->Mk[ (k * nd) + id ] = 0 ;
        }

        /* Do cumulated sums over all points */
        for ( ipt = 0 ; ipt < npt ; ipt ++ )
        {
            double   uik = Cih_nk[ (ipt * K) + k ] ;

            sizek += uik ;
            for ( id = 0 ; id < nd ; id ++ )
            {
                double   yid = DataP->PointsM[ (ipt * nd) + id ] ;
                double   mkd = NoiseParaP->Mk[ (k * nd) + id ] ;

                mkd         = mkd + uik * yid ;
                trWk        = trWk + uik * yid * yid ;

                NoiseParaP->Mk[ (k * nd) + id ] = mkd ;
            }
        }

        /* Normalize the sums by the cardinal of the class */
        if ( sizek != 0.0 )
        {
            trWk /= sizek ;

            for ( id = 0 ; id < nd ; id ++ )
            {
                double   mkd = NoiseParaP->Mk[ (k * nd) + id ] ;

                mkd         = mkd / sizek ;
                trWk        = trWk - ( mkd * mkd ) ;

                NoiseParaP->Mk[ (k * nd) + id ] = mkd ;
            }
            NoiseParaP->Vk[ k ] = trWk / nd ;
        }
        else
        {
            (*EmptyK_P) = k + 1 ;
        }

        NoiseParaP->Pk[ k ] = sizek / npt ;

        /* Set normalized variance matrix to identity */
        SetIdMatrix( nd , & NoiseParaP->Ck[ k*nd*nd ] ) ;  /*V1.03-a*/

    } /* end for ( k = 0 , (*EmptyK_P) = 0 ; k < K ; k ++ ) */

    if ( (*EmptyK_P) == 0 )
      return STS_OK ;
    else
      return STS_W_EMPTYCLASS ;

}   /* end of ParaPkVkI_Complete() */


/* ------------------------------------------------------------------- */
static StatusET             /* ret : OK, W_EMPTYCLASS or E_MEMORY *//*V1.05-c*/
ParaPkVkI_Missing
 (
  const double   *Cih_nk,    /* I : classification matrix (N,K) */
  const DataT   *DataP,     /* I : observed points */
  int           K,          /* I : number of classes */
  MissET        MissMode,   /* I : how to treat missing data */ /*V1.05-b*/

  int           *EmptyK_P,  /* O : which empty class (1..K) or 0 *//*V1.05-c*/
  NoiseParaT    *NoiseParaP /* O : estimated parameters */
 ) 
/* ------------------------------------------------------------------- */
{
  const char* func = "ParaPkVkI_Missing" ;

  int      N = DataP->NbPts ;
  int      D = DataP->NbVars ;

  int      h ;         /* current class    0..K-1 */
  int      j ;         /* current variable 0..D-1 */

  double*   nh_k ;      /* size / class (K) local alloc */
  double*   nhj_kd ;    /* size / obs class and variable (K,D) local alloc */
  double*   disphj_kd ; /* dispersion / class and variable (K,D) local alloc */

  double    sumj_nhj ;  /* size of obs class */

  StatusET sts ;


  /* Allocate local data */
  nh_k      = GenAlloc( K,     sizeof( double ), 1, func, "nh_k" ) ;
  nhj_kd    = GenAlloc( K * D, sizeof( double ), 1, func, "nhj_kd" ) ;
  disphj_kd = GenAlloc( K * D, sizeof( double ), 1, func, "disphj_kd" ) ;

  /* Common computation for any diagonal gaussian model with missing data */
  sts = CommonGaussDiagMissing
    ( DataP->PointsM, DataP->NbPts, DataP->NbVars, Cih_nk, K, MissMode,
      NoiseParaP->Vk, NoiseParaP->Ck, 
      NoiseParaP->Mk, NoiseParaP->Pk, EmptyK_P, nh_k, nhj_kd, disphj_kd ) ;

  /* For each class h */
  for ( h = 0; h < K ; h ++ )
    {
      /* If not empty class */
      if ( nh_k[ h ] > 0 )
	{
	  /* Reestimate its volume: sum dispersions over the variables 
	     then divide by all/observed data size */
	  sumj_nhj = 0 ;
	  NoiseParaP->Vk[ h ] = 0.0 ;  /* use as sum first */

	  for ( j = 0 ; j < D ; j ++ )
	    {
	      sumj_nhj            += nhj_kd[ _HJ ] ;
	      NoiseParaP->Vk[ h ] += disphj_kd[ _HJ ] ;
	    }

	  if ( MissMode == MISSING_REPLACE )
	    NoiseParaP->Vk[ h ] /= nh_k[ h ] ;
	  else
	    NoiseParaP->Vk[ h ] /= sumj_nhj ;
	}
      else /* class h is empty */
	{
	  NoiseParaP->Vk[ h ] = 0 ;
	}

      /* Set its normalized covariance to identity */
      SetIdMatrix( D, & NoiseParaP->Ck[ h * D * D ] ) ;
    }

  /* Free local data */
  GenFree( disphj_kd ) ;
  GenFree( nhj_kd ) ;
  GenFree( nh_k ) ;

  return sts ;
}



/*V1.05-d*/
/* ------------------------------------------------------------------- */
static StatusET                /* Returns : OK, W_EMPTYCLASS or E_MEMORY */
CommonGaussDiagMissing
 (
  const double*  Xij_nd,        /* I : data matrix (N,D) */
  int           N,             /* I : nb of observation vectors */
  int           D,             /* I : nb of variables */
  const double*  Cih_nk,        /* I : classification matrix (N,K) */
  int           K,             /* I : number of clusters */
  MissET        Miss,          /* I : how to treat missing data */
  const double*  OldVh_k,       /* I : old volumes (K) */
  const double*  OldChjj_kdd,   /* I : old normalized covariances (K,D,D) */

  double*        OldNewMhj_kd,  /* I/O : old then updated means (K,D) */
  double*        NewPh_k,       /* O : updated proportions (K) */
  int*          EmptyK_P,      /* O : index of empty class (0 or 1 ..K) */
  double*        Nh_k,          /* O : size of a class (K) */
  double*        Nhj_kd,        /* O : size of a class and variable (K,D) */
  double*        Disphj_kd      /* O : dispersion of a class/variable (K,D) */
 )
/*\

   This function does the common computation that is needed at the
   M-step, in all the diagonal Gaussian models with missing data.

   The routine takes as input a data matrix Xij_nd (possibly
   containing NaN values), a classification matrix Cih_nk, and the
   kind of missing data processing Miss (REPLACE or IGNORE) ; and also
   the old volumes OldVh_k, covariance matrices OldChjj_kdd, and means
   OldNewMhj_kd of the classes, to be used in the REPLACE mode.

   It produces as output the reestimated mean vectors in OldNewMhj_kd;
   reestimated proportions in NewP_k; the last encountered empty class
   in EmptyK_P (or 0 if no empty class); the {sum[i] cih} in Nh_k, the
   {sum[i] cih rij} in Nhj_kd (number of present observations in class h
   and variable j); and the "missing data dispersion" of each
   class/variable in Disphj_kd (i.e. the missing data equivalent of
   {sum[i] Cih (xij - mhj)^2}).

\*/
/* ------------------------------------------------------------------- */
{
  const char* func = "CommonGaussDiagMissing" ;

  double*   sumdata_hj_kd ;    /* sum[i] cih rij xij   (K,D), local alloc */
  double*   sumsquare_hj_kd ;  /* sum[i] cih rij xij^2 (K,D), local alloc */
  double*   oldmean_hj_kd ;    /* saved previous means (K,D), local alloc */
  int      h ;                /* current class    0..K-1 */
  int      j ;                /* current variable 0..D-1 */
  int      i ;                /* current object   0..N-1 */
  StatusET sts ;              /* return status */


  sts = STS_OK ;

  /* Allocate local structures */
  sumdata_hj_kd   = GenAlloc( K * D, sizeof(double), 1, func, "sumdata" ) ;
  sumsquare_hj_kd = GenAlloc( K * D, sizeof(double), 1, func, "sumsquare" );
  oldmean_hj_kd   = GenAlloc( K * D, sizeof(double), 1, func, "oldmean" ) ;

  /* Save old mean value */
  memcpy( oldmean_hj_kd , OldNewMhj_kd , K * D * sizeof( double ) ) ;

  /* Set no empty class by default */
  (*EmptyK_P) = 0 ;

  /* For each class h */
  for ( h = 0 ; h < K ; h ++ )
    {
      /* For each variable j */
      for ( j = 0 ; j < D ; j ++ )
	{
	  /* Initialize the sum[i] quantities to 0 */
	  Nh_k[ h ]               = 0.0 ;
	  Nhj_kd[ _HJ ]           = 0.0 ;
	  sumdata_hj_kd[ _HJ ]    = 0.0 ;
	  sumsquare_hj_kd[ _HJ ]  = 0.0 ;
	  OldNewMhj_kd[ _HJ ]     = 0.0 ; /* use Mhj as sum first */

	  /* For each object i */
	  for ( i = 0 ; i < N ; i ++ )
	    {
	      double cih = Cih_nk[ _IH ] ;
	      double xij = Xij_nd[ _IJ ] ;

	      /* Increment class size */
	      Nh_k[ h ] += cih ;

	      /* If this object's j variable is observed */
	      if ( ! isnan( xij ) )
		{
		  /* Increment all sum[i] of observed quantities */
		  Nhj_kd[ _HJ ]           += cih ;
		  sumdata_hj_kd[ _HJ ]    += cih * xij ;
		  sumsquare_hj_kd[ _HJ ]  += cih * xij * xij ;
		}
	      /* Else xij is missing, don't change sum[i] of observed data */
	    }
	  /* End  For each object i (all sums[i] are now done) */

	  /* If this class size > 0 */
	  if ( Nh_k[ h ] > 0 )
	    {
	      /* 
	       * Compute means Mhj and dispersions Dhj from computed sums,
	       * depending on REPLACE or IGNORE missing mode 
	       */
	      if ( Miss == MISSING_REPLACE )
		{
		  /* 
		   * new_mhj = 
		   *  ( sum[i] cih rij xij + ( nh - nhj ) old_mhj ) / nh 
		   *
		   * disp_hj = 
		   *  "observed dispersion"_hj +
		   *  ( nh - nhj ) [ (new_mhj - old_mhj)^2 + old_vh old_chjj ]
		   */
		  double obsdisphj ;

		  OldNewMhj_kd[ _HJ ] = 
		    ( sumdata_hj_kd[ _HJ ] + 
		      ( Nh_k[ h ] - Nhj_kd[ _HJ ] ) * oldmean_hj_kd[ _HJ ] )
		    / Nh_k[ h ] ;

		  /*
		   * "observed dispersion"_hj = 
		   *      sum[i] cih rij xij^2 - nhj new_mhj^2 
		   */
		  obsdisphj = sumsquare_hj_kd[ _HJ ] - 
		    Nhj_kd[ _HJ ] * sqr( OldNewMhj_kd[ _HJ ] ) ;

		  Disphj_kd[ _HJ ] = 
		    obsdisphj +
		    ( Nh_k[ h ] - Nhj_kd[ _HJ ] ) *
		    ( sqr( OldNewMhj_kd[ _HJ ] - oldmean_hj_kd[ _HJ ] ) +
		      OldVh_k[ h ] * OldChjj_kdd[ h * D * D + j * D + j ] ) ;
		}
	      else  /* assume Miss == MISSING_IGNORE */
		{
		  double obsdisphj ;

		  /* 
		   * new_mhj = 
		   *  . if  nhj <> 0,  ( sum[i] cih rij xij ) / nhj  
		   *  . if  nhj == 0,  old_mhj
		   *
		   * disp_hj = 
		   *  "observed dispersion"_hj 
		   */

		  if ( Nhj_kd[ _HJ ] > 0 )
		    OldNewMhj_kd[ _HJ ] = sumdata_hj_kd[ _HJ ] / Nhj_kd[ _HJ ];
		  else
		    OldNewMhj_kd[ _HJ ] = oldmean_hj_kd[ _HJ ] ;

		  /*
		   * "observed dispersion"_hj = 
		   *      sum[i] cih rij xij^2 - nhj new_mhj^2 
		   */
		  obsdisphj = sumsquare_hj_kd[ _HJ ] - 
		    Nhj_kd[ _HJ ] * sqr( OldNewMhj_kd[ _HJ ] ) ;

		  Disphj_kd[ _HJ ] = obsdisphj ;
		}

	    } 
	  else /* then this class size == 0 */
	    {
	      /* signal empty class and which one */
	      sts = STS_W_EMPTYCLASS ;
	      (*EmptyK_P) = h + 1 ;
	    }
	}
      /* end  For each variable j */

      /* Update proportions as  ph = nh / n */
      NewPh_k[ h ] = Nh_k[ h ] / N ;
    }
  /* end  For each class h */


  /* Free local structures */
  GenFree( oldmean_hj_kd ) ;
  GenFree( sumsquare_hj_kd ) ;
  GenFree( sumdata_hj_kd  ) ;

  return sts ;

}  /*  end of CommonGaussDiagMissing() */





/* ------------------------------------------------------------------- */
void SetIdMatrix ( int Nd , double* M )
/* Set matrix M to identity */ /*V1.03-a*/
/* ------------------------------------------------------------------- */
{
  int l , c ;  /* line and column counters : 0..Nd-1 */

  for ( l = 0 ; l < Nd ; l ++ )
    {
      for ( c = 0 ; c < Nd ; c ++ )
        {
          if ( c == l )
            M[ (l * Nd) + c ] = 1.0 ;
          else
            M[ (l * Nd) + c ] = 0.0 ;
        }
    }
}





#ifdef 0
/* old versions before V1.05-a*/

/* ------------------------------------------------------------------- */
int DensPkVkI          /* ret : 0 if OK, -1 if zero density */
        (
            int                Nd,         /* I : point dimension */
            int                Ik,         /* I : class number : 0..K-1 */
            const NoiseParaT*  NoiseP,     /* I : noise parameters */
            const double*       XV,         /* I : point (dim d) */
            double*             FkP,        /* O : density for class Ik */
            double*             LogFkP      /* O : log of density */
        )
/* ------------------------------------------------------------------- */
{
    int     d ;     /* 0..Nd-1 : current variable */
    double   dk ;    /* >= 0.0  : squared distance of XV to mean Ik */

    double   vk = NoiseP->Vk[ Ik ] ;

    /* Gaussian mixture density, model [ Pk , Vk . I ] :
       fk(x) = [(2*pi)^Nd * vk ]^(-0.5) * exp( -0.5 * dist_y_mk )
       dist_y_mk = sum[d=1,Nd]( (y(d) - mk(d))^2 ) / vk
    */

    for ( d = 0, dk = 0.0 ; d < Nd ; d ++ )
    {
        double dif = XV[ d ] - NoiseP->Mk[ (Ik * Nd) + d ] ;
        dk = dk + (dif * dif) ;
    }

    if ( vk > 0 )
    {
        *LogFkP = -0.5 * (  Nd * log( TWO_PI * vk )  +  dk / vk  ) ;/*V1.04-a*/
        *FkP = exp( *LogFkP ) ;
        return 0 ;
    }
    else
    {
        *LogFkP = - MAXDOUBLE ;
        *FkP = 0.0 ;
        return -1 ;
    }


}   /* end of DensPkVkI() */




/* ------------------------------------------------------------------- */
int             ParaPkVkI          /* ret : nb of empty classes */
                (
                  const double       *Cih_nk,        /* I : classification matrix */
                  const DataT       *DataP,     /* I : observed points */
                  int               K,         /* I */
		  MissET            MissMode,   /* I */
                  NoiseParaT        *NoiseParaP /* O : estimated parameters */
                ) 
/* ------------------------------------------------------------------- */
{
    int         k ;
    int         numempty ;

    void        SetIdMatrix ( int Nd , double* M ) ; /*V1.03-a*/


    /* For each class */
    for ( k = 0 , numempty = 0 ; k < K ; k ++ )
    {
        int     ipt ;                  
        int     id ;
        int     nd  = DataP->NbVars ;
        int     npt = DataP->NbPts ;    /* = n nb points */
        double   sizek = 0.0 ;            /* = nk fuzzy cardinal [0,n] */
        double   trWk = 0.0 ;

        /* Compute its 'fuzzy' cardinal : nk = sum[i=1,n]( uik ) */
        /* Compute its proportion : pk = nk / n */
        /* Compute its mean (mk) : mk = sum[i=1,n]( uik * Y(i,:) ) / nk */
        /* Compute its intra-class inertia matrix : 
           Wk = sum[i=1,n]( uik * ( Y(i,:) - mk ) * ( Y(i,:) - mk )' ) / nk */
        /* Compute its volume : Vk = tr(Wk) / d */

        /* Initialize means to zero */
        for ( id = 0 ; id < nd ; id ++ )
        {
            NoiseParaP->Mk[ (k * nd) + id ] = 0 ;
        }

        /* Do cumulated sums over all points */
        for ( ipt = 0 ; ipt < npt ; ipt ++ )
        {
            double   uik = Cih_nk[ (ipt * K) + k ] ;

            sizek += uik ;
            for ( id = 0 ; id < nd ; id ++ )
            {
                double   yid = DataP->PointsM[ (ipt * nd) + id ] ;
                double   mkd = NoiseParaP->Mk[ (k * nd) + id ] ;

                mkd         = mkd + uik * yid ;
                trWk        = trWk + uik * yid * yid ;

                NoiseParaP->Mk[ (k * nd) + id ] = mkd ;
            }
        }

        /* Normalize the sums by the cardinal of the class */
        if ( sizek != 0.0 )
        {
            trWk /= sizek ;

            for ( id = 0 ; id < nd ; id ++ )
            {
                double   mkd = NoiseParaP->Mk[ (k * nd) + id ] ;

                mkd         = mkd / sizek ;
                trWk        = trWk - ( mkd * mkd ) ;

                NoiseParaP->Mk[ (k * nd) + id ] = mkd ;
            }
            NoiseParaP->Vk[ k ] = trWk / nd ;
        }
        else
        {
            numempty = k + 1 ;
        }

        NoiseParaP->Pk[ k ] = sizek / npt ;

        /* Set normalized variance matrix to identity */
        SetIdMatrix( nd , & NoiseParaP->Ck[ k*nd*nd ] ) ;  /*V1.03-a*/

    } /* end for ( k = 0 , numempty = 0 ; k < K ; k ++ ) */

    return numempty ;

}   /* end of ParaPkVkI() */



/* ------------------------------------------------------------------- */
int             ParaPkV_I          /* ret : nb of empty classes */
                (
                  const double       *Cih_nk,        /* I : classification matrix */
                  const DataT       *DataP,     /* I : observed points */
                  int               K,         /* I */
		  MissET            MissMode,   /* I */
                  NoiseParaT        *NoiseParaP /* O : estimated parameters */
                ) 
/* ------------------------------------------------------------------- */
{
    int         k ;         /* class counter : 0 .. K - 1 */
    int         numempty ;  /* index of empty class, 0 if none */
    double       sumk_trTk ; /* sum over k of class k's dispersion matrix's trace */
    double       V ;         /* common volume = sum[k] trTk / (N * Nd) */
    int         nd  = DataP->NbVars ;   /* Nd : nb of variables */
    int         npt = DataP->NbPts ;    /* N : nb of points */

    void        SetIdMatrix ( int Nd , double* M ) ;  /*V1.03-a*/


    /* For each class */
    for ( k = 0 , numempty = 0, sumk_trTk = 0.0 ; 
          k < K ; 
          k ++ )
    {
        int     ipt ;
        int     d ;
        double   sumi_uik ;
        double   sumi_uik_yi2 ;

        /* 'fuzzy' cardinal : sizek = sum[i]( uik ) */
        /* proportion :       pk    = 1 / K */
        /* mean :             mk(d) = sum[i]( uik * yi(d) ) / sizek */
        /* trace of its dispersion matrix : 
           trTk = sum[i]( uik * sum[d] yi(d)? ) - sizek * sum[d] mk(d)? */
        /* volume :           Vk = V = sum[k] trTk / (N * Nd) */

        /* Initialize means to zero */
        for ( d = 0 ; d < nd ; d ++ )
        {
            NoiseParaP->Mk[ ( k * nd ) + d ] = 0 ;
        }

        /* Do cumulated sums over all points */
        for ( ipt = 0, sumi_uik = 0.0, sumi_uik_yi2 = 0.0 ; 
              ipt < npt ; 
              ipt ++ )
        {
            double   sumd_uik_yid2 ;
            double   uik = Cih_nk[ ( ipt * K ) + k ] ;

            for ( d = 0, sumd_uik_yid2 = 0.0 ; 
                  d < nd ; 
                  d ++ )
            {
                double   yid   = DataP->PointsM[ (ipt * nd) + d ] ;

                NoiseParaP->Mk[ ( k * nd ) + d ]    += uik * yid ;
                sumd_uik_yid2                       += uik * yid * yid ;
            }

            sumi_uik        += uik ;
            sumi_uik_yi2    += sumd_uik_yid2 ;
        }

        /* Normalize the sums by the cardinal of the class */
        if ( sumi_uik != 0.0 )
        {
            double   sumd_mkd2 ;
            double   trTk ;
            double   invsizek = 1 / sumi_uik ;

            for ( d = 0, sumd_mkd2 = 0.0 ; 
                  d < nd ; 
                  d ++ )
            {
                double   mkd         = NoiseParaP->Mk[ (k * nd) + d ] ;

                mkd                 *= invsizek ;
                sumd_mkd2           += mkd * mkd ;

                NoiseParaP->Mk[ (k * nd) + d ]  = mkd ;
            }

            trTk                    = sumi_uik_yi2 - sumi_uik * sumd_mkd2 ;
            sumk_trTk               += trTk ;
        }
        else
        {
            numempty = k + 1 ;
        }

        NoiseParaP->Pk[ k ] = sumi_uik / npt ;

        /* Set normalized variance matrix to identity */
        SetIdMatrix( nd , & NoiseParaP->Ck[ k*nd*nd ] ) ;  /*V1.03-a*/

    } /* end for ( k = 0 , numempty = 0 ; k < K ; k ++ ) */

    V = sumk_trTk / ( npt * nd ) ;

    for ( k = 0 ; k < K ; k ++ )
    {
        NoiseParaP->Vk[ k ] = V ;
    }

    return numempty ;

}   /* end of ParaPkV_I() */

/* end of old versions before V1.05-a*/
#endif



/* ------------------------------------------------------------------- */

/* Tests

#define DIM 5
main(void)
{
  double     A[ DIM * DIM ] ;

  SetIdMatrix( DIM , A ) ;

  return 0 ;
}
*/

/* ~~~~~~~~~~~~~~~~~~~~~~~~ END OF FILE ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
