#ifndef SEMMS_TYP_H
#define SEMMS_TYP_H

/*\
    SEMMS_TYP.H

    Programme SEMMS (Spatial EM for Markov Segmentation) : tous les types echanges entre modules

    Van Mo DANG       Janvier 96


    Vers-mod  Date         Who Description

    V1.03-a   01-NOV-1996  MD  Add log-likelihood criterion in CriterT
    V1.04-a   04-OCT-1997  MD  Add BetaET for beta estimation mode
    V1.04-b   04-OCT-1997  MD  Add BtaStep and BtaMode in NemParaT
    V1.04-c   10-OCT-1997  MD  Add CvThres in NemParaT
    V1.04-d   10-OCT-1997  MD  Add Seed in NemParaT
    V1.04-e   10-OCT-1997  MD  Type OrderET and add VisitOrder in NemParaT
    V1.04-f   13-OCT-1997  MD  Debug and RefName in NemParaT
    V1.04-g   13-OCT-1997  MD  Error in CriterT
    V1.04-h   13-OCT-1997  MD  CritET and Crit in NemParaT
    V1.05-a   12-JAN-1998  MD  NbMissing in DataT, MisModeET, MissMode in NemParaT
    V1.05-b   12-JAN-1998  MD  MissMode in parameters of EstimNoiseFT
    V1.05-c   16-JAN-1998  MD  EstimNoiseFT : EmptyK_P as parameter, return status
    V1.05-d   05-FEB-1998  MD  Add Markov fuzzy pseudo-like. (CritET and CriterT)
    V1.06-a   17-JUN-1998  MD  NoiseModel -> StatModel, structure changed
    V1.06-b   23-JUN-1998  MD  LEN_LINE <- nem_exe.c
    V1.06-c   30-JUN-1998  MD  Add define EPSILON
    V1.06-d   03-AUG-1998  MD  Add INIT_MIXFIX and INIT_MIXINI
    V1.06-e   10-SEP-1998  MD  Add SiteUpdate in NemParaT and UpdET
    V1.06-f   10-SEP-1998  MD  Add TieRule in NemParaT and TieET
    V1.06-g   15-SEP-1998  MD  Add BtaPsGrad in NemParaT and BtaPsGradT struct
    V1.06-h   20-SEP-1998  MD  Add ErrInfo in CriterT and ErrInfoT struct
    V1.06-i   21-SEP-1998  MD  Change CvTest in CriterT and add CvemET
    V1.06-j   30-NOV-1998  MD  Add EPSILON_INV
    V1.06-k   01-DEC-1998  MD  FkP double* instead of *float in compudensft
              1999 -----   NP  Add ALGO_ {MEANFF, HOMMEANF,
	                       MARGMEANF, MARGCONDMEANF, SIMULF,
                               SIMULF2, MODE, ICM}
              10-OCT-1999  NP  Add MSCritT (BIC, ICL)
              04-FEV-2000  NP  Add InitfpeET for fpe initialization type
              15-FEV-2000  NP  Add ParamET for non-beta estimation type
              16-FEV-2000  NP  Add AlgoPara->NbSimuls and VarParamT (used when
                               simulf or simulf2 are run sevral times)
              27-AVR-2000  NP  Add BETA_ZEROIN in BetaET
              27-AVR-2000  NP  Add ComputeCritGrad (for zeroin)
              28-AVR-2000  NP  Add AlgoParaP->True, to compute ErrorRate
              03-MAI-2000  NP  Replace float by double
              23-MAI-2000  NP  Add the CPU option
              04-JAN-2000  NP  Add ALGO_ {GSF, MCEM}. Now SIMULF2 is included
	                       in GSF.
                               Suppress SIMULF.
	      22-JUN-2001  FC  Add SINGL_EM in enumerated type SinglET.
	      02-JUL-2001  FC  Add SINGL_BETA in enumerated type SinglET,
	                       allows a second method for potential estimation
			       on singletons.
              12-JUL-2001  FC  Add ParameterET and ParameterT for zeroin :
	                       defines the parameter we want to estimate
			       (Beta or Potential no xx)
	      12-JUL-2001  FC  Modify the ComputeCritGrad profile : replace
	                       x with a pointer to Beta and added a ParameterT
			       parameter
              17-JUL-2001  NP  suppress the option "estimparam" in
	                       AlgoParaP->Opt, this case is included
			       in the option "modelselec"
              17-JUL-2001  NP  add the option "AddGaussianNoise"
	                       in AlgoParaT.Opt
              23-JUL-2001  FC  add the option "GibbsSampler"
	                       in AlgoParaT.Opt
	      24-JUL-2001  FC  added all the structures necessary for the
	                       loop : ModeET and SaveTypeET
			       + LoopParamT and StatParamT
			       added Mode and LoopParam to AlgoParaT
	      26-JUL-2001  FC  NeighDataT : Union replaced by a Struct
	      26-JUL-2001  FC  EXT_OUTNAMENOISE and EXT_OUTNAMEPOTTS
	                       added for saving files in the loop.
	      27-JUL-2001  FC  EXT_OUTNAMESTATS added for saving statistics
	                       during the loop
	      07-AUG-2001  FC  SaveNow added to LoopParamT for saving pictures
	      16-AUG-2001  FC  TrueT : ErrorRate changed in double*
	                       to allow computations of error rates for
			       each class
	      16-AUG-2001  FC  StatParamT : MeanError and VarError
                               changed in double* too
	      16-AUG-2001  FC  Added Prop to TrueT and MeanProp and VarProp to
	                       StatParamT to know the proportions between the
			       different classes
              23-OCT-2001  NP  remove SIMULF since we always use  GSF now
              28-JAN-2005  NP  replace 'nem' by 'semms'
\*/

/*
 *  Constant definitions
 */

#define    TRUE             1
#define    FALSE            0

#define    LEN_NOISEMODEL   20
#define    LEN_FILENAME     100
#define    LEN_LINE         500  /*V1.06-b*/


#define    MAX_PTS          1000000L
#define    MAX_VARS         100

#define    CONV_THRES       0.001

#define    EXT_OUTNAMEHARD  ".cf"
#define    EXT_OUTNAMENOISE ".nf"
#define    EXT_OUTNAMEPOTTS ".pf"
#define    EXT_OUTNAMESTATS ".sf"
#define    EXT_OUTNAMEFUZZY ".uf"
#define    EXT_MFNAME       ".mf"
#define    EXT_LOGNAME      ".log"

#define    EPSILON          1e-20  /* to check for FP zero or equality */
#define    EPSILON_INV      1e20   /* multiply by this for small doubles */

/*
 *  Enumerated types
 */

typedef enum {
        STS_OK = 0,
        STS_I_DONE,
        STS_W_EMPTYCLASS,
        STS_E_ARG,
        STS_E_MEMORY,
        STS_E_FILEIN,
        STS_E_FILEOUT,
        STS_E_FILE,
        STS_E_FUNCARG
        }
        StatusET ;


typedef enum {
        ALGO_MEANF,  
        ALGO_GSF, 
        ALGO_ICM,
        ALGO_NEM ,
        ALGO_NCEM ,
	ALGO_NB,
        }
        AlgoET;

typedef enum {
        RESTO_ICM,
        RESTO_MPM,
        RESTO_NB,
        }
        RestoET;



typedef enum {
        BETA_FIX ,        /* Use given fixed beta */
        BETA_PSGRAD ,     /* Estimate beta using a descent gradient algorithm*/
        BETA_ZEROIN ,     /* Estimate beta as zero of the function
			     of interest'gradient */
        BETA_HEUD,        /* Estimate beta using Hathaway heuristic */
        BETA_HEUL,        /* Estimate beta using likelihood heuristic */
        BETA_SA,          /* beta gradually increased from zero to beta_max */
	BETA_NB
        }
        BetaET;           /* Beta estimation mode */ /*V1.04-a*/

typedef enum {
        SINGL_FIX ,      /* Singleton potentiels fixed */
        SINGL_ZEROIN ,   /* Singleton potentiels estimated as
			    zero of a function */
        SINGL_NONE,      /* Singleton potentiels
		 	    set to zero ( to use for mixture model and Potts
                            model with no external field) */
	SINGL_EM,        /* Singleton potentiels estimated before Beta
			    and Theta estimation, using the EM algorithm. */
	SINGL_BETA,      /* Singleton potentiels and noise parameters
			    estimated with NEM, after Beta estimation and
			    before restauration */
        SINGL_NB
	}
        SinglET;         /*Singleton potentiels estimation mode */


typedef enum {
        PARAM_FIX ,
        PARAM_ESTIM,
	PARAM_NB
        }
        ParamET;           /* non-Beta params estimation mode */

typedef enum {
        CRIT_U ,    /* Use NEM criterion */
	CRIT_M ,    /* Use markovian fuzzy log pseudo-likelihood */ /*V1.05-d*/
	CRIT_D ,    /* Use Hathaway criterion */
	CRIT_L ,    /* Use Mixture log-likelihood */
	CRIT_NB
        }
        CritET;     /* Which criterion to choose local max */ /*V1.04-h*/



typedef enum {
        TYPE_SPATIAL,
        TYPE_IMAGE,
        TYPE_NONSPATIAL,
        TYPE_NB
        }
        TypeET;




/*V1.06-a*/
#if 0
typedef enum {
	   MODEL_P_VkI,
	   MODEL_PkVkI,
	   MODEL_P_V_I,
	   MODEL_PkV_I,
	   MODEL_NB
	   }
	   ModelET ;
#endif

typedef enum {
  FAMILY_NORMALGEN, /* the cov matrix is full */
  FAMILY_NORMALDIAG,/* the cov matrix is diagonal */
  FAMILY_LAPLACE,
  FAMILY_NB
}
FamilyET;

typedef enum {
  DISPER___,  /* 1 dispersion for all classes and variables (V.I) */
  DISPER_K_,  /* 1 dispersion for each class, same in all variables (Vk.I),
                 not available if FAMILY_NORMALGEN */
  DISPER__D,  /* 1 dispersion for each variable, same in all classes (V.B),
                 not available if FAMILY_NORMALGEN */
  DISPER_KD,  /* 1 dispersion for each class and variable (VkBk) */
  DISPER_NB
}
DisperET;

typedef enum {
  PROPOR__,  /* equal proportion for all classes = 1/K (P) */
  PROPOR_K,  /* 1 proportion for each class (Pk) */
  PROPOR_NB
}
ProporET;


typedef enum {
        FORMAT_HARD,
        FORMAT_FUZZY,
        FORMAT_UF,
        FORMAT_NB
        }
        FormET ;

typedef enum {
        INIT_SORT,
        INIT_RANDOM,
	INIT_PARAM,       /* init with given params */
	INIT_MIXINI,      /* EM mixture estimate at start */
        INIT_MIXFIX,      /* EM mixture estimate with fixed value */
        INIT_FILE,
	INIT_LABEL,
        INIT_EMIND,       /* segmentation obtained with EM for indep mixture*/
        INIT_NB
        }
        InitET ;

typedef enum {
        INITfpe_HARD,
        INITfpe_GIVEN,
        INITfpe_NB
        }
        InitfpeET;


typedef enum
{
  MISSING_REPLACE,  /* Replace missing statistics with expectation as in EM */
  MISSING_IGNORE,   /* Ignore missing statistics as in CEM */
  MISSING_NB
}
MissET ;   /*V1.05-a*/


typedef enum {
        NEIGH_FOUR,
        NEIGH_FILE,
        NEIGH_NB
        }
        NeighET ;

typedef enum {
        ORDER_DIRECT,
        ORDER_RANDOM,
        ORDER_NB
        }
        OrderET ;       /*V1.04-e*/


typedef enum {
        UPDATE_SEQ,
        UPDATE_PARA,
        UPDATE_NB
        }
        UpdET ;         /*V1.06-e*/


typedef enum {
        TIE_RANDOM,
        TIE_FIRST,
        TIE_NB
        }
        TieET ;         /*V1.06-f*/


typedef enum {
        CVTEST_NONE,
        CVTEST_CLAS,
        CVTEST_HARDCLASS,
	CVTEST_CRIT,
        CVTEST_BETA,
        CVTEST_PARAMS,
        CVTEST_FUNC,
        CVTEST_NB
        }
        CvemET ;        /*V1.06-g*/

typedef enum {
  ModelSelec,
  GibbsSampler,
  AddGaussianNoise,
  Opt_NB
}
Option ;

typedef enum {
  BETA,
  POTENTIAL
}
ParameterET;

typedef enum {
  NORMAL,
  LOOP
}
ModeET;

typedef enum {
  ALL,
  LAST,
  SAVETYPE_NB
}
SaveTypeET;

/*
 *  Structured types
 */

typedef struct
{
  double  MeanBeta;
  double  VarBeta;
  double* MeanUk;          /* Mean value of each class */
  double* VarUk;
  double* MeanSigmaK;      /* Variance of each class */
  double* VarSigmaK;
  double* MeanPotK;        /* Potential of each singleton (~class) */
  double* VarPotK;
  double* MeanErr;         /* Error of classification */
  double* VarErr;
  double* MeanProp;        /* Proportion of each class */
  double* VarProp;
}
StatParamT;

typedef struct
{
    int         NbPts ;       /* number of observation vectors */
    int         NbVars ;      /* number of variables */
    int         NbMiss ;      /* number of missing data 0..Npts*NbVars */
                              /*V1.05-a*/
    double       *PointsM ;    /* observations (NbPts,NbVars) to allocate */
    int         *LabelV ;     /* fixed labels (NbPts) to allocate: 0..k */
    int         *SiteVisitV ; /* site to visit (NbPts) to allocate: 0..Npts-1*/
    int         *SortPos_ND ; /* PointsM[ SortPos_ND[i*D+d]*D+d ] : +++*/
}
DataT ;         /* Matrix of observed data (each line = 1 point) */

typedef struct
{
  int          K;          /* true K if true segm known, 0 otherwise */
  double*      DataTrue;
  double*      ErrorRate;  /*  misclassified pixels (in percent)
			       total misclassified pixels +
			       misclassified pixes for each class
			       -> K+1 Elements
                          */
  double*      Prop;       /* Proportions of each class of the true image
			    */
}
TrueT ;       /* Matrix of the true data, if known */

/*V1.06-g*/
typedef struct
{
  int     NbIter ;     /* Max number of iterations of gradient ascent */
  double   ConvThres ;  /* Convergence if gradient <  threshold * N */
  double   Step ;       /* >0 : bta += grad*(step/N), 0: bta += grad/dsec */
  int     RandInit ;   /* 1 = random initial beta, 0 = specified by -b */
}
BtaPsGradT ;   /* parameters of beta pseudo-likelihood gradient estimation */


typedef struct
{
  double bmin;  /* first value of beta */
  double bmax;  /* last value of beta */
  double bstep; /* incrementation step for beta */
  int Kmin;  /* first value of K */
  int Kmax;  /* last value of K */
  double*  BICp ;
  double*  BICw ;     /* percentage of  selection of each nb of class by */
  double*  ICLp ;     /* the BIC and ICL MF-like approx */
  double*  ICLw ;     /* when algo = NEM, NCEM, NEM, BICp = BIC, ICLp = ICL */
  StatParamT* StatParam; /* size of Kmax - Kmin */
}
MSparamT; /* parameters for simultaneous selection of beta and K */


typedef struct
{
  double*  unique; /* labels in the original values */
  double*  stdev;  /*  vector of the stdevs of the N(0,sigma^2) noise added to
		    the original values according to their label */
}
GaussNoiseT ;

typedef struct
{
  int        NbIters;      /* Nb of iterations of the
			      simulation/add-noise/estimisation-restauration
			      loop */
  int        SavePotts;    /* Shall we save the simulated Potts image ? */
  int        SaveNoise;    /*   "  the Potts image with noise added ?   */
  int        SaveRest;     /*   "  the restaurated image ?              */
  SaveTypeET SaveType;     /* Shall we save all the images or the last one,
			      or ... ? */
  int        SaveNow;      /* If SaveType = LAST, SaveNow is activated at
			      the last iteration of the loop */
}
LoopParamT;

typedef struct
{
  ModeET        Mode;           /* if we use the program as a loop
				   simulation-noise-restauration */
  LoopParamT    LoopParam;      /* Parameters for the loop */
  GaussNoiseT   GaussNoise;     /* if using the program with the option
			           "AddGaussNoise" */
  Option        Opt;            /* Type of application done */
  AlgoET        Algo;           /* Type of algorithm */
  RestoET       Resto;          /* Type of restoration */
  ParamET       ParamMode;      /* type of non-beta parameters estimation */
  SinglET       SinglMode;      /* type of sinlgeton estimation */
  MSparamT      MSparam;        /* parameters for simultaneous selection of
			           beta and K */
  double        BtaHeuStep;     /* step of beta for heuristic estimation */
                                /*V1.04-b*/
  double        BtaHeuMax;      /* maximum beta for heuristic */
  double        BtaHeuDDrop;    /* drop of Hathaway slope for beta heuristic */
  double        BtaHeuDLoss;    /* proportion of Hathaway loss for
			           beta heuristic */
  double        BtaHeuLLoss;    /* proportion of likelihood loss
			           for beta heuristic */
  BtaPsGradT    BtaPsGrad;      /* parameters of beta gradient estimation */
  CritET        Crit;           /* criterion to choose local max */ /*V1.04-h*/
  double        CvThres;        /* convergence threshold */    /*V1.04-c*/
  CvemET        CvTest;         /* which convergence test to use */
  int           NbIters;        /* nb of iterations for NEM or FIELD algos*/
  int           NbEIters;       /* nb of iterations for E-step
                                   or simulateZ */
  int           NbEMIters;      /* nb of iterations for EM with the
				   approximated model within
				   each iteration of the mean field-like
				   algos */
  int           NbfpeIters;     /* nb of iterations to solve
				   the marg mean field fpe */
  int           NbGibbsIters;   /* nb of iteration when creating a
				   Potts model representation with CreateGibbs
				*/
  InitfpeET     InitFPE;        /* initialization for the FPE
				   in ComputePriorMean */
  int           NbSimuls;       /* nb of runs for the simulated algo */
  int           NbGibbs;        /* nb of Gibbs sampler iterations
				   at each step of the generalized simulated
				   algo */
  int           NbRandomInits;  /* nb of random initializations */
  long          Seed;           /* random generator seed */   /*V1.04-d*/
  FormET        Format;         /* output file format (hard or fuzzy) */
  InitET        InitMode;       /* initialization mode (histogram, random,
				   file) */
  MissET        MissMode;       /* how to process missing statistics */
                                /*V1.05-a*/
  int           SortedVar;      /* variable to be sorted : 0..NbVars */
  NeighET       NeighSpec;      /* neighborhood specification */
  OrderET       VisitOrder;     /* order of visit at E-step */ /*V1.04-e*/
  UpdET         SiteUpdate;     /* site update scheme at E-step */
  TieET         TieRule;        /* rule for equal probabilities
				   when computing MAP */
  TrueT         True;           /* true segmentation, to compute error rate */
  int           Label;          /* 1 if some labels are known, 0 else */
  int           CPUmax;
  int           CompF;          /* TRUE if F' is evaluated during the
				   FIELD algos */
  int           Debug;          /* TRUE if in debug mode */    /*V1.04-f*/
  char OutBaseName[LEN_FILENAME+1];   /* base name of output file */
  char OutName    [LEN_FILENAME+1];   /* name of output file */
  char LogName    [LEN_FILENAME+1];   /* name of log file ("" = no log) */
  char StartName  [LEN_FILENAME+1];   /* name of initial partition file */
  char StartParamName  [LEN_FILENAME+1];   /* name of initial params file */
  char ListName   [LEN_FILENAME+1];   /* name of E and Beta list file for
					 InvAlgo */
  char NeighName  [LEN_FILENAME+1];   /* name of neighborhood file */
  char LabelName  [LEN_FILENAME+1];   /* name of fixed labels file */
  char RefName    [LEN_FILENAME+1];   /* name of reference labels file */
                                      /*V1.04-f*/
  char ParamName  [LEN_FILENAME+1];   /* name of non-Beta parameter's values
					 file */
                                      /* when they are fixed */
  char SinglName  [LEN_FILENAME+1];   /* name of singleton potentiels fiel */
                                      /* when they are fixed */
  char TrueName   [LEN_FILENAME+1];   /* name of the file with true data,
					 if known */
  /*V1.06-a*/
  //     double   Beta ;      /* context weight : >= 0 */
  //    BetaET  BtaMode ;   /* type of beta estimation */  /*V1.04-b*/
}
AlgoParaT ;      /* ALGO running parameters */


typedef struct
{
    int     Dl ;    /* neighbour shift in line :     -2 .. 2 */
    int     Dc ;    /* neighbour shift in column :   -2 .. 2 */
    double   Weight ; /* neighbour weight >= 0 */
}
INeighT ;       /* one neighbour (in image configurations) */

typedef struct
{
    int     Nl ;        /* Image number of lines (height) : > 0 */
    int     Nc ;        /* Image number of columns (width) : > 0 */
    int     NbNeigh ;   /* nb of allocated neighbours : >= 0 */
    INeighT *NeighsV ;  /* to be allocated : pixel's neighbours */
}
ImageNeighT ;   /* neighbourhood system (in image configurations) */

typedef struct
{
    int         Index ; /* index of neighbour : 0 .. Npt-1 */
    double       Weight ;/* weight of neighbour (default : 1.0) */
}
NeighT ;        /* one neighbour (in non-image configurations) */

typedef struct
{
    int     NbNeigh ;   /* nb of allocated neighbours */
    NeighT  *NeighsV ;  /* to be allocated : point's neighbours */
}
PtNeighsT;      /* one point's neighbours (in non-image configuration) */

typedef struct
{
    ImageNeighT Image ;
    PtNeighsT   *PtsNeighsV ;   /* to be allocated : all points' neighbours */
}
NeighDataT ;    /* generic neighbourhood system */

typedef struct
{
    NeighDataT  NeighData ; /* generic neighbourhood system */
    int         MaxNeighs ; /* maximum number of neighbours (>= 0) */
    TypeET      Type ;      /* type of spatial configuration */
}
SpatialT ;


/*V1.06-a*/

#if 0
typedef struct
{
    double   *Pk ;   /* proportions (K) */   /* to be allocated */
    double   *Vk ;   /* volumes (K) */       /* to be allocated */
    double   *Ck ;   /* shapes (d*d*K) */    /* to be allocated */
    double   *Mk ;   /* means (d*K) */       /* to be allocated */
}
NoiseParaT ;


typedef struct
{
    ModelET     ModelNum ;
    int         Nk ;
    NoiseParaT  NoisePara ;
}
NoiseModelT ;
#endif

typedef struct
{
  int       K ;                /* number of classes */
  FamilyET  ClassFamily ;
  DisperET  ClassDisper ;
  ProporET  ClassPropor ;
  BetaET    BetaModel ;
}
ModelSpecT ;  /* Model specification */

typedef struct
{
  double*    Ztilde;      /* z~ */
  double*    Singl;      /* singleton potentiels */
  double     Beta ;
  double*    Center_KD ;  /* Center in each class and variable (K*D) */
  double*    Disp_KD ;    /* Dispersion (variance) in each class and variable (K*D*D) */
  double*    Prop_K ;     /* Proportion of each class (in ]0,1[) */

  double*    NbObs_K ;    /* Nb of observations in each class (K) */
  double*    NbObs_KD ;   /* Nb of observations in each class/variable (K*D) */
  double*    NbObs_KDD ;  /* Nb of observations in each class/variable/variable (K*D*D) */
  double*    Iner_KD ;    /* Inertia = sum_i cik * Dist(xid, mkd) (K*D*D) */
}
ModelParaT ;  /* Model parameters */

typedef struct
{
  double*     DispSam_D ; /* Dispersion of whole sample in each variable (D) */
  double*     MiniSam_D ; /* Minimum of whole sample in each variable (D) */
  double*     MaxiSam_D ; /* Maximum of whole sample in each variable (D) */
}
SampleDesT ;  /* Sample description */

typedef struct
{
  ModelSpecT   Spec ;
  ModelParaT   Para ;
  SampleDesT   Desc ;
}
StatModelT ;  /* Model description */


typedef struct
{
  int      Kc ;            /* # of classes in computed classification */
  int      Kr ;            /* # of classes in reference classification */
  int      Km ;            /* max( Kr, Kc ) */
  int      Kmfac ;         /* Km! */
  TieET    TieRule ;       /* same value as in SemmsPara */
  double*   Refclas_N_Kr ;  /* reference classification : 0/1 */
  int*     Perm_Kmfac_Km ; /* all permutation of Km classes */
}
ErrinfoT ;    /* Classification error information */


typedef struct
{
  double*   Agree_Km_Km ;   /* #common elements between found and ref classes */
  double*   Loclas_N_Kc ;   /* local copy of computed classification */
  int      Ibestpermut ;   /* index of best agreement permutation 0..Km!-1 */
  double    Errorrate ;     /* # misclassified objects of best permut / N */
}
ErrcurT ;    /* Currently computed classification error */


typedef struct
{
  double    D ; /* hathaway crit. D = sum[i]sum[k] cik (log pkfki-log cik) */
  double    E ; /* entropy. E = sum[i]sum[k] cik log cik */
  double    G ; /* geog. cohesion G = sum[i]sum[j]sum[k] cik cjk wij */
  double    U ; /* NEM maximized criterion U = D + 0.5 * beta * G */
  double    M ; /* markovian fuzzy class. like. M = D + beta * G - Z */
  double    L ; /* mixture likelihood crit. L = sum[i] log sum[k] pkfki ) */
  double    Z ; /* log pseudo-l. Z =-sum[i]log(sum[k]e(bta*sum[j~i]wij cjk)) */
  ErrinfoT Errinfo ; /* information to compute error */   /*V1.06-h*/
  ErrcurT  Errcur ;  /* current error rate */   /*V1.06-h*/
} /*V1.05-d*/
CriterT ;       /*V1.03-a*/


typedef struct
{ /* for Potts model, H=Nb of homogen cliques */
 double   LogWmf;    /* MF like approximation of the normalization constant */
 double   C;      /* logC when the lower bound of W is Wmf*exp(C) */
 double   Hmf;       /* MF like approximation of the Potts NRJ : prod[i]Hi */
 double   ICLHmfWmf; /* 2*sum[i] sum[k] zhati*log[fik*P(zi=ek| ztildeN(i))] -d*logNpt */
 double   ICLHWinf;  /* 2*sum[i] log(fik) + beta*H(zhat) + beta*H(ztilde)-
			sum[i] log[sum[k] exp(beta zi sum[j~i]ztildej)] -d*logNpt */
 double   ICLHWmf;   /* 2*sum[i] log(fik) + beta*H(zhat)-
			sum[i] log[sum[k] exp(beta zi sum[j~i]ztildej)] -d*logNpt */
}
ICLArgT ;


typedef struct
{
  double  BICPmf; /* sum[i] log sum[k] fik*P(zi=ek| ztildeN(i)) -0.5*d*logNpt */
  double  BICWinf;//BICWmf; /* log(WinfCond) -log(WinfMarg) -0.5*d*logNpt */
 double   BIC2;
 double   EICL;
 ICLArgT ICLMarg; /* pour le moment pas etendu a mode ou simul field approx */
 ICLArgT ICLZtilde;
 double   LogVCond; /* log P(y|z) */
 double   H; /* NRJ of the segmentation. For Potts model, H=Nb of homogen cliques */
}
MSCritT ;


typedef struct
{
    int*        KmaxesV ;   /* values of ex-aequo MAP labels [nk] */
    double*      FikM;       /* fk(xi) [npt,nk] */
    double*      PkFkiM;     /* pk fk(xi) [npt,nk] */ /*V1.06-t*/
    double*      LogPkFkiM;  /* log( pk fk(xi) ) [npt,nk] */
    double*      LogFikM;    /* log( fk(xi) ) [npt,nk] */
    double*      CtmpM ;     /* last clas. for parallel update [npt,nk] */
    double*      ColdM ;     /* clas. of previous XEM iteration [npt,nk] */
    double*      CiNumV;     /* numerator of cik's for one i [nk] */ /*V1.06-t*/
    double*      Ztilde;     /* z~ in the field algos */
    PtNeighsT   Neighs;     /* neigh. ind/wei, NeighsV [SpatialP->MaxNeighs] */
}
WorkingT ;

typedef struct
{
 double     VarBeta ;
 double*    VarCenter_KD ;
 double*    VarStd_KD ; /* variance of the st. dev. over the runs */
 double     VarBICPmf;
 double     VarBICWinf;
 double     VarICLp;
 double     VarICLw;
 double     VarErrorRate;
  /* variance of the estimated parameters and MSCritT when the simulated algo */
  /* is run several times */

}
VarParamT ;

typedef struct
{
  ParameterET Type; /* Beta or potential */
  int         No;   /* If parameter is a potential */
}
ParameterT;

/****************************************************************************/

typedef StatusET            /* ret : OK, W_EMPTYCLASS or E_MEMORY *//*V1.05-c*/
EstimNoiseFT
 (
  const double   *Cih_nk,    /* I : classification matrix (N,K) */
  const DataT   *DataP,     /* I : observed points */
  int           K,          /* I : number of classes */
  MissET        MissMode,   /* I : how to treat missing data */ /*V1.05-b*/

  int           *EmptyK_P,  /* O : which empty class (1..K) or 0 *//*V1.05-c*/
  StatModelT    *StatModelP /* O : estimated parameters */
 ) ;

typedef int CompuDensFT         /* ret : 0 if OK, -1 if zero density */
        (
            int                Nd,      /* I : point dimension */
            int                Ik,      /* I : class number : 0..Nk-1 */
            const ModelParaT*  ParaP,   /* I : noise parameters *//*V1.06-a*/
            const double*      XV,      /* I : point (dim d) */
            const ModelSpecT*  SpecP,
            double*            FkP,     /* O : density for class Ik */
            double*            LogFkP,   /* O : log of density */
            const double*      InvChol,
            const double       logdet
        ) ;


typedef double ComputeCritGrad
          (
             const int           Npt,          /* nb of points */
             const int           Nk,           /* nb of classes */
             const double*       Beta,
             const double*       Ztilde,
             const double*       Singl,
             const double*       PostM,
             const SpatialT*     SpatialP,     /* neiborhood system */
             PtNeighsT*          NeighsP,
	     ParameterT          ParameterP    /* which singleton to use ?*/

          );

/* Old version : remove or restaure it if necessary */
/*   typedef double ComputeCritGrad */
/*            ( */
/*               const int           Npt,           nb of points */
/*               const int           Nk,            nb of classes */
/*               const double           x,             point where compute */
/*                                                  the function */
/*               const double*        Ztilde, */
/*               const double*        Singl, */
/*               const double*        PostM,   */
/*               const SpatialT*     SpatialP,      neiborhood system */
/*               PtNeighsT*    NeighsP  */

/*            ); */


typedef int GetNeighFT         /* ret : nb of neighbours */
                (
                  int               Ipt ,           /* I : index of point */
                  const NeighDataT  *NeighDataP,    /* I : neighborhood data */
                  PtNeighsT         *PtNeighsP      /* O : indices/weights of neighbours */
                ) ;


#endif

/* ~~~~~~~~~~~~~~~~~~~~~~~~ END OF FILE ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */





