/* dans les algos mise a jour de l'image sauf pixels du bord
  ( peut-etre pas si mise a jour en parallele, a verifier) */


/*
     SEMMS_ALG.C

    Program SEMMS (Spatial EM for Markovian Segmentation): routines de calcul algorithmique


Vers-mod  Date         Who Description

1.03-a    07-JAN-1997  MD  Print criteria to screen, compute them
                           at each iteration only if logfile requested
1.03-b    07-JAN-1997  MD  Write to logfile only if not NULL (bug fix)
1.03-c    22-AUG-1997  MD  Removed ComputePartitionParallel() (outdated)
1.03-d    22-AUG-1997  MD  Add MakeParaFromLabeled() called by ClassifyByNem()
1.03-e    22-AUG-1997  MD  Fix error - Pk was not init in MakeRandomPara()
1.03-f    22-AUG-1997  MD  ClassifyByNem() now returns criteria
1.03-g    30-SEP-1997  MD  Criterion L renamed D, and add new criterion L
1.04-a    05-OCT-1997  MD  Heuristic estimation of beta
1.04-b    10-OCT-1997  MD  Convergence threshold as modifiable option
1.04-c    10-OCT-1997  MD  Sites visit order specified
1.04-d    13-OCT-1997  MD  Debug mode for beta heuristic
1.04-e    13-OCT-1997  MD  Compute classification error for 2 clusters
1.04-f    13-OCT-1997  MD  Choice of criterion to select best run
1.04-g    13-OCT-1997  MD  Save all criteria of best try in random init mode
1.04-h    01-DEC-1997  MD  Add ComputePartitionGEM (Monte-Carlo to get cih)
1.05-a    12-JAN-1998  MD  Add arg MissMode to ParaP_VkI() and fEstimNoise()
1.05-b    12-JAN-1998  MD  In MakeParaFromLabeled() Pk do not vary any more
1.05-c    12-JAN-1998  MD  In MakeParaFromLabeled() process missing data
1.05-d    12-JAN-1998  MD  In MakeRandomPara() process missing data
1.05-e    17-JAN-1998  MD  emptyk as arg to P*V*I and fEstimNoise
1.05-f    26-JAN-1998  MD  GenAlloc/GenFree instead of malloc/calloc/free
1.05-g    26-JAN-1998  MD  Changed Iterations appearance
1.05-h    27-JAN-1998  MD  Fix bug: clasiniM was freed *before* reading it
1.05-i    03-FEB-1998  MD  Initial vol = whole vol / K (was whole vol before)
1.05-j    03-FEB-1998  MD  Fix bug: L computed even for i/h with zero cih !
1.05-k    05-FEB-1997  MD  Add Markov fuzzy pseudo-like. (CritET and CriterT)
1.05-l    05-FEB-1997  MD  call RandomFloat() instead of drand48()
1.05-m    05-FEB-1997  MD  check that drawn center does not exist
1.06-a    03-JUN-1998  MD  Initial vol = whole vol / K (was whole vol before)
1.06-b    28-JUN-1998  MD  Reflect changes to nem_typ.h (NoiseModel)
1.06-c    28-JUN-1998  MD  fGetNeigh and fCompuDens fetched local to their use
1.06-d    29-JUN-1998  MD  MakeParaFromLabeled uses EstimPara to compute param.
1.06-e    30-JUN-1998  MD  Call MakeParaFromLabeled in INIT_SORT mode
1.06-f    30-JUN-1998  MD  Add ModelPreprocess
1.06-g    01-JUL-1998  MD  Call MakeParaFromLabeled in INIT_FILE mode
1.06-h    17-JUL-1998  MD  Bug fix : CompSortValue now puts NaN at end
1.06-i    19-JUL-1998  MD  Change heuristic, save classif. just before L drops
1.06-j    03-AUG-1998  MD  Add INIT_MIXINI and INIT_MIXFIX
1.06-k    10-SEP-1998  MD  ComputePartitionNEM/GEM call ComputeLocalProba
1.06-l    10-SEP-1998  MD  Add SiteUpdate parallel or sequential
1.06-m    15-SEP-1998  MD  Add EstimBeta
1.06-n    21-SEP-1998  MD  Add TieRule arg to ComputeMAP
1.06-o    21-SEP-1998  MD  Change convergence test (test criterion also)
1.06-p    05-OCT-1998  MD  MakeParaFromLabeled draws random if empty class
                           and INIT_LABEL does not fail in that case
1.06-q    22-OCT-1998  MD  Bug : RandNemAlgo must save/restore best cen./disp.
1.06-r    03-NOV-1998  MD  Process case 0 < cumnum < EPSILON
1.06-t    01-DEV-1998  MD  pkfki and cinumv now double (was float)
          JAN-1999     NP  Add main functions for the condmf anf gsf algos:
              to           ClassifiedByFieldAlgo, FieldAlgo, SimulateZ,
	                   SimulateAlgo,
          AVR-1999         EstimBetaForFieldAlgo,
	                   ComputePostPartition(FromPara),
                           ComputeDensity
          01-OCT-1999  NP  Add main function to compute the
	                   BIC and ICL criteria:
                           ComputeBIC, ComputePriorPartition, HardClassif
          01-FEV-2000  NP  Add ComputePriorMean, ComputeProps,
	                   for margmf and mixedmf algos
          01-FEV-2000  NP  Add mode algo (ALGO_MODE)
          01-MAR-2000  NP  Add ComputePartitionICM
	                   and ComputeLocalPostPartition,
                           for icm algo
          01-AVR-2000  NP  Add ReadParamFile
          27-AVR-2000  NP  Add zeroin and ComputeCritGradForFieldAlgo
	                   for estimating Beta
                           (psgrad has artefacts, forget it)
          03-MAI 2000  NP  replace float by double
                           (not only in nem_alg.c actually)
          04-MAI-2000  NP  dans les algos mise a jour de l'image sauf pixels du
                           bord ( peut-etre pas si mise a jour en parallele,
                           a verifier, et pas fait dans ComputePartitionGEM() )
          10-JAN-2001  NP  add the possibility of more than 1 iter of EM
	                   to estimate
                           the parameters in the estimation step of the
			   MFL algos
	  02-JUL-2001  FC  Add FieldAlgoSingl to allow a first method of
	                   estimation of the potentials on singletons.
			   ClassifyByFieldAlgo modified to call
			   FieldAlgoSingl.
	  03-JUL-2001  FC  Add a second method for estimation of the
	                   potentials on singletons in FieldAlgoSingl
	  10-JUL-2001  FC  Complete reecriture of ClassifyByFieldAlgo so as to
	                   support easily many methods for evaluating
			   potentials on singletons
          12-JUL-2001  FC  ComputeCritGradForField removed and replaced by
	                      - ComputeGradBeta for Beta estimation
			      - ComputeGradPotential for potentials estimation
	  12-JUL-2001  FC  NemAlgo and FieldAlgo modified so as to take
	                   into account new calls to zeroin
	  12-JUL-2001  FC  FieldAlgo modified so as to compute singletons
	                   potentials using zeroin
	  12-JUL-2001  NP  ComputeBIC function reinserted : this time it is
	                   the good function
          17-JUL-2001  NP  add function ComputeErrorRate
                           now the error rate is computed at the end of the
                           algo and not anymore in SaveMSResults
          18-JUL-2001  NP  change SimulatedAlgo according to the new strategy
                           (i.e. only one run of gsf or mcem but mean of the
                           10 last estimators)
	  06-AUG-2001  FC  Added CreateGibbs and ComputeGibbsPartition for
	                   simulation of Potts models using Gibbs Sampler.
	  16-AUG-2001  FC  Changed ComputeErrorRate to compute error
	                   within each class
	  17-AUG-2001  FC  Display f when call of zeroin for potentials
	  18-MAY-2004  BA  in ComputeErrorRate possibiliy to have one row or one column.
          27-JAN-2005  NB  suppression of DoLog
          27-JAN-2005  NB  suppression of useless variant of mean-field like EM 
\*/

#include "semms_typ.h"    /* DataT, ... */
#include "semms_arg.h"    /* CritStrVC */
#include "semms_nei.h"    /* FuncNeighNone, ... */
#include "semms_mod.h"    /* EstimPara, ... */
#include "semms_rnd.h"    /* RandomInteger */
#include "semms_zeroin.h" /* Zeroin function for Beta ad singleton estimation */
#include "genmemo.h"    /* GenAlloc, GenFree */
#include <stdio.h>      /* printf, ... */
#include <string.h>     /* memcpy */
#include <math.h>       /* exp */
#include <time.h>       /* time */
#include <values.h>     /* MINDOUBLE */
#ifdef __TURBOC__
#include <alloc.h>      /* coreleft, ... */
#endif

#include "semms_alg.h"    /* prototypes */


#define NBCYCLES_STARTUP   1
#define NBCYCLES_RECORD    2


#define MAX_BETA           20   /* To check against divergence of beta */
#define MIN_BETA           -5   /* To check against divergence of beta */


typedef struct
{
    int     Index ;
    double   Value ;
}
SortPtT ;






static double mknan()
{
   return atof( "nan" ) ;
}


/* ==================== LOCAL FUNCTION PROTOTYPING =================== */


/* Called by ClassifyByAlgo */

  static void ModelPreprocess
        (
	  const ModelSpecT*   SpecP,            /* I */
	  DataT*              DataP             /* I/O */
	) ;



  static int ClassifyByNemHeuBeta                             /*V1.04-a*/
        (
          const DataT         *DataP,           /* I */
          AlgoParaT            *AlgoParaP,        /* I[M] */
          const SpatialT      *SpatialP,        /* I */
          StatModelT          *StatModelP,      /* I/O */
          double               *ClassifM,        /* I/O */
          WorkingT*           WorkingP ,
	  CriterT             *CriterP,          /* O */  /*V1.03-f*/
          MSCritT             *MSCritP
        ) ;


/* Called by ModelPreprocess */

	static int CompSortValue( const void* elt1P, const void* elt2P ) ;



/* Called by ClassifyByNemHeuBeta */

  static int ClassifyByNemOneBeta                               /*V1.04-a*/
        (
          const DataT         *DataP,           /* I */
          const AlgoParaT      *AlgoParaP,        /* I */
          const SpatialT      *SpatialP,        /* I */
          StatModelT          *StatModelP,      /* I/O */
          double              *ClassifM,        /* I/O */
          WorkingT*           WorkingP ,
	  CriterT             *CriterP,          /* O */  /*V1.03-f*/
          MSCritT             *MSCritP
        ) ;


/* Called by ClassifyByNemOneBeta */

 

    static void InitPara
        (
         const DataT*       DataP,       /* I */
	 const SampleDesT*  DescP,       /* I */
	 const ModelSpecT*  SpecP,       /* I */
         const AlgoParaT*   AlgoParaP,
	 ModelParaT*        ParaP,       /* O */
         double*             C_NK         /* T */
        ) ;



    static void ComputePartitionFromPara
        (
	 const int          Needinit,    /* I : 1 if need init, 0 if normal */
	 const DataT*       DataP,       /* I */
	 const AlgoParaT*    AlgoParaP,    /* I */
	 const ModelSpecT*  SpecP,       /* I */
	 ModelParaT*        ParaP,       /* I (temporary modification) */
         const SpatialT*    SpatialP,    /* I */
         double*             C_NK,        /* O */
         CriterT*           CriterP,     /* O */
	 WorkingT*          WorkingP     /* T */
	) ;


    static int NemAlgo
        (
         const DataT*       DataP,           /* I */
         const AlgoParaT*    AlgoParaP,        /* I */
         const SpatialT*    SpatialP,        /* I */
	 const ModelSpecT*  SpecP,     	     /* I */
	 const SampleDesT*  DescP,     	     /* I */
	 ModelParaT*        ParaP,     	     /* I/O */
         double*             CM,              /* I/O */
         WorkingT*          WorkP,           /* O */
         CriterT*           CriterP,          /* O */
          MSCritT             *MSCritP
        ) ;

    static int MakeParaFromLabeled  /*V1.03-d*/
        (
            const AlgoParaT*    AlgoParaP,
            const DataT*        DataP,      /* I */
	    const double*        C_NK,       /* I */
	    const ModelSpecT*   SpecP,      /* I */
            const SampleDesT*   DescP,      /* I */
	    ModelParaT*         ParaP,      /* O */
	    int                 *misskP,    /* O */
	    int                 *missdP     /* O */
        ) ;

    static StatusET MakeRandomPara
        (
            const DataT*        DataP,      /* I */
	    const ModelSpecT*   SpecP,      /* I */
            const SampleDesT*   DescP,      /* I */
	    ModelParaT*         ParaP       /* O */
        ) ;

    static StatusET ComputePkFkiM
        (
            const DataT*        DataP,      /* I */
            const ModelSpecT*   SpecP,      /* I */
            const ModelParaT*   ParaP,      /* I */
            double*             PkFkiM,     /* O */
            double*              LogPkFkiM   /* O */
        ) ;

    static int ComputePartition
        (
          const ModelSpecT*   SpecP,      /* I */
          const ModelParaT*   ParaP,      /* I */
          const DataT*        DataP,      /* I */
          const SpatialT*     SpatialP,   /* I */
          const AlgoParaT*     AlgoParaP,   /* I */
          double*              CM,         /* I/O */
          WorkingT*           WorkP,      /* I:PkFkiM O:CtmpM,Neighs,CiNumV */
          CriterT*            CriterP     /* O */
        ) ;

    static int ComputePottsSinglMargMeanField
      (
          const ModelSpecT*   SpecP,      /* I */
          const ModelParaT*   ParaP,      /* I */
          const DataT*        DataP,      /* I */
          const SpatialT*     SpatialP,   /* I */
          const AlgoParaT*    AlgoParaP,   /* I */
          double*              CM,         /* I/O */
          WorkingT*           WorkP,      /*  O:CtmpM,Neighs,CiNumV */
          CriterT*            CriterP     /* O */
      );

 


     /*V1.04-f*/
    static double ChosenCrit( const CriterT* CriterP , CritET Which ) ;


    static StatusET RandNemAlgo
        (
         const DataT*       DataP,           /* I */
         const AlgoParaT*    AlgoParaP,        /* I */
         const SpatialT*    SpatialP,        /* I */
	 const ModelSpecT*  SpecP,     	     /* I */
	 const SampleDesT*  DescP,     	     /* I */
	 ModelParaT*        ParaP,     	     /* I/O */
         double*             CM,              /* I/O */
         WorkingT*          WorkP,           /* I/O */
         CriterT*           CriterP,          /* O */
          MSCritT             *MSCritP
        ) ;


/* Called by InitPartitionSort */

    static int CompSortValue( const void* elt1P, const void* elt2P ) ;



/* Called by NemAlgo */

   static  int HasConverged       /* ret : TRUE if convergence test satisfied */
        (
           const CvemET    CvTest,     /* I : which convergence test to use */
           const double     CvThres,    /* I : convergence threshold to use */
           const TieET     TieRule,    /* for case = CVTEST_HARDCLASS */
           const double*    ColdM,      /* I : previous classification matrix */
           const double*    CM,         /* I : current  classification matrix */
           const int       Npt,        /* I : number of objects */
           const int       Nk,         /* I : number of classes */
           const int       Nd,         /* I : number of variables */
           const double     OldFunc,    /* previous value of F' in field algos */
           const double     NewFunc,    /* current value of F' in field algos */
           const double     OldCrit,    /* I : previous criterion */
           const double     OldBeta,    /* I : previous value of beta */
           const double*    OldCenter,  /* I : previous value of the centers */
           const double*    OldDisp,    /* I : previous value of the dispersions */
           const ModelParaT*  ParaP,    /* current values of beta, centers and dispersions */
           const AlgoParaT* AlgoParaP,
           const ModelSpecT*  SpecP,
           const SpatialT* SpatialP,   /* I : spatial info to compute criterion */
           const DataT*       DataP,
           const CritET    WhichCrit,  /* I : what criterion to use */
           CriterT*        CriterP,    /* I/O : current criterion */
           WorkingT*       WorkP       /* I: LogPkFkiM, T: Neighs */
        ) ;




#if 0
    static StatusET ComputePkFkiM
        (
            const DataT*        DataP,      /* I */
            const ModelSpecT*   SpecP,      /* I */
            const ModelParaT*   ParaP,      /* I */
            double*             PkFkiM,     /* O */
            double*              LogPkFkiM   /* O */
        ) ;

    static int ComputePartition
        (
          const ModelSpecT*   SpecP,      /* I */
          const ModelParaT*   ParaP,      /* I */
          const DataT*        DataP,      /* I */
          const SpatialT*     SpatialP,   /* I */
          const AlgoParaT*     AlgoParaP,   /* I */
          double*              CM,         /* I/O */
          WorkingT*           WorkP,      /* I:PkFkiM O:CtmpM,Neighs,CiNumV */
          CriterT*            CriterP     /* O */
        ) ;

#endif


   static int  /* 0 = OK,  1 = beta diverges to infinity */
    EstimBeta
       (
         const BetaET       BetaModel,  /* I : beta estimation mode */
         const BtaPsGradT*  BtaPsGradP, /* I : gradient ascent parameters */
         const SpatialT*    SpatialP,   /* I : neighborhood system */
         const double*       C_NK,       /* I : classification matrix (Npt, NbK)*/
         const int          Npt,        /* I : number of points */
         const int          Nbk,        /* I : number of classes */
         double*            BetaP,      /* I/O : previous and new beta */
         PtNeighsT*         NeighsP     /* W : to store a site's neighbors */
       ) ;




    static StatusET ComputeCrit   /*V1.03-a*/
     (
        int                 Npt,         /* I */
        int                 Nk,          /* I */
        double               Beta,        /* I : spatial coefficient */
        const double*        CM,          /* I : class. matrix [Npt*Nk] */
        const SpatialT*     SpatialP,    /* I : neighborhood system */
        const AlgoParaT*    AlgoParaP,
        const DataT*              DataP,
        WorkingT*           WorkP,       /* I :PkFkiM, O:Neighs */
        CriterT*            CriterP      /* O : computed criteria */
     ) ;

    static void CalcError                /*V1.04-e*/
      (
       const double*     CM,       /* I : found classification */
       const int        N,        /* I : number of points */
       const int        Harden,   /* I : 1 use hardened CM, 0 use CM itself */
       const ErrinfoT*  ErrinfoP, /* I : info to compute error */
       ErrcurT*         ErrcurP   /* O : computed error and agreement */
      ) ;


/* Called by ComputePartition */

 static int ComputePartitionICM

      (
          const ModelSpecT*   SpecP,      /* I */
          const ModelParaT*   ParaP,      /* I */
          const DataT*        DataP,      /* I */
          const SpatialT*     SpatialP,   /* I */
          const AlgoParaT*    AlgoParaP,   /* I */
          double*              CM,         /* I/O */
          WorkingT*           WorkP,      /* I:PkFkiM O:CtmpM,Neighs,CiNumV */
          CriterT*            CriterP     /* O */
      );

  static int ComputePartitionGEM
      (
          const ModelSpecT*   SpecP,      /* I */
          const ModelParaT*   ParaP,      /* I */
          const DataT*        DataP,      /* I */
          const SpatialT*     SpatialP,   /* I */
          const AlgoParaT*     AlgoParaP,   /* I */
          double*              CM,         /* I/O */
          WorkingT*           WorkP,      /* I:PkFkiM O:CtmpM,Neighs,CiNumV */
          CriterT*            CriterP     /* O */
      ) ;

  static int ComputePartitionNEM
      (
          const ModelSpecT*   SpecP,      /* I */
          const ModelParaT*   ParaP,      /* I */
          const DataT*        DataP,      /* I */
          const SpatialT*     SpatialP,   /* I */
          const AlgoParaT*     AlgoParaP,   /* I */
          double*              CM,         /* I/O */
          WorkingT*           WorkP,      /* I:PkFkiM O:CtmpM,Neighs,CiNumV */
          CriterT*            CriterP     /* O */
      ) ;


/* Called by ComputePartitionNEM */

    static int ComputeLocalProba
    (
     const ModelSpecT*   SpecP,
     const int            Ipt,        /* I: current site */
     const int            Nk,         /* I: nb of classes */
     const ModelParaT*    ParaP,      /* I: parameters (use Beta) */
     const DataT*         DataP,      /* I */
     const NeighDataT*    NeighDataP, /* I: neighborhood system */
     GetNeighFT*          FGetNeigh,  /* I: function to fetch neighbors */
     const double*        PkfkiM,     /* I: pk fk(xi) */
     const double*         Cin_NK,     /* I: class of neighbors */
     double*               Cout_K,     /* O: class of site Ipt */
     PtNeighsT*           NeighsP,    /* W: to store fetched neighbors */
     double*              Cinum_K     /* W: to store cik's numerators */
    );

 

/* Called by ComputePartitionGEM */

    static int Multinomial(int km, const double *tk) ;

#if 0

    static int ComputeLocalProba
    (
     const ModelSpecT*   SpecP,
     const int            Ipt,        /* I: current site */
     const int            Nk,         /* I: nb of classes */
     const ModelParaT*    ParaP,      /* I: parameters (use Beta) */
     const DataT*         DataP,      /* I */
     const NeighDataT*    NeighDataP, /* I: neighborhood system */
     GetNeighFT*          FGetNeigh,  /* I: function to fetch neighbors */
     const double*        PkfkiM,     /* I: pk fk(xi) */
     const double*         Cin_NK,     /* I: class of neighbors */
     double*               Cout_K,     /* O: class of site Ipt */
     PtNeighsT*           NeighsP,    /* W: to store fetched neighbors */
     double*              Cinum_K     /* W: to store cik's numerators */
    );




#endif


/* Called by ComputeLocalProba */

    static double SumNeighsOfClass
     (
        int             K ,     /* I : class to search, 0..Nk-1 */
        int             Nbn ,   /* I : number of neighbours */
        int             Nk,     /* I : number of classes */
        const NeighT*   NeiV ,  /* I : neighbours, [Nbn] */
        const double*    CM      /* I : classification matrix [Npt*Nk] */
     ) ;


/* Called by WriteLogCrit */

    static StatusET ComputeCrit
     (
        int                 Npt,         /* I */
        int                 Nk,          /* I */
        double               Beta,        /* I : spatial coefficient */
        const double*        CM,          /* I : class. matrix [Npt*Nk] */
        const SpatialT*     SpatialP,    /* I : neighborhood system */
        const AlgoParaT*    AlgoParaP,
        const DataT*        DataP,
        WorkingT*           WorkP,       /* I :PkFkiM, O:Neighs */
        CriterT*            CriterP      /* O : computed criteria */
     ) ;


/* Called by ComputeCrit */

#if 0
    static double SumNeighsOfClass
     (
        int             K ,     /* I : class to search, 0..Nk-1 */
        int             Nbn ,   /* I : number of neighbours */
        int             Nk,     /* I : number of classes */
        const NeighT*   NeiV ,  /* I : neighbours, [Nbn] */
        const double*    CM      /* I : classification matrix [Npt*Nk] */
     ) ;

static void CalcError                /*V1.04-e*/
      (
       const double*     CM,       /* I : found classification */
       const int        N,        /* I : number of points */
       const int        Harden,   /* I : 1 use hardened CM, 0 use CM itself */
       const ErrinfoT*  ErrinfoP, /* I : info to compute error */
       ErrcurT*         ErrcurP   /* O : computed error and agreement */
      ) ;

#endif


/* Called by ClassifyByAlgo */



static int   ClassifyByFieldAlgo
           (
             const AlgoParaT     *AlgoParaP,
             const SpatialT      *SpatialP,
             DataT               *DataP,
             StatModelT          *StatModelP,
             MSCritT             *MSCritP ,
             VarParamT           *VarParamP ,
             double              *Increase,
             double              *PostM,
             WorkingT*           WorkingP,
             double              *EBlist
           ) ;

/* Called by ClassifyByFieldAlgo */

int FieldAlgoSingl /* estimates first potential on singletons with an EM
		      algorithm */
(
 const DataT*        DataP,
 const AlgoParaT*    AlgoParaP,
 const SpatialT*     SpatialP,
 StatModelT*         StatModelP,
 double*             PostM,
 MSCritT*            MSCritP,
 VarParamT*          VarParamP,
 double*             Increase,
 WorkingT*           WorkP,
 int*                LastCPUtime
);

/* Called by FieldAlgoSingl */

int Algo /* Calls either SimulatedAlgo
	     or FieldAlgo. It depends on AlgoParaP->Algo */
(
 const DataT*        DataP,
 const AlgoParaT*    AlgoParaP,
 const SpatialT*     SpatialP,
 StatModelT*         StatModelP,
 double*             PostM,
 MSCritT*            MSCritP,
 VarParamT*          VarParamP,
 double*             Increase,
 WorkingT*           WorkP,
 int*                LastCPUtime
);


/* int FieldAlgo */
/*              ( */
/*                const DataT*        DataP, */
/*                const AlgoParaT*    AlgoParaP, */
/*                const SpatialT*     SpatialP, */
/*                const ModelSpecT*   SpecP, */
/*                const SampleDesT*   DescP, */
/*                ModelParaT*         ParaP, */
/*                double*              PostM, */
/*                MSCritT*            MSCritP, */
/*                double*              Increase, */
/*                WorkingT*           WorkP, */
/*                int*            LastCPUtime */
/* 	       ); */

 int FieldAlgoNeighUpdating
             (
               const DataT*        DataP,
               const AlgoParaT*    AlgoParaP,
               const SpatialT*     SpatialP,
               const ModelSpecT*   SpecP,
               const SampleDesT*   DescP,
               ModelParaT*         ParaP,
               double*             PostM,
               MSCritT*            MSCritP,
               WorkingT*           WorkP
	       );



void Proportion2Potentials /* Transforms pi_k to potentials V_k using
			      logarithm */
(
 ModelSpecT* SpecP, /* To get the number of classes */
 double* prop,      /* The proportions pi_k */
 double* singl      /* The potentials V_k */
 );


 static int ComputePostPartition
           (
            const ModelSpecT*    SpecP,       /* I */
            const ModelParaT*    ParaP,
            const DataT*         DataP,       /* I */
            const SpatialT*      SpatialP,
	    const AlgoParaT*     AlgoParaP,    /* I */
            const double*         ZtildeM,      /* z~ik matrix */
            WorkingT*      WorkP,
            double*               PostM       /* P~ik matrix */

           ) ;


static int ComputeGibbsPartition (
  const ModelSpecT*  SpecP,
  const ModelParaT*  ParaP,                 /* To catch the parameters */
  const DataT*       DataP,                 /* Current values of the pixels */
  const SpatialT*    SpatialP,              /* Spatial information */
  PtNeighsT*         NeighsP,               /* also for spatial information */
  int                ipt,                   /* Current point of computation */
  double*            PostM                  /* Posterior probabilities of all
					       the points belongs to each
					       class */
);


 static int ComputeLocalPostPartition
     (
              const ModelSpecT*    SpecP,
              const ModelParaT*    ParaP,
              const DataT*         DataP,        /* I */
              const SpatialT*    SpatialP,
              const int          ipt,
              const int          Nk,
              PtNeighsT*   NeighsP,
              const double*       ZtildeM,
              const double        Beta,
              const double*       Singl,
              const double*       FikM,
              double*             PostM

     );

 static int ComputePriorPartition
           (
            const ModelSpecT*    SpecP,       /* I */
            const ModelParaT*    ParaP,
            const DataT*         DataP,       /* I */
            const SpatialT*      SpatialP,
	    const AlgoParaT*     AlgoParaP,    /* I */
            const double*         ZtildeM,      /* z~ik matrix */
            WorkingT*            WorkP,
            double*               PriorM

           ) ;

 static double ComputeFunc
       (
        const ModelSpecT*    SpecP,       /* I */
        const ModelParaT*    ParaP,
        const DataT*         DataP,       /* I */
        const SpatialT*      SpatialP,
  	const AlgoParaT*     AlgoParaP,    /* I */
        WorkingT*            WorkP,
        const double*         PostM,         /* P~ in the field algorithms */
        const double*         Ztilde

        );

  static int ComputeNRJCond
       (
        const int          Npt,
        const int          Nk,
        const double*      ClassifM, /* must be hard classification */
        const double*      ZtildeM,
        const double*      Singl,
        const double       beta,
        const double*      LogFikM,
        const SpatialT*    SpatialP,
        const int          a,
        WorkingT*          WorkP,
        double*            Ecompute
       );

 static int ComputeNRJMarg
       (
        const int          Npt,
        const int          Nk,
        const double*      ClassifM,
        const double*      ZtildeM,
        const double*      Singl,
        const double       beta,
        const SpatialT*    SpatialP,
        const int          a, /* specify if H or Hmf must be computed */
        WorkingT*          WorkP,
        double*            Ecompute
       );
 

 static StatusET ComputeDensity
           (
             const ModelParaT*   ParaP,
             const DataT*        DataP,
             const ModelSpecT*   SpecP,
             double*              FikM,
             double*              LogFkiM
           );


static double ComputeLogWMarg
    (
             const int           Npt,          /* nb of points */
             const int           Nk,           /* nb of classes */
             const double         b,            /* point where compute the function */
             const double*        Ztilde,
             const double*       Singl,
             const SpatialT*     SpatialP,     /* neiborhood system */
             PtNeighsT*          NeighsP
    /* compute the  mean field like approximation of the marg normalization constant W  */
          );

static double ComputeLogWCond
    (
             const int           Npt,          /* nb of points */
             const int           Nk,           /* nb of classes */
             const double        b,            /* point where compute the function */
             const double*       Ztilde,
             const double*       Singl,
             const double*       FikM,
             const SpatialT*     SpatialP,     /* neiborhood system */
             PtNeighsT*          NeighsP
    /* compute the  mean field like approximation of the cond normalization constant W  */
          );

/* New version of ComputeCritGradforField */

static double ComputeGradBeta
(
 const int           Npt,          /* nb of points */
 const int           Nk,           /* nb of classes */
 const double*       Beta,
 const double*       Ztilde,
 const double*       Singl,
 const double*       PostM,
 const SpatialT*     SpatialP,     /* neiborhood system */
 PtNeighsT*          NeighsP,
 ParameterT          ParameterP    /* nearly useless here but...*/
 );

static double ComputeGradPotential
(
 const int           Npt,          /* nb of points */
 const int           Nk,           /* nb of classes */
 const double*       Beta,
 const double*       Ztilde,
 const double*       Singl,
 const double*       PostM,
 const SpatialT*     SpatialP,     /* neiborhood system */
 PtNeighsT*          NeighsP,
 ParameterT          ParameterP    /* To know which is the number of the
				      potential that will serve as
				      derivation variable */
 );



 static void SimulateZ
           (
	     const AlgoParaT*     AlgoParaP,
             const int*          LabelV,
             const double*        PostM,
             const int           Npt,          /* nb of points */
             const int           Nk,           /* nb of classes */
             double*              simulZ       /* result of max(F') in z~ */
           );

 static void SimulateZSeq
           (
	     const AlgoParaT*     AlgoParaP,
             const int*          LabelV,
             const SpatialT*     SpatialP,
             const ModelSpecT*   SpecP,
             const ModelParaT*    ParaP,
             const DataT*         DataP,
             const double        Beta,
             const double*       Singl,
             WorkingT*           WorkP,
             const int           Npt,
             const int           Nk,
             double*             ZtildeM,
             double*             PostM
            );

void CreateGibbs(
		 DataT*      DataP,
		 AlgoParaT*  AlgoParaP,
		 StatModelT* StatModelP,
		 SpatialT*   SpatialP
);


 static int EstimBetaForFieldAlgo
           (
             const BetaET        BetaModel,  /* I : beta estimation mode */
             const BtaPsGradT    BtaPsGradP,   /* gradient ascent  parameters */
             const SpatialT*     spatialP,     /* neiborhood system */
             const double*       ZtildeM,       /* current z~ */
             const double*       PostM,
             const double*       Singl,
             const int           Npt,          /* nb of points */
             const int           Nk,           /* nb of classes */
             double*             BetaP,        /* previous and next Beta */
             PtNeighsT*          NeighsP      /* to store a site's neighbors */
            );


 static int SimulatedAlgo
           (
             const DataT*        DataT,
             const AlgoParaT*    AlgoParaP,
             const SpatialT*     SpatialP,
             StatModelT*         StatModelP,
             double*              PostM,
             MSCritT*            MSCritP,
             VarParamT*          VarParamP,
             double*              Increase,
             WorkingT*           WorkP,
             int*            LastCPUtime
           );




  static double IncreaseTest
   (
             const ModelSpecT*    SpecP,       /* I */
             const ModelParaT*    ParaP,
             const DataT*         DataP,       /* I */
             const SpatialT*      SpatialP,
             const AlgoParaT*     AlgoParaP,   /* I */
             WorkingT*            WorkP,       /* z~ and P~ at iteration q-1 */
             double*               newz         /* z~ at iteration q */

           );

  static int ComputeProps
   (
     const int*      Hclassif,
     const int         nk,
     const int         npt,
     double*            InitProps
   );

static void ComputeErrorRate (
			      const double*      Hclassif,
			      const double*      DataTrue,
			            double*      ErrorRate,
			      const int       nk,
			      const int       npt,
			      const int       nl,
			      const int       nc
			      );


/* ==================== GLOBAL FUNCTION DEFINITION =================== */

/* ------------------------------------------------------------------- */
int ClassifyByAlgo                                             /*V1.04-a*/
        (
          const AlgoParaT     *AlgoParaP,        /* I */
          const SpatialT      *SpatialP,        /* I */
          DataT               *DataP,           /* I/O */
          StatModelT          *StatModelP,      /* I/O */
          double              *ClassifM,
          WorkingT            *WorkingP,
          MSCritT             *MSCritP,
          VarParamT           *VarParamP,
          double              *Increase,
	  CriterT             *CriterP,          /* O */  /*V1.03-f*/
          double              *EBlist
        )
/* ------------------------------------------------------------------- */
{
  int  sts  ;


  /*V1.06-f*/
  ModelPreprocess( & StatModelP->Spec, DataP ) ;

  switch ( AlgoParaP->Algo)
   {
   case ALGO_MEANF:
   case ALGO_GSF:
   case ALGO_ICM:
     sts = ClassifyByFieldAlgo( AlgoParaP, SpatialP, DataP, StatModelP,
                  MSCritP, VarParamP,Increase, ClassifM,WorkingP,  EBlist);
     break;


   case ALGO_NEM:
   case ALGO_NCEM:

       switch( StatModelP->Spec.BetaModel )
        {
          case BETA_FIX:
          case BETA_ZEROIN:
          case BETA_PSGRAD:
             sts = ClassifyByNemOneBeta( DataP, AlgoParaP, SpatialP,
				    StatModelP, ClassifM,WorkingP, CriterP,MSCritP) ;
          break;

         case BETA_HEUD:
         case BETA_HEUL:
            sts = ClassifyByNemHeuBeta( DataP, (AlgoParaT*) AlgoParaP, SpatialP,
				  StatModelP, ClassifM,WorkingP, CriterP,MSCritP ) ;
         break ;

         default:
            fprintf( stderr , "Error : unkwnown beta estimation mode %d\n",
	       StatModelP->Spec.BetaModel ) ;
            sts = STS_E_FUNCARG ;
         }

        break;

   default:
     fprintf( stderr, "Error, unknown algorithm %d\n", AlgoParaP->Algo );
     sts = STS_E_FUNCARG ;
  }


  return sts ;
}



/* ------------------------------------------------------------------- */
int ReadParamFile
         (
	     const ModelSpecT ModelSpecP,
             const char  *FileName,
             const int   K,  /* nb of classes */
             const int   D,   /* nb of variables */
             double*       Center,
             double*       Disp
         )
     /* check if it's OK for models other than norm */
/* ------------------------------------------------------------------- */
{
    FILE        *fp ;
    float      m = 0.0;
    float      v = 0.0 ;
    int         nbscan = 0;
    int         ik = 0;
    int         id = 0;
    int         idd = 0;

if ( strcmp( FileName , "-" ) != 0 )   /*V1.04-c*/
      {
	if ( ( fp = fopen( FileName, "r" ) ) == NULL )
	  {
	    fprintf( stderr, "File %s does not exist\n", FileName ) ;

	    return STS_E_FILEIN ;
	  }
      }
else
      fp = stdin ;


 if (ModelSpecP.ClassFamily == FAMILY_NORMALDIAG)
   {
    for (ik = 0; ik<K; ik++)
           {
            for (id = 0; id<D; id++)
	     {
              nbscan= fscanf( fp, "%f", &m) ;
              if (nbscan == 1)
                {
                 Center[ik*D + id] = m ;
                }
	     }
            for (id = 0; id<D; id++)
	     {
               nbscan= fscanf( fp, "%f", &v) ;
               if (nbscan == 1)
                {
                   Disp[ik*D +  id] = v ;
                }
	     }

	   }

   }



 if ( (ModelSpecP.ClassFamily == FAMILY_NORMALGEN) ||
      (ModelSpecP.ClassFamily == FAMILY_LAPLACE ) )
   {
     for (ik = 0; ik<K; ik++)
     {
       for ( id = 0; id<D; id++)
	 {
          nbscan= fscanf( fp, "%f", &m) ;
          if (nbscan == 1)
                 Center[ik*D + id] = m ;
	 }

       for ( id = 0; id<D; id++)
	 {
           for ( idd = id; idd<D; idd++)
	   {
            nbscan= fscanf( fp, "%f", &v) ;
            if (nbscan == 1)
	      {
		if ((ModelSpecP.ClassDisper == DISPER_KD ) ||
                       (ModelSpecP.ClassFamily == FAMILY_LAPLACE ) )
                 Disp[ik*D*(D+1)/2+ id*D+idd-id*(id+1)/2] = v ;

                if (ModelSpecP.ClassDisper == DISPER__D )
                   Disp[id*D+idd-id*(id+1)/2] = v ;
	      }
	   }
	 }
     }

   }

if ( strcmp( FileName , "-" ) != 0 )   /*V1.04-c*/
      fclose( fp ) ;

return STS_OK;

} /* end of ReadParamFile() */

/* ------------------------------------------------------------------- */
 int  ReadListFile
         (
          const char  *ListName,
          double       **EBlist
         )

/* ------------------------------------------------------------------- */
{
    double       *List ;
    FILE        *fp ;
    int         l ; /* nb of couples (E,Beta) of the list */
    int         i ;
    float       e,b ;
    int         nbscan = 0;



    if ( strcmp( ListName , "-" ) != 0 )   /*V1.04-c*/
      {
	if ( ( fp = fopen( ListName, "r" ) ) == NULL )
	  {
	    fprintf( stderr, "File %s does not exist\n", ListName ) ;
	    return STS_E_FILEIN ;
	  }
      }
    else
      fp = stdin ;

    nbscan = fscanf( fp, "%d", &l) ;
    if ( ( *EBlist = GenAlloc( (l+1)*2, sizeof( double ),
			     0, "ReadListFile", ListName ) ) == NULL )
      return STS_E_MEMORY ;

    List = *EBlist ;


   /* EBlist is a (l+1) * 2 matrix. The 0eme line is (l,0).  */
   /*   The ieme line is  the couple (E(i), Beta(i)). */
   /* They are store in  the increasing order of the Beta's values. */

              if (nbscan == 1)
                {
		  List[0] = l ;}

      for ( i = 1; i<l+1; i++)
        {
             nbscan= fscanf( fp, "%f", &e) ;
              if (nbscan == 1)
                {
                 List[i*2] = e ;

                 nbscan= fscanf( fp, "%f", &b) ;
                 if (nbscan == 1)
                  {
                   List[i*2+1] = b ;
                  }
		}
              else  { fprintf( stderr, "PB") ;}
	}

   if ( strcmp( ListName , "-" ) != 0 )   /*V1.04-c*/
      fclose( fp ) ;

    return STS_OK;

} /* end of ReadListFile */

/* ------------------------------------------------------------------- */
int ComputeMAP                     /* ret : MAP label in 0..Nk-1 */
        (
          const double* ClassifM,   /* I */
          int          Ipt,        /* I */
          int          Nk,         /* I */
	  TieET        TieRule,    /* I */
          int*         kmaxesV     /* I/O [Nk] */
        )
/* ------------------------------------------------------------------- */
{
    int   k ;
    int   nequal ;
    int   kmax = 0;
    double ukmax = ClassifM[ Ipt * Nk + 0 ] ;


    /* ... First pass : find out the first maximum uik */
    for ( k = 1 ; k < Nk ; k ++ ) {

        double uik = ClassifM[ (Ipt * Nk) + k ] ;
        if ( uik > ukmax )
          {
            ukmax = uik ;
            kmax = k ;
          }
    } /* end for ( k = 1 ... ) */

    if ( TieRule == TIE_RANDOM ) {
      /* ... Second pass : check if there are other k / uik = ukmax */
      kmaxesV[ 0 ] = kmax ;
      for ( k = kmax + 1, nequal = 0 ; k < Nk ; k ++ ) {

        double uik = ClassifM[ (Ipt * Nk) + k ] ;
        if ( uik == ukmax )
          {
            nequal ++ ;
            kmaxesV[ nequal ] = k ;
          }
      }

      /* ... If more than one k / uik = ukmax, choose randomly one of them */
      if ( nequal > 0 ) {
        /* Choose randomly one of the k */
        int  imax ;   /* index of chosen k : 0 .. nequal */

        imax = RandomInteger( 0, nequal ) ;
        kmax = kmaxesV[ imax ] ;
      }
      else {
        kmax = kmaxesV[ 0 ] ;
      }
    } /* else TieRule == TIE_FIRST : keep first chosen maximum */

    return kmax ;

}  /* end of ComputeMAP() */


/* ------------------------------------------------------------------- */
void LabelToClassVector
 (
  const int Nk,    /* I: number of classes */
  const int Label, /* I: class of interest 0..Nk-1 */
  double* Cout_K    /* O: classification vector [Nk] */
 )
/* ------------------------------------------------------------------- */
{
  int k;

  for ( k = 0 ; k < Nk ; k ++ )
    Cout_K[ k ] = 0.0 ;

  if ( ( 0 <= Label ) && ( Label < Nk ) )
    Cout_K[ Label ] = 1.0 ;
}






/* ------------------------------------------------------------------- */
int  InitPartitionSort
        (
	 const DataT*    DataP,      /* I */
	 int             Nk,         /* I */
	 int             SortedVar,  /* I : sorted variable : 0..dim-1 */
	 double           *ClassifM   /* O (to allocate before call) */
	)
/* ------------------------------------------------------------------- */
{
    int         ipt ;
    SortPtT*    tabptV ;
    int         npt = DataP->NbPts ;
    int         nd  = DataP->NbVars ;


    /* Allocate the local sorting table to be sorted */
    if ( ( tabptV = GenAlloc( npt, sizeof( SortPtT ),
			      0, "InitPartitionSort", "tabptV" ) ) == NULL )
      return STS_E_MEMORY ;


    /* Fill the local sorting table with data's selected variable */
    for ( ipt = 0 ; ipt < npt ; ipt ++ )
    {
        tabptV[ ipt ].Index = ipt ;
        tabptV[ ipt ].Value = DataP->PointsM[ ( ipt * nd ) + SortedVar ] ;
    }

    /* Sort table by ascending variable value */
    qsort( (void *) tabptV, npt, sizeof( SortPtT ), CompSortValue ) ;

    /* Use table to partition the data into Nk clusters of equal size */
    for ( ipt = 0 ; ipt < npt ; ipt ++ )
    {
        int index   = tabptV[ ipt ].Index ; /* true index of point */
        int ki      = ( ipt * Nk ) / npt ;  /* class to affect to point */
        int otherk ;                        /* other classes */

        for ( otherk = 0 ; otherk < Nk ; otherk ++ )
        {
            ClassifM[ (index * Nk) + otherk ] = 0.0 ;
        }

        ClassifM[ (index * Nk) + ki ] = 1.0 ;
    }

    /* Free the local sorting table */
    GenFree( tabptV ) ;

    return STS_OK ;

}  /* end of InitPartitionSort() */


/* ==================== LOCAL FUNCTION DEFINITION =================== */




/* ------------------------------------------------------------------- */
static void ModelPreprocess
        (
	  const ModelSpecT*   SpecP,            /* I */
	  DataT*              DataP             /* I/O */
	)
/* ------------------------------------------------------------------- */
{
  const char* func = "ModelPreprocess" ;

  SortPtT*    tabptV ;
  int         npt = DataP->NbPts ;
  int         nd  = DataP->NbVars ;
  int         d ;
  int         ipt ;


  switch( SpecP->ClassFamily )
    {
    case FAMILY_LAPLACE:  	/* Compute sorted index */
      {
	/* Allocate globally the sorted index */
	if ((DataP->SortPos_ND = GenAlloc( npt * nd, sizeof( int ),
				      1, func, "SortPos_ND" )) == NULL)
         fprintf(stderr," impossible to allocate DataP->SortPos_ND in ModelPreprocess\n") ;

	/* Allocate the local sorting table to be sorted */
	if ((tabptV = GenAlloc( npt, sizeof( SortPtT ), 1, func, "tabptV" )) == NULL)
         fprintf(stderr," impossible to allocate tabptV in ModelPreprocess\n")  ;

	/* For each variable */
	for( d = 0 ; d < nd ; d ++ )
	  {
	    /* Fill the local sorting table with data's selected variable */
	    for ( ipt = 0 ; ipt < npt ; ipt ++ )
	      {
		tabptV[ ipt ].Index = ipt ;
		tabptV[ ipt ].Value = DataP->PointsM[ ( ipt * nd ) + d ] ;
	      }

	    /* Sort table by ascending variable value */
	    qsort( (void *) tabptV, npt, sizeof( SortPtT ), CompSortValue ) ;

	    /* Save the sorted indices */
	    for ( ipt = 0 ; ipt < npt ; ipt ++ )
	      DataP->SortPos_ND[ ( ipt * nd ) + d ] = tabptV[ ipt ].Index ;
	  }

	/* Free the local sorting table */
	GenFree( tabptV ) ;
      }
      break ;

    default:
      DataP->SortPos_ND = NULL ;
    }
}



/* ------------------------------------------------------------------- */
static int ClassifyByNemHeuBeta                             /*V1.04-a*/
        (
          const DataT         *DataP,           /* I */
          AlgoParaT            *AlgoParaP,        /* I[M] */
          const SpatialT      *SpatialP,        /* I */
          StatModelT          *StatModelP,      /* I/O */
          double               *ClassifM,        /* I/O */
          WorkingT*           WorkingP ,
	  CriterT             *CriterP,          /* O */  /*V1.03-f*/
          MSCritT             *MSCritP
        )
/* ------------------------------------------------------------------- */
{
  const char* func = "ClassifyByNemHeuBeta" ;

  int    npt = DataP->NbPts ;
  int    nk = StatModelP->Spec.K ;
  int    nbtamax ;  /* number of tested beta values */
  int    nbta ;     /* number of valid beta values */
  double* btaV ;     /* valid beta values [1..nbtamax] */
  double* criV ;     /* criteria value for each valid beta [1..nbtamax] */
  double* clasiniM ; /* initial classification matrix */
  double* clasbestM ;/* "best" obtained classification matrix */
  double  btatest ;  /* currently tested beta */
  double  Dmin ;     /* minimum criterion found so far */
  int    Dincreas ;  /* TRUE if increase of criterion already occurred */
  int    Ddrop ; /* TRUE if drop of criterion slope detected */
  double  prevSlope ;/* previously computed criterion slope */
  double  thisSlope ;/* currently computed criterion slope */
  double  DdropThres ;/* threshold of criterion slope change to detect */
  double  LlossThres ;
  double  Lmax ;
  int    Lfound ;
  int    stop ;
  double  btaEst ;   /* estimated beta */
  char   namecri[ LEN_FILENAME + 1 ] ; /* name of file to save criteria */
  FILE*  fcri ;     /* pointer to file to save criteria */


  /* Initialize to NaN to shut gcc warnings (and catch errors) */
  thisSlope = mknan() ;
  prevSlope = mknan() ;
  Lmax      = mknan() ;
  btaEst    = mknan() ;

  nbtamax = (int) ( AlgoParaP->BtaHeuMax / AlgoParaP->BtaHeuStep ) + 1 ;

  if ((btaV = GenAlloc( nbtamax + 1 , sizeof( double ), 0, func, "btaV" )) == NULL )
      return STS_E_MEMORY ;
  if ((criV = GenAlloc( nbtamax + 1 , sizeof( double ), 0, func, "criV"  )) == NULL )
      return STS_E_MEMORY  ;
  if ((clasiniM = GenAlloc( npt * nk, sizeof( double ), 0, func, "clasiniM" )) == NULL )
      return STS_E_MEMORY  ;
  if ((clasbestM = GenAlloc( npt * nk, sizeof( double ), 0, func, "clasbestM" )) == NULL )
      return STS_E_MEMORY  ;

  if ( ( btaV == NULL ) || ( criV == NULL ) || ( clasiniM == NULL ) )
    return STS_E_MEMORY ;

  /* Save initial classification matrix */
  memcpy( clasiniM, ClassifM,
	  npt * nk * sizeof( double ) ) ;

  /* For tests */  /*V1.04-d*/
  if ( AlgoParaP->Debug )
    {
      strncpy( namecri, AlgoParaP->OutBaseName, LEN_FILENAME ) ;
      strncat( namecri, ".bta", LEN_FILENAME ) ;
      fcri = fopen( namecri , "w" ) ;
    }
  else
    fcri = NULL ;



  /* While bta less than btamax and stopping condition not met */

  Dincreas = FALSE;  /* Initialize values for Hathaway heuristic */
  Dmin = 0.0;
  Ddrop = FALSE ;
  DdropThres = - AlgoParaP->BtaHeuDDrop * DataP->NbPts ;

  Lfound = FALSE ;   /* Initialize values for likelihood heuristic */
  LlossThres= AlgoParaP->BtaHeuLLoss * DataP->NbPts;

  fprintf( stderr, "* * Starting heuristic * *\n" ) ;
  fprintf( stderr, "    Parameters : " ) ;
  if ( StatModelP->Spec.BetaModel == BETA_HEUD )
    fprintf( stderr, "Drop < %3.1f  or  (Dmax - D) < %3.1f (Dmax - Dmin)\n" ,
	     DdropThres, AlgoParaP->BtaHeuDLoss ) ;
  else
    fprintf( stderr, "L < Lmax - %3.1f \n" , LlossThres ) ;

  for ( btatest = 0.0, nbta = 0, stop = FALSE ;
	( btatest <= AlgoParaP->BtaHeuMax ) && ( ! stop ) ;
	btatest += AlgoParaP->BtaHeuStep )
    {
      StatusET sts ;

      fprintf( stderr, "\n * * Testing beta = %5.2f * * \n" , btatest ) ;

      StatModelP->Para.Beta = btatest ;

      /* Reload initial classification matrix */
      memcpy( ClassifM, clasiniM, npt * nk * sizeof( double ) ) ;

      sts = ClassifyByNemOneBeta( DataP, AlgoParaP, SpatialP,
				    StatModelP, ClassifM, WorkingP,CriterP,MSCritP ) ;
      if ( sts == STS_OK )
	{
	  nbta ++ ;

	  btaV[ nbta ] = btatest ;

	  if ( StatModelP->Spec.BetaModel == BETA_HEUD )
	    {
	      criV[ nbta ] = CriterP->D ;

	      if ( CriterP->D < Dmin )
		Dmin = CriterP->D ;

    	      if ( nbta >= 2 )
    		{
    		  prevSlope = thisSlope ;
    		  thisSlope = ( criV[ nbta ] - criV[ nbta - 1 ] ) /
    		    ( btaV[ nbta ] - btaV[ nbta - 1 ] ) ;

    		  if ( thisSlope >= 0.5 * DataP->NbPts ) /* positive slope */
    		    Dincreas = TRUE ;
    		}

    	      if ( nbta >= 3 )
    		{
    		  fprintf( stderr, "    * Drop : %5.1f (threshold %5.1f) *\n",
    			       ( thisSlope - prevSlope ) , DdropThres ) ;

    		  if ( (! Ddrop) && (! Dincreas) ) {

		    if ( (thisSlope - prevSlope) < DdropThres ) {

    		      Ddrop = TRUE ;
		      if ( ! AlgoParaP->Debug )  stop = TRUE ;
    		      fprintf( stderr, "    * ---> Drop detected *\n" ) ;
    		      /* btaEst = btaV[ nbta - 2 ] ; V1.06-i*/
		      btaEst = btaV[ nbta - 1 ] ;

    		    }
		    else /* no drop of slope detected, save this class. */ {

		      memcpy( clasbestM, ClassifM,
			      npt * nk * sizeof( double ) ) ;

		    }

		  } /* Else drop or increase already detected */

    		}
	      else /* 1st or 2nd tested beta -> save classification */
		memcpy( clasbestM, ClassifM,
			npt * nk * sizeof( double ) ) ;

	    }
	  else  /* else Likelihood heuristic */
	    {
	      criV[ nbta ] = CriterP->L ;

    	      if ( nbta < 2 )  /* 1st beta */ {

		  Lmax = CriterP->L ;
		  fprintf( stderr, " * *  L threshold  :  %3.1f  * * \n" ,
			   Lmax - LlossThres ) ;
		  memcpy( clasbestM, ClassifM, npt * nk * sizeof( double ) ) ;
		}
	      else /* >= 2nd beta */
		{
		  if ( CriterP->L > Lmax )
		    {
		      Lmax = CriterP->L ;
		      fprintf( stderr, " * *  L threshold  :  %3.1f  * * \n" ,
			       Lmax - LlossThres ) ;
		    }

    		  if (! Lfound) {

		    if ( CriterP->L < Lmax - LlossThres ) {

    		      Lfound = TRUE ;
		      if ( ! AlgoParaP->Debug )  stop = TRUE ;
    		      fprintf( stderr, "    * ---> L loss detected *\n" ) ;
    		      /* btaEst = btaV[ nbta - 2 ] ; V1.06-i*/
		      btaEst = btaV[ nbta - 1 ] ;
    		    }
		    else /* no drop yet detected -> save classification */ {
		      memcpy( clasbestM, ClassifM,
			      npt * nk * sizeof( double ) ) ;
		    }
		  }
		}
	    }
	  /* end-else Likelihood heuristic */

	  if ( fcri != NULL )
	    fprintf( fcri, "%5.2f  %10.1f  %6.3f  %d\n",
		     btatest ,
		     StatModelP->Spec.BetaModel == BETA_HEUD ?
		     CriterP->D : CriterP->L,
		     CriterP->Errcur.Errorrate,
		     StatModelP->Spec.BetaModel == BETA_HEUD ?
		     Ddrop : Lfound );
	}
      /* end-if ClassifyByNemOneBeta OK */
    }
  /* end-While bta less than btamax and stopping condition not met */

  if ( ( StatModelP->Spec.BetaModel == BETA_HEUD ) && ( ! Ddrop ) ) {
    int ibta ;
    int found ;
    double  DThres ; /* threshold of criterion loss to detect */

    DThres = criV[ 1 ] - ( criV[ 1 ] - Dmin ) * AlgoParaP->BtaHeuDLoss ;

    for ( ibta = 1, found = FALSE ;
	  ( ibta <= nbta ) && ( ! found ) ; ibta ++ )
      found = criV[ ibta ] <= DThres ;

    if ( found )
      btaEst = btaV[ ibta - 2 ] ;
    else {

      fprintf( stderr , "Warning : heuristic failed to detect beta\n" ) ;
      btaEst = 0.0 ;
    }

    /* Reload initial classification matrix */
    memcpy( ClassifM, clasiniM, npt * nk * sizeof( double ) ) ;
  }
  else /* likelihood heuristic, or D heuristic and drop detected */ {

    /* restore saved "best" classification, and use INIT_FILE */ //*V1.06-i*/
    memcpy( ClassifM, clasbestM, npt * nk * sizeof( double ) ) ;
    AlgoParaP->InitMode = INIT_FILE ;
  }

  if ( ( StatModelP->Spec.BetaModel == BETA_HEUL ) && ( ! Lfound ) )
    {
      /* btaEst = btaV[ nbta - 1 ] ; V1.06-i*/
      btaEst = btaV[ nbta ] ;
      fprintf( stderr, "    * L loss not detected *\n" ) ;
    }

  if ( fcri != NULL )
    fclose( fcri ) ;

  fprintf( stderr , "\n * * *  Estimated beta : %3.2f * * *\n" , btaEst ) ;
  if ( StatModelP->Spec.BetaModel == BETA_HEUD )
    fprintf( stderr , " * * *   Using %s * * *\n\n" ,
	     Ddrop ? "drop detection" : "loss thresholding" ) ;


  GenFree( btaV ) ;     /*V1.05-h*/
  GenFree( criV ) ;
  GenFree( clasiniM ) ;
  GenFree( clasbestM ) ;

  StatModelP->Para.Beta = btaEst ;
  return ClassifyByNemOneBeta( DataP, AlgoParaP, SpatialP,
				 StatModelP, ClassifM,WorkingP, CriterP,MSCritP ) ;
} /* end of ClassifyByNemHeuBeta() */



/* ------------------------------------------------------------------- */
static int ClassifyByNemOneBeta                             /*V1.04-a*/
        (
          const DataT         *DataP,           /* I */
          const AlgoParaT      *AlgoParaP,        /* I */
          const SpatialT      *SpatialP,        /* I */
          StatModelT          *StatModelP,      /* I/O */
          double               *ClassifM,       /* I/O */
          WorkingT*           WorkingP ,
	  CriterT             *CriterP,          /* O */  /*V1.03-f*/
          MSCritT             *MSCritP
        )
/* ------------------------------------------------------------------- */
{
    /* const char* func        = "ClassifyByNemOneBeta" ;
    WorkingT    working     = {NULL};    working variables to allocate */
    StatusET    err         = STS_OK ;
    /*int         npt         = DataP->NbPts;*/
    int         nk          = StatModelP->Spec.K ;
    int         k ;
    int         missk ;  /* 0..K-1 : class with missing data (-1 if none) */
    int         missd ;  /* 0..D-1 : variable with missing data (-1 if none) */




    switch( AlgoParaP->InitMode )
    {
    case INIT_EMIND:
      fprintf( stderr," chose another init mode when using EM for mixture for estimating the parameters\n");
      return(STS_E_ARG);

    case INIT_SORT:    /* initial partition to be computed by sorting */
        fprintf( stderr,
                 "Computing initial partition (sort variable %d)\n",
                  AlgoParaP->SortedVar + 1 ) ;

        InitPara( DataP, & StatModelP->Desc, & StatModelP->Spec, AlgoParaP,
		  & StatModelP->Para,
		  WorkingP->CtmpM ) ;

        if ( ( err = InitPartitionSort( DataP, nk,
                                        AlgoParaP->SortedVar,
                                        ClassifM ) ) != STS_OK )
	  return err ;

	/*V1.06-e*/
	err = MakeParaFromLabeled( AlgoParaP, DataP, ClassifM, & StatModelP->Spec,
				   & StatModelP->Desc,
				   & StatModelP->Para, & missk, & missd ) ;
        if ( err != STS_OK )
	  return err ;

     

           switch (AlgoParaP->ParamMode )
             {
           case PARAM_ESTIM:
               break;

           case PARAM_FIX:


              if ( ( err = ReadParamFile( StatModelP->Spec,AlgoParaP->ParamName,        /*V1.04-a*/
                                 StatModelP->Spec.K,
                                 DataP->NbVars,
                                 StatModelP->Para.Center_KD,
                                 StatModelP->Para.Disp_KD
                                 ) ) != STS_OK )

               return err ;

           break;

           default:
                err= STS_OK;

                  } /* end switch ( AlgoParaP->ParamMode ) */



        err = NemAlgo( DataP, AlgoParaP, SpatialP, & StatModelP->Spec,
		       & StatModelP->Desc,
		        & StatModelP->Para, ClassifM,
		       WorkingP, CriterP,MSCritP ) ;
        break ;


    case INIT_FILE:    /* initial classification is given by a file, already loaded */
        

	/*V1.06-g*/
        InitPara( DataP, & StatModelP->Desc, & StatModelP->Spec, AlgoParaP,
		  & StatModelP->Para,
		  WorkingP->CtmpM ) ;

	err = MakeParaFromLabeled( AlgoParaP, DataP, ClassifM, & StatModelP->Spec,
				   & StatModelP->Desc,
				   & StatModelP->Para, & missk, & missd ) ;
        if ( err != STS_OK )
	  return err ;

           switch (AlgoParaP->ParamMode )
             {
           case PARAM_ESTIM:
               break;

           case PARAM_FIX:
              if ( ( err = ReadParamFile(StatModelP->Spec, AlgoParaP->ParamName,        /*V1.04-a*/
                                 StatModelP->Spec.K,
                                 DataP->NbVars,
                                 StatModelP->Para.Center_KD,
                                 StatModelP->Para.Disp_KD
                                 ) ) != STS_OK )
               return err ;

           break;

           default:
                err= STS_OK;

                  } /* end switch ( AlgoParaP->ParamMode ) */
        err = NemAlgo( DataP, AlgoParaP, SpatialP, & StatModelP->Spec,
		       & StatModelP->Desc,
		        & StatModelP->Para, ClassifM,
		       WorkingP, CriterP,MSCritP ) ;
        break ;


    case INIT_PARAM: /* initialization with given params, already loaded */

         /* Proportions are set equal */
         for ( k = 0 ; k < nk ; k ++ )
          StatModelP->Para.Prop_K[ k ] = 1.0 / nk ; /*V1.03-e*/

           switch (AlgoParaP->ParamMode )
             {
           case PARAM_ESTIM:
               break;

           case PARAM_FIX:
              if ( ( err = ReadParamFile(StatModelP->Spec, AlgoParaP->ParamName,        /*V1.04-a*/
                                 StatModelP->Spec.K,
                                 DataP->NbVars,
                                 StatModelP->Para.Center_KD,
                                 StatModelP->Para.Disp_KD
                                 ) ) != STS_OK )
               return err ;

           break;

           default:
                err= STS_OK;

                  } /* end switch ( AlgoParaP->ParamMode ) */

	ComputePartitionFromPara( 1, DataP, AlgoParaP, & StatModelP->Spec,
				  & StatModelP->Para, SpatialP,
				  ClassifM, CriterP,  WorkingP ) ;

        err = NemAlgo( DataP, AlgoParaP, SpatialP, & StatModelP->Spec,
		       & StatModelP->Desc,
		        & StatModelP->Para, ClassifM,
		       WorkingP, CriterP,MSCritP ) ;
        break ;



    case INIT_MIXINI:  /*V1.06-j*/
    case INIT_MIXFIX:
        fprintf( stderr,
		 "Initializing parameters from given value\n");


           switch (AlgoParaP->ParamMode )
             {
           case PARAM_ESTIM:
               break;

           case PARAM_FIX:
              if ( ( err = ReadParamFile(StatModelP->Spec, AlgoParaP->ParamName,        /*V1.04-a*/
                                 StatModelP->Spec.K,
                                 DataP->NbVars,
                                 StatModelP->Para.Center_KD,
                                 StatModelP->Para.Disp_KD
                                 ) ) != STS_OK )
               return err ;

           break;

           default:
                err= STS_OK;

                  } /* end switch ( AlgoParaP->ParamMode ) */

	ComputePartitionFromPara( 1, DataP, AlgoParaP, & StatModelP->Spec,
				  & StatModelP->Para, SpatialP,
				  ClassifM, CriterP,  WorkingP ) ;

	/* Run NEM with this initial partition */
        err = NemAlgo( DataP, AlgoParaP, SpatialP, & StatModelP->Spec,
		       & StatModelP->Desc,
		        & StatModelP->Para, ClassifM,
		       WorkingP, CriterP,MSCritP ) ;
	break ;

    default:   /* Else randomly generate initial classifications */


           switch (AlgoParaP->ParamMode )
             {
           case PARAM_ESTIM:
               break;

           case PARAM_FIX:
              if ( ( err = ReadParamFile(StatModelP->Spec, AlgoParaP->ParamName,        /*V1.04-a*/
                                 StatModelP->Spec.K,
                                 DataP->NbVars,
                                 StatModelP->Para.Center_KD,
                                 StatModelP->Para.Disp_KD
                                 ) ) != STS_OK )
               return err ;

           break;

           default:
                err= STS_OK;

                  } /* end switch ( AlgoParaP->ParamMode ) */
        err = RandNemAlgo ( DataP, AlgoParaP, SpatialP, & StatModelP->Spec,
			    & StatModelP->Desc,
			     & StatModelP->Para, ClassifM,
			    WorkingP, CriterP,MSCritP ) ;

    } /* end switch( AlgoParaP->InitMode ) */



    return err ;

}  /* end of ClassifyByNemOneBeta() */






/* ------------------------------------------------------------------- */
static void InitPara
        (
         const DataT*       DataP,       /* I */
	 const SampleDesT*  DescP,       /* I */
	 const ModelSpecT*  SpecP,       /* I */
         const AlgoParaT*   AlgoParaP,
	 ModelParaT*        ParaP,       /* O */
         double*             C_NK         /* T */
        )
/*\

    This function initializes to NaN all class parameters, computes
    the volume of the whole sample, and minimum and maximum within
    each variable.

\*/
/* ------------------------------------------------------------------- */
{
  StatusET        err;
  int             npt         = DataP->NbPts;
  int             nd          = DataP->NbVars ;
  int             nk          = SpecP->K ;
  int             ipt ;
  int             d, k ;
  int             emptyk ;  /* index of empty class, 0 if none
			       (not used here)*/


  /* Compute minimum and maximum within each variable */
  /*V1.05-d*/
  for ( d = 0 ; d < nd ; d ++ )
    {
      DescP->MiniSam_D[ d ] = MAXDOUBLE ;
      DescP->MaxiSam_D[ d ] = -MAXDOUBLE ;
      for ( ipt = 0 ; ipt < npt ; ipt ++ )
	{
	  double x = DataP->PointsM[ ipt * nd + d ] ;

	  if ( ! isnan( x ) )
	    {
	      if ( x < DescP->MiniSam_D[ d ] )
		DescP->MiniSam_D[ d ] = x ;

	      if ( x > DescP->MaxiSam_D[ d ] )
		DescP->MaxiSam_D[ d ] = x ;
	    }
	}
    }

  /* Compute dispersion of whole sample
     by assigning all data to first class */
  for ( ipt = 0 ; ipt < npt ; ipt ++ )
    {
      C_NK[ ipt * nk + 0 ] = 1.0 ;
      for ( k = 1 ; k < nk ; k ++ )
	{
	  C_NK[ ipt * nk + k ] = 0.0 ;
	}
    }

  err = EstimPara( C_NK, DataP, AlgoParaP, nk, MISSING_IGNORE, SpecP,
	     &emptyk, ParaP ) ;

  for ( d = 0 ; d < nd ; d ++ )
    {
      if (SpecP->ClassFamily == FAMILY_NORMALDIAG)
        DescP->DispSam_D[ d ] =
	         ParaP->Disp_KD[ d ] ;
      if (SpecP->ClassFamily == FAMILY_NORMALGEN)
        DescP->DispSam_D[ d ] =
	         ParaP->Disp_KD[ d*(2*nd+1-d)/2 ] ;
    }

  /* Set to nan all class parameters */
  for ( k = 0 ; k < nk ; k ++ )
    {
      ParaP->Prop_K[ k ]    = mknan() ;
      ParaP->NbObs_K [ k ]  = mknan() ;
      for ( d = 0 ; d < nd ; d ++ )
	{
	  ParaP->Center_KD[ k * nd + d ] = mknan() ;
	  ParaP->NbObs_KD[ k * nd + d ]  = mknan() ;
	}
      switch (SpecP->ClassFamily)
	{
	case FAMILY_NORMALDIAG:
        for ( d = 0 ; d < nd ; d ++ )
	{
	  ParaP->Disp_KD[k * nd + d  ]   = mknan() ;
        }
        break;

	case FAMILY_NORMALGEN:
	  if (SpecP->ClassDisper == DISPER_KD)
	  {
           for ( d = 0 ; d < nd*(nd+1)/2 ; d ++ )
	   {
	     ParaP->Disp_KD[k * nd*(nd+1)/2 + d  ]   = mknan() ;
	     ParaP->NbObs_KDD[ k * nd*(nd+1)/2 + d ]  = mknan() ;
	   }
	  }
         if (SpecP->ClassDisper == DISPER__D)
         {
           for ( d = 0 ; d < nd*(nd+1)/2 ; d ++ )
	   {
	    ParaP->Disp_KD[ d  ]   = mknan() ;
	    ParaP->NbObs_KDD[  d ]  = mknan() ;
	   }
         }
        break;

	case FAMILY_LAPLACE:
        for ( d = 0 ; d < nd*(nd+1)/2 ; d ++ )
	{
	  ParaP->Disp_KD[k * nd*(nd+1)/2 + d  ]   = mknan() ;
	  ParaP->NbObs_KDD[ k * nd*(nd+1)/2 + d ]  = mknan() ;
	}
        break;

	default: break;

	}

    }


} /* end of InitPara */



/* ------------------------------------------------------------------- */

static double ChosenCrit( const CriterT* CriterP , CritET Which )  /*V1.04-f*/
{
  switch( Which )
    {
    case CRIT_U: return CriterP->U ;
    case CRIT_M: return CriterP->M ;  /*V1.05-k*/
    case CRIT_D: return CriterP->D ;
    case CRIT_L: return CriterP->L ;
    default:     return CriterP->D ;
    }
}


/* ------------------------------------------------------------------- */
/*V1.03-d*/
static int MakeParaFromLabeled
        (
            const AlgoParaT*    AlgoParaP,
            const DataT*        DataP,      /* I */
	    const double*        C_NK,       /* I */
	    const ModelSpecT*   SpecP,      /* I */
            const SampleDesT*   DescP,      /* I */
	    ModelParaT*         ParaP,      /* O */
	    int                 *misskP,    /* O */
	    int                 *missdP     /* O */
        )
/*\

     This function initializes the parameters given a classification.
     If a class and variable has no observation, the mean is drawn randomly.
     If a class and variable has < 3 observation,
     the dispersion is set to ( whole sample dispersion / nk ).

     In the latter case, the function returns ( k * D + d ), where k and
     d are the indices of the empty class / variable (0..K-1, 0..D-1).
     If there is no empty class / variable, -1 is returned.

\*/
/* ------------------------------------------------------------------- */
{
    int             nd          = DataP->NbVars ;
    int             nk          = SpecP->K ;
    int             sts         = STS_OK ;
    int             d, k ;
    int             emptyk ; /* index of empty class, 0 if none */

    /*V1.06-d*/

   *misskP = -1 ;
   *missdP = -1 ;

    /* Estimate parameters using known labels */
    sts = EstimPara( C_NK, DataP, AlgoParaP, nk, MISSING_IGNORE, SpecP,
		     &emptyk, ParaP ) ;

    /* If some class is empty, produce an error message */
    if ( sts != STS_OK )
      {
	if ( sts == STS_W_EMPTYCLASS )
	  fprintf( stderr, "Class %d has no labeled observation\n", emptyk ) ;
	return sts ;
      }

    /* Check mean and dispersion of each class and variable */
    for ( k = 0 ; k < nk ; k ++ )
      for ( d = 0 ; d < nd ; d ++ )
	{
	  /* If there is no observation, draw mean randomly in min-max range */
	  if ( ParaP->NbObs_KD[ k * nd + d ] < EPSILON )
	    {
	      fprintf( stderr ,
		       "Warning: no data in class k=%d, variable=%d\n" ,
		       k + 1 , d + 1 );
	      *misskP = k ;
	      *missdP = d ;

	      ParaP->Center_KD[ k * nd + d ] =
		RandomDouble( DescP->MiniSam_D[ d ],
			     DescP->MaxiSam_D[ d ] ) ;
	    }
	 /*  (Else, this mean could be computed from labeled data) */

	  /* If this class/variable had less than 3 observations,
	     set its dispersion to ( whole sample dispersion / nk ) */
	  if ( ParaP->NbObs_KD[ k * nd + d ] < 3 )
	    {
	      switch (SpecP->ClassFamily)
	      {
	        case FAMILY_NORMALDIAG:
	        ParaP->Disp_KD[k * nd +d ] = DescP->DispSam_D[ d ] / nk ;
                break;

	        case FAMILY_NORMALGEN:
		  if (SpecP->ClassDisper == DISPER_KD)
                  ParaP->Disp_KD[k * nd*(nd+1)/2 +d*(2*nd+1-d)/2  ] =
                                 DescP->DispSam_D[ d ] / nk ;
		  if (SpecP->ClassDisper == DISPER__D)
                  ParaP->Disp_KD[ d*(2*nd+1-d)/2  ] =
                                 DescP->DispSam_D[ d ] / nk ;
                break;

                case FAMILY_LAPLACE:
	        ParaP->Disp_KD[k * nd*(nd+1)/2 +d*(2*nd+1-d)/2  ] =
                                     DescP->DispSam_D[ d ] / nk ;
                break;

	      default: break;
	      }

	    }
	}

    return sts ;

}  /* end of MakeParaFromLabeled() */


/* ------------------------------------------------------------------- */
static StatusET  MakeRandomPara
        (
            const DataT*        DataP,      /* I */
	    const ModelSpecT*   SpecP,      /* I */
            const SampleDesT*   DescP,      /* I */
	    ModelParaT*         ParaP       /* O */
        )
/* ------------------------------------------------------------------- */
{
    int             npt         = DataP->NbPts;
    int             nd          = DataP->NbVars ;
    int             nk          = SpecP->K ;
    int             k ;
    int             d ;


    /* For each class and variable */
    for ( k = 0 ; k < nk ; k ++ )
      for ( d = 0 ; d < nd ; d ++ )
	{
	  /* set its dispersion to ( whole sample dispersion / nk )
	   */
        switch (SpecP->ClassFamily)
	      {
	        case FAMILY_NORMALDIAG:
	        ParaP->Disp_KD[k * nd +d ] = DescP->DispSam_D[ d ] / nk ;
                break;

	        case FAMILY_NORMALGEN:
		  if (SpecP->ClassDisper == DISPER_KD)
                  ParaP->Disp_KD[k * nd*(nd+1)/2 +d*(2*nd+1-d)/2  ] =
                                 DescP->DispSam_D[ d ] / nk ;
		  if (SpecP->ClassDisper == DISPER__D)
                  ParaP->Disp_KD[ d*(2*nd+1-d)/2  ] =
                                 DescP->DispSam_D[ d ] / nk ;
                break;

                case FAMILY_LAPLACE:
	        ParaP->Disp_KD[k * nd*(nd+1)/2 +d*(2*nd+1-d)/2  ] =
                                     DescP->DispSam_D[ d ] / nk ;
                break;

	      default: break;
	      }


	}

    /* Proportions are set equal */
    for ( k = 0 ; k < nk ; k ++ )
      ParaP->Prop_K[ k ] = 1.0 / nk ; /*V1.03-e*/


    /* For each class */
    for ( k = 0 ; k < nk ; k ++ )
    {
        int     ipt = 0 ;  /* = 0 to shut gcc warning */
	int     again ;
	int     ndraw ;

	/* Its mean is picked randomly from the data */

	/*V1.05-m*/
	/* Draw an object again while identical to a previously
	   drawn center */
	for ( ndraw = 0, again = 1 ; again && ( ndraw < 100 ) ; ndraw ++ )
	  {
	    int h ;

	    ipt = RandomInteger( 0 , npt - 1 ) ;

	    /* Compare with a previous class center until one is identical */
	    for ( h = 0, again = 0 ; ( h < k ) && ( ! again ) ; h ++ )
	      {
		int different = 0 ; /* assume no difference until a different
				       coordinate is found */
		for ( d = 0 ; d < nd ; d ++ )
		  {
		    if ( ! isnan( DataP->PointsM[ ipt * nd + d ] )  )
		      {
			if ( ParaP->Center_KD[ h*nd + d ] !=
			     DataP->PointsM[ ipt*nd + d ] )
			  different = 1 ;
		      }
		    else
		      different = 1 ;  /* if this point has a missing
					  coordinate, assume it's different */
		  }

		if ( ! different )
		  again = 1 ;
	      }
	  }


	/* Get mean coordinates from drawn object non-NaN coordinates ;
	   or by uniform sampling if drawn object's coodinate is NaN */
        for ( d = 0 ; d < nd ; d ++ )
	  {
	    /* Use drawn point's dth coordinate if not NaN */
	    if( ! isnan( DataP->PointsM[ ipt * nd + d ] )  )  /*V1.05-d*/
	      {
		ParaP->Center_KD[ k*nd + d ] = DataP->PointsM[ ipt*nd + d ] ;
	      }
	    else  /* else draw randomly in min-max range of variable d */
	      {
		ParaP->Center_KD[ k*nd + d ] =
		  RandomDouble( DescP->MiniSam_D[ d ],
			       DescP->MaxiSam_D[ d ] ) ;
	      }
	  }
    }

    return STS_OK ;

}  /* end of MakeRandomPara() */

/* ------------------------------------------------------------------- */





static int CompSortValue( const void* elt1P, const void* elt2P )
{
    const SortPtT*  p1 = elt1P ;
    const SortPtT*  p2 = elt2P ;

    /*V1.06-h*/
    if ( isnan( p1->Value ) )           return 1 ;  /* 1 ">" 2 NaN at end */
    else if ( isnan( p2->Value ) )      return -1 ; /* 2 ">" 1 NaN at end */
    else if ( p1->Value < p2->Value )   return -1 ;
    else if ( p1->Value > p2->Value )   return 1 ;
    else                                return 0 ;
}




/* ------------------------------------------------------------------- */
static StatusET RandNemAlgo
        (
         const DataT*       DataP,           /* I */
         const AlgoParaT*    AlgoParaP,        /* I */
         const SpatialT*    SpatialP,        /* I */
	 const ModelSpecT*  SpecP,     	     /* I */
	 const SampleDesT*  DescP,     	     /* I */
	 ModelParaT*        ParaP,     	     /* I/O */
         double*             ClassifM,        /* I/O */
         WorkingT*          WorkP,           /* I/O */
         CriterT*           CriterP,          /* O */
         MSCritT*           MSCritP
        )
/* ------------------------------------------------------------------- */
{
  const char* func      = "RandNemAlgo" ;
  StatusET  err         = STS_OK ;
  StatusET  error;
  int       npt         = DataP->NbPts;
  int       nd          = DataP->NbVars;
  int       nk          = SpecP->K ;

  int       irandom ;
  int       nbsucc ;
  int       bestTry = -1 ;  /* stays -1 if no success
			       (to shut gcc warning) */
  CriterT   bestCritS ;    /*V1.04-g*/
  double*    bestClM ;      /* to allocate */
  double*    bestCen_KD ;   /* to allocate */ /*V1.06-q*/
  double*    bestDis_KD ;   /* to allocate */
  double     initbeta ;     /* specified initial value of beta */
  double     bestbeta    = 0 ; /* beta of best classification */


  InitPara( DataP, DescP, SpecP, AlgoParaP,
	    ParaP, WorkP->CtmpM ) ;

  if ( ( bestClM = GenAlloc( npt * nk, sizeof(double),
			     0, func, "bestClM" ) ) == NULL )
    return STS_E_MEMORY ;

  if ( ( bestCen_KD = GenAlloc( nk * nd, sizeof(double),
			     0, func, "bestCen_KD" ) ) == NULL )
    return STS_E_MEMORY ;

  if ( (SpecP->ClassFamily == FAMILY_NORMALGEN )
                          && (SpecP->ClassDisper == DISPER__D) )
    {
     if ( ( bestDis_KD = GenAlloc( nd *( nd +1 )/2, sizeof(double),
			     0, func, "bestDis_KD" ) ) == NULL )
     return STS_E_MEMORY ;
    }

  if ( (SpecP->ClassFamily == FAMILY_NORMALGEN )
                          && (SpecP->ClassDisper == DISPER_KD) )
    {
     if ( ( bestDis_KD = GenAlloc(nk* nd *( nd +1 )/2, sizeof(double),
			     0, func, "bestDis_KD" ) ) == NULL )
     return STS_E_MEMORY ;
    }

  if (SpecP->ClassFamily == FAMILY_NORMALDIAG )
    {
     if ( ( bestDis_KD = GenAlloc( nk * nd, sizeof(double),
			     0, func, "bestDis_KD" ) ) == NULL )
     return STS_E_MEMORY ;
    }

  if (SpecP->ClassFamily == FAMILY_LAPLACE )
    {
     if ( ( bestDis_KD = GenAlloc(nk* nd *( nd +1 )/2, sizeof(double),
			     0, func, "bestDis_KD" ) ) == NULL )
     return STS_E_MEMORY ;
    }


  initbeta = ParaP->Beta ;

  /* For each random start */
  for ( irandom = 0, nbsucc = 0 ;
      irandom < AlgoParaP->NbRandomInits ;
      irandom ++ )
  {
    fprintf( stderr,
             "\nRandom initial partition  %d\n", irandom + 1 ) ;


    /* Generate randomly initial parameters */
    err = MakeRandomPara( DataP , SpecP, DescP,
			  ParaP );
    if ( err != STS_OK )
	return err ;

    if (SpecP->BetaModel == BETA_PSGRAD ) {
      if ( AlgoParaP->BtaPsGrad.RandInit ) /* random initial beta */
	ParaP->Beta = RandomDouble( 0.0, 2.0 );
      else   /* Specified initial beta */
	ParaP->Beta = initbeta ;
    }

    /* Initialize classification with zeros to remove eventual NaN/Inf from
       previous runs */
    {
      int ipt, k ;
      for ( ipt = 0 ; ipt < npt ; ipt ++ ) {
	for ( k = 1 ; k < nk ; k ++ ) {
	  ClassifM[ ipt * nk + k ] = 0.0 ;
	}
      }
    }

    ComputePartitionFromPara( 1, DataP, AlgoParaP, SpecP,
			      ParaP, SpatialP,
			      ClassifM, CriterP,  WorkP ) ;

    /* Run NEM with this initial partition */
    err = NemAlgo( DataP, AlgoParaP, SpatialP, SpecP, DescP,
		    ParaP, ClassifM,
		   WorkP, CriterP,MSCritP ) ;

    /* If NEM was successful */
    if ( err == STS_OK )
    {
        /* Count this success */
        nbsucc += 1 ;

        /* If it is the first success */
        if ( nbsucc == 1 )
        {
            /* Memorize the results as the currently best ones */
            memcpy( bestClM, ClassifM, npt * nk * sizeof(double) ) ;
	    bestCritS = *CriterP ;
            bestTry   = irandom ;
	    bestbeta  = ParaP->Beta ;
        }
        else    /* Else it is the 2nd or more success */     {
	  /* If the criterion is better than the current one,
	     Memorize the results */
	  if ( ChosenCrit( CriterP , AlgoParaP->Crit ) >
	       ChosenCrit( & bestCritS , AlgoParaP->Crit ) )  {

	    bestCritS = *CriterP ;
	    bestTry   = irandom ;
	    bestbeta  = ParaP->Beta ;
	    memcpy( bestClM, ClassifM, npt * nk * sizeof(double) ) ;
	    memcpy( bestCen_KD, ParaP->Center_KD, nk * nd * sizeof(double) );
            if ( (SpecP->ClassFamily == FAMILY_NORMALGEN )
                          && (SpecP->ClassDisper == DISPER__D) )
	       memcpy( bestDis_KD, ParaP->Disp_KD,nd *( nd +1 )/2*sizeof(double) );

            if ( (SpecP->ClassFamily == FAMILY_NORMALGEN )
                          && (SpecP->ClassDisper == DISPER_KD) )
	       memcpy( bestDis_KD, ParaP->Disp_KD,nk * nd *( nd +1 )/2* sizeof(double) );

            if (SpecP->ClassFamily == FAMILY_NORMALDIAG )
               memcpy( bestDis_KD, ParaP->Disp_KD, nk * nd * sizeof(double) );

            if (SpecP->ClassFamily == FAMILY_LAPLACE )
	       memcpy( bestDis_KD, ParaP->Disp_KD,nk * nd *( nd +1 )/2* sizeof(double) );
	  }
        }
    } /* end if NEM was successful */
  } /* end for each random start */

  /* If at least one success, return OK,
     and restore best classification results */
  if ( nbsucc > 0 ) {
    int emptyk ; /* index of empty class, 0 if none */

    err = STS_OK ;
    memcpy( ClassifM, bestClM, npt * nk * sizeof(double) ) ;

    /* restore estimated centers/disp,
       used if missing data and replace mode */
    memcpy( ParaP->Center_KD, bestCen_KD, nk * nd * sizeof(double) );
    if ( (SpecP->ClassFamily == FAMILY_NORMALGEN )
                          && (SpecP->ClassDisper == DISPER__D) )
       memcpy(ParaP->Disp_KD, bestDis_KD, nd *( nd +1 )/2* sizeof(double) );

    if ( (SpecP->ClassFamily == FAMILY_NORMALGEN )
                          && (SpecP->ClassDisper == DISPER_KD) )
       memcpy(ParaP->Disp_KD, bestDis_KD,nk * nd *( nd +1 )/2* sizeof(double) );

    if (SpecP->ClassFamily == FAMILY_NORMALDIAG )
       memcpy( ParaP->Disp_KD, bestDis_KD, nk * nd * sizeof(double) );

    if (SpecP->ClassFamily == FAMILY_LAPLACE )
	memcpy(ParaP->Disp_KD, bestDis_KD ,nk * nd *( nd +1 )/2* sizeof(double) );

    error = EstimPara( ClassifM, DataP,AlgoParaP, nk, AlgoParaP->MissMode, SpecP,
	       &emptyk, ParaP ) ;
    ParaP->Beta = bestbeta ;

    *CriterP = bestCritS ;

    fprintf( stderr,
             "Best start was %d (%s = %g)\n", bestTry+1,
	     CritStrVC[ AlgoParaP->Crit ] ,
	     ChosenCrit( CriterP , AlgoParaP->Crit ) ) ;

    if ( CriterP->Errinfo.Kr != 0 )
      fprintf( stderr, "Error of best start = %5.1f\n",
	       100.0 * CriterP->Errcur.Errorrate ) ;

  }
  /* otherwise, return last error status */

  GenFree( bestDis_KD ) ;
  GenFree( bestCen_KD ) ;
  GenFree( bestClM ) ;

  return err ;

} /* end RandNemAlgo() */


/* ------------------------------------------------------------------- */
static int NemAlgo
        (
         const DataT*       DataP,           /* I */
         const AlgoParaT*    AlgoParaP,        /* I */
         const SpatialT*    SpatialP,        /* I */
	 const ModelSpecT*  SpecP,     	     /* I */
	 const SampleDesT*  DescP,     	     /* I */
	 ModelParaT*        ParaP,     	     /* I/O */
         double*             CM,              /* I/O */
         WorkingT*          WorkP,           /* I/O */
         CriterT*           CriterP,          /* O */
         MSCritT*           MSCritP
        )
/* ------------------------------------------------------------------- */
{
    StatusET        err         = STS_OK ;
    int             npt         = DataP->NbPts;
    int             nk          = SpecP->K ;
    int             result;     /* to store EstimBeta result */
    int             ipt ;       /* 0..npt-1 : index of current point */
    int             ik ;        /* 0..nk-1  : index of current class */
    int             iter ;      /* 0..NbIters-1 : current iteration */
    int             converged ; /* TRUE if convergence reached */
    double          oldcrit ;   /* criterion at previous iteration */
    double          oldbeta ;   /* beta at previous iteration */
    /*clock_t  */
    int             CPUtime = 0;
    int             emptyk ;    /* 1..nk = index of empty class (0 = none) */
    double*         Hclassif;
    ParameterT      Parameter;  /* For call to ZeroIn */

    ComputeCritGrad*  f;       /* for Beta estimation by ZeroIn */

   /*  fprintf(stderr,"Entry of NemAlgo.\n"); */

    if ((Hclassif = GenAlloc( npt * nk, sizeof(double), 0, "NemAlgo",
			      "Hclassif")) == NULL)
      {
        fprintf(stderr,"impossible to allocate Hclassif in function NemAlgo\n");
        return STS_E_MEMORY;
      }

    // fprintf(stderr,"NemAlgo : Memory allocated.\n");

    f = ComputeGradBeta; //f= ComputeCritGradforField;

    //fprintf(stderr,"NemAlgo : f assignated.\n");
    /* Initialize old partition and criterion to extreme values
       (to make sure we skip the first convergence test)
    */
    //fprintf(stderr,"NemAlgo : npt = %d\n",npt);
    for( ipt = 0 ; ipt < npt ; ipt ++ ) {
      for ( ik = 0 ; ik < nk ; ik ++ ) {
	//fprintf(stderr,"NemAlgo : loop %d %d.\n",ipt,ik);
	WorkP->ColdM[ ( ipt * nk ) + ik ] = 0.0 ;
      }
    }
    //fprintf(stderr,"Leaving the loop\n");
    oldcrit = -MAXDOUBLE ;



   /*  fprintf(stderr,"NemAlgo : initialisations done.\n"); */
    fprintf( stderr, "  Iterations : " ) ;
    fprintf( stderr, "%4d ", 0 ) ;              /*V1.05-g*/



   /*********BEGINNING OF THE LOOP********/
  /* For each iteration of NEM (until :
       convergence, or iteration count reached, or empty class) */
    for ( iter = 1, converged = FALSE ;         /*V1.05-g*/
          ( iter <= AlgoParaP->NbIters ) &&
          ( ! converged ) &&
	  ( CPUtime < AlgoParaP->CPUmax)  &&
          ( err == STS_OK ) ;
          iter ++ )
    {
        fprintf( stderr, "\b\b\b\b\b" ) ;       /*V1.05-g*/
        fprintf( stderr, "%4d ", iter ) ;       /*V1.05-g*/


	/* Save last computed partition as old partition */
        memcpy( WorkP->ColdM, CM, npt * nk * sizeof( double ) ) ;
        oldcrit = ChosenCrit( CriterP , AlgoParaP->Crit ) ;

        /* Save last value of beta */
        oldbeta = ParaP->Beta;

        /* M-Step (Maximization) : compute current model parameters
           from old partition */
	if ( AlgoParaP->InitMode != INIT_MIXFIX )
          {
          if (AlgoParaP->ParamMode == PARAM_ESTIM)
	    {
	  err = EstimPara( CM, DataP,AlgoParaP, nk,  AlgoParaP->MissMode, SpecP,
			   &emptyk, ParaP ) ;
            }
          }

         switch ( SpecP->BetaModel )
	      {
              case BETA_PSGRAD:
                result = EstimBeta( SpecP->BetaModel, &AlgoParaP->BtaPsGrad, SpatialP, CM,
		   npt, nk,
		   & ParaP->Beta, & WorkP->Neighs ) ;
                break;

              case BETA_ZEROIN:
		Parameter.Type = BETA;
		result = ZeroIn(-10,10,f,0.001,npt,nk,&ParaP->Beta,
                                 CM ,ParaP->Singl, CM,SpatialP,&WorkP->Neighs,
				 Parameter);
                break;

              default: break;
              }


        /* If no empty class */
        if ( err == STS_OK )
        {
            /* E-Step (Expectation) : compute current partition from
               current model parameters and old partition */

	    ComputePartitionFromPara( 0, DataP, AlgoParaP, SpecP, ParaP,
				      SpatialP,
				      CM, CriterP,  WorkP ) ;

            /* Test convergence */
	    converged = HasConverged( AlgoParaP->CvTest, AlgoParaP->CvThres,
                                      AlgoParaP->TieRule,
				      WorkP->ColdM, CM, npt, nk, DataP->NbVars,0,0, oldcrit,
				       oldbeta,0,0, ParaP,AlgoParaP, SpecP,
                                      SpatialP,DataP, AlgoParaP->Crit,
				      CriterP, WorkP ) ;  /*V1.06-o*/
        }
        else if ( err == STS_W_EMPTYCLASS ) /* Else -> class emptyk is empty */
        {
            fprintf( stderr, "Class %d empty at iteration %d\n",
                     emptyk, iter ) ;
           
        }


    CPUtime = clock()/ CLOCKS_PER_SEC;

    } /* end for ( iter = 1, converged = FALSE  ) */
    /**********END OF THE LOOP*****/



    iter = iter - 1 ;  /*V1.05-g*/

    /* fprintf( stderr," CPU = %d \n", CPUtime); */

    /* Compute and display value of criteria */  /*V1.03-a*/
      if ( iter == 0 )     /*  model parameters estimation not yet done */
      {
         err =  EstimPara( CM, DataP, AlgoParaP, nk, AlgoParaP->MissMode, SpecP,
         &emptyk, ParaP ) ;
         err =  ComputePkFkiM( DataP, SpecP, ParaP,
  		       WorkP->PkFkiM, WorkP->LogPkFkiM ) ;

          ComputePartitionFromPara( 0, DataP, AlgoParaP, SpecP, ParaP,
  				      SpatialP,
  				      CM, CriterP, WorkP ) ;
      }
     err = ComputeCrit( npt, nk, ParaP->Beta, CM, SpatialP, AlgoParaP,DataP, WorkP, CriterP ) ;

      CalcError( CM, npt, 1, & CriterP->Errinfo, & CriterP->Errcur );

      fprintf( stderr, "\n" ) ;  /* -> to end iterations count line */
      //fprintf( stderr,
      //	     "  criterion NEM = %6.3f / Ps-Like = %6.3f / Lmix = %6.3f\n" ,
      //	     CriterP->U , CriterP->M , CriterP->L ) ;  /*  V1.03-g*/ 
      if ( CriterP->Errinfo.Kr != 0 )
        fprintf( stderr, "  error = %5.3f\n", CriterP->Errcur.Errorrate ) ;

    /* If convergence test was requested and no error occurred */
    if ( ( AlgoParaP->CvTest != CVTEST_NONE ) && ( err == STS_OK ) )
    {
        /* Display message about convergence */
        if ( converged )
        {
            fprintf( stderr, "  NEM converged after %d iterations\n", iter ) ;
        }
        else
        {
            fprintf( stderr, "  NEM did not converge after %d iterations\n",
                     iter ) ;
        }
    }

     err = HardClassif(AlgoParaP,DataP,npt,nk,CM,Hclassif);

    if ( nk == AlgoParaP->True.K)
      {
	ComputeErrorRate(Hclassif,AlgoParaP->True.DataTrue,
			 AlgoParaP->True.ErrorRate, nk, npt,
			 SpatialP->NeighData.Image.Nl,
			 SpatialP->NeighData.Image.Nc);
      }

    err = ComputeBIC( SpecP,ParaP, DataP, SpatialP, AlgoParaP,Hclassif,
                CM , CriterP, MSCritP, WorkP);

    GenFree(Hclassif);

    return err ;

}  /* end of NemAlgo() */




/* ------------------------------------------------------------------- */
static void ComputePartitionFromPara
        (
	 const int          Needinit,    /* I : 1 if need init, 0 if normal */
	 const DataT*       DataP,       /* I */
	 const AlgoParaT*    AlgoParaP,    /* I */
	 const ModelSpecT*  SpecP,       /* I */
	 ModelParaT*        ParaP,       /* I (temporary modification) */
         const SpatialT*    SpatialP,    /* I */
         double*             C_NK,        /* I/O */
         CriterT*           CriterP,     /* O */
	 WorkingT*          WorkingP     /* T */
	)
/* ------------------------------------------------------------------- */
{

  StatusET   err;
  int        result;

  /* Compute density of each point relatively to each class */
  err = ComputePkFkiM( DataP, SpecP, ParaP,
                WorkingP->PkFkiM,  WorkingP->LogPkFkiM ) ;

  if ( Needinit ) {
    /* Compute initial partition by "blind" segmentation */
    double beta = ParaP->Beta ;
    ParaP->Beta = 0.0 ;
    result = ComputePartition( SpecP, ParaP, DataP, SpatialP, AlgoParaP,
		       C_NK, WorkingP, CriterP ) ;
    ParaP->Beta = beta ;
  }

  /* Compute partition from parameters and previous partition */
  result = ComputePartition( SpecP, ParaP, DataP, SpatialP, AlgoParaP,
		     C_NK, WorkingP, CriterP ) ;

 

}   /* end of ComputePartitionFromPara() */




/* ------------------------------------------------------------------- */
static int  /* 0 = OK,  1 = beta diverges to infinity, 2 = call bug */
EstimBeta
(
  const BetaET       BetaModel,  /* I : beta estimation mode */
  const BtaPsGradT*  BtaPsGradP, /* I : gradient ascent parameters */
  const SpatialT*    SpatialP,   /* I : neighborhood system */
  const double*       C_NK,       /* I : classification matrix (Npt, NbK)*/
  const int          Npt,        /* I : number of points */
  const int          Nbk,        /* I : number of classes */
  double*            BetaP,      /* I/O : previous and new beta */
  PtNeighsT*         NeighsP     /* W : to store a site's neighbors */
)
/* ------------------------------------------------------------------- */
{
  int    iter ;           /* current iteration : 0..NbIter-1 */
  int    conv ;           /* TRUE if convergence reached */
  int    ipt ;            /* current site  : 0..Npt-1 */
  int    k ;              /* current class : 0..Nbk-1 */
  double  con_ik ;         /* sum_{j\neighbor{i}} v_{ij} c_{jk} */
  double  exp_ik ;         /* exp{ beta * con_ik } */
  double  sumk_exp ;       /* sum_k exp_ik */
  double  sumk_con_exp ;   /* sum_k con_ik exp_ik */
  double  sumk_con2_exp ;  /* sum_k (con_ik)^2 exp_ik */
  double  sumk_cik_con ;   /* sum_k c_ik con_ik */
  double  crit ;           /* log pseudo-likelihood */
  double  grad ;           /* first deriv of log pseudo-likelihood */
  double  dsec ;           /* - second deriv of log pseudo-likelihood */

  GetNeighFT* fGetNeigh ; /* function fetching neighbors of a site */


  if ( ( BetaModel != BETA_PSGRAD ) || ( SpatialP->Type == TYPE_NONSPATIAL ) )
    return 0 ;



  /* Get function specific to chosen spatial model */
  if ( GetSpatialFunc( SpatialP->Type,
			       & fGetNeigh ) != STS_OK )
    return 2 ;

  /* For each iteration of gradient ascent */
  for ( iter = 0,
	  conv = FALSE ;
	( iter < 100 ) &&
	  ( ! conv ) ;
	iter ++, conv = (fabs( grad ) < ( BtaPsGradP->ConvThres  )) ) {

    /* Compute crit, grad and dsec by summation over all sites */
    for ( ipt = 0, crit = 0.0, grad = 0.0, dsec = 0.0 ;
	  ipt < Npt ; ipt ++ ) {

      /* Get site's neighbours */
      int nbn = fGetNeigh( ipt, &SpatialP->NeighData, NeighsP ) ;

      for ( k = 0,
	      sumk_exp      = 0.0,
	      sumk_con_exp  = 0.0,
	      sumk_con2_exp = 0.0,
	      sumk_cik_con  = 0.0 ;
	    k < Nbk ; k ++ )  {

	con_ik = SumNeighsOfClass( k, nbn, Nbk, NeighsP->NeighsV, C_NK );
	exp_ik = exp( (*BetaP) * con_ik ) ;

	sumk_exp        += exp_ik ;
	sumk_con_exp    += con_ik * exp_ik ;
	sumk_con2_exp   += con_ik * con_ik * exp_ik ;
	sumk_cik_con    += C_NK[ ( ipt * Nbk ) + k ] * con_ik ;
      }

      crit += (*BetaP) * sumk_cik_con - log( sumk_exp ) ;
      grad += sumk_cik_con - sumk_con_exp / sumk_exp ;

      /* might get NaN because of high sumk_exp*/
      dsec += (double) ( (double)
	( sumk_con2_exp * sumk_exp - sumk_con_exp * sumk_con_exp )
	/ (double) ( sumk_exp * sumk_exp ) ) ;
    }

    /* Update beta */
    if ( BtaPsGradP->Step <= 0.0 ) {
      /* dsec = dsec * 4 ;  reduce the step to avoid overstepping */
      if ( dsec < ( Npt / 10 ) )
	dsec = Npt / 10 ;
       {
	(*BetaP) += grad / dsec ;}
    }
    else {
      (*BetaP) += grad * ( BtaPsGradP->Step / Npt ) ;
    }

  } /* end  For each iteration of gradient ascent */

  if ( (*BetaP) > MAX_BETA ) {
    (*BetaP) = MAX_BETA ;
    return 1 ;
  }

  if ( (*BetaP) < MIN_BETA ) {
    (*BetaP) = MIN_BETA ;
    return 1 ;
  }

  if ( isnan(*BetaP) ) {
    (*BetaP) = 0 ;
    return 1 ;
  }
   /* fprintf(stderr,"beta=%f\n",*BetaP); */

  return 0 ;
  /*???*/
}   /* end of EstimBeta() */



/* ------------------------------------------------------------------- */
static StatusET ComputePkFkiM
        (
            const DataT*        DataP,      /* I */
            const ModelSpecT*   SpecP,      /* I */
            const ModelParaT*   ParaP,      /* I */
            double*              PkFkiM,     /* O */
            double*              LogPkFkiM   /* O */
        )
/* ------------------------------------------------------------------- */
{
    StatusET        err = STS_OK ;
    int             result;
    int             i,j,s ;
    int             k ;
    int             npt   = DataP->NbPts;
    int             nk    = SpecP->K ;
    int             Nd    = DataP->NbVars ;
    double* p=0;     /* for the cholesky decomposition */
    double* cholDISP=0; /* for the cholesky decomposition */
    double logdet = 0;
    double sum = 0;
    CompuDensFT*    fCompuDens ;      /*V1.06-c*/

    /* Get functions specific to chosen noise model */
    if ( ( err = GetDensityFunc( SpecP ,
                                 & fCompuDens ) ) != STS_OK )     /*V1.06-c*/
    {
        return err ;
    }

 if ( (SpecP->ClassFamily == FAMILY_NORMALGEN) &&
     (SpecP->ClassDisper == DISPER__D) )
    /* same cov matrix for each class, so the cholesky decomp  is computed
     only once */
  {
    if (( p = GenAlloc( Nd, sizeof(double), 0, "ComputePkFkiM", "p")) == NULL )
         return STS_E_MEMORY;
    if (( cholDISP = GenAlloc(Nd*Nd, sizeof(double), 0, "ComputePkFkiM","cholDISP")) == NULL )
         return STS_E_MEMORY;

    for (i = 0; i<Nd; i++)
      {
	for (j=i; j<Nd; j++)
	{
	    cholDISP[i*Nd+j] = ParaP->Disp_KD[ i*Nd +j - i*(i+1)/2];
	}
      }

      /* compute the cholesky decomposition of the covariance matrix: L.L^t */
      /* the diagonal elements of L are in the vector p */
      /* the lij, j<i, are in the lower triangle of cholDISP */
      err = choldc(cholDISP,Nd,p);
      if (err == -1)
         fprintf(stderr,"cholesky failed\n");

      /* compute L^-1 and put the result in the lower triangle of cholDISP */
      for (i=0; i<Nd; i++)
      {
	cholDISP[i*Nd+i]  = 1.0/p[i];
	for (j=i+1; j<Nd; j++)
	{
	  sum =0;
          for (s=i;s<j;s++)
          sum-=cholDISP[j*Nd+s]*cholDISP[s*Nd+i];
          cholDISP[j*Nd+i] = sum/p[j];
	}
      }


     /* compute log(sqrt(det)) of the covariance matrix */
        for ( i=0; i<Nd; i++)
        {
          if ( p[i] > EPSILON )
           logdet += log(p[i]);

          else
           {
             err = -1 ;
             return err;
	   }
	}

  } /* end of case normalgen + disper__D */



    /* Compute densities of each point relative to each class */
    for ( k = 0 ; k < nk ; k ++ )
    {
        double pk    = ParaP->Prop_K[ k ] ;
        double logpk ;

 if ( (SpecP->ClassFamily == FAMILY_NORMALGEN) &&
        (SpecP->ClassDisper == DISPER_KD) )
     {
      if ((p = GenAlloc( Nd, sizeof(double), 0, "ComputePkFkiM", "p")) == NULL )
         return STS_E_MEMORY;
      if ((cholDISP = GenAlloc(Nd*Nd, sizeof(double), 0, "ComputePkFkiM","cholDISP")) == NULL )
         return STS_E_MEMORY;

      for (i = 0; i<Nd; i++)
      {
	for (j=i; j<Nd; j++)
	{
	    cholDISP[i*Nd+j] = ParaP->Disp_KD[k*Nd*(Nd+1)/2 + i*Nd +j - i*(i+1)/2];
	}
      }

      /* compute the cholesky decomposition of the covariance matrix: L.L^t */
      /* the diagonal elements of L are in the vector p */
      /* the lij, j<i, are in the lower triangle of cholDISP */
      err = choldc(cholDISP,Nd,p);
      if (err == -1)
         fprintf(stderr,"cholesky failed\n");

      /* compute L^-1 and put the result in the lower triangle of cholDISP */
      for (i=0; i<Nd; i++)
      {
	cholDISP[i*Nd+i]  = 1.0/p[i];
	for (j=i+1; j<Nd; j++)
	{
	  sum =0;
          for (s=i;s<j;s++)
          sum-=cholDISP[j*Nd+s]*cholDISP[s*Nd+i];
          cholDISP[j*Nd+i] = sum/p[j];
	}
      }


       /* compute log(sqrt(det)) of the covariance matrix */
        for ( i=0; i<Nd; i++)
        {
          if ( p[i] > EPSILON )
           logdet += log(p[i]);

          else
           {
             err = -1 ;
             return err;
	   }
	}

     } /* end of case normalgen + disper_KD */


	if ( pk > EPSILON )
	  logpk = log( pk ) ;
	else
	  {
	    logpk = atof( "-Inf" ) ;
	    err = STS_W_EMPTYCLASS ;
	  }

        for ( i = 0 ; i < npt ; i ++ )
        {
            double fki ;
	    double  logfki ;

            result = fCompuDens( DataP->NbVars, k, ParaP,
			& (DataP->PointsM[ i * DataP->NbVars ]) , SpecP,
                        & fki , & logfki,cholDISP,logdet );

            PkFkiM   [ ( i * nk ) + k ] = pk * fki ;
            LogPkFkiM[ ( i * nk ) + k ] = logpk + logfki ;
        }
    }

    GenFree(cholDISP);
    GenFree(p);
    return err ;

}   /* end of ComputePkFkiM() */



/* ------------------------------------------------------------------- */
static int ComputePartition
      (
          const ModelSpecT*   SpecP,      /* I */
          const ModelParaT*   ParaP,      /* I */
          const DataT*        DataP,      /* I */
          const SpatialT*     SpatialP,   /* I */
          const AlgoParaT*     AlgoParaP,   /* I */
          double*              CM,         /* I/O */
          WorkingT*           WorkP,      /* I:PkFkiM O:CtmpM,Neighs,CiNumV */
          CriterT*            CriterP     /* O */
      )
/* ------------------------------------------------------------------- */
{
  StatusET sts ;


      sts = ComputePartitionNEM( SpecP, ParaP, DataP, SpatialP, AlgoParaP,
				  CM, WorkP, CriterP ) ;


  return sts ;

}   /* end of ComputePartition() */

/* ------------------------------------------------------------------- */
static int ComputePartitionICM

      (
          const ModelSpecT*   SpecP,      /* I */
          const ModelParaT*   ParaP,      /* I */
          const DataT*        DataP,      /* I */
          const SpatialT*     SpatialP,   /* I */
          const AlgoParaT*    AlgoParaP,   /* I */
          double*              CM,         /* I/O */
          WorkingT*           WorkP,      /* I:PkFkiM O:CtmpM,Neighs,CiNumV */
          CriterT*            CriterP     /* O */
      )
     /* ICM's restoration step  */
/* ------------------------------------------------------------------- */
{
    int             result;
    int             il, ic ;  /*V1.04-c*/
    int             itere ;
    int             Nl = SpatialP->NeighData.Image.Nl ;
    int             Nc = SpatialP->NeighData.Image.Nc ;
    const double*   fikM   = WorkP->FikM;
    double*         ctmpM  = WorkP->CtmpM;
    int             npt    = DataP->NbPts;
    int             nk     = SpecP->K ;
    int             ik;
    int*            kmaxesV;
    int             kmax;
    double*         hardCM;



    /* Iteratively compute new partition */
    for ( itere = 0 ; itere < AlgoParaP->NbEIters ; itere ++ ) {

        /* Save last computed value of partition in temporary buffer */
        memcpy( ctmpM, CM, npt * nk * sizeof( double ) ) ;

    switch (AlgoParaP->SiteUpdate)
      {
       case UPDATE_PARA:
           if ( ( hardCM = GenAlloc( npt*nk, sizeof( double ),
				   0, "computePartitionICM", "hardCM" ) ) == NULL )
            return STS_E_MEMORY ;

           ComputePostPartition(SpecP,ParaP,DataP,SpatialP,AlgoParaP,
                                                  ctmpM,WorkP,CM);
           result = HardClassif(AlgoParaP,DataP,npt,nk,CM,hardCM);
           memcpy( CM, hardCM, npt * nk * sizeof( double ) ) ;

           GenFree(hardCM);
           break;

       case UPDATE_SEQ:
        if ( ( kmaxesV = GenAlloc( nk, sizeof( int ),
				   0, "computePartitionICM", "kmaxesV" ) ) == NULL )
            return STS_E_MEMORY ;

        /* For each point xi  exepted on the boarder*/
        for ( il = 1; il< (Nl-1); il++ ) {
	  for (ic = 1; ic<(Nc-1); ic++) {

	    int ipt = DataP->SiteVisitV[ il*Nc+ic ] ;  /*V1.04-c*/

            /* If  label is not known */
	    if ( ( AlgoParaP->Label   != 1 ) ||
		 ( DataP->LabelV[ ipt ] == 0 ) ) {

                    result = ComputeLocalPostPartition(SpecP,ParaP,DataP,SpatialP,ipt,nk,&(WorkP->Neighs),
                                                       CM, ParaP->Beta, ParaP->Singl, fikM,ctmpM);
                    kmax = ComputeMAP(ctmpM,ipt,nk,AlgoParaP->TieRule,kmaxesV);
                    for (ik = 0; ik<nk; ik++){
                         if (ik == kmax) CM[ipt*nk+ik] =1;
                         else CM[ipt*nk+ik] =0;
		    }
	    }
            /* Else (label known) -> do not recompute its classification */

	  }  /* end for (ic = 0;ic<Nc; ic++) */
        }  /* end for ( il = 0 ; il<Nl ; il ++ ) */
        GenFree(kmaxesV);
          break;

      default:
       fprintf(stderr,"unkown typ for AlgoParaP->SiteUpdate\n");
      } /* end of switch */

    } /* end for ( itere = 0 ; itere < AlgoParaP->NbEIters ; itere ++ ) */


    return STS_OK ;



} /* end of ComputePartitionICM */

/* ------------------------------------------------------------------- */
static int ComputePartitionNEM

      (
          const ModelSpecT*   SpecP,      /* I */
          const ModelParaT*   ParaP,      /* I */
          const DataT*        DataP,      /* I */
          const SpatialT*     SpatialP,   /* I */
          const AlgoParaT*    AlgoParaP,   /* I */
          double*              CM,         /* I/O */
          WorkingT*           WorkP,      /* I:PkFkiM O:CtmpM,Neighs,CiNumV */
          CriterT*            CriterP     /* O */
      )
/* ------------------------------------------------------------------- */
{
    StatusET        err ;
    int             result;
    int             il,ic;
    int             itere ;
    int             Nl = SpatialP->NeighData.Image.Nl ;
    int             Nc = SpatialP->NeighData.Image.Nc ;
    const double*    pkfkiM = WorkP->PkFkiM ;
    const double*    fikM   = WorkP->FikM;
    double*          ctmpM  = WorkP->CtmpM;
    int             npt    = DataP->NbPts;
    int             nk     = SpecP->K ;
    GetNeighFT*     fGetNeigh ;     /*V1.06-c*/


    /* Get function specific to chosen spatial model */
    if ( ( err = GetSpatialFunc( SpatialP->Type,
                                 & fGetNeigh ) ) != STS_OK )     /*V1.06-c*/
      return err ;


    /* Iteratively compute new partition */
    for ( itere = 0 ; itere < AlgoParaP->NbEIters ; itere ++ ) {

        /* Save last computed value of partition in temporary buffer */
        memcpy( ctmpM, CM, npt * nk * sizeof( double ) ) ;


        /* For each point xi */
        for (il = 0;il<Nl; il++) {
	  for (ic = 0;ic<Nc; ic++){

	    int ipt = DataP->SiteVisitV[ il*Nc+ic ] ;  /*V1.04-c*/

            /* If  label is not known */
	    if ( ( AlgoParaP->Label  != 1 ) ||
		 ( DataP->LabelV[ ipt ] == 0 ) ) {


               switch ( AlgoParaP->Algo )
		 {
                  case ALGO_MEANF:
                           result =  ComputeLocalProba(SpecP, ipt, nk, ParaP, DataP, &(SpatialP->NeighData),
				   fGetNeigh, fikM,
				   (AlgoParaP->SiteUpdate == UPDATE_SEQ)?
				   CM : ctmpM ,
				   &CM[ipt*nk],
				   &WorkP->Neighs, WorkP->CiNumV );
                           break;

                 case ALGO_NEM:
		           result = ComputeLocalProba(SpecP, ipt, nk, ParaP, DataP, &(SpatialP->NeighData),

			   fGetNeigh, pkfkiM,
				   (AlgoParaP->SiteUpdate == UPDATE_SEQ)?
				   CM : ctmpM,
				   &CM[ipt*nk],
				   &WorkP->Neighs, WorkP->CiNumV );

                          break;
                 default: break;

		 }
	       /*fprintf(stderr,"z%d = %f, %f\n",ipt,CM[ipt*nk+0],CM[ipt*nk+1]);*/

		/* Eventual C-step */
		if ( AlgoParaP->Algo == ALGO_NCEM ) {
		  int kmap   = ComputeMAP( CM , ipt , nk , AlgoParaP->TieRule,
					   WorkP->KmaxesV ) ;

		  LabelToClassVector( nk, kmap, &CM[ ipt * nk ] ) ;
		}
	    }
            /* Else (label known) -> do not recompute its classification */

        } /* end for (ic = 0;ic<Nc; ic++ ) */
        } /* end for (il = 0;il<Nl; il++ )  */



    } /* end for ( itere = 0 ; itere < AlgoParaP->NbEIters ; itere ++ ) */

    return STS_OK ;

}   /* end of ComputePartitionNEM() */



/* ------------------------------------------------------------------- */
static int ComputePottsSinglMargMeanField
      (
          const ModelSpecT*   SpecP,      /* I */
          const ModelParaT*   ParaP,      /* I */
          const DataT*        DataP,      /* I */
          const SpatialT*     SpatialP,   /* I */
          const AlgoParaT*    AlgoParaP,   /* I */
          double*              CM,         /* I/O */
          WorkingT*           WorkP,      /*  O:CtmpM,Neighs,CiNumV */
          CriterT*            CriterP     /* O */
      )
/* ------------------------------------------------------------------- */
{
 StatusET        err ;
    int             result;
    int             il,ic,i,k;
    int             itere ;
    int             Nl = SpatialP->NeighData.Image.Nl ;
    int             Nc = SpatialP->NeighData.Image.Nc ;
    double*         zerofikM   ;
    double*         ctmpM  = WorkP->CtmpM;
    int             npt    = DataP->NbPts;
    int             nk     = SpecP->K ;
    GetNeighFT*     fGetNeigh ;     /*V1.06-c*/

     if (( zerofikM = GenAlloc(nk*npt, sizeof(double), 1, "ComputePottsSinglMargMeanField", "zerofikM")) == NULL)
     {
       fprintf(stderr, "impossible to allocate Mean in function ComputePottsSinglMargMeanField\n");
       return STS_E_MEMORY;
     }

     for (i=0; i<npt; i++){
       for(k=0; k<nk; k++){
	 zerofikM[i*nk+k]=1;}}

    /* Get function specific to chosen spatial model */
    if ( ( err = GetSpatialFunc( SpatialP->Type,
                                 & fGetNeigh ) ) != STS_OK )     /*V1.06-c*/
      return err ;

 

    /* Iteratively compute new partition */
    for ( itere = 0 ; itere < AlgoParaP->NbfpeIters ; itere ++ ) {

        /* Save last computed value of partition in temporary buffer */
        memcpy( ctmpM, CM, npt * nk * sizeof( double ) ) ;


        /* For each point xi */
        for (il = 0;il<Nl; il++) {
	  for (ic = 0;ic<Nc; ic++){

	    int ipt = DataP->SiteVisitV[ il*Nc+ic ] ;  /*V1.04-c*/

            /* If  label is not known */
	    if ( ( AlgoParaP->Label  != 1 ) ||
		 ( DataP->LabelV[ ipt ] == 0 ) ) {


                           result =  ComputeLocalProba(SpecP, ipt, nk, ParaP, DataP, &(SpatialP->NeighData),
				   fGetNeigh, zerofikM,
				   (AlgoParaP->SiteUpdate == UPDATE_SEQ)?
				   CM : ctmpM ,
				   &CM[ipt*nk],
				   &WorkP->Neighs, WorkP->CiNumV );


	  } /* Else (label known) -> do not recompute its classification */

        } /* end for (ic = 0;ic<Nc; ic++ ) */
        } /* end for (il = 0;il<Nl; il++ )  */



    } /* end for ( itere = 0 ; itere < AlgoParaP->NbEIters ; itere ++ ) */

    GenFree(zerofikM);
    return STS_OK ;

}   /* end of ComputePottsSinglPriorMean */


/* ------------------------------------------------------------------- */
/* Compute fuzzy (or hard) partition with fixed parameters, using
   Monte-Carlo simulations */ /*1.04-h*/

static int ComputePartitionGEM /* +++ */
      (
          const ModelSpecT*   SpecP,      /* I */
          const ModelParaT*   ParaP,      /* I */
          const DataT*        DataP,      /* I */
          const SpatialT*     SpatialP,   /* I */
          const AlgoParaT*     AlgoParaP,   /* I */
          double*              CM,         /* I/O */
          WorkingT*           WorkP,      /* I:PkFkiM O:CtmpM,Neighs,CiNumV */
          CriterT*            CriterP     /* O */
      )
/* ------------------------------------------------------------------- */
{
    StatusET        err ;
    int             result;
    const double*    pkfkiM = WorkP->PkFkiM ;
    double*          ctmpM = WorkP->CtmpM;
    int             npt   = DataP->NbPts;
    int             nk    = SpecP->K ;
    GetNeighFT*     fGetNeigh ;      /*V1.06-c*/

    double*          z_nk ;     /* currently simulated partition */
    int*            occur_nk ; /* occurence count for class h at site i */
    int             icycle ;   /* current Monte-Carlo cycle */
    int             kdraw ;    /* drawn class : 0..nk-1 */
    int             ivis ;


    if ( ( err = GetSpatialFunc( SpatialP->Type,
                                 & fGetNeigh ) ) != STS_OK )     /*V1.06-c*/
        return err ;


    /* Allocate all memory */
    if ( ( z_nk = GenAlloc( npt * nk, sizeof( double ),
		     0, "ComputePartitionGEM", "z_nk" ) ) == NULL )
	return STS_E_MEMORY ;

    if ( ( occur_nk = GenAlloc( npt * nk, sizeof( int ),
			 0, "ComputePartitionGEM", "occur_nk") ) == NULL )
	return STS_E_MEMORY ;

    /* Initialize values */
    {
      int ipt, k ;

      /* Initialize hard classification from fuzzy classification, by MPM */
      for ( ipt = 0 ; ipt < npt ; ipt ++ )
	{
	  int kmap   = ComputeMAP( CM , ipt , nk , AlgoParaP->TieRule,
				   WorkP->KmaxesV ) ;

	  LabelToClassVector( nk, kmap, &z_nk[ ipt * nk ] ) ;
	}

      /* Initialize to zero all occurence counts */
      for ( ipt = 0 ; ipt < npt ; ipt ++ )
	for ( k = 0 ; k < nk ; k ++ )
	  occur_nk[ ( ipt * nk ) + k ] = 0 ;
    }


    /* Iteratively run Monte-Carlo simulation with startup and
       recording durations */
    for ( icycle = 0 ;
	  icycle < AlgoParaP->NbEIters * (NBCYCLES_STARTUP + NBCYCLES_RECORD) ;
	  icycle ++ ) {

        /* Save last computed value of partition in temporary buffer */
        memcpy( ctmpM, z_nk, npt * nk * sizeof( double ) ) ;

	/* Run through each site */
        for ( ivis = 0 ; ivis < npt ; ivis ++ ) {
	    int ipt = DataP->SiteVisitV[ ivis ] ;  /*V1.04-c*/

            /* If  label is not known */
	    if ( ( AlgoParaP->Label   != 1 ) ||
		 ( DataP->LabelV[ ipt ] == 0 ) ) {

		  result = ComputeLocalProba(SpecP, ipt, nk, ParaP, DataP, &(SpatialP->NeighData),
				   fGetNeigh, pkfkiM,
				   (AlgoParaP->SiteUpdate == UPDATE_SEQ)?
				   z_nk : ctmpM,
				   &CM[ipt*nk],
				   &WorkP->Neighs, WorkP->CiNumV );

		/* Update zi by multinomial draw from ti's k proportions */
		kdraw = Multinomial( nk , & CM[ ipt * nk ] ) - 1 ;

		LabelToClassVector( nk, kdraw, &z_nk[ ipt * nk ] ) ;

		/* Update occurrence counter if in record phase */
		if ( icycle >= AlgoParaP->NbEIters * NBCYCLES_STARTUP )
		  occur_nk[ ( ipt * nk ) + kdraw ] += 1 ;
	    }
            /* Else (label known) -> do not recompute its classification */

        } /* end for ( ivis = 0 ; ivis < npt ; ivis ++ ) */

      } /* end for ( icycle = 0 ; ... ) */


    /* Compute frequency of classes at each site */
    if ( AlgoParaP->NbEIters * NBCYCLES_RECORD > 0 ) {
      int ipt, k ;
      int nbrecord = AlgoParaP->NbEIters * NBCYCLES_RECORD ;

      for ( ipt = 0 ; ipt < npt ; ipt ++ )
	{
	  for ( k = 0 ; k < nk ; k ++ )
	    CM[ ( ipt * nk ) + k ] =
	      ( (double) occur_nk[ ( ipt * nk ) + k ] ) / nbrecord ;
	}
    }

  

    /* Free all memory */
    GenFree( z_nk ) ;
    GenFree( occur_nk ) ;


    return STS_OK ;

}   /* end of ComputePartitionGEM() */



/*-------------------------------------------------------*/
static int ComputeLocalProba
 (
  const ModelSpecT*   SpecP,
  const int            Ipt,        /* I: current site */
  const int            Nk,         /* I: nb of classes */
  const ModelParaT*    ParaP,      /* I: parameters (use Beta) */
  const DataT*         DataP,
  const NeighDataT*    NeighDataP, /* I: neighborhood system */
  GetNeighFT*          FGetNeigh,  /* I: function to fetch neighbors */
  const double*         PkfkiM,     /* I: pk fk(xi) if NEM or fk(xi) if MeanF, GSF */
  const double*         Cin_NK,     /* I: class of neighbors */
  double*               Cout_K,     /* O: class of site Ipt */
  PtNeighsT*           NeighsP,    /* W: to store fetched neighbors */
  double*              Cinum_K     /* W: to store cik's numerators */
 )
/*-------------------------------------------------------*/
{
  static int first=TRUE; /* FALSE once a zero density point has occurred */

  const char*   func = "ComputeLocalProba";
  int     k ;      /* current class : 0..Nk-1 */
  int     nbn ;    /* nb of neighbours of ipt */
  double  cumnum ; /* for point i, sum over k of pkfki*context */


  if (( Cinum_K = GenAlloc( Nk, sizeof( double ), 0, func, "Cinum_K")) == NULL)
    fprintf(stderr,"impossible to allocate Cinum_K in function ComputeLocalProba\n")
;
  /* Get point's neighbours */
  nbn = FGetNeigh( Ipt, NeighDataP, NeighsP ) ;

  /* For each class k :
     - compute the contribution of xi's neighbors
     (j), based on their last cjk
     - based on the neighbors contribution and on the density for
     class k at xi, compute cik's numerator
     - increment the cumulated sum of ci's K numerators
  */
  for ( k = 0, cumnum = 0.0 ; k < Nk ; k ++ ) {
    /* i's class k context contrib. */
    double   context = SumNeighsOfClass( k, nbn, Nk, NeighsP->NeighsV, Cin_NK );


    Cinum_K[ k ] = PkfkiM[( Ipt * Nk ) + k] * exp( (double) ParaP->Beta * context+ ParaP->Singl[k] ) ;

    cumnum = cumnum + Cinum_K[ k ] ;
  } /* end for ( k = 0, cumnum = 0.0 ; k < nk ; k ++ ) */

  /* cik = numerator of cik / sum of ci's K numerators */
  if ( cumnum > 0 ) {
    if ( cumnum > EPSILON ) {
      double   invZ = ( 1 / cumnum ) ;

      for ( k = 0 ; k < Nk ; k ++ )
	{

         if ((SpecP->ClassFamily ==FAMILY_NORMALDIAG  )
               &&(ParaP->Disp_KD[k] == 0 )){
                  int diff = 0;
                  diff = diff + ParaP->Center_KD[k ]
                                             - DataP->PointsM[Ipt*DataP->NbVars+1];
                  if(diff == 0){ Cout_K[ k ] =1;}
                  else {Cout_K[ k ] = 0 ;}
	 }
        else {Cout_K[ k ] = invZ * Cinum_K[ k ] ;}
	}
    }
    else {    /*V1.06-r*/
      double   invZ = ( 1 / ( cumnum / EPSILON ) ) ;

      for ( k = 0 ; k < Nk ; k ++ )
	Cout_K[ k ] = invZ * ( Cinum_K[ k ] / EPSILON ) ;
    }
  }

  else {
    double   invZ = ( 1.0 / Nk ) ;

    for ( k = 0 ; k < Nk ; k ++ )
      Cout_K[ k ] = invZ ;

    if ( first ) {
      first = FALSE ;
      fprintf( stderr, "Warning : pt %d density = 0\n", Ipt ) ;
    }
  }
GenFree (Cinum_K);
 return ( cumnum > 0.0 ) ;
}




/*-------------------------------------------------------*/
static int Multinomial(int km, const double *tk)
/*
 Entrees
     km           nombre de modalites
     tk           probabilites de chaque modalites

 Retour
     valeur retenue
*/

{
    int        k ;
    double      pk,x;

    x = RandomDouble( 0.0, 1.0 ) ;  /*V1.05-l*/
    pk = 0.0;
    for ( k = 0 ; k < km ; k ++ )
    {
	pk = pk + tk[ k ] ;
	if ( x <= pk ) return ( k + 1 ) ;
    }
    return km;
}



/* ------------------------------------------------------------------- */
static StatusET ComputeCrit
     (

        int                 Npt,         /* I */
        int                 Nk,          /* I */
        double               Beta,        /* I : spatial coefficient */
        const double*        CM,          /* I : class. matrix [Npt*Nk] */
        const SpatialT*     SpatialP,    /* I : neighborhood system */
        const AlgoParaT*    AlgoParaP,
        const DataT*              DataP,
        WorkingT*           WorkP,       /* I :LogPkFkiM, T:Neighs */
        CriterT*            CriterP      /* O : computed criteria */
     )
/* ------------------------------------------------------------------- */
{
    StatusET    err ;
    int         i ;
    int         k ;
    NeighT*     neiV = WorkP->Neighs.NeighsV ;
    GetNeighFT* fGetNeigh ;     /*V1.06-c*/

    if ( ( err = GetSpatialFunc( SpatialP->Type,
                                 & fGetNeigh ) ) != STS_OK )    /*V1.06-c*/
    {
        return err ;
    }



    CriterP->D = 0.0 ; /* D = sum[i]sum[k] cik (log pkfki-log cik) */
    CriterP->E = 0.0 ; /* E = sum[i]sum[k] cik log cik */
    CriterP->G = 0.0 ; /* G = sum[i]sum[j]sum[k] cik cjk wij */
    CriterP->U = 0.0 ; /* U = D + 0.5 * beta * G */
    CriterP->M = 0.0 ; /* M = D + beta * G + Z */
    CriterP->L = 0.0 ; /* L = sum[i] log( sum[k] pkfki ) = sum[i] log fi */
    CriterP->Z = 0.0 ; /* Z =-sum[i] log(sum[k] exp(bta * sum[j~i] wij cjk)) */

    /* For each point i */
    for ( i = 0 ; i < Npt ; i ++ )
    {
        double fi ;  /* fi = sum[k] pkfki */
	double zi ;  /* zi = sum[k] exp(beta * sum[j~i] wij cjk) */

        /* Get point i's neighbors */
        int nbn = fGetNeigh( i, & SpatialP->NeighData, &(WorkP->Neighs) ) ;

        /* For each class k */
        for ( k = 0 , fi = 0.0, zi = 0.0 ;
	      k < Nk ;
	      k ++ )
        {
            double   cik      = CM[ i * Nk + k ] ;
	    double   pik      = SumNeighsOfClass( k, nbn, Nk, neiV, CM ) ;

            /* If point i has non-null membership in class k */
            if ( cik > 0.0 )
            {
                /* Add to criteria the contribution of i's membership in k */
	        double logpkfki = WorkP->LogPkFkiM[ i * Nk + k ] ;
                double dik  = cik * ( logpkfki - log( cik ) ) ;
                double gik  = cik * pik ;
                double eik  = cik * log(cik) ;

                CriterP->D  = CriterP->D + dik ;
                CriterP->E  = CriterP->E + eik ;
                CriterP->G  = CriterP->G + gik ;
            }
            /* Else point i for class k has no contribution to criteria D/G*/

	    fi = fi + WorkP->PkFkiM[ i * Nk + k ] ;/*V1.05-j*/
	    zi = zi + exp( Beta * pik ) ;

        } /* end   For each class k */

	CriterP->L = CriterP->L + log( fi ) ;
	CriterP->Z = CriterP->Z - log( zi ) ;

    } /* end   For each point i */

    /* Criterion U is a combination of criteria D and G */
    CriterP->U = CriterP->D + 0.5 * Beta * CriterP->G ; /*V1.05-k*/
    CriterP->M = CriterP->D + Beta * CriterP->G + CriterP->Z ; /*V1.05-k*/

    CalcError( CM, Npt, 1, & CriterP->Errinfo, & CriterP->Errcur );



    return STS_OK ;

}   /* end of ComputeCrit() */



/* ------------------------------------------------------------------- */
static void CalcError                /*V1.04-e*/
      (
       const double*     Cla_N_Kc, /* I : found classification */
       const int        N,        /* I : number of points */
       const int        Harden,   /* I : 1 use hardened CM, 0 use CM itself */
       const ErrinfoT*  ErrinfoP, /* I : info to compute error */
       ErrcurT*         ErrcurP   /* O : computed error and agreement */
      )
/* ------------------------------------------------------------------- */
{
  int    Kc = ErrinfoP->Kc ;
  int    Kr = ErrinfoP->Kr ;
  int    Km = ErrinfoP->Km ;
  double* Loclas_N_Kc = ErrcurP->Loclas_N_Kc ;

  int*   kmaxes_Kc ;  /* used by ComputeMAP */
  int    kmap ;       /* MAP class for current object : 0..Kc-1 */
  int    in ;         /* cur object : 0..N-1 */
  int    ikc ;        /* cur class of current classification : 0..Km-1 */
  int    ikr ;        /* cur class of ref. classification : 0..Km-1 */

  double  bestagree ;  /* cur max agreement over checked permutations : 0..N */
  int    ikmfact ;    /* cur permutation : 0..Km!-1 */
  double  sumagree ;   /* summed diagonal of agreement matrix : 0..N */
  int    ikm ;        /* cur diagonal element of agreement matrix : 0..Km-1 */
  int    kperm ;      /* permuted ikm using permutation #ikmfact  : 0..Km-1 */


  if ( Kr == 0 ) {
    ErrcurP->Errorrate = mknan() ;
    return ;
  }

  kmaxes_Kc = GenAlloc( Kc, sizeof( int ), 0, "CalcError", "kmaxes_Kc" ) ;
  if ( kmaxes_Kc == NULL )
    return ;

  /* Copy classification locally, eventually harden */
  memcpy( Loclas_N_Kc, Cla_N_Kc, N * Kc * sizeof(double) ) ;
  if ( Harden ) {
    for ( in = 0 ; in < N ; in ++ ) {
      kmap = ComputeMAP( Loclas_N_Kc, in, Kc, ErrinfoP->TieRule, kmaxes_Kc ) ;
      LabelToClassVector( Kc, kmap, & Loclas_N_Kc[ in * Kc ] ) ;
    }
  }

  /*
   *  Compute agreement matrix : for each class ikc of current partition
   *  and each class ikr of reference partition, their agreement is the
   *  scalar product of their [N] classification vectors.
   *  For added classes (Kc to Km-1 and Kr to Km-1) : agreement = 0.
   */
  for ( ikc = 0 ; ikc < Km ; ikc ++ ) {
    for ( ikr = 0 ; ikr < Km ; ikr ++ ) {
      if ( ( ikc < Kc ) && ( ikr < Kr ) ) {
	ErrcurP->Agree_Km_Km[ ikc * Km + ikr ] = 0.0 ;
	for ( in = 0 ; in < N ; in ++ )
	  ErrcurP->Agree_Km_Km[ ikc * Km + ikr ] +=
	    Loclas_N_Kc[ in * Kc + ikc ] *
	    ErrinfoP->Refclas_N_Kr[ in * Kr + ikr ] ;
      }
      else
	ErrcurP->Agree_Km_Km[ ikc * Km + ikr ] = 0.0 ;
    }
  }

  /* Find permutation of lines with maximum sum of diagonal */
  bestagree            = 0.0 ;
  ErrcurP->Ibestpermut = 0 ;
  for ( ikmfact = 0 ; ikmfact < ErrinfoP->Kmfac ; ikmfact ++ ) {
    /* sum diagonal of permuted matrix and compare to best agreement */
    for ( ikm = 0, sumagree = 0.0 ; ikm < Km ; ikm ++ ) {
      kperm     = ErrinfoP->Perm_Kmfac_Km[ ( ikmfact * Km ) + ikm ] ;
      sumagree += ErrcurP->Agree_Km_Km[ kperm * Km + ikm ] ;
    }
    if ( sumagree > bestagree ) {
      bestagree            = sumagree ;
      ErrcurP->Ibestpermut = ikmfact ;
    }
  }

  ErrcurP->Errorrate = ( N - bestagree ) / N ;

  GenFree( kmaxes_Kc ) ;
}


/* ------------------------------------------------------------------- */
static double SumNeighsOfClass
     (
        int             K ,     /* I : class to search, 0..Nk-1 */
        int             Nbn ,   /* I : number of neighbours */
        int             Nk,     /* I : number of classes */
        const NeighT*   NeiV ,  /* I : neighbours, [Nbn] */
        const double*    CM      /* I : classification matrix [Npt*Nk] */
     )
/* ------------------------------------------------------------------- */
{
    double   Sumj_wij_cjk ; /* sum of neighbours'membership in class K */

    int     n ;            /* neighbour counter : 0..nbn-1 */
    double   sumj_wij ;     /* divide sumj_wij_cjk by this normalizing factor */

    for ( n = 0, Sumj_wij_cjk = 0.0, sumj_wij = 0.0 ;
          n < Nbn ;
          n ++ )
    {
        int     j   = NeiV[ n ].Index ;
        double   wij = NeiV[ n ].Weight ;
        double   cjk = CM[ (j * Nk) + K ] ;
        Sumj_wij_cjk  = Sumj_wij_cjk + ( wij * cjk ) ;
        sumj_wij      = sumj_wij + wij ;
    }

    if ( sumj_wij != 0.0 )
    {
        /* sumj_wij_cjk  = sumj_wij_cjk / sumj_wij ; +++*/
    }

    return Sumj_wij_cjk ;

}   /* end of SumNeighsOfClass() */


/* ------------------------------------------------------------------- */
static int HasConverged       /* ret : TRUE if convergence test satisfied */
(
  const CvemET    CvTest,     /* I : which convergence test to use */
  const double     CvThres,    /* I : convergence threshold to use */
  const TieET     TieRule,    /* for case = CVTEST_HARDCLASS */
  const double*    ColdM,      /* I : previous classification matrix */
  const double*    CM,         /* I : current  classification matrix */
  const int       Npt,        /* I : number of objects */
  const int       Nk,         /* I : number of classes */
  const int       Nd,         /* I : number of variables */
  const double     OldFunc,    /* previous value of F' in field algos */
  const double     NewFunc,    /* current value of F' in field algos */
  const double     OldCrit,    /* I : previous criterion */
  const double     OldBeta,    /* I : previous value of beta */
  const double*    OldCenter,  /* I : previous value of the centers */
  const double*    OldDisp,    /* I : previous value of the dispersions */
  const ModelParaT*  ParaP,    /* current values of beta, centers and dispersions */
  const AlgoParaT*   AlgoParaP,
  const ModelSpecT*  SpecP,
  const SpatialT*    SpatialP,   /* I : spatial info to compute criterion */
  const DataT*       DataP,
  const CritET       WhichCrit,  /* I : what criterion to use */
  CriterT*           CriterP,    /* I/O : current criterion */
  WorkingT*          WorkP       /* I: LogPkFkiM, T: Neighs */
)
/* ------------------------------------------------------------------- */
{

  switch( CvTest ) {
    case CVTEST_CLAS: {
      /* cv criterium based on the approximated a post probas */
      int ipt, ik ;
      double maxdif = 0.0 ;

      for ( ipt = 0 ; ipt < Npt ; ipt ++ )
	for ( ik = 0 ; ik < Nk ; ik ++ ) {
	  int i = (ipt * Nk) + ik ;
	  double dif = CM[ i ] - ColdM[ i ] ;
	  if ( dif < 0 ) dif = -dif ;
	  if ( dif > maxdif ) maxdif = dif ;
	}

      return ( maxdif < CvThres ) ;
      break;

    }

  case CVTEST_HARDCLASS: {
    /*cv criterium based on the difference between two consecutive segmentations */
      int     dif = 0;
      int*     oldhard;
      int*     newhard;
      int     ipt;
      int*    kmaxesV;


      if ((oldhard = GenAlloc(Npt, sizeof( int ), 0, "HasConverged","oldhard") ) == NULL )
            return STS_E_MEMORY ;
      if ((newhard = GenAlloc(Npt, sizeof( int ), 0, "HasConverged","newhard") ) == NULL )
            return STS_E_MEMORY ;
      if ( ( kmaxesV = GenAlloc( Nk, sizeof( int ),
				   0, "HasConverged", "kmaxesV" ) ) == NULL )
            return STS_E_MEMORY ;

      for (ipt = 0; ipt < Npt; ipt ++)
	{
          oldhard[ipt] = ComputeMAP(ColdM, ipt, Nk, TieRule, kmaxesV);
           newhard[ipt] = ComputeMAP(CM, ipt, Nk, TieRule, kmaxesV);
           if (  oldhard[ipt] != newhard[ipt])
             dif = dif +1;
 	}

       GenFree(oldhard);
       GenFree(newhard);
       GenFree(kmaxesV);

       return ( dif == 0);  /* <= Npt * 0.05 ); */
       break;
      }

     case CVTEST_CRIT: {
       double curcrit ;
       double critdif ;

 

       curcrit = ChosenCrit( CriterP , WhichCrit ) ;
       if ( curcrit != 0 )
 	critdif = fabs( ( curcrit - OldCrit ) / curcrit ) ;
       else
 	critdif = MAXDOUBLE ;

       return ( critdif < CvThres ) ;
       break;
     }

    case CVTEST_BETA: {
       double maxdif = 0.0 ;
       double dif = ParaP->Beta - OldBeta ;
 	  if ( dif < 0 ) dif = -dif ;
 	  if ( dif > maxdif ) maxdif = dif ;
       return ( maxdif < CvThres ) ;
       break;

     }

    case CVTEST_FUNC: {
      /*cv criterium for the field algos, based on the value of F' */
       double maxdif = 0.0 ;
       double dif = NewFunc - OldFunc ;
 	  if ( dif < 0 ) dif = -dif ;
 	  if ( dif > maxdif ) maxdif = dif ;
       return ( maxdif < CvThres ) ;
       break;

     }

  case CVTEST_PARAMS: {
    /* cv criterium based on beta and (mu, sigma) */
      int ik, id ;
      double maxdif = 0.0 ;

      double difbeta = ParaP->Beta - OldBeta;
	  if ( difbeta < 0 ) difbeta = -difbeta ;
 	  if ( difbeta > maxdif ) maxdif = difbeta ;

	for ( ik = 0 ; ik < Nk ; ik ++ )
        {
	  for ( id = 0 ; id < Nd ; id ++ )
           {
          double dif1 = ParaP->Center_KD[ik*Nd +id]- OldCenter[ik*Nd +id];

          if ( dif1 < 0 ) dif1 = -dif1 ;
          if ( dif1 > maxdif ) maxdif = dif1 ;
	   }
	}

	for ( ik = 0 ; ik < Nk ; ik ++ )
        {
	  for ( id = 0 ; id < Nd ; id ++ )
           {
	    double dif2 = 0 ;
	    switch (SpecP->ClassFamily)
	    {
	     case FAMILY_NORMALDIAG:
             dif2 = ParaP->Disp_KD[ik*Nd* + id]
                                 - OldDisp[ik*Nd +id];
             break;

	     case FAMILY_NORMALGEN:
	     if (SpecP->ClassDisper == DISPER_KD)
	          dif2 = ParaP->Disp_KD[ik*Nd*(Nd+1)/2 + id*(2*Nd +1 - id)/2]
                                 - OldDisp[ik*Nd*(Nd+1)/2+ id*(2*Nd +1 - id)/2];
	     if (SpecP->ClassDisper == DISPER__D)
	          dif2 = ParaP->Disp_KD[ id*(2*Nd +1 - id)/2]
                                 - OldDisp[ id*(2*Nd +1 - id)/2];
	       break;

	     case FAMILY_LAPLACE:
	     dif2 = ParaP->Disp_KD[ik*Nd*(Nd+1)/2 + id*(2*Nd +1 - id)/2]
                                 - OldDisp[ik*Nd*(Nd+1)/2+ id*(2*Nd +1 - id)/2];
             break;

	     default:
               return -1;

	    }
          if ( dif2 < 0 ) dif2 = -dif2 ;
          if ( dif2 > maxdif ) maxdif = dif2 ;
	   }
	}


      return ( maxdif < CvThres ) ;
       break;
  }


    case CVTEST_NONE: return FALSE ;
     default:          return FALSE ;
   }

 }   /* end of HasConverged() */


/*  ------------------------------------------------------------------- */
 static  int   ClassifyByFieldAlgo
            (
              const AlgoParaT*    AlgoParaP,
              const SpatialT*     SpatialP,
              DataT*              DataP,
              StatModelT*         StatModelP,
              MSCritT*            MSCritP,
              VarParamT*          VarParamP,
              double*             Increase,
              double*             PostM,
              WorkingT*           WorkingP,
              double*             EBlist
            )
/*  ------------------------------------------------------------------- */

 {
   /* Local variables declaration */
  const char*    func    = "ClassifyByFieldAlgo";
  StatusET       err     = STS_OK;
  int            npt     = DataP->NbPts;
  int            nk      = StatModelP->Spec.K;
  int            missk, missd;
  int*           LastCPUtime ;
  ModelSpecT* NewSpec     = (ModelSpecT*) malloc(sizeof(ModelSpecT)); /* for the case */
  ModelParaT* NewPara     = (ModelParaT*) malloc(sizeof(ModelParaT)); /* init mode = EMIND */
  AlgoParaT*  NewAlgo     = (AlgoParaT*) malloc(sizeof(AlgoParaT));
  CriterT*  FCriterP      = (CriterT*) malloc(sizeof(CriterT));

  // fprintf(stderr, "Entry of ClassifyByFieldAlgo.\n");
  if ((LastCPUtime  = GenAlloc(1, sizeof(int), 0, func,"LastCPUtime"))  == NULL ) return STS_E_MEMORY ;



    if (AlgoParaP->SinglMode != SINGL_EM)    {

      if  (AlgoParaP->InitMode != INIT_PARAM)
         InitPara( DataP, &(StatModelP->Desc), &(StatModelP->Spec),
	              AlgoParaP, &(StatModelP->Para), WorkingP->CtmpM);

    switch (AlgoParaP->InitMode)
      {
      case INIT_EMIND:  /*  initial classification is given by EM
                            for independent mixture */

            memcpy(NewSpec,&(StatModelP->Spec),sizeof(ModelSpecT));
            memcpy(NewPara,&(StatModelP->Para),sizeof(ModelParaT));
            memcpy(NewAlgo,AlgoParaP,sizeof(AlgoParaT));

            NewSpec->BetaModel   = BETA_FIX;
            NewSpec->ClassPropor = PROPOR_K;
            NewPara->Beta        = 0;
            NewAlgo->Algo        = ALGO_NEM;
            NewAlgo->SinglMode   = SINGL_NONE;
            NewAlgo->NbIters     = 100;

            fprintf( stderr, "Computing initial segmentation with EM for independent mixture \n");
            err = InitPartitionSort( DataP, nk, 1,PostM);
            err = NemAlgo(DataP,NewAlgo,SpatialP,NewSpec,&(StatModelP->Desc),
		  NewPara,PostM,WorkingP,FCriterP,MSCritP);

            if (err != 0){
	                  fprintf( stderr, "TRUBLES to compute initialization with EM for independent mixture (may an empty class?): PROGRAM STOPPED\n ");
                          return err;
	                }
            else  fprintf( stderr, "Initialization with EM for independent mixture computed \n");

            break ;


      case INIT_FILE:   /*  initial classification is given by a file   */
	break; /* end case (INIT_FILE) */
      case INIT_SORT:   /* initial partition to be computed by sorting   */
	fprintf( stderr,
		 "Computing initial partition (sort variable %d)\n",
		 AlgoParaP->SortedVar + 1 ) ;
	err = InitPartitionSort( DataP, nk,
				 AlgoParaP->SortedVar,
				 PostM);
	
	if (err != STS_OK) return err;
	break; /* end case (INIT_SORT) */

      case INIT_PARAM: /* read the given init params */
	  break;/* end case (INIT_PARAM) */

      default:
	/* to be treated */
	err= STS_OK;
      }

    if (AlgoParaP->InitMode!= INIT_PARAM)
               err = MakeParaFromLabeled( AlgoParaP, DataP, PostM,
			       &(StatModelP->Spec), &(StatModelP->Desc),
			       &(StatModelP->Para), & missk, & missd);


    if (err != STS_OK) return err;
    switch (AlgoParaP->ParamMode)
      {
      case PARAM_ESTIM:
	break;
      case PARAM_FIX:
	if ( ( err = ReadParamFile(StatModelP->Spec,
				   AlgoParaP->ParamName, /*V1.04-a*/
				   StatModelP->Spec.K,
				   DataP->NbVars,
				   StatModelP->Para.Center_KD,
				   StatModelP->Para.Disp_KD
				   ) ) != STS_OK )
	  return err ;
	break;
      default:
	err= STS_OK;
      } /* end switch ( AlgoParaP->ParamMode ) */
  }

   /* Initialize Ztilde  */
   if (AlgoParaP->InitMode!= INIT_PARAM)
   memcpy( StatModelP->Para.Ztilde , PostM , npt*nk * sizeof( double ) ) ;



  /* Now simply call FieldAlgoSingl */
  err=FieldAlgoSingl(DataP, AlgoParaP, SpatialP, StatModelP,
		      PostM, MSCritP, VarParamP, Increase,
		     WorkingP, LastCPUtime);





  GenFree(LastCPUtime);

  /* fprintf(stderr, "Leaving ClassifyByFieldAlgo.\n"); */
  return err;
 } /* end of classifyByFieldAlgo */

/*------------------------------------------------------------------------*/
int Algo /* Calls either SimulatedAlgo
	     or FieldAlgo. It depends on AlgoParaP->Algo */
(
 const DataT*        DataP,
 const AlgoParaT*    AlgoParaP,
 const SpatialT*     SpatialP,
 StatModelT*         StatModelP,
 double*             PostM,
 MSCritT*            MSCritP,
 VarParamT*          VarParamP,
 double*             Increase,
 WorkingT*           WorkP,
 int*                LastCPUtime
)
/*-----------------------------------------------------------------------*/
{
  /* for the errors */
  int err;
  /* fprintf(stderr,"AlgoParaP->Algo : %d.\n", AlgoParaP->Algo); */
  if ((AlgoParaP->Algo==ALGO_GSF) ) {
    /* fprintf(stderr,"Calling SimulatedAlgo.\n"); */
    err = SimulatedAlgo(DataP, AlgoParaP, SpatialP, StatModelP,
		    PostM, MSCritP, VarParamP, Increase,
		  WorkP, LastCPUtime);
  } else {
    /* fprintf(stderr,"Calling FieldAlgo.\n"); */
    err = FieldAlgo(DataP,AlgoParaP,SpatialP,&(StatModelP->Spec),&(StatModelP->Desc),
		    &(StatModelP->Para),PostM,
		    MSCritP,Increase,WorkP,LastCPUtime);
  }
  /* fprintf(stderr,"End of Algo.\n"); */
  return err;
}

/*-----------------------------------------------------------------------*/
void Proportion2Potentials /* Transforms pi_k to potentials V_k using
			      logarithm */
(
 ModelSpecT* SpecP, /* To get the number of classes */
 double* prop,      /* The proportions pi_k */
 double* singl      /* The potentials V_k */
)
/*-----------------------------------------------------------------------*/
{
  int i;
  for(i=0; i<SpecP->K; i++) {
    singl[i] = log(prop[i]);
  }
}

/*-----------------------------------------------------------------------*/
int FieldAlgoSingl /* estimates first potential on singletons with an EM
		      algorithm */
(
 const DataT*        DataP,
 const AlgoParaT*    AlgoParaP,
 const SpatialT*     SpatialP,
 StatModelT*         StatModelP,
 double*             PostM,
 MSCritT*            MSCritP,
 VarParamT*          VarParamP,
 double*             Increase,
 WorkingT*           WorkP,
 int*                LastCPUtime
 )
/*-----------------------------------------------------------------------*/
{
  /* Local Variables */
  /* Error ? */
  int err;
  /* We define new parameters so as not to modify the old ones
     when exiting the function */
  ModelSpecT* NewSpec     = (ModelSpecT*) malloc(sizeof(ModelSpecT));
  ModelParaT* NewPara     = (ModelParaT*) malloc(sizeof(ModelParaT));
  AlgoParaT*  NewAlgo     = (AlgoParaT*) malloc(sizeof(AlgoParaT));
  StatModelT* NewStat     = (StatModelT*) malloc(sizeof(StatModelT));
  /* Don't know what to do with CriterP
     -> get him into fantom variables...
     Beeh. Not really clean but. */
  CriterT*  FCriterP = (CriterT*) malloc(sizeof(CriterT));

  // fprintf(stderr,"Entry of FieldAlgoSingl.\n");
  /* Do we need to compute potential on singletons, and how do
     we want to compute them ?? */
  switch(AlgoParaP->SinglMode) {
  case SINGL_EM :
    /* Yes, we need... We compute them first using NemAlgo */
    /* Uses NEM algo to compute proportions between the potentials : pi_k */
    /* Some parameters to modify first so as to compute the good pi_k...*/
    memcpy(NewSpec,&(StatModelP->Spec),sizeof(ModelSpecT));
    memcpy(NewPara,&(StatModelP->Para),sizeof(ModelParaT));
    memcpy(NewAlgo,AlgoParaP,sizeof(AlgoParaT));
    NewSpec->BetaModel   = BETA_FIX;
    NewSpec->ClassPropor = PROPOR_K;
    NewPara->Beta        = 0;
    NewAlgo->Algo        = ALGO_NEM;
    err = NemAlgo(DataP,NewAlgo,SpatialP,NewSpec,&(StatModelP->Desc),
		  NewPara,PostM,WorkP,FCriterP,MSCritP);
    if (err != 0) return err;
    /* We get the pi_k from ParaP->Prop_K
       We transform them into potentials */
    Proportion2Potentials(&(StatModelP->Spec),NewPara->Prop_K,StatModelP->Para.Singl);
    
    /*     } */
    /* Now call Algo */
    err = Algo(DataP,AlgoParaP,SpatialP,StatModelP,PostM,
	       MSCritP,VarParamP,Increase,WorkP,LastCPUtime);
    /* That's all folks */
    break;
  case SINGL_BETA :
    /* Yes, we need. We compute first Beta using SimulatedAlgo and
       then we compute the singletons and the noise parameters using
       Nem. */
    /* Creating the Statmodel we need to call SimulatedAlgo */
    /* Not very beautiful method but no other choice */
    memcpy(NewStat,StatModelP,sizeof(StatModelT));
    /* A first call to Algo to compute Beta */
    err = Algo(DataP,AlgoParaP,SpatialP,NewStat,PostM,
	       MSCritP,VarParamP,Increase,WorkP,LastCPUtime);
    if (err != 0) return err;
    /* Before we call NemAlgo, setting the good parameters. */
    memcpy(NewSpec,&(StatModelP->Spec),sizeof(ModelSpecT));
    memcpy(NewAlgo,AlgoParaP,sizeof(AlgoParaT));
    NewSpec->BetaModel   = BETA_FIX;
    NewSpec->ClassPropor = PROPOR_K;
    NewAlgo->Algo        = ALGO_NEM;
    StatModelP->Para.Beta = NewStat->Para.Beta;
    /* Now we call NemAlgo */
    err = NemAlgo(DataP,NewAlgo,SpatialP,NewSpec,&(StatModelP->Desc),&(StatModelP->Para),
		  PostM,WorkP,FCriterP,MSCritP);
    
    break;

  default:
    /* No we needn't. Simply call Algo. */
    err = Algo(DataP,AlgoParaP,SpatialP,StatModelP,PostM,
	       MSCritP,VarParamP,Increase,WorkP,LastCPUtime);
  }
  return err;
} /* end of FieldAlgoSingl */




/*   -------------------------------------------------------------------    */
static int SimulatedAlgo
           (
             const DataT*        DataP,
             const AlgoParaT*    AlgoParaP,
             const SpatialT*     SpatialP,
             StatModelT*         StatModelP,
             double*              PostM,
             MSCritT*            MSCritP,
             VarParamT*          VarParamP,
             double*              Increase,
             WorkingT*           WorkP,
             int*            LastCPUtime
           )
/* run the  gsf and mcem algos Nbiters iterations,
   then do 20 more iters and compute the
     mean of the parameters estimations over these 20 iter ( namely \hat{Psi}),
   then simulate ztilde for these estimations \hat{Psi},
   then compute a segmentation by  MAP of Pztilde(z|y,\hat{Psi})  . */
/*   -------------------------------------------------------------------    */
{

     StatusET       err        = STS_OK;
     double*        meanCenter_KD ;
     double*        meanDisp_KD ;
     double*        Hclassif;
     int            Nk = StatModelP->Spec.K ;
     int            Nd = DataP->NbVars ;
     int            Npt = DataP->NbPts;

     /* initialisation of mean Para*/

     if ((meanCenter_KD  = GenAlloc(Nk*Nd , sizeof(double), 0,
				    "SimulatedAlgo", "meanCenter_KD"))  == NULL )
            return STS_E_MEMORY ;

     if ( (StatModelP->Spec.ClassFamily == FAMILY_NORMALGEN )
                          && (StatModelP->Spec.ClassDisper == DISPER__D) )
     {
        if ((meanDisp_KD  = GenAlloc(Nd , sizeof(double), 0,
				     "SimulatedAlgo", "meanDisp_KD"))  == NULL )
            return STS_E_MEMORY ;
     }
     else
     {
        if ((meanDisp_KD  = GenAlloc(Nd*Nk , sizeof(double), 0,
                                   "SimulatedAlgo", "meanStd_KD"))  == NULL )
            return STS_E_MEMORY ;
     }

     /* initialisation of Hclassif */
     if ((Hclassif = GenAlloc( Npt * Nk, sizeof(double), 0, "SimulatedAlgo", "Hclassif"))  == NULL )
	return STS_E_MEMORY ;


      err=FieldAlgo(DataP, AlgoParaP, SpatialP, &StatModelP->Spec,
	                               &StatModelP->Desc, 
	                               &StatModelP->Para, PostM, MSCritP, Increase, WorkP,LastCPUtime);

   /************/
     /* old version of SimulFieldAlgo, corresponding of doing 10 more runs of
            gsf and take an average of the parameters over these 10 runs
        there is a bug which induces bad BIC and ICL */
      /*  AlgoParaP->NbIters = 1;
   for (iter = 0 ; iter < 10 ; iter ++ )
     {
      err=FieldAlgo(DataP, AlgoParaP, SpatialP, &StatModelP->Spec,
	                               &StatModelP->Desc, Flog,
	                               &StatModelP->Para, PostM, MSCritP, Increase, WorkP,LastCPUtime);

      for (ik = 0; ik<Nk ; ik++)
	 {
	   for (id = 0; id<Nd ; id++)
	     {
	       (meanCenter_KD)[ik*Nd + id] = (meanCenter_KD)[ik*Nd + id]+
                               (StatModelP->Para.Center_KD)[ik*Nd + id] ;

              if ( (StatModelP->Spec.ClassFamily == FAMILY_NORMALGEN )
                          && (StatModelP->Spec.ClassDisper == DISPER__D) )
               {
                  if (ik == 0 )
                  (meanDisp_KD)[id] = (meanDisp_KD)[id]+
                          StatModelP->Para.Disp_KD[id*(2*Nd+1-id)/2] ;
	       }

               if ( (StatModelP->Spec.ClassFamily == FAMILY_NORMALGEN )
                          && (StatModelP->Spec.ClassDisper == DISPER_KD) )
               {
                  (meanDisp_KD)[ik * Nd +id] = (meanDisp_KD)[ik * Nd +id] +
                     StatModelP->Para.Disp_KD[ik * Nd*(Nd+1)/2 +id*(2*Nd+1-id)/2] ;
	       }

               if (StatModelP->Spec.ClassFamily == FAMILY_LAPLACE )
               {
                  (meanDisp_KD)[ik * Nd +id] = (meanDisp_KD)[ik * Nd +id] +
                      StatModelP->Para.Disp_KD[ik * Nd*(Nd+1)/2 +id*(2*Nd+1-id)/2] ;
	       }

               if (StatModelP->Spec.ClassFamily == FAMILY_NORMALDIAG )
               {
                  (meanDisp_KD)[ik * Nd +id] = (meanDisp_KD)[ik * Nd +id] +
                            StatModelP->Para.Disp_KD[ik * Nd +id] ;
	       }


	       } */ /* end of for (id =0, ...) */
      /* }*/ /* end of for (k=0, ...) */

      /*  meanBeta = meanBeta + StatModelP->Para.Beta;
	  } */ /* end of for (ik = 0; ik<Nk ; ik++) */


   /* compute the mean over the 10 last iters of each parameter leading to an estimator hat{Psi} */
      /* for (ik = 0; ik<Nk ; ik++)
	 {
	   for (id = 0; id<Nd ; id++)
	     {
	       (StatModelP->Para.Center_KD)[ik*Nd + id] = (meanCenter_KD)[ik*Nd + id]/ 10 ;

              if ( (StatModelP->Spec.ClassFamily == FAMILY_NORMALGEN )
                          && (StatModelP->Spec.ClassDisper == DISPER__D) )
               {
                  if (ik == 0 )
		    StatModelP->Para.Disp_KD[id*(2*Nd+1-id)/2] = (meanDisp_KD)[id]/10 ;
	       }

               if ( (StatModelP->Spec.ClassFamily == FAMILY_NORMALGEN )
                          && (StatModelP->Spec.ClassDisper == DISPER_KD) )
               {
                  StatModelP->Para.Disp_KD[ik * Nd*(Nd+1)/2 +id*(2*Nd+1-id)/2] =
                                  (meanDisp_KD)[ik * Nd +id]/10 ;
	       }

               if (StatModelP->Spec.ClassFamily == FAMILY_LAPLACE )
               {
                  StatModelP->Para.Disp_KD[ik * Nd*(Nd+1)/2 +id*(2*Nd+1-id)/2] =
                            (meanDisp_KD)[ik * Nd +id]/10  ;
	       }

               if (StatModelP->Spec.ClassFamily == FAMILY_NORMALDIAG )
               {
                    StatModelP->Para.Disp_KD[ik * Nd +id] = (meanDisp_KD)[ik * Nd +id]/10 ;
	       }


	       }*/ /* end of for (id =0, ...) */
      /* }*/ /* end of for (k=0, ...) */

      /*   StatModelP->Para.Beta = meanBeta/10 ;*/

   /* simulate ztilde from the cond. markovian distrib with the parameter = hat{Psi} */
      /*    AlgoParaP->NbEIters = 50 ;
       SimulateZSeq(AlgoParaP,DataP->LabelV,SpatialP,&StatModelP->Spec,
                    &StatModelP->Para,DataP, StatModelP->Para.Beta,StatModelP->Para.Singl,
                    WorkP, Npt,Nk,StatModelP->Para.Ztilde,PostM);*/

   /* compute the segmentation, the error rate and the criteria */
      /*  err = HardClassif(AlgoParaP,DataP,Npt,Nk,PostM,Hclassif);

      if ( Nk == AlgoParaP->True.K)
	{
	  ComputeErrorRate(Hclassif,
			   AlgoParaP->True.DataTrue,
			   AlgoParaP->True.ErrorRate,
			   Nk,
			   Npt,
			   SpatialP->NeighData.Image.Nl,
			   SpatialP->NeighData.Image.Nc);
	}

     err = ComputeBIC( &StatModelP->Spec, &StatModelP->Para, DataP, SpatialP, AlgoParaP,Hclassif,
                PostM , 0, MSCritP, WorkP);

     GenFree(Hclassif);
     return err ;*/



      return(err);


} /* end of SimulatedAlgo */


/*   -------------------------------------------------------------------    */
   int FieldAlgo
             (
               const DataT*        DataP,
               const AlgoParaT*    AlgoParaP,
               const SpatialT*     SpatialP,
               const ModelSpecT*   SpecP,
               const SampleDesT*   DescP,
               ModelParaT*         ParaP,
               double*              PostM,
               MSCritT*            MSCritP,
               double*              Increase,
               WorkingT*           WorkP,
               int*            LastCPUtime
             )
     /* common body for confmf, gsf, mcem, mode,icm, meanf2, mixedmf
      and marmf algos */
  /* ------------------------------------------------------------------- */
  {

     StatusET       err        = STS_OK;
     int            result;
     int            npt        = DataP->NbPts;
     int            nk         = SpecP->K;
     int            nd         = DataP->NbVars;
     int            ipt;       /*  index of current point    */
     int            ik;        /*  index of current class */
     int            iter;      /*  index of current iteration */
     int            Iters = AlgoParaP->NbIters*AlgoParaP->NbEMIters;
     int            i;
     int            converged; /*  TRUE if converged reached */
     int            emptyk;    /*  index of empty class, 0 if none */
     double          oldbeta;   /* last value of beta */
     double         Fnnn = 0;
     double         Fono = 0;
     double         oldFono = 0;
     double         Fonn = 0;
     double*         oldcenter;     /* previous centers */
     double*         oldisp;      /* previous dispersions */
     double*         Hclassif;
     int             CPUtime;    /* CPU used by the algo */
     double*         priormean; /* MF approx of the mean a priori */
     double*         postMtemp;
     double*         ztildetemp; /* for GSF*/
     ParameterT      Parameter; /* For call to ZeroIn */

     ComputeCritGrad*  f;  /* for Beta estimation by ZeroIn */

      if ((Hclassif = GenAlloc( npt * nk, sizeof(double), 0, "FieldAlgo", "Hclassif"))  == NULL )
	return STS_E_MEMORY ;
     if ((ztildetemp = GenAlloc( npt * nk, sizeof(double), 0, "FieldAlgo", "ztildetemp"))  == NULL )
            return STS_E_MEMORY ;
     if ((priormean = GenAlloc(nk, sizeof(double), 0, "FieldAlgo", "priormean"))  == NULL )
            return STS_E_MEMORY ;
     if ((postMtemp = GenAlloc(npt*nk, sizeof(double), 0, "FieldAlgo", "postMtemp"))  == NULL )
            return STS_E_MEMORY ;



     /* initialize the time CPU used */
       CPUtime = 0;

 

      /* the loop starts by updating the params so it is  OK if init with
	 classif, but we will have to adapt this
	 if init with params */
      switch (AlgoParaP->InitMode)
     {
      case INIT_PARAM: /*init with param, so we need to run one iter of the
			 step (2) to  compute Ztilde and postM */
        

       result = FieldAlgoNeighUpdating(DataP, AlgoParaP, SpatialP, SpecP, DescP,   ParaP, PostM,
              MSCritP,  WorkP);
       break;

      default : /* init with a classif */

      /* Compute P_ztilde(. |y, theta_0, beta_0) */

	 result = ComputeDensity( ParaP, DataP, SpecP, WorkP->FikM, WorkP->LogFikM );
         result = ComputePostPartition( SpecP, ParaP,DataP, SpatialP,AlgoParaP,
                                       ParaP->Ztilde,WorkP,PostM);

      }

      if (result != STS_OK)
          {
	    GenFree(Hclassif);
            GenFree(priormean);
            GenFree(postMtemp);
            GenFree(ztildetemp);
	    return -1;
	  }

      else {

      /* Initialize WorkP */
        memcpy( WorkP->ColdM, PostM, npt * nk * sizeof( double ) ) ;
  	memcpy( WorkP->Ztilde, ParaP->Ztilde, npt * nk * sizeof( double ) ) ;

      fprintf( stderr, "  Iterations : " ) ;
      fprintf( stderr, "%4d ", 0 ) ;

     if (SpecP->BetaModel == BETA_SA) ParaP->Beta=0;



     /**********BEGINING OF THE  LOOP**********/
     /* For each iteration of FieldAlgo (until :
         convergence, or iteration count reached or empty class)   */
      for ( iter = 1, converged = FALSE ;
            ( iter <= Iters ) &&
	    ( ! converged) &&
	    ( CPUtime <AlgoParaP->CPUmax ) &&
            ( err == STS_OK ) ;
            iter ++ )
      {


	 /* save last value of centers and dispersions */
         oldcenter = ParaP->Center_KD;
         oldisp = ParaP->Disp_KD;

          /* Save last value of beta */
          oldbeta = ParaP->Beta ;



	  /****  step (1.1):theta estimation */
	  switch (AlgoParaP->ParamMode)
	   {
           case PARAM_FIX:
                break;

           case PARAM_ESTIM:
               err=EstimPara(PostM, DataP, AlgoParaP, nk, AlgoParaP->MissMode,
                                  SpecP, &emptyk, ParaP);
               break;

           default:
                fprintf( stderr, "Error, unknown non-Beta parameters estimation mode\n" );
                err = STS_E_FUNCARG ;
	   }

	  

	  /****  step (1.2): beta estimation */
         switch ( SpecP->BetaModel )
	      {
              case BETA_PSGRAD:
                result = EstimBetaForFieldAlgo( SpecP->BetaModel, AlgoParaP->BtaPsGrad,
                             SpatialP, ParaP->Ztilde, PostM, ParaP->Singl, npt, nk,
                             &ParaP->Beta, &WorkP->Neighs);
                break;

              case BETA_ZEROIN:
		Parameter.Type = BETA;
		f = ComputeGradBeta;
                result = ZeroIn(0,10,f,0.0001,npt,nk,&ParaP->Beta,
                                ParaP->Ztilde,ParaP->Singl,PostM,SpatialP,
				&WorkP->Neighs,Parameter);
                break;

	      case BETA_SA:
                ParaP->Beta = iter*1/Iters;
                break;

              default:/* BETA_FIX*/ break;
	      }


	/**** Step (1.3) : Potentials estimation */
	if (AlgoParaP->SinglMode == SINGL_ZEROIN) {
	  Parameter.Type = POTENTIAL;
	  f              = ComputeGradPotential;
	  for (Parameter.No = 0; Parameter.No < nk; Parameter.No++)
	    {
	      ZeroIn(-10,10,f,0.0001,npt,nk,&ParaP->Beta,
		     ParaP->Ztilde,ParaP->Singl,PostM,SpatialP,
		     &WorkP->Neighs,Parameter);

	    }
	}

	 /* save oldFono */
         oldFono = Fono ;

         /* compute F'(Ptilde(q-1), Psi(q), ztilde(q-1)) */
         if (AlgoParaP->CompF == TRUE ) {
         Fono = ComputeFunc(SpecP,ParaP, DataP, SpatialP, AlgoParaP, WorkP ,
			    WorkP->ColdM,WorkP->Ztilde);}


     /* If no empty class  or ParamMode = Fix */
     if(err == STS_OK)
       {
	 result = ComputeDensity( ParaP, DataP, SpecP, WorkP->FikM, WorkP->LogFikM );
         /* for multivariate data, cholesky decomp may failed */
         if ( result == STS_OK)
	   {

	 /****  step (2): updating of z~  */
	 if ((iter % (AlgoParaP->NbEMIters)) == 0 )
	 {
          switch ( AlgoParaP->Algo )
  	  {
            case ALGO_MEANF:
                 result = ComputePartitionNEM ( SpecP, ParaP, DataP, SpatialP, AlgoParaP, 
                                        ParaP->Ztilde, WorkP, NULL);
                 break;

            case ALGO_GSF:
              for (ik = 0; ik<nk; ik++){
		   for (ipt = 0; ipt<npt; ipt++) {
                       postMtemp[ipt*nk + ik]= 0;
                   }
                 }
	      /* the GS is initialized  with the current value of Ztilde */
              for ( i = 0 ; i <AlgoParaP->NbGibbs  ; i ++ )
              {
                switch (AlgoParaP->SiteUpdate)
                {
                 case UPDATE_PARA:
	        /* computation of P'(z|y,z~(q-1),Psi(q)) */
                result = ComputePostPartition( SpecP, ParaP,DataP, SpatialP,AlgoParaP,
                                       ParaP->Ztilde,WorkP,PostM);

                /* realization of P'(z|y,z~(q-1),Psi(q)) */
                 SimulateZ(AlgoParaP,DataP->LabelV,PostM, DataP->NbPts, SpecP->K,ParaP->Ztilde );
                 break;

		 case UPDATE_SEQ:
                 SimulateZSeq(AlgoParaP,DataP->LabelV,SpatialP,SpecP,ParaP,DataP,ParaP->Beta,ParaP->Singl,WorkP,
                                     npt,nk,ParaP->Ztilde,PostM);
                 break;

		 default: break;
                }
                result = ComputePostPartition( SpecP, ParaP,DataP, SpatialP,AlgoParaP,
                                       ParaP->Ztilde,WorkP,PostM);
                for (ik = 0; ik<nk; ik++){
		   for (ipt = 0; ipt<npt; ipt++) {
                       postMtemp[ipt*nk + ik]= postMtemp[ipt*nk + ik] + PostM[ipt*nk + ik];
                   }
                 }
              }

            break;

	    case ALGO_ICM:
                 result = ComputePartitionICM(SpecP, ParaP, DataP, SpatialP, AlgoParaP, 
                                        ParaP->Ztilde, WorkP, NULL);

                 break;



            default:
                fprintf( stderr, "Error, unknown algorithm %d\n", AlgoParaP->Algo );
                err = STS_E_FUNCARG ;
  	  }

         /* compute F'(Ptilde(q-1), Psi(q), ztilde(q)) */
         if (AlgoParaP->CompF == TRUE ) {
         Fonn = ComputeFunc(SpecP,ParaP, DataP, SpatialP, AlgoParaP, WorkP ,
			    WorkP->ColdM,ParaP->Ztilde);}
	 }

	 /*****  step (3): computation of Postik */
           switch ( AlgoParaP->Algo )
  	  {
            case ALGO_MEANF:
	    case ALGO_ICM:
                 memcpy( PostM ,ParaP->Ztilde, npt*nk * sizeof( double ) ) ;
                 break;

	    case ALGO_GSF:
              for (ik = 0; ik<nk; ik++){
	       for (ipt = 0; ipt<npt; ipt++) {
                       PostM[ipt*nk + ik]= postMtemp[ipt*nk + ik]/AlgoParaP->NbGibbs;
                   }
	      }
              break;

	   default:
              fprintf( stderr, "Error, unknown algorithm %d\n", AlgoParaP->Algo );
              err = STS_E_FUNCARG ;
  	  }



         /* compute F'(Ptilde(q), Psi(q), ztilde(q)) */
         if (AlgoParaP->CompF == TRUE ) {
         Fnnn = ComputeFunc(SpecP,ParaP, DataP, SpatialP, AlgoParaP, WorkP ,
			    PostM, ParaP->Ztilde );

         }


	 if ((iter % (AlgoParaP->NbEMIters)) == 0 )
	 {
         converged = HasConverged( AlgoParaP->CvTest, AlgoParaP->CvThres,
                                    AlgoParaP->TieRule, WorkP->ColdM,
                                    PostM, npt, nk,nd ,oldFono, Fono, 0.0,oldbeta,
                                    oldcenter,oldisp , ParaP,AlgoParaP, SpecP,SpatialP,
                                    DataP, AlgoParaP->Crit, 0 , WorkP);
	 }
          /* Save last computed partition as old partition */
          memcpy( WorkP->ColdM, PostM, npt * nk * sizeof( double ) ) ;

          /* Save last computed z~ as old z~ */
  	memcpy( WorkP->Ztilde, ParaP->Ztilde, npt * nk * sizeof( double ) ) ;

	   } /* end of successfull ComputeDensity */

       else /* ComputeDensity  failed */
	 iter = AlgoParaP->NbIters;

       } /* end for no empty class */

     /* If empty class */
       else if (err ==STS_W_EMPTYCLASS)
	 {
           fprintf( stderr, "class %d is empty at iteration %d\n", emptyk, iter);
         }



        fprintf( stderr, "\b\b\b\b\b" ) ;
        fprintf( stderr, "%4d ", iter ) ;


	 /* updatind the CPU time */
      CPUtime = clock()/ CLOCKS_PER_SEC - *LastCPUtime;


      } /* end for (iter = 1 ....) */
      /**********END OF THE LOOP***********/


    iter = iter -1;
    *LastCPUtime = *LastCPUtime + CPUtime;


   /* If convergence test was requested and no error occurred */
      if ( ( AlgoParaP->CvTest != CVTEST_NONE ) && ( err == STS_OK ) )
      {
          /* Display message about convergence */
          if ( converged )
          {
              fprintf( stderr, "  Algo converged after %d iterations\n", iter ) ;
          }
          else
          {
              fprintf( stderr, "  Algo did not converge after %d iterations\n",
                       iter ) ;
          }
      }


     err = HardClassif(AlgoParaP,DataP,npt,nk,PostM,Hclassif);


      if ( nk == AlgoParaP->True.K)
	{
	  ComputeErrorRate(Hclassif,
			   AlgoParaP->True.DataTrue,
			   AlgoParaP->True.ErrorRate,
			   nk,
			   npt,
			   SpatialP->NeighData.Image.Nl,
			   SpatialP->NeighData.Image.Nc);
	}

     err = ComputeBIC( SpecP, ParaP, DataP, SpatialP, AlgoParaP,Hclassif,
                PostM , 0, MSCritP, WorkP);



      }

      GenFree(Hclassif);
      GenFree(priormean);
      GenFree(postMtemp);
      GenFree(ztildetemp);

      return err;

 } /* end for FieldAlgo */


/*   -------------------------------------------------------------------    */
   int FieldAlgoNeighUpdating
             (
               const DataT*        DataP,
               const AlgoParaT*    AlgoParaP,
               const SpatialT*     SpatialP,
               const ModelSpecT*   SpecP,
               const SampleDesT*   DescP,
               ModelParaT*         ParaP,
               double*             PostM,
               MSCritT*            MSCritP,
               WorkingT*           WorkP
             )
     /* "almost" Ztilde and PostM updating step for confmf, gsf, mcem, mode,icm, meanf2, mixedmf
      and marmf algos. ised if init with param instead of classif */
  /* ------------------------------------------------------------------- */
{

     int            result;
     int            ik, ipt, i;
     int            npt        = DataP->NbPts;
     int            nk         = SpecP->K;
     double*         priormean; /* MF approx of the mean a priori */
     double*         postMtemp;
     StatusET       err        = STS_OK;



     if ((priormean = GenAlloc(nk, sizeof(double), 0, "FieldAlgo", "priormean"))  == NULL )
            return STS_E_MEMORY ;
     if ((postMtemp = GenAlloc(npt*nk, sizeof(double), 0, "FieldAlgo", "postMtemp"))  == NULL )
            return STS_E_MEMORY ;


	 result = ComputeDensity( ParaP, DataP, SpecP, WorkP->FikM, WorkP->LogFikM );
         /* for multivariate data, cholesky decomp may failed */
         if ( result == STS_OK)
	   {

	 /****   updating of z~  */

          switch ( AlgoParaP->Algo )
  	  {
            case ALGO_MEANF:
                 result = ComputePartitionNEM ( SpecP, ParaP, DataP, SpatialP, AlgoParaP, 
                                        ParaP->Ztilde, WorkP, NULL);
                 break;

            case ALGO_GSF:
              for (ik = 0; ik<nk; ik++){
		   for (ipt = 0; ipt<npt; ipt++) {
                       postMtemp[ipt*nk + ik]= 0;
                   }
                 }
	      /* the GS is initialized  with the current value of Ztilde */
              for ( i = 0 ; i <AlgoParaP->NbGibbs  ; i ++ )
              {
                switch (AlgoParaP->SiteUpdate)
                {
                 case UPDATE_PARA:
	        /* computation of P'(z|y,z~(q-1),Psi(q)) */
                result = ComputePostPartition( SpecP, ParaP,DataP, SpatialP,AlgoParaP,
                                       ParaP->Ztilde,WorkP,PostM);

                /* realization of P'(z|y,z~(q-1),Psi(q)) */
                 SimulateZ(AlgoParaP,DataP->LabelV,PostM, DataP->NbPts, SpecP->K,ParaP->Ztilde );
                 break;

		 case UPDATE_SEQ:
                 SimulateZSeq(AlgoParaP,DataP->LabelV,SpatialP,SpecP,ParaP,DataP,ParaP->Beta,ParaP->Singl,WorkP,
                                     npt,nk,ParaP->Ztilde,PostM);
                 break;

		 default: break;
                }
                result = ComputePostPartition( SpecP, ParaP,DataP, SpatialP,AlgoParaP,
                                       ParaP->Ztilde,WorkP,PostM);
                for (ik = 0; ik<nk; ik++){
		   for (ipt = 0; ipt<npt; ipt++) {
                       postMtemp[ipt*nk + ik]= postMtemp[ipt*nk + ik] + PostM[ipt*nk + ik];
                   }
                 }
              }

            break;

	    case ALGO_ICM:
                 result = ComputePartitionICM(SpecP, ParaP, DataP, SpatialP, AlgoParaP, 
                                        ParaP->Ztilde, WorkP, NULL);

                 break;



            default:
                fprintf( stderr, "Error, unknown algorithm %d\n", AlgoParaP->Algo );
                err = STS_E_FUNCARG ;
  	  }




	 /*****   computation of Postik */
           switch ( AlgoParaP->Algo )
  	  {
            case ALGO_MEANF:
	    case ALGO_ICM:
                 memcpy( PostM ,ParaP->Ztilde, npt*nk * sizeof( double ) ) ;
                 break;

	    case ALGO_GSF:
              for (ik = 0; ik<nk; ik++){
	       for (ipt = 0; ipt<npt; ipt++) {
                       PostM[ipt*nk + ik]= postMtemp[ipt*nk + ik]/AlgoParaP->NbGibbs;
                   }
	      }
              break;

	   default:
              fprintf( stderr, "Error, unknown algorithm %d\n", AlgoParaP->Algo );
              err = STS_E_FUNCARG ;
  	  }




          /* Save last computed partition as old partition */
          memcpy( WorkP->ColdM, PostM, npt * nk * sizeof( double ) ) ;

          /* Save last computed z~ as old z~ */
  	memcpy( WorkP->Ztilde, ParaP->Ztilde, npt * nk * sizeof( double ) ) ;

       } /* end of successfull ComputeDensity */

       else /* ComputeDensity  failed */
	 err =  -1;


	 return err ;





      GenFree(postMtemp);
      GenFree(priormean);

   } /* end of NeighUpdating() */



  /* ------------------------------------------------------------------- */
  static void SimulateZ
             (
	       const AlgoParaT*     AlgoParaP,
               const int*          LabelV,
	       const double*       PostM,  /* not modified */
               const int           Npt,           /* nb of points */
               const int           Nk,            /* nb of classes */
               double*             simulZ         /*result of max(F') in z~ */
             )
  /* simulation of a field simulZ according to the proba PostM */
  /* ------------------------------------------------------------------- */
  {
    int     ipt;       /* current point */
    int     ik;        /* current class */
    int     classi;    /* class of z~i */
    double   u;         /* random */
    double  F=0;       /* sum of the PostIik, k=0..Nk-1 */


    for ( ipt = 0; ipt < Npt; ipt++ )
    {
      /* If label is not known */
      if ( ( AlgoParaP->Label   != 1 ) ||
		 ( LabelV[ ipt ] == 0 ) )
      {
       F=PostM[ipt*Nk];
       u=RandomDouble(0,1);
       ik=0;
       while (u > F)
        {
          ik=ik+1;
          F= F + PostM[(ipt*Nk)+ik];
        }
       classi=ik;
       for ( ik = 0; ik < Nk; ik++)
        {
          if (ik == classi)
  	 {simulZ[(ipt*Nk)+ik]=1;}
          else
           {simulZ[(ipt*Nk)+ik]=0;}
        }

      }
      /* else label known, remains unchanged */
    } /* end for for ( ipt = 0; ... */


  } /* end for SimulateZ */

/* ------------------------------------------------------------------- */
static void SimulateZSeq (
  const AlgoParaT*     AlgoParaP,
  const int*           LabelV,
  const SpatialT*      SpatialP,
  const ModelSpecT*    SpecP,
  const ModelParaT*    ParaP,
  const DataT*         DataP,
  const double         Beta,
  const double*        Singl,
  WorkingT*            WorkP,
  const int            Npt,
  const int            Nk,
  double*              SimulZ, /* simulated according to PostM
				  (sequential version)*/
  double*              PostM   /* modified after each local
				  updating of ZtildeM */
  )
/* Gibbs Sampling, if run until convergence                            */
/* ------------------------------------------------------------------- */
{
  int          itere;      /* current iteration */
  int          ipt;
  double       F=0;
  int          ik;
  double       u;
  int          classi;


 /* Iteratively compute new partition */
 for (itere = 0 ; itere < AlgoParaP->NbEIters ; itere ++)
   {
     for (ipt=0; ipt<Npt;ipt++)
       {
	 /* If label is not known */
	 if ((AlgoParaP->Label   != 1) || (LabelV[ ipt ] == 0))
	   {
	     ComputeLocalPostPartition(SpecP,ParaP,DataP,SpatialP,ipt,Nk,
				       &(WorkP->Neighs),
				       SimulZ,Beta,Singl, WorkP->FikM,PostM);
	     F=PostM[ipt*Nk];
	     u=RandomDouble(0,1);
	     ik=0;
	     while (u > F)
	       {
		 ik=ik+1;
		 F= F + PostM[ipt*Nk +ik];
	       }
	     classi=ik;
	     for ( ik = 0; ik < Nk; ik++)
	       {
		 if (ik == classi)
		   SimulZ[(ipt*Nk)+ik]=1;
		 else
		   SimulZ[(ipt*Nk)+ik]=0;
	       }
	   }
	 /* else label known, remains unchanged */
       } /* end for (ipt=0....*/
   } /* end for (itere = 0 ...) */
}  /* end for SimulateZSeq */

/* ------------------------------------------------------------------- */
void CreateGibbs(
		 DataT*      DataP,
		 AlgoParaT*  AlgoParaP,
		 StatModelT* StatModelP,
		 SpatialT*   SpatialP
)
/* ------------------------------------------------------------------- */
{
  /* Local declarations */
  /* Loops */
  int iIter;
  int iPts;
  /* Number of classes */
  int nK   = StatModelP->Spec.K;
  /* Number of points */
  int nPts = DataP->NbPts;
  /* To compute the class knowing a posteriori probability */
  double sumProb;
  double uniform;
  int    class;
  /* Create The PostM matrix */
  double* PostM;
  PostM = (double*) malloc (sizeof(double)*nK*nPts);
  /* Make as enough iterations as necessary */
  for(iIter = 0; iIter<AlgoParaP->NbGibbsIters; iIter++) {
    // fprintf(stderr,"Gibbs sampler : iteration %d.\n",iIter);
    for(iPts = 0; iPts<DataP->NbPts; iPts++) {
      // fprintf(stderr,"Treating point no %d.\n",iPts);
      ComputeGibbsPartition(&(StatModelP->Spec),&(StatModelP->Para),
			    DataP,SpatialP,
			    SpatialP->NeighData.PtsNeighsV,
			    iPts,PostM);
      sumProb = PostM[iPts*nK];
      uniform = RandomDouble(0,1);
      class   = 0;
      while (uniform > sumProb)
	{
	  class++;
	  sumProb += PostM[iPts*nK + class];
	}
      DataP->PointsM[iPts] = class;
    } /* end For iPts */
  } /* end For iIter  */
}

/* ------------------------------------------------------------------- */

   static int ComputeNRJCond
       (
        const int          Npt,
        const int          Nk,
        const double*      ClassifM, /* must be hard classification */
        const double*      ZtildeM,
        const double*      Singl,
        const double       beta,
        const double*      LogFikM,
        const SpatialT*    SpatialP,
        const int          a,
        WorkingT*          WorkP,
        double*            Ecompute
       )
     /* if a = 0, compute the energy function H of the Potts model with external field */
     /* if a = 1, compute the mean field like approximation of
                  the energy function H of the Potts model with external field */
 /* ------------------------------------------------------------------- */
{
  int    ipt,ik;
  double  NRJ = 0;
  PtNeighsT*   NeighsP = &(WorkP->Neighs);

  GetNeighFT*  fGetNeigh ; /* function fetching neighbors of a site */

   /* Get function specific to chosen spatial model */
   if ( GetSpatialFunc( SpatialP->Type,
			       & fGetNeigh ) != STS_OK )
    return 2 ;

   for (ipt = 0; ipt<Npt; ipt++)
     {
       int nbn = fGetNeigh(ipt,&(SpatialP->NeighData),NeighsP);
       for (ik = 0; ik<Nk; ik++)
	 {
          if (a == 0)
          NRJ = NRJ + ClassifM[ipt*Nk + ik]*(0.5*beta*SumNeighsOfClass(ik,nbn,Nk,NeighsP->NeighsV,ZtildeM)
                    + Singl[ik]+ LogFikM[ipt*Nk + ik]);

         if (a == 1)
          NRJ = NRJ + ClassifM[ipt*Nk + ik]*(beta*SumNeighsOfClass(ik,nbn,Nk,NeighsP->NeighsV,ZtildeM)
                    + Singl[ik]+ LogFikM[ipt*Nk + ik]);
         }
     }
  *Ecompute = NRJ;
  return STS_OK;


} /* end of ComputeNRJCond */

  /* ------------------------------------------------------------------- */
   static int ComputeNRJMarg
       (
        const int          Npt,
        const int          Nk,
        const double*      ClassifM, /* must be hard classification */
        const double*      ZtildeM,
        const double*      Singl,
        const double       beta,
        const SpatialT*    SpatialP,
        const int          a,
        WorkingT*          WorkP,
        double*            Ecompute
       )
     /* if a = 0, compute the energy function H of the Potts model with external field */
     /* if a = 1, compute the mean field like approximation of
                  the energy function H of the Potts model with external field */
 /* ------------------------------------------------------------------- */
{
  int    ipt,ik;
  double  NRJ = 0;
  PtNeighsT*   NeighsP = &(WorkP->Neighs);

  GetNeighFT*  fGetNeigh ; /* function fetching neighbors of a site */

   /* Get function specific to chosen spatial model */
   if ( GetSpatialFunc( SpatialP->Type,
			       & fGetNeigh ) != STS_OK )
    return 2 ;

   for (ipt = 0; ipt<Npt; ipt++)
     {
       int nbn = fGetNeigh(ipt,&(SpatialP->NeighData),NeighsP);
       for (ik = 0; ik<Nk; ik++)
	 {
          if (a == 0)
          NRJ = NRJ + ClassifM[ipt*Nk + ik]*(0.5*beta*SumNeighsOfClass(ik,nbn,Nk,NeighsP->NeighsV,ZtildeM)
                    + Singl[ik]);

         if (a == 1)
          NRJ = NRJ + ClassifM[ipt*Nk + ik]*(beta*SumNeighsOfClass(ik,nbn,Nk,NeighsP->NeighsV,ZtildeM)
                    + Singl[ik]);
         }
     }
  *Ecompute = NRJ;
  return STS_OK;


} /* end of ComputeNRJMarg */

  /* ------------------------------------------------------------------- */
   static int ComputePostPartition
             (
              const ModelSpecT*    SpecP,       /* I */
              const ModelParaT*    ParaP,
             const DataT*         DataP,        /* I */
              const SpatialT*      SpatialP,
  	    const AlgoParaT*     AlgoParaP,    /* I */
              const double*         ZtildeM,      /* z~ik matrix */
              WorkingT*            WorkP,
              double*               PostM       /* P~ik matrix   */
             )
    /* compute an approximation of the a posteriori distribution of the markov
	field, setting neighborhood to ZtildeM */
  /* ------------------------------------------------------------------- */
  {
    /* const char*     func = " ComputePostPartition"; */

    StatusET err;
    int          Nk      = SpecP->K;
    int          Npt     = DataP->NbPts;
    int          ik;     /* current class */
    int          ivis;   /* current point */
    int          nbn;    /* nb of neighbours of ipt */
    int          n;      /* current neighbour n=0..nbn */
    double        Beta    = ParaP->Beta;
    double*       FikM    = WorkP->FikM;
    double        numik;  /* fk(yi)* exp(-beta sum i's neighbours potentials */
    double        sumi;   /* sum of numik, k=O..Nk */
    double        inv ;
    double        potik;  /* i's neibourghs potentials */
    double*       posterior;
    PtNeighsT*   NeighsP = &(WorkP->Neighs);
    GetNeighFT*  FGetNeigh;   /* function to get neighbors */



    if ((posterior = GenAlloc( Npt * Nk, sizeof(double), 0, "ComputePostPartition", "posterior"))==NULL)
         fprintf(stderr,"impossible to allocate posterior in function ComputePostPartition\n");
    if ( ( err = GetSpatialFunc( SpatialP->Type,
                                   & FGetNeigh ) ) != STS_OK )
          return err ;


    for ( ivis = 0 ; ivis < Npt ; ivis ++)
      {

       int ipt = DataP->SiteVisitV[ ivis ] ;  /*V1.04-c*/


       /* If label is not known
       if ( ( AlgoParaP->Label   != 1 ) ||
       ( DataP->LabelV[ ipt ] == 0 ) ) { */

        nbn= FGetNeigh(ipt, &(SpatialP->NeighData),NeighsP );

        for ( ik = 0, sumi = 0; ik<Nk; ik++)
        {
          if ((SpecP->ClassFamily ==FAMILY_NORMALDIAG  )
               &&(ParaP->Disp_KD[ik] == 0 )){
                  int diff = 0;
                  diff = diff + ParaP->Center_KD[ik ]
                                             - DataP->PointsM[ipt*DataP->NbVars+1];

                if(diff == 0){ posterior[(ipt*Nk) + ik] = 1;
                               sumi = 1; }
                else { posterior[(ipt*Nk) + ik] = 0 ;
                               sumi = 1; }
	  }

          else{
          for ( n = 0, potik = 0; n<nbn; n++)
           {
           int   j = NeighsP->NeighsV[n].Index;
           double wij = NeighsP->NeighsV[n].Weight;
           double zjk = ZtildeM[(j*Nk)+ik];
           potik = potik + wij * zjk;
            }
        numik = FikM[(ipt*Nk) + ik] * exp(Beta * potik + ParaP->Singl[ik]);
        sumi = sumi+numik;
        posterior[(ipt*Nk) + ik] = numik;
	  }

        }  /* enf of for ( ik = 0, ... */

	/*  for ( ik = 0; ik<Nk; ik++)
        {
         PostM[(ipt*Nk) +ik] = posterior[ik] / sumi;
	 }*/

    if (sumi >0)
     {
      if(sumi>EPSILON)
      {
        for ( ik = 0; ik<Nk; ik++)
        {
         PostM[(ipt*Nk) +ik] = posterior[(ipt*Nk)+ik] / sumi;
        }
      }
      else
      {
        for ( ik = 0; ik<Nk; ik++)
        {
	  inv = 1/(sumi/EPSILON);
          PostM[(ipt*Nk) +ik] = inv*posterior[(ipt*Nk)+ik] / EPSILON;
        }
      }
     }
     else
     {
       
       for ( ik = 0; ik<Nk; ik++)
        {
          PostM[(ipt*Nk) +ik] = 1/Nk;
        }
     }


    /* } end of if  ( ( AlgoParaP->Label ...

      else
      {
	  LabelToClassVector( Nk, DataP->LabelV[ ipt ] - 1,
			      &(PostM[ ipt * Nk ]) ) ;
      } */

     }   /*end of for ( ivis = 0 ; ... */

   GenFree( posterior);

   return STS_OK;

  }  /* end of ComputePostPartition */


/* ------------------------------------------------------------------------ */
static int ComputeGibbsPartition (
  const ModelSpecT*  SpecP,
  const ModelParaT*  ParaP,                 /* To catch the parameters */
  const DataT*       DataP,                 /* Current values of the pixels */
  const SpatialT*    SpatialP,              /* Spatial information */
  PtNeighsT*         NeighsP,               /* also for spatial information */
  int                ipt,                   /* Current point of computation */
  double*            PostM                  /* Posterior probabilities of all
					       the points belongs to each
					       class */
)
/* ------------------------------------------------------------------------ */
{
  /* Local Variables declaration */
  /* Parameters */
  StatusET    err;                    /* Function return statement */
  int         nK      = SpecP->K;     /* Number of classes */
  int         nbPts   = DataP->NbPts; /* Number of points */
  int         ik;                     /* Current class */
  int         nbn;                    /* Nb of neighbours of ipt */
  int         n;                      /* current neighbour n=0..nbn */
  double      beta    = ParaP->Beta;
  /* For temporary computation */
  double      numik;                  /* exp(-beta sum i's neighbours pots. */
  double      sumi;                   /* sum of numik, k = O..nK */
  double      inv;
  double      potik;                  /* i's neibourghs potentials */
  double*     posterior;              /* before putting the real result
					 into PostM */
  GetNeighFT* FGetNeigh;              /* function to get neighbors */

  /* Initialisations */
  if ((posterior = GenAlloc( nbPts * nK, sizeof(double), 0,
			     "ComputePostPartition", "posterior"))
      == NULL)
    fprintf(stderr,"Impossible to allocate posterior in function ComputePostPartition\n");
  if ((err = GetSpatialFunc( SpatialP->Type,
			     & FGetNeigh )) != STS_OK)
    return err;
  nbn= FGetNeigh(ipt, &(SpatialP->NeighData),NeighsP);
  /* Main loop */
  for (ik = 0, sumi = 0; ik<nK ; ik++)
    {
      for (n = 0, potik = 0; n<nbn ; n++)
	{
	  int    j   = NeighsP->NeighsV[n].Index;
	  double wij = NeighsP->NeighsV[n].Weight;
	  double zjk;
	  /* To be checked */
	  if (DataP->PointsM[j] == ik)
	    zjk = 1;
	  else
	    zjk = 0;
	  potik+= wij * zjk;
	}
      numik = exp(beta * potik + ParaP->Singl[ik]);
      sumi += numik;
      posterior[(ipt*nK) + ik] = numik;
    }  /* enf of for ( ik = 0, ... */

  /* Compute PostM */
  if (sumi > 0)
    {
      if(sumi > EPSILON)
	{
	  for (ik = 0; ik<nK; ik++)
	    PostM[(ipt*nK)+ik] = posterior[(ipt*nK)+ik] / sumi;
	}
      else
	{
	  for (ik = 0; ik<nK; ik++)
	    {
	      inv = 1/(sumi/EPSILON);
	      PostM[(ipt*nK)+ik] = inv*posterior[(ipt*nK)+ik] / EPSILON;
	    }
	}
    }
  else
    {
      for ( ik = 0; ik<nK; ik++)
	PostM[(ipt*nK)+ik] = 1/nK;
    }

  GenFree( posterior);
  return STS_OK;
}

/* ------------------------------------------------------------------- */
static int ComputeLocalPostPartition (
       const ModelSpecT*    SpecP,
       const ModelParaT*    ParaP,
       const DataT*         DataP,
       const SpatialT*      SpatialP,
       const int            ipt,
       const int            Nk,
       PtNeighsT*           NeighsP,
       const double*        ZtildeM,
       const double         Beta,
       const double*        Singl,
       const double*        FikM,
       double*              PostM
       )
/* compute the approximation of the local proba P_G(Zi|yi),
   setting the neighborhood of i to ZtildeM_N(i) */
/* ------------------------------------------------------------------- */
{
   StatusET      err;
   double        numik;     /* fk(yi)*
			       exp(-beta sum i's neighbours potentials) */
   double        sumi;      /* sum of numik, k=O..Nk */
   double        inv;
   double        potik;     /* i's neighbours potentials */
   int           ik;
   int           nbn;       /* nb of neighbours of ipt */
   int           n;         /* current neighbour*/
   double*       posterior;
   GetNeighFT*   FGetNeigh; /* function to get neighbours */

   if ((err = GetSpatialFunc( SpatialP->Type,
			      & FGetNeigh)) != STS_OK)
     return err ;

   if ((posterior = GenAlloc( Nk, sizeof(double), 0,
			      "ComputeLocalPostPartition", "posterior"))
       == NULL)
     fprintf(stderr,"impossible to allocate posterior in function ComputeLocalPostPartition\n");

   nbn = FGetNeigh(ipt, &(SpatialP->NeighData),NeighsP);

   for ( ik = 0, sumi = 0; ik<Nk; ik++)
     {
       if ((SpecP->ClassFamily ==FAMILY_NORMALDIAG)
	   &&(ParaP->Disp_KD[ik] == 0 ))
	 {
	   int diff = 0;
	   diff = diff + ParaP->Center_KD[ik]
	     - DataP->PointsM[ipt*DataP->NbVars+1];
	   if(diff == 0)
	     {
	       posterior[ ik] = 1;
	       sumi = 1;
	     }
	   else
	     {
	       posterior[ ik] = 0 ;
	       sumi = 1;
	     }
	 } else {
	   for (n = 0, potik = 0; n<nbn; n++)
	     {
	       int   j    = NeighsP->NeighsV[n].Index;
	       double wij = NeighsP->NeighsV[n].Weight;
	       double zjk = ZtildeM[(j*Nk)+ik];
	       potik = potik + wij * zjk;
	     }
	   numik = FikM[(ipt*Nk) +ik] * exp(Beta * potik + Singl[ik]);
	   sumi = sumi+numik;
	   posterior[ik] = numik;
	 }
     }  /* enf of for ( ik = 0, ... */

       /*   for ( ik = 0; ik<Nk; ik++)
        {
         PostM[(ipt*Nk) +ik] = posterior[ik] / sumi;
        }
       */

   if (sumi >0)
     {
       if(sumi>EPSILON)
	 {
	   for ( ik = 0; ik<Nk; ik++)
	     {
	       PostM[(ipt*Nk) +ik] = posterior[ik] / sumi;
	     }
	 }
       else
	 {
	   for ( ik = 0; ik<Nk; ik++)
	     {
	       inv = 1/(sumi/EPSILON);
	       PostM[(ipt*Nk) +ik] = inv*posterior[ik] / EPSILON;
	     }
      }
     }
   else
     {
       fprintf(stderr,"pt %d density=0 sumi =%f\n",ipt,sumi);
       for ( ik = 0; ik<Nk; ik++)
	 {
	   PostM[(ipt*Nk) +ik] = 1/Nk;
	 }
     }

   GenFree( posterior);

   return STS_OK;

  }  /* end of ComputeLocalPostPartition */


  /* ------------------------------------------------------------------- */
   static int ComputePriorPartition
             (
              const ModelSpecT*    SpecP,       /* I */
              const ModelParaT*    ParaP,
              const DataT*         DataP,       /* I */
              const SpatialT*      SpatialP,
  	      const AlgoParaT*     AlgoParaP,    /* I */
              const double*         ZtildeM,      /* z~ik matrix */
              WorkingT*            WorkP,
              double*               PriorM
             )
     /*compute an approximation of the a priori gibbs distribution,
       setting the neighbors to ZtildeM */
  /* ------------------------------------------------------------------- */
  {
    /* const char*     func = " ComputePriorPartition"; */

    StatusET err;
    int          Nk      = SpecP->K;
    int          Npt     = DataP->NbPts;
    int          ik;     /* current class */
    int          ipt;    /* current point */
    int          nbn;    /* nb of neighbours of ipt */
    int          n;      /* current neighbour n=0..nbn */
    double        Beta    = ParaP->Beta;
    double        numik;  /* exp(-beta sum i's neighbours potentials */
    double        sumi;   /* sum of numik, k=O..Nk */
    double        potik;  /* i's neibourghs potentials */
    double*       prior;
    PtNeighsT*   NeighsP = &(WorkP->Neighs);
    GetNeighFT*  FGetNeigh;   /* function to get neighbors */



    if ((prior = GenAlloc( Npt * Nk, sizeof(double), 0, "ComputePriorPartition", "prior"))==NULL)
         fprintf(stderr,"impossible to allocate prior in function ComputePriorPartition\n");
    if ( ( err = GetSpatialFunc( SpatialP->Type,
                                   & FGetNeigh ) ) != STS_OK )
          return err ;


    for ( ipt = 0 ; ipt < Npt ; ipt ++)
      {
       /* If label is not known
       if ( ( AlgoParaP->Label   != 1 ) ||
		 ( DataP->LabelV[ ipt ] == 0 ) )
		 { */

        nbn= FGetNeigh(ipt, &(SpatialP->NeighData),NeighsP );

        for ( ik = 0, sumi = 0; ik<Nk; ik++)
        {

          for ( n = 0, potik = 0; n<nbn; n++)
           {
           int   j = NeighsP->NeighsV[n].Index;
           double wij = NeighsP->NeighsV[n].Weight;
           double zjk = ZtildeM[(j*Nk)+ik];
           potik = potik + wij * zjk;
            }
        numik = exp(Beta * potik + ParaP->Singl[ik]);
        sumi = sumi+numik;
        prior[(ipt*Nk) + ik] = numik;

         }  /* enf of for ( ik = 0, ... */


       for ( ik = 0; ik<Nk; ik++)
       {
         PriorM[(ipt*Nk) + ik] = prior[(ipt*Nk) +ik] / sumi;
        }


    /* }
       else
       {
	  LabelToClassVector( Nk, DataP->LabelV[ ipt ] - 1,
			      &(PriorM[ ipt * Nk ]) ) ;
			      }*/

      } /* end of for ( ipt = 0 ; ... */

   GenFree( prior);

   return STS_OK;

  }  /* end of ComputePriorPartition */


 /* ------------------------------------------------------------------- */
 static double ComputeFunc
       (
        const ModelSpecT*    SpecP,       /* I */
        const ModelParaT*    ParaP,
        const DataT*         DataP,       /* I */
        const SpatialT*      SpatialP,
  	const AlgoParaT*     AlgoParaP,    /* I */
        WorkingT*            WorkP,
        const double*         PostM,  /* P~ in the field algorithms */
        const double*         Ztilde

        )
     /* Compute F'(PostM,Center,Disp,Beta,Ztilde)  */
/* ------------------------------------------------------------------- */
{
  StatusET  err;
  int          Nk      = SpecP->K;
  int          Npt     = DataP->NbPts;
  int       ipt ;           /* current site */
  int       ik  ;           /* current class */
  double     func = 0 ;      /* the F' function on which are based the fields algos */
  double*    FikM ;
  double*    LogFikM ;
  double*    priorM;

  if ((FikM = GenAlloc( Npt * Nk, sizeof(double), 0, "ComputeFunc", "FikM"))==NULL)
         fprintf(stderr,"impossible to allocate FikM in function ComputeFunc\n");

  if ((LogFikM = GenAlloc( Npt * Nk, sizeof(double), 0, "ComputeFunc", "LogFikM"))==NULL)
         fprintf(stderr,"impossible to allocate LogFikM in function ComputeFunc\n");

   err = ComputeDensity( ParaP, DataP, SpecP, FikM, LogFikM );

  if (( priorM = GenAlloc( Npt * Nk, sizeof(double), 0, "ComputeFunc", "priorM"))==NULL)
         fprintf(stderr,"impossible to allocate priorM in function ComputeFunc\n");

   err = ComputePriorPartition( SpecP, ParaP, DataP, SpatialP, AlgoParaP,
                                 Ztilde , WorkP, priorM);

  for ( ipt = 0; ipt <Npt; ipt ++) {
      for (ik = 0; ik < Nk; ik++)
	{

	  double dens = LogFikM[ipt*Nk + ik];
	  double logprior = log(priorM[ipt*Nk + ik]);
	  double logpost = log(PostM[ipt*Nk + ik]);
	  if (PostM[ipt*Nk + ik] == 0){ logpost = 1;}
	  func = func + PostM[ipt*Nk + ik] * ( dens + logprior
					       - logpost );

	}
			          }

  GenFree(priorM);
  GenFree(FikM);
  GenFree(LogFikM);
  return(func);

      } /* end of ComputeFunc */

 /* ------------------------------------------------------------------- */
static double ComputeLogWMarg
    (
             const int           Npt,          /* nb of points */
             const int           Nk,           /* nb of classes */
             const double         b,            /* point where compute the function */
             const double*        Ztilde,
             const double*        Singl,
             const SpatialT*     SpatialP,     /* neiborhood system */
             PtNeighsT*          NeighsP
    /* compute the  mean field like approximation of the marg. normalization constant W  */
          )
  /* ------------------------------------------------------------------- */
{
 int     ipt ;            /* current site  : 0..Npt-1 */
    int     k = 0;              /* current class : 0..Nk-1 */
    double  con_ik = 0;          /* sum_{j\neighbor{i}} v_{ij} c_{jk} */
    double  sumk_exp = 0 ;       /* sum_k exp_ik */
    double  LogW = 0;           /* first deriv of the criterium to maximize */

    GetNeighFT*  fGetNeigh ; /* function fetching neighbors of a site */

   /* Get function specific to chosen spatial model */
    if ( GetSpatialFunc( SpatialP->Type,
  			       & fGetNeigh ) != STS_OK )
      fprintf(stderr," Pb to find spatial model");


  for ( ipt = 0, LogW = 0 ;
  	  ipt < Npt ; ipt ++ ) {

        /* Get site's neighbours */
        int nbn = fGetNeigh( ipt, &SpatialP->NeighData, NeighsP ) ;

        for ( k = 0,
  	      sumk_exp      = 0.0;
  	    k < Nk ; k ++ )  {



  	con_ik = SumNeighsOfClass( k, nbn, Nk, NeighsP->NeighsV, Ztilde);

        sumk_exp  += exp( (b) * con_ik + Singl[k]) ;
                          }

        LogW += log(sumk_exp) ;
        /* exp(LogW)=W  might get inf because of high sumk_exp*/
         }

 return(LogW);


}  /* end of ComputeLogWMarg */

 /* ------------------------------------------------------------------- */
static double ComputeLogWCond
    (
             const int           Npt,          /* nb of points */
             const int           Nk,           /* nb of classes */
             const double        b,            /* point where compute the function */
             const double*       Ztilde,
             const double*       Singl,
             const double*       FikM,
             const SpatialT*     SpatialP,     /* neiborhood system */
             PtNeighsT*          NeighsP
    /* compute the  mean field like approximation of the cond. normalization constant W  */
          )
  /* ------------------------------------------------------------------- */
{
 int     ipt ;            /* current site  : 0..Npt-1 */
    int     k = 0;              /* current class : 0..Nk-1 */
    double  con_ik = 0;          /* sum_{j\neighbor{i}} v_{ij} c_{jk} */
    double  sumk_exp = 0 ;       /* sum_k exp_ik */
    double  LogW = 0;           /* first deriv of the criterium to maximize */

    GetNeighFT*  fGetNeigh ; /* function fetching neighbors of a site */

   /* Get function specific to chosen spatial model */
    if ( GetSpatialFunc( SpatialP->Type,
  			       & fGetNeigh ) != STS_OK )
      fprintf(stderr," Pb to find spatial model");


  for ( ipt = 0, LogW = 0 ;
  	  ipt < Npt ; ipt ++ ) {

        /* Get site's neighbours */
        int nbn = fGetNeigh( ipt, &SpatialP->NeighData, NeighsP ) ;

        for ( k = 0,
  	      sumk_exp      = 0.0;
  	    k < Nk ; k ++ )  {



  	con_ik = SumNeighsOfClass( k, nbn, Nk, NeighsP->NeighsV, Ztilde);

        sumk_exp  += FikM[ipt*Nk+k ]*exp( (b) * con_ik + Singl[k]) ;
                          }

        LogW += log(sumk_exp) ;
        /* exp(LogW)=W  might get inf because of high sumk_exp*/
         }

 return(LogW);


}  /* end of ComputeLogWCond */

/*----------------------------------------------------------------------*/
static double ComputeGradBeta
(
 const int           Npt,          /* nb of points */
 const int           Nk,           /* nb of classes */
 const double*       Beta,
 const double*       Ztilde,
 const double*       Singl,
 const double*       PostM,
 const SpatialT*     SpatialP,     /* neiborhood system */
 PtNeighsT*          NeighsP,
 ParameterT          ParameterP    /* nearly useless here but...*/
 )
/*---------------------------------------------------------------------*/
{
  int     ipt ;                /* current site  : 0..Npt-1 */
  int     k = 0;               /* current class : 0..Nk-1 */
  double  con_ik = 0;          /* sum_{j\neighbor{i}} v_{ij} c_{jk} */
  double  exp_ik = 0 ;         /* exp{ beta * con_ik + Vk} */
  double  sumk_exp = 0 ;       /* sum_k exp_ik */
  double  sumk_con_exp = 0 ;   /* sum_k con_ik exp_ik */
  double  sumk_cik_con = 0 ;   /* sum_k c_ik con_ik */
  double  grad = 0;            /* first deriv of the criterium to maximize */
  GetNeighFT*  fGetNeigh ;     /* function fetching neighbors of a site */

  /* Get function specific to chosen spatial model */
  if ( GetSpatialFunc( SpatialP->Type, & fGetNeigh ) != STS_OK )
    fprintf(stderr," Pb to find spatial model");

  for ( ipt = 0, grad = 0.0; ipt < Npt ; ipt ++ )
    {
      /* Get site's neighbours */
      int nbn = fGetNeigh( ipt, &SpatialP->NeighData, NeighsP ) ;
      for ( k = 0,
  	      sumk_exp      = 0.0,
  	      sumk_con_exp  = 0.0,
  	      sumk_cik_con  = 0.0;
  	    k < Nk ; k ++ )
	{
	  con_ik = SumNeighsOfClass( k, nbn, Nk, NeighsP->NeighsV, Ztilde);
	  exp_ik = exp((*Beta) * con_ik + Singl[k]);
	  sumk_exp        += exp_ik ;
	  sumk_con_exp    += con_ik * exp_ik ;
	  sumk_cik_con    += PostM[ ( ipt * Nk ) + k ] * con_ik ;
        }
      grad += sumk_cik_con - sumk_con_exp / sumk_exp ;
      /* might get NaN because of high sumk_exp*/
    }
  return (grad);
} /* end of ComputeGradBeta */

/*---------------------------------------------------------------------*/
static double ComputeGradPotential
(
 const int           Npt,          /* nb of points */
 const int           Nk,           /* nb of classes */
 const double*       Beta,
 const double*       Ztilde,
 const double*       Singl,
 const double*       PostM,
 const SpatialT*     SpatialP,     /* neiborhood system */
 PtNeighsT*          NeighsP,
 ParameterT          ParameterP    /* To know which is the number of the
				      potential that will serve as
				      derivation variable */
 )
/*---------------------------------------------------------------------*/
{
  int     ipt ;                /* current site  : 0..Npt-1 */
  int     l        = 0;        /* for computing the denominator */
  double  con_il   = 0;        /* sum_{j\neighbor{i}} v_{ij} c_{jl} */
  double  exp_il   = 0;        /* exp{ beta * con_il + Vl} */
  double  exp_ik   = 0;        /* exp{ beta * con_ik + Vk} */
  double  suml_exp = 0;        /* sum_l exp_il */
  double  grad     = 0;        /* first deriv of the criterium to maximize */
  GetNeighFT*  fGetNeigh ;     /* function fetching neighbors of a site */

  /* Get function specific to chosen spatial model */
  if ( GetSpatialFunc( SpatialP->Type, & fGetNeigh ) != STS_OK )
    fprintf(stderr," Pb to find spatial model");

  for (ipt = 0, grad = 0.0; ipt < Npt ; ipt ++)
    {
      /* Get site's neighbours */
      int nbn = fGetNeigh(ipt, &SpatialP->NeighData, NeighsP);
      for (l = 0,
	     suml_exp      = 0.0;
	    l < Nk; l++)
	{
	  con_il = SumNeighsOfClass(l, nbn, Nk, NeighsP->NeighsV, Ztilde);
	  exp_il = exp((*Beta) * con_il + Singl[l]);
	  if (l == ParameterP.No) exp_ik = exp_il;
	  suml_exp += exp_il;
        }
      grad += PostM[(ipt * Nk ) + ParameterP.No] - exp_ik / suml_exp;
      /* might get NaN because of high suml_exp*/
    }
  return grad;
} /* end of ComputeGradPotential */



  /* ------------------------------------------------------------------- */
   static int EstimBetaForFieldAlgo
             (
  	       const BetaET        BetaModel,    /* beta estimation mode */
               const BtaPsGradT    BtaPsGradP,   /* gradient ascent  parameters */
               const SpatialT*     SpatialP,     /* neiborhood system */
               const double*       Ztilde,       /* current z~ */
               const double*       Post,
               const double*       Singl,        /* singleton potentiels */
               const int           Npt,          /* nb of points */
               const int           Nk,           /* nb of classes */
               double*              Beta,        /* previous and next Beta */
               PtNeighsT*          NeighsP      /* to store a site's neighbors */
              )
  /* beta estimation using gradient descent: to avoid */
  /* ------------------------------------------------------------------- */
  {
    int    iter ;           /* current iteration : 0..NbIter-1 */
    int    conv ;           /* TRUE if convergence reached */
    int    ipt ;            /* current site  : 0..Npt-1 */
    int    k ;              /* current class : 0..Nk-1 */
    int    n;               /* current neighbour n=0..nbn */
    double  con_ik;          /* sum_{j\neighbor{i}} v_{ij} c_{jk} */
    double  exp_ik ;         /* exp{ beta * con_ik + Vk } */
    double  sumk_exp ;       /* sum_k exp_ik */
    double  sumk_con_exp ;   /* sum_k con_ik exp_ik */
    double  sumk_con2_exp ;  /* sum_k (con_ik)^2 exp_ik */
    double  sumk_cik_con ;   /* sum_k c_ik con_ik */
    double  sumk_cik_Hik  ;   /* sum_k c_ik * (beta * con_ik + Vk) */
    double  crit ;           /* log pseudo-likelihood */
    double  grad ;           /* first deriv of log pseudo-likelihood */
    double  dsec ;           /* - second deriv of log pseudo-likelihood */

    GetNeighFT*  fGetNeigh ; /* function fetching neighbors of a site */

    if ( ( BetaModel != BETA_PSGRAD  ) || ( SpatialP->Type == TYPE_NONSPATIAL) )
        return 0;

    /* Get function specific to chosen spatial model */
    if ( GetSpatialFunc( SpatialP->Type,
  			       & fGetNeigh ) != STS_OK )
      return 2 ;



    /* For each iteration of gradient ascent */
    for ( iter = 0,
  	  conv = FALSE ;
  	( iter < 100 )  &&
		  ( ! conv )  ;
  	iter ++, conv = (fabs( grad ) < ( BtaPsGradP.ConvThres )) )  {

      /* Compute crit, grad and dsec by summation over all sites */
      for ( ipt = 0, crit = 0.0, grad = 0.0, dsec = 0.0 ;
  	  ipt < Npt ; ipt ++ ) {

        /* Get site's neighbours */
        int nbn = fGetNeigh( ipt, &SpatialP->NeighData, NeighsP ) ;

        for ( k = 0,
  	      sumk_exp      = 0.0,
  	      sumk_con_exp  = 0.0,
  	      sumk_con2_exp = 0.0,
  	      sumk_cik_con  = 0.0 ,
              sumk_cik_Hik   = 0.0  ;
  	    k < Nk ; k ++ )  {

  	    for ( n = 0, con_ik = 0.0; n<nbn; n++)
  	{
  	     int   j   = NeighsP->NeighsV[n].Index;
               double wij = NeighsP->NeighsV[n].Weight;
  	     con_ik    = con_ik + wij * Ztilde[(j*Nk) + k];
          }

  	con_ik = SumNeighsOfClass( k, nbn, Nk, NeighsP->NeighsV, Ztilde);
  	exp_ik = exp( (*Beta) * con_ik + Singl[k]) ;

  	sumk_exp        += exp_ik ;
  	sumk_con_exp    += con_ik * exp_ik ;
  	sumk_con2_exp   += con_ik * con_ik * exp_ik ;
  	sumk_cik_con    += Post[ ( ipt * Nk ) + k ] * con_ik ;
        sumk_cik_Hik    += Post[ ( ipt * Nk ) + k ]*((*Beta) * con_ik + Singl[k]);
        }

        crit +=  sumk_cik_Hik - log( sumk_exp ) ;
        grad += sumk_cik_con - sumk_con_exp / sumk_exp ;
        dsec += (double) ( (double)
  	( sumk_con2_exp * sumk_exp - sumk_con_exp * sumk_con_exp )
  	/ (double) ( sumk_exp * sumk_exp ) ) ;


        /* might get NaN because of high sumk_exp*/

      }
      if (iter == 0) {fprintf(stderr,"crit= %f\n  ",crit);}

        /*  Update beta */
      if ( BtaPsGradP.Step <= 0.0 ) {
	/* dsec = dsec * 4 ;   reduce the step to avoid overstepping */
        if ( dsec < ( Npt / 10 ) )
  	 {dsec = Npt / 10 ;
	  }
        (*Beta) += grad / dsec ;
      }
      else {
        (*Beta) += grad * ( BtaPsGradP.Step / Npt ) ;
      }

     if ( isnan(*Beta) ) {
       (*Beta) = 1 ;
     }

    } /* end  For each iteration of gradient ascent */
     fprintf(stderr, "iter %d\n",iter);

    if ( (*Beta) > MAX_BETA ) {
      (*Beta) = MAX_BETA ;
      return 1 ;
    }

    if ( (*Beta) < MIN_BETA ) {
      (*Beta) = MIN_BETA ;
      return 1 ;
    }

    if ( isnan(*Beta) ) {
      (*Beta) = 0 ;
      return 1 ;
    }


    return 0 ;
    /*???*/
  }   /* end of EstimBetaForFieldAlgo */



  /* ------------------------------------------------------------------- */
   static StatusET ComputeDensity
             (
               const ModelParaT*   ParaP,
               const DataT*        DataP,
               const ModelSpecT*   SpecP,
               double*              FikM,
               double*              LogFikM
             )

  /* ------------------------------------------------------------------- */
  {
   StatusET          err = STS_OK;
   int               result;
   int               ipt;
   int               ik;
   int               i,j,s;
   int               npt = DataP->NbPts;
   int               nk  = SpecP->K;
   int               Nd  = DataP->NbVars ;
   double* p=0;     /* for the cholesky decomposition */
   double* cholDISP=0; /* for the cholesky decomposition */
   double  logdet = 0;
   double  sum;
   CompuDensFT*      fcompuDens;

   if ( ( err = GetDensityFunc( SpecP ,
                                   &fcompuDens ) ) != STS_OK )     /*V1.06-c*/
      {
          return err ;
      }

  if ( (SpecP->ClassFamily == FAMILY_NORMALGEN) &&
     (SpecP->ClassDisper == DISPER__D) )
    /* same cov matrix for each class, so the cholesky decomp  is computed
     only once */
  {
    if ((p = GenAlloc( Nd, sizeof(double), 0, "ComputeDensity", "p"))==NULL)
         fprintf(stderr,"impossible to allocate p in function ComputeDensity\n");

    if ((cholDISP = GenAlloc(Nd*Nd, sizeof(double), 0, "ComputeDensity","cholDISP"))==NULL)
         fprintf(stderr,"impossible to allocate cholDISP in function ComputeDensity\n");
    for (i = 0; i<Nd; i++)
      {
	for (j=i; j<Nd; j++)
	{
	    cholDISP[i*Nd+j] = ParaP->Disp_KD[ i*Nd +j - i*(i+1)/2];
	}
      }

      /* compute the cholesky decomposition of the covariance matrix: L.L^t */
      /* the diagonal elements of L are in the vector p */
      /* the lij, j<i, are in the lower triangle of cholDISP */
      err = choldc(cholDISP,Nd,p);
      if (err == -1){
         fprintf(stderr,"cholesky failed\n");
         return err;
      }

      /* compute L^-1 and put the result in the lower triangle of cholDISP */
      for (i=0; i<Nd; i++)
      {
	cholDISP[i*Nd+i]  = 1.0/p[i];
	for (j=i+1; j<Nd; j++)
	{
	  sum =0;
          for (s=i;s<j;s++)
          sum-=cholDISP[j*Nd+s]*cholDISP[s*Nd+i];
          cholDISP[j*Nd+i] = sum/p[j];
	}
      }


     /* compute log(sqrt(det)) of the covariance matrix */
        for ( i=0; i<Nd; i++)
        {
          if ( p[i] > EPSILON )
           logdet += log(p[i]);

          else
           {
             err = -1 ;
             return err;
	   }
	}

  } /* end of case normalgen + disper__D */


  for ( ik = 0; ik<nk; ik++)
   {
     if ( (SpecP->ClassFamily == FAMILY_NORMALGEN) &&
        (SpecP->ClassDisper == DISPER_KD) )
     {
      if ((p = GenAlloc( Nd, sizeof(double), 0, "ComputeDensity", "p"))==NULL)
         fprintf(stderr,"impossible to allocate p in function ComputeDensity\n");

      if ((cholDISP = GenAlloc(Nd*Nd, sizeof(double), 0, "ComputeDensity","cholDISP"))==NULL)
         fprintf(stderr,"impossible to allocate cholDISP in function ComputeDensity\n");

      for (i = 0; i<Nd; i++)
      {
	for (j=i; j<Nd; j++)
	{
	    cholDISP[i*Nd+j] = ParaP->Disp_KD[ik*Nd*(Nd+1)/2 + i*Nd +j - i*(i+1)/2];
	}
      }

      /* compute the cholesky decomposition of the covariance matrix: L.L^t */
      /* the diagonal elements of L are in the vector p */
      /* the lij, j<i, are in the lower triangle of cholDISP */
      err = choldc(cholDISP,Nd,p);
      if (err == -1){
         fprintf(stderr,"cholesky failed\n");
         return err;
      }


      /* compute L^-1 and put the result in the lower triangle of cholDISP */
      for (i=0; i<Nd; i++)
      {
	cholDISP[i*Nd+i]  = 1.0/p[i];
	for (j=i+1; j<Nd; j++)
	{
	  sum =0;
          for (s=i;s<j;s++)
          sum-=cholDISP[j*Nd+s]*cholDISP[s*Nd+i];
          cholDISP[j*Nd+i] = sum/p[j];
	}
      }


       /* compute log(sqrt(det)) of the covariance matrix */
        for ( i=0; i<Nd; i++)
        {
          if ( p[i] > EPSILON )
           logdet += log(p[i]);

          else
           {
             err = -1 ;
             return err;
	   }
	}

     } /* end of case normalgen + disper_KD */


     for ( ipt = 0; ipt<npt; ipt++)
     {
         double    fik;
         double    logfik;
         result = fcompuDens(Nd, ik, ParaP, &(DataP->PointsM[ipt*DataP->NbVars]),
                       SpecP, &fik, &logfik,cholDISP,logdet);


	 FikM[ipt*nk +ik]    = fik;
	 LogFikM[ipt*nk +ik] = logfik;
     }
   }

    GenFree(cholDISP);
    GenFree(p);
    return err;

  }    /* end of ComputeDensity */


  /* ------------------------------------------------------------------- */
   static double IncreaseTest
     (
       const ModelSpecT*    SpecP,       /* I */
       const ModelParaT*    ParaP,
       const DataT*         DataP,       /* I */
       const SpatialT*      SpatialP,
       const AlgoParaT*     AlgoParaP,    /* I */
       WorkingT*            WorkP,      /* z~ and P~ at iteration q-1 */
       double*               newz       /* z~ at iteration q */

   )
  /* ------------------------------------------------------------------- */
   {
    StatusET            err;
    int                 Nk = SpecP->K;
    int                 Npt = DataP->NbPts;
    int                 ik;
    int                 in;
    double*              oldPriorM;
    double*              newPriorM;
    double               difprior;
    double               dif = 0;

    if ((oldPriorM = GenAlloc( Npt * Nk, sizeof(double), 0, "IncreaseTest", "oldPriorM"))==NULL)
         fprintf(stderr,"impossible to allocate oldPriorM in function IncreaseTest\n");
    if ((newPriorM = GenAlloc( Npt * Nk, sizeof(double), 0, "IncreaseTest", "newPriorM"))==NULL)
         fprintf(stderr,"impossible to allocate newPriorM in function IncreaseTest\n");

    err = ComputePriorPartition( SpecP, ParaP, DataP, SpatialP, AlgoParaP,
                                  WorkP->Ztilde, WorkP, oldPriorM);
    err = ComputePriorPartition( SpecP, ParaP, DataP, SpatialP, AlgoParaP,
                                  newz, WorkP, newPriorM);

    for ( in = 0; in <Npt; in ++) {
      for (ik = 0; ik < Nk; ik++) {
        difprior = log( newPriorM[in *Nk + ik]) -log(oldPriorM[in *Nk + ik]);
        dif = dif + WorkP->ColdM[in *Nk + ik] * difprior;
      }
    }

    GenFree(newPriorM);
    GenFree(oldPriorM);
    /*  if ( difprior >= 0) */
    /*  { return 1;} */

    /*  else{ return 0;} */

     return(dif);

   }    /* end of IncreaseTest */


  /* ------------------------------------------------------------------- */
   StatusET ComputeBIC
   (
       const ModelSpecT*    SpecP,       /* I */
       const ModelParaT*    ParaP,
       const DataT*         DataP,       /* I */
       const SpatialT*      SpatialP,
       const AlgoParaT*     AlgoParaP,
       const double*        Hclassif,
       const double*        ICMclassif,
       const CriterT*       CriterP,
       MSCritT*             MSCritP,
       WorkingT*            WorkP
  )
     /* for spatial model,
       HP(z) = NRJ of the Potts model = sum[i] 0.5*beta*zi*sum[j~i]zj
       HPG(z) = NRJ of the Generalized Potts model = sum[i] (0.5*beta*zi*sum[j~i]zj + Vi(zi))
       zhat = final segmentation

       0.5*BICPmf = sum[i] log sum[k] fik*P(zi=ek| ztildeN(i)) -0.5*d*logNpt
       0.5*BICWinf = log(WinfCond) -log(WinfMarg) -0.5*d*logNpt

       ICL with mean field like approx using Ztilde
       0.5*ICLHmfWmf = sum[i] sum[k] zhati*log[fik*PG(zi=ek| ztildeN(i))] -0.5*d*logNpt
       0.5*ICLHWinf = sum[i] sum[k] zhati*log(fik) + HPG(zhat) - E_ztilde[HP- HP_ztilde] -
                  sum[i] log[sum[k] exp(beta*ek*sum[j~i]ztildej+ Vi(ek))] -0.5*d*logNpt
       0.5*ICLHWmf = sum[i] log(fik) + HPG(zhat)-
                  sum[i] log[sum[k] exp(beta*ek*sum[j~i]ztildej+ Vi(ek))] -0.5*d*logNpt

       ICL with mean field approximation of the marginal Gibbs distrib (->zmf)
       0.5*ICLHmfWmf = sum[i] sum[k] zhati*log[fik*PG(zi=ek| zmfN(i))] -0.5*d*logNpt
       0.5*ICLHWinf = sum[i] sum[k] zhati*log(fik) + HPG(zhat) + HP(zmf)-
                  sum[i] log[sum[k] exp(beta*ek*sum[j~i]zmf+ Vi(ek))] -0.5*d*logNpt
       0.5*ICLHWmf = sum[i] log(fik) + HPG(zhat)-
                  sum[i] log[sum[k] exp(beta*ek*sum[j~i]zmfj+ Vi(ek))] -0.5*d*logNpt

       0.5*EICL = BICPmf + sum[i] sum[k] P(zi=ek| ztildeN(i), yi) *log[P(zi=ek| ztildeN(i), yi)]
                = E[ICLHmfWmf | y]

      rmq: ICM + BICPmf =ICM + BIC2 = Dereck PLIC */
  /* ------------------------------------------------------------------- */
  {
    StatusET err;
    int          Nk        = SpecP->K;
    int          Npt       = DataP->NbPts;
    int          Nd        = DataP->NbVars;
    int          ik;       /* current class */
    int          in;       /* current point */
    int          d = 0;        /* number of parameters in the model */
    double*      zeroSingl; /* nul singleton potentiel for Potts model*/
    double       LogWmfCond = 0; /* MF  approx of Wcond for HPG, neighbors set to the cond mean field */
    double       LogWmfMarg = 0; /* MF  approx of Wmarg for HPG, neighbors set to the marg mean field */
    double       HPGmfMarg = 0; /* MF  approx of HPGmarg(z), z set to the final segmentation
                                                 and neighbors set the marg mean field */
    double       LogWmflMarg = 0; /* MF like approx of Wmarg for HPG, with neighbors set to Ztilde */
    double       HPGmflMarg = 0; /* MF  approx of HPGmarg(z), z set to the final segmentation
                                                 and neighbors set to Ztilde */
    double       HPG = 0; /*HPG of the final segmentation */
    double       LogVCond = 0;
    double       HPzmfMarg = 0;  /*  HP(z), z  set to the marg mean field */
    double       HPzmfCond = 0;  /*  HP(z), z  set to the cond mean field */
    double       C1,C2;   /* C2 - C1  =  - E_ztilde[HP- HP_ztilde] */
    double       HPzmflMarg = 0;  /* HP(z), z  set to Ztilde */
    double        sumikmf; /* sum[k]f(yi | ek)P(zi = ek | ztilde_N(i) )  */
    double        sumikmode;
    double        sumikEICL     = 0;
    /*double        sumiICL       = 0;*/
    double        loglikemf     = 0;
    double        loglikemode   = 0;
    double        dik           = 0; /* for the computation in the case of indep. mixture model */
    double*       priorMmf;   /* Zhang's prior */
    double*       priorMmode;
    double*       MargField; /* marginal mean field */
    double*       priormean;
    double*       CondField; /* conditional mean field */
    double*       tempo;
    double        p;
    int          label;
    int*         kmaxesV ;  /* classes having same maximum probabilites */

   if ( ( kmaxesV = GenAlloc( Nk, sizeof( int ),
				   0, "ComputeBIC", "kmaxesV" ) ) == NULL )
            return STS_E_MEMORY ;


   if ( (zeroSingl    = GenAlloc( Nk, sizeof(double),
                                         0, "ComputeBIC", "zeroSingl")) == NULL )
             return STS_E_MEMORY;

   if ( (priormean    = GenAlloc( Nk, sizeof(double),
                                         0, "ComputeBIC", "priormean")) == NULL )
             return STS_E_MEMORY;

   if ( (priorMmf   = GenAlloc( Npt * Nk, sizeof(double), 0, "ComputeBIC", "priorMmf"))==NULL)
         fprintf(stderr,"impossible to allocate priorMmf  in function ComputeBIC\n");
   if ( (priorMmode = GenAlloc( Npt * Nk, sizeof(double), 0, "ComputeBIC", "priorMmode"))==NULL)
         fprintf(stderr,"impossible to allocate priorMmode in function ComputeBIC\n");
   if ( (MargField      = GenAlloc( Npt * Nk, sizeof(double), 0, "ComputeBIC", "MargField"))==NULL)
         fprintf(stderr,"impossible to allocate tempo in function ComputeBIC\n");
   if ( (CondField      = GenAlloc( Npt * Nk, sizeof(double), 0, "ComputeBIC", "CondField"))==NULL)
         fprintf(stderr,"impossible to allocate tempo in function ComputeBIC\n");
   if ( (tempo      = GenAlloc( Npt * Nk , sizeof(double), 0, "ComputeBIC", "tempo"))==NULL)
         fprintf(stderr,"impossible to allocate tempo in function ComputeBIC\n");


   memcpy(tempo, Hclassif, Npt * Nk * sizeof( double ) ) ;

   err = ComputePriorPartition( SpecP, ParaP, DataP, SpatialP, AlgoParaP,
                                 ParaP->Ztilde , WorkP, priorMmf);
   err = ComputePriorPartition( SpecP, ParaP, DataP, SpatialP, AlgoParaP,
                                 tempo, WorkP, priorMmode);
   err = ComputeDensity( ParaP, DataP, SpecP, WorkP->FikM, WorkP->LogFikM );

   memcpy(MargField, ParaP->Ztilde, Npt * Nk * sizeof( double ) ) ;
   /* initialize the fpe */
   err = ComputePottsSinglMargMeanField(SpecP, ParaP, DataP, SpatialP, AlgoParaP, 
     MargField, WorkP, NULL);





   memcpy(CondField, ParaP->Ztilde, Npt * Nk * sizeof( double ) ) ;
   if (AlgoParaP->Algo != ALGO_MEANF)
     {
       err = ComputePartitionNEM ( SpecP, ParaP, DataP, SpatialP, AlgoParaP, 
                                        CondField, WorkP, NULL);
     }


  switch (AlgoParaP->Algo)
    {
     case ALGO_GSF:

      if (AlgoParaP->NbGibbs == 1)
	{
          ComputePostPartition(SpecP, ParaP, DataP, SpatialP, AlgoParaP,
                                     ParaP->Ztilde, WorkP, WorkP->ColdM );
	}

      else /* BIC, ICL and EICL not defined when MCMC like algos are used to
                           estimate the comditional distribution */
	{

          fprintf(stderr," Warning, Criteria not implemented for GSF and MCEM when nbGibbs >1\n");
            return 0;
	}
      break;

    case ALGO_NEM: /* Ztilde has no meaning in that case */

      break;

    case ALGO_MEANF:
         ComputePostPartition(SpecP, ParaP, DataP, SpatialP, AlgoParaP,
                                ParaP->Ztilde, WorkP, WorkP->ColdM );
     break;

    case ALGO_ICM:
        ComputePostPartition(SpecP, ParaP, DataP, SpatialP, AlgoParaP,
                                ParaP->Ztilde, WorkP, WorkP->ColdM );
      break;

    default:
          fprintf(stderr,"Criteria not implemented for HOMMEANF, MARGCONDMEANF,MARGMEANF,INV,PLMF, NCEM, GEM \n");
          return -1;

    }

   if ( (AlgoParaP->Algo == ALGO_MEANF )||
        (AlgoParaP->Algo == ALGO_ICM )||
        (AlgoParaP->Algo == ALGO_GSF ))
  {
   /* compute MF like approximation of the Potts NRJ of the configuration Hclassif: prod[i]Hi */
   err = ComputeNRJMarg(Npt, Nk, Hclassif,ParaP->Ztilde, ParaP->Singl,
                        ParaP->Beta, SpatialP, 1, WorkP, &HPGmflMarg);
   err = ComputeNRJMarg(Npt, Nk, Hclassif,MargField , ParaP->Singl,
                        ParaP->Beta,SpatialP, 1, WorkP, &HPGmfMarg);

   /* compute Potts NRJ of the configuration Hclassif: H */
   err = ComputeNRJMarg(Npt, Nk, Hclassif,Hclassif, ParaP->Singl,
                        ParaP->Beta, SpatialP, 0, WorkP, &HPG);

   /* compute the MF like approximation of the marginal normalization constant */
   LogWmflMarg = ComputeLogWMarg(Npt, Nk, ParaP->Beta , ParaP->Ztilde, ParaP->Singl,
                          SpatialP, &WorkP->Neighs );

   /* compute the MF  approximation of the conditional and marginal normalization constants */
   LogWmfCond = ComputeLogWCond(Npt, Nk, ParaP->Beta ,CondField,
                                ParaP->Singl,WorkP->FikM, SpatialP, &WorkP->Neighs );
   LogWmfMarg = ComputeLogWMarg(Npt, Nk, ParaP->Beta ,MargField ,
     ParaP->Singl, SpatialP, &WorkP->Neighs );

  /* compute logC when the lower bound of W is Wmf*C */
   err = ComputeNRJMarg(Npt, Nk, ParaP->Ztilde, ParaP->Ztilde, zeroSingl ,
        ParaP->Beta, SpatialP, 0, WorkP, &HPzmflMarg);
   err = ComputeNRJMarg(Npt, Nk,priorMmf,priorMmf,  zeroSingl ,
                          ParaP->Beta,SpatialP, 0, WorkP, &C1);
   err = ComputeNRJMarg(Npt, Nk,priorMmf, ParaP->Ztilde,  zeroSingl ,
                          ParaP->Beta,SpatialP, 1, WorkP, &C2);
   err = ComputeNRJMarg(Npt, Nk,MargField  ,MargField , zeroSingl ,
                          ParaP->Beta,SpatialP, 0, WorkP, &HPzmfMarg);
   err = ComputeNRJMarg(Npt, Nk,CondField  ,CondField , zeroSingl,
                          ParaP->Beta, SpatialP, 0,WorkP, &HPzmfCond);
  }

   for ( in = 0; in <Npt; in ++) {

      for (ik = 0, sumikmf = 0,sumikmode = 0; ik < Nk; ik++) {

        double logpkfki = WorkP->LogPkFkiM[ in * Nk + ik ] ;
        dik  = dik + Hclassif[in*Nk + ik] * logpkfki;
	if ( WorkP->FikM[in *Nk + ik] >0){
                 sumikmf   = sumikmf + WorkP->FikM[in *Nk + ik] * priorMmf[in *Nk + ik] ;
                 sumikmode = sumikmode + WorkP->FikM[in *Nk + ik] * priorMmode[in *Nk + ik] ;
	         }
        p = WorkP->ColdM[in*Nk + ik];
        if (p !=0) { sumikEICL =  sumikEICL + p*log(p);  }

      }


       if (AlgoParaP->Resto == RESTO_ICM)  label = ComputeMAP(ICMclassif, in, Nk, AlgoParaP->TieRule,
			       kmaxesV );
       else
          {
              if (AlgoParaP->Algo == ALGO_NEM)
                  label = ComputeMAP(Hclassif, in, Nk, AlgoParaP->TieRule,
			       kmaxesV ) ;
              else label = ComputeMAP(WorkP->ColdM, in, Nk, AlgoParaP->TieRule,
			       kmaxesV ) ;
	  }

       if (sumikmf == 0 ) return -1;
       if (sumikmode == 0 ) return -1;
       loglikemf = loglikemf + log(sumikmf);
       loglikemode = loglikemode + log(sumikmode);
       LogVCond += WorkP->LogFikM[in *Nk + label];

    }
  /* Remark: for mean mode and simul field, Hclassif = label, but not for ICM */

  /* compute the number of parameters of the model  */

   if (AlgoParaP->ParamMode ==PARAM_ESTIM)
   {
    if (SpecP->ClassFamily == FAMILY_NORMALDIAG)
    {
     switch( SpecP->ClassDisper ) {
     case DISPER___:
       d = Nd*Nk +1 ;
       break ;

     case DISPER_K_:
       d = Nd*Nk + Nk ;
       break ;

     case DISPER__D:
       d = Nd*Nk + Nd ;
       break ;

     case DISPER_KD:
       d = Nd*Nk + Nk*Nd ;
       break ;

     default: break;

     }
    }

    if (SpecP->ClassFamily == FAMILY_NORMALGEN)
    {
     switch( SpecP->ClassDisper ) {
     case DISPER__D:
       d = Nd*Nk + Nd*Nd ;
       break ;

     case DISPER_KD:
       d = Nd*Nk + Nd*Nd*Nk ;
       break ;

     default: break;

     }
    }
   }


   if (SpecP->BetaModel != BETA_FIX) {d = d+1;}
   if (AlgoParaP->SinglMode == SINGL_ZEROIN) {d = d+Nk;}

 if ((AlgoParaP->Algo == ALGO_NCEM) )
   {
        fprintf(stderr,"for NCEM and GEM, the computation of BIC, ICL and EICL is not implemented \n");
   }
 else
  {
   if (AlgoParaP->Algo == ALGO_NEM)
     {
      /* add the mixture proportions */
      if (SpecP->ClassPropor == PROPOR_K )
                 d = d+Nk-1;

      MSCritP->BICPmf =  2* CriterP->L  - d*log(Npt);
      MSCritP->ICLZtilde.ICLHmfWmf  =  2* dik - d*log(Npt); /* no MF like appro in that case */
      MSCritP->LogVCond = LogVCond;
      MSCritP->EICL =  2* CriterP->D + 2*CriterP->E - d*log(Npt) ;

     }

   else
    {
     MSCritP->BICPmf =  2* loglikemf - d*log(Npt);
     MSCritP->BICWinf = 2*LogWmfCond  -2*HPzmfCond  -2*LogWmfMarg +2*HPzmfMarg - d*log(Npt);
     MSCritP->BIC2 =  2* loglikemode - d*log(Npt);


     MSCritP->LogVCond = LogVCond;
     MSCritP->H = HPG;

     /* ICLHmfWmf computed with logP(z) ~= Hmf(z) -logWmf  and  MF like approx using Ztilde.*/
     MSCritP->ICLZtilde.Hmf = HPGmflMarg;
     MSCritP->ICLZtilde.LogWmf = LogWmflMarg;
     MSCritP->ICLZtilde.ICLHmfWmf = 2*(LogVCond +HPGmflMarg - LogWmflMarg ) - d*log(Npt);

     /* ICLHWinf computed with logP(z) ~= HPG(z) -logWmf + - E_ztilde[HP- HP_ztilde]
                                             and  MF like approx using Ztilde.. */
     MSCritP->ICLZtilde.C = C1 - C2;
     MSCritP->ICLZtilde.ICLHWinf = 2*(LogVCond +HPG + C2 - C1 - LogWmflMarg )
                                     - d*log(Npt);

     /* ICLHWmf computed with logP(z) ~= HPG(z) -logWmf and  MF like approx using Ztilde. */
     MSCritP->ICLZtilde.ICLHWmf = 2*(LogVCond +HPG  - LogWmflMarg ) - d*log(Npt);

     /* ICLHmfWmf computed with logP(z) ~= Hmf(z) -logWmf  and marg. MF  approx.*/
     MSCritP->ICLMarg.Hmf = HPGmfMarg;
     MSCritP->ICLMarg.LogWmf = LogWmfMarg;
     MSCritP->ICLMarg.ICLHmfWmf = 2*(LogVCond +HPGmfMarg - LogWmfMarg ) - d*log(Npt);

     /* ICLHWinf computed with logP(z) ~= HPG(z) -logWmf + HP(zmf) and marg. MF  approx. */
     MSCritP->ICLMarg.C =  - HPzmfMarg;
     MSCritP->ICLMarg.ICLHWinf = 2*(LogVCond +HPG +HPzmfMarg - LogWmfMarg )
                                     - d*log(Npt);

     /* ICLHWmf computed with logP(z) ~= HPG(z) -logWmf and Marg. MF  approx. */
     MSCritP->ICLMarg.ICLHWmf = 2*(LogVCond +HPG  - LogWmfMarg ) - d*log(Npt);


     MSCritP->EICL =  MSCritP->BICPmf + 2*sumikEICL;


     }
  }

  for ( ik = 0 ; ik < Nk ; ik ++ )
	{

         if ((SpecP->ClassFamily ==FAMILY_NORMALDIAG  )
               &&(ParaP->Disp_KD[ik] == 0 )){
                  int diff = 0;
                  diff = diff + ParaP->Center_KD[ik ]
                                             - DataP->PointsM[Npt*DataP->NbVars+1];
                  if(diff == 0)
                      { fprintf(stderr, "there are unbounded fik   \n") ;}
	 }
        }


 GenFree(priorMmf);
 GenFree(priorMmode);
 GenFree(MargField);
 GenFree(priormean);
 GenFree(CondField);
 GenFree(tempo);
 GenFree(kmaxesV ) ;
 GenFree(zeroSingl ) ;


 return err;

} /* end of ComputeBIC */





/* ------------------------------------------------------------------- */
 StatusET HardClassif
  (
   const AlgoParaT*    AlgoParaP,
   const DataT*        DataP,
   const int           Npt,
   const int           Nk,
   const double*       Fclassif, /* fuzzy classif, i.e probas */
   double*             Hclassif  /* hard classif provided by MAP on Fclassif */


   )

/* ------------------------------------------------------------------- */
{
 StatusET        err         = STS_OK ;
 int    ipt;
 int    ik;
 int*   kmaxesV;  /* classes having same maximum probabilites */

 /* Hclassif initialization: zero everywhere */
 for ( ipt = 0 ; ipt < Npt  ; ipt ++ ) {
   for(ik = 0; ik < Nk ; ik++){
          Hclassif[ipt*Nk + ik] = 0;
                              }
                                        }

 if ( ( kmaxesV = GenAlloc( Nk, sizeof( int ),
				   0, "HardClassif", "kmaxesV" ) ) == NULL )
            return STS_E_MEMORY ;

 /* For each record */
 for ( ipt = 0 ; ( ipt < Npt ) && ( err == STS_OK ) ; ipt ++ )
          {
            int   kmax = 0 ;

            /* If label is not known */
            if ( ( AlgoParaP->Label   != 1 ) ||
		 ( DataP->LabelV[ ipt ] == 0 ) )
            {
              /* Compute its MAP class */
              kmax = ComputeMAP( Fclassif, ipt, Nk, AlgoParaP->TieRule,
			       kmaxesV ) ;

              /* Save the record's class */
               Hclassif[ipt*Nk + kmax] = 1;
	    }
            else
	    { LabelToClassVector( Nk, (DataP->LabelV)[ ipt ] - 1,
				  &Hclassif[ ipt * Nk ] ) ; }

          } /* end for ( ipt ... ) */

        GenFree( kmaxesV ) ;

  return err;

} /* end of HardClassif */

/* ------------------------------------------------------------------- */
static int ComputeProps
  (
     const int*      Hclassif,
     const int         nk,
     const int         npt,
     double*            InitProps
  )
/* compute the proportion of each class in Hclassif */
/* ------------------------------------------------------------------- */
{
  int       ipt,ik;

  for ( ik = 0; ik<nk; ik++)
  {
      for (ipt = 0; ipt<npt;ipt++)
      {
	  if (Hclassif[ipt*nk + ik] == 1)
          {
            InitProps[ik] = InitProps[ik] +1;
          }
      }
      InitProps[ik] = InitProps[ik]/npt;
  }


 return STS_OK;


} /* end of ComputeProps */

/* ------------------------------------------------------------------- */
/* ------------------------------------------------------------------- */
static void ComputeErrorRate
(
 const double*      Hclassif,
 const double*      DataTrue,
       double*      ErrorRate,
 const int       nk,
 const int       npt,
 const int       nl,
 const int       nc
 )
     /* compute the percentage of misclassified pixels when the true
	segmentation is known */
/* ------------------------------------------------------------------- */
{
  int     il,ic,ik;
  int* NbPixels   = (int*) GenAlloc(nk+1,sizeof(int),0,
					"ComputeErrorRate","NbPixels");
  double* ErrorClass = (double*) GenAlloc(nk+1,sizeof(double),0,
					  "ComputeErrorRate","Error");


  /* For each record */
  for ( il = 0 ;  il < nl   ; il ++ )
    {
      for ( ic = 0 ;  ic < nc  ; ic ++ )
	{
	  int i = il*nc +ic;
	  /* do not take into account the boarder */
	  if (((il != 0) && (il!=nl-1)) || nl == 1)
	    {
	      if (((ic != 0) && (ic!=nl-1)) || nc == 1)
		{
		  /* Update the number of pixels for each class */
		  NbPixels[(int) (DataTrue[i])+1]++;
		  /* Update errors */
		  for (ik = 0; ik<nk ; ik ++)
		    if ((Hclassif[i*nk + ik] == 1) &&  (DataTrue[i] != ik))
		      {
			/* Overall error */
			ErrorClass[0]++;
			/* Error for the true class of the pixel */
			ErrorClass[(int) (DataTrue[i])+1]++;
		      }

		}
	    }
	}
    }
  /* Now the real errors */
  ErrorRate[0] = 100 * ErrorClass[0] / (npt-2*nl-2*nc+4);
  for (ik = 1; ik<=nk; ik++)
  	ErrorRate[ik] = 100 * ErrorClass[ik] / NbPixels[ik];

  GenFree(NbPixels) ;
  GenFree(ErrorClass) ;
} /* end of ComputeErrorRate */

/* ------------------------------------------------------------------- */
/* ~~~~~~~~~~~~~~~~~~~~~~~~ END OF FILE ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
